program ArrayExTest;

{$APPTYPE CONSOLE}

{$WARN GARBAGE OFF}

{$R *.res}

uses
  SysUtils,
  Classes,
  Windows,
  Indexes,
  Generics.Collections,
  Rapid.Generics,
  ShaDictionaries,
  TypInfo;

var
  List1 : TStringList;
  List2 : TStringList;
  Stage : integer;



Procedure WriteHeader(Text: TSymbolName);
begin
  inc(Stage);
  Writeln;
  Writeln(IntToStr(Stage)+'. '+StringReplace(string(Text),'System.','',[rfReplaceAll,rfIgnoreCase]));
end;


procedure PrepareConsoleWindow;
var
  Rect  : TSmallRect;
  Coord : TCoord;
begin
  Rect.Left:=0;
  Rect.Top:=0;
  Rect.Right:=110;
  Rect.Bottom:=50;
  Coord.X:=Rect.Right+1-Rect.Left;
  Coord.Y:=Rect.Bottom+1-Rect.Top;
  SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE), Coord);
  SetConsoleWindowInfo(GetStdHandle(STD_OUTPUT_HANDLE), True, Rect);

  SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE), Coord);

end;



function GetGuid: string; inline;
var
  MyGUID : TGUID;
begin
  CreateGUID(MyGUID);
  Result:=GUIDToString(MyGUID);
end;

procedure TArrayExFeatures;
type
  TDict = TArrayEx<Integer>;
var
  A,B,C : TArrayEx<Integer>;
  i     : Integer;
  n     : Integer;
  X     : TArray<integer>;
const
  SearchVal = 5;
begin
  WriteHeader('Logic Tests '+PTypeInfo(TypeInfo(TDict)).Name);

  Randomize;

  for i:=1 to 50 do begin
    A.Add(random(10));
  end;
  A.Sort;
  A.Delete(A.Low);
  for i:=1 to 8 do begin
    A.Delete(Random(A.Count));
  end;
  A.Delete(A.High);

  Writeln(A.ToString);
  Writeln('Length of array A = '+IntToStr(A.Count));

  for i:=1 to 10 do begin
    B.Insert(0,10-i);
  end;
  Writeln(B.ToString);

  Writeln('Length of array B = '+IntToStr(B.Count));

  Writeln('These are equal: ',BoolToStr(A=B));

  {$IF CompilerVersion>27}
  Writeln('B in A: ',BoolToStr(B in A));
  {$ENDIF}

  Writeln('C:=A+B');
  C:=A+B;
  Writeln(C.ToString);
  Writeln('Length of array C = '+IntToStr(C.Count));

  {$IF CompilerVersion>27}
  n:=0;
  for i in C.Collection do begin
    if i=SearchVal then inc(n);
  end;
  {$ELSE}
  n:=0;
  for i:=0 to C.High do begin
    if C[i]=SearchVal then inc(n);
  end;
  {$ENDIF}

  if C.Exists(SearchVal) then begin
    Writeln('Value "'+IntToStr(SearchVal)+'" found '+IntToStr(n)+' times in array C, at indexes = ['+C.IndexesOf(SearchVal).ToString+']');
  end;

  X:=C;
  Assert(Length(x)=C.Count);
  Writeln('Implicit cast to array passed.');
end;


procedure TArrayExTestInteger;
type
  TDict = TArrayEx<Integer>;
var
  A : TArrayEx<Integer>;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A.Clear;
  Write('Add 10.000.000 integers. ');
  T:=GetTickCount;
  for i:=0 to 9999999 do begin
    A.Add(i);
  end;
  T:=GetTickCount-T;
  Write(IntToStr(T)+' msec.');

  A.Clear;
  T:=GetTickCount;
  A.Count:=10000000;
  for i:=0 to 9999999 do begin
    A.Items[i]:=i;
  end;
  T:=GetTickCount-T;
  Writeln(' '+IntToStr(T)+' msec (optimised).');


  A.Clear;
  Write('Add 10.000 integers in 10.000.000 iterations. ');
  T:=GetTickCount;
  for i:=0 to 9999999 do begin
    A.AddUnique(Random(10000));
  end;
  T:=GetTickCount-T;
  WriteLn(IntToStr(T)+' msec.');

  Write('Locate 10.000 integers in 10.000.000 iterations. ');
  T:=GetTickCount;
  for i:=0 to 9999999 do begin
    A.Exists(Random(10000));
  end;
  T:=GetTickCount-T;
  Writeln(IntToStr(T)+' msec.');
end;


procedure THashTableTestInteger;
type
  TDict = THashTable<integer,integer>;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A:=TDict.Create;
  try
    Write('Add 10.000.000 integers. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.Add(i);
    end;
    T:=GetTickCount-T;
    Write(IntToStr(T)+' msec.');

    A.Clear;
    A.OnDuplicateKeyRule:=rmIgnore;
    T:=GetTickCount;
    A.BeginUpdates;
    for i:=0 to 9999999 do begin
      A.Add(i);
    end;
    A.EndUpdates;
    T:=GetTickCount-T;
    WriteLn(' '+IntToStr(T)+' msec (optimised).');

    A.Clear;
    Write('Add 10.000 integers in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.AddUnique(Random(10000),0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Locate 10.000 integers in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.Exists(Random(10000));
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

  finally
    A.Free;
  end;
end;


procedure TDictionaryTestInteger;
type
  TDict = Generics.Collections.TDictionary<integer,integer>;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A:=TDict.Create;
  try
    Write('Add 10.000.000 integers. ');
    T:=GetTickCount;

    for i:=0 to 9999999 do begin
      A.Add(i,0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    A.Clear;
    Write('Add 10.000 integers in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.AddOrSetValue(Random(10000),0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Locate 10.000 integers in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.ContainsKey(Random(10000));
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

  finally
    A.Free;
  end;
end;


procedure TShaDictionaryTestInteger;
type
  TDict = ShaDictionaries.TShaObjectDictionary;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A:=TDict.Create;
  try
    Write('Add 10.000.000 integers. ');
    T:=GetTickCount;

    for i:=0 to 9999999 do begin
      A.Add(i,0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    A.Clear;
    Write('Add 10.000 integers in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.Add(Random(10000),0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Locate 10.000 integers in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.Find(Random(10000));
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

  finally
    A.Free;
  end;
end;


procedure TRapidDictionaryTestInteger;
type
  TDict = Rapid.Generics.TRapidDictionary<integer,integer>;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A:=TDict.Create;
  try
    Write('Add 10.000.000 integers. ');
    T:=GetTickCount;

    for i:=0 to 9999999 do begin
      A.Add(i,0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');


    A.Clear;
    Write('Add 10.000 integers in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.AddOrSetValue(Random(10000),0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Locate 10.000 integers in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.ContainsKey(Random(10000));
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

  finally
    A.Free;
  end;
end;

procedure TRapidDictionaryTestInteger2;
type
//  TDict = THashTable<integer,integer>;
  TDict = Rapid.Generics.TRapidDictionary<integer,integer>;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A:=TDict.Create(256*1024);
  try
    Write('Add i. ');
    T:=GetTickCount;

    for i:=0 to 3*64*1024-1 do begin
      A.Add(i,0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Add i*256. ');
    A.Clear;
    T:=GetTickCount;

    for i:=0 to 3*64*1024-1 do begin
      A.Add(i*255,0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Add random(3*64*1024). ');
    A.Clear;
    T:=GetTickCount;
    for i:=0 to 3*64*1024-1 do begin
      A.AddOrSetValue(random(256*1024),0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

  finally
    A.Free;
  end;
end;

procedure TArrayExTestString;
type
  TDict = TArrayEx<string>;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A.Clear;
  Write('Add 1.000.000 GUIDs. ');
  T:=GetTickCount;
  for i:=0 to 999999 do begin
    A.Add(List1[i]);
  end;
  T:=GetTickCount-T;
  Write(IntToStr(T)+' msec.');

  A.Clear;
  T:=GetTickCount;
  A.Count:=1000000;
  for i:=0 to 999999 do begin
    A.Items[i]:=List1[i];
  end;
  T:=GetTickCount-T;
  WriteLn(' '+IntToStr(T)+' msec (optimised).');

  A.Clear;
  Write('Add '+IntToStr(List2.Count)+' strings in 10.000.000 iterations. ');
  T:=GetTickCount;
  for i:=0 to 9999999 do begin
    A.AddUnique(List2[Random(List2.Count)]);
  end;
  T:=GetTickCount-T;
  WriteLn(IntToStr(T)+' msec.');

  Write('Locate '+IntToStr(List2.Count)+' strings in 10.000.000 iterations. ');
  T:=GetTickCount;
  for i:=0 to 9999999 do begin
    A.Exists(List2[Random(List2.Count)]);
  end;
  T:=GetTickCount-T;
  WriteLn(IntToStr(T)+' msec.');
end;

procedure THashTableTestString;
type
  TDict = THashTable<String,integer>;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A:=TDict.Create;
  try
    Write('Add 1.000.000 GUIDs. ');
    T:=GetTickCount;
    for i:=0 to 999999 do begin
      A.Add(List1[i]);
    end;
    T:=GetTickCount-T;
    Write(IntToStr(T)+' msec.');

    A.Clear;
    A.OnDuplicateKeyRule:=rmIgnore;
    T:=GetTickCount;
    A.BeginUpdates;
    for i:=0 to 999999 do begin
      A.Add(List1[i]);
    end;
    A.EndUpdates;
    T:=GetTickCount-T;
    WriteLn(' '+IntToStr(T)+' msec (optimised).');

    A.Clear;
    Write('Add '+IntToStr(List2.Count)+' strings in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.AddUnique(List2[Random(List2.Count)],0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Locate '+IntToStr(List2.Count)+' strings in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.Exists(List2[Random(List2.Count)]);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

  finally
    A.Free;
  end;
end;


procedure TDictionaryTestString;
type
  TDict = Generics.Collections.TDictionary<String,integer>;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A:=TDict.Create;
  try
    Write('Add 1.000.000 GUIDs. ');
    T:=GetTickCount;
    for i:=0 to 999999 do begin
      A.Add(List1[i],i);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    A.Clear;
    Write('Add '+IntToStr(List2.Count)+' strings in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.AddOrSetValue(List2[Random(List2.Count)],0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Locate '+IntToStr(List2.Count)+' strings in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.ContainsKey(List2[Random(List2.Count)]);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');
  finally
    A.Free;
  end;
end;


procedure TRapidDictionaryTestString;
type
  TDict = Rapid.Generics.TRapidDictionary<String,integer>;
var
  A : TDict;
  i : Integer;
  T : Cardinal;
const
  SearchVal = 5;
begin
  WriteHeader('Benchmark '+PTypeInfo(TypeInfo(TDict)).Name);

  A:=TDict.Create;
  try
    Write('Add 1.000.000 GUIDs. ');
    T:=GetTickCount;
    for i:=0 to 999999 do begin
      A.Add(List1[i],i);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    A.Clear;
    Write('Add '+IntToStr(List2.Count)+' strings in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.AddOrSetValue(List2[Random(List2.Count)],0);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');

    Write('Locate '+IntToStr(List2.Count)+' strings in 10.000.000 iterations. ');
    T:=GetTickCount;
    for i:=0 to 9999999 do begin
      A.ContainsKey(List2[Random(List2.Count)]);
    end;
    T:=GetTickCount-T;
    WriteLn(IntToStr(T)+' msec.');
  finally
    A.Free;
  end;
end;

procedure Init(Full: boolean = True);
var
  i : integer;
begin
  ReportMemoryLeaksOnShutdown:=True;
  PrepareConsoleWindow;

  if not Full then Exit;

  Write('Preparing GUIDs...');
  List1:=TStringList.Create;
  for i:=0 to 999999 do begin
    List1.Add(GetGuid);
  end;
  WriteLn('Done.');
  Write('Loading Dictionary...');
  List2:=TStringList.Create;
  List2.LoadFromFile('    (.  ).txt',TEncoding.UTF8);

  WriteLn('Done.');
end;

procedure Finalize;
begin
  Writeln;
  Write('Tests complete.');
  Readln;

  List1.Free;
  List2.Free;
end;

const
  TestIntegers = True;
  TestStrings = False;
begin
  Init(TestStrings);
  try
//    TArrayExFeatures;

    if TestIntegers then begin
      TArrayExTestInteger;
      THashTableTestInteger;
      TDictionaryTestInteger;
      TRapidDictionaryTestInteger;
      TShaDictionaryTestInteger;
    end;

    if TestStrings then begin
      TArrayExTestString;
      THashTableTestString;
      TDictionaryTestString;
      TRapidDictionaryTestString;
    end;

  except
    on E: Exception do begin
      WriteLn('Error: '+E.Message);
    end;
  end;
  Finalize;
end.


