unit Rapid.Generics;

{******************************************************************************}
{ Copyright (c) 2017 Dmitry Mozulyov                                           }
{                                                                              }
{ Permission is hereby granted, free of charge, to any person obtaining a copy }
{ of this software and associated documentation files (the "Software"), to deal}
{ in the Software without restriction, including without limitation the rights }
{ to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    }
{ copies of the Software, and to permit persons to whom the Software is        }
{ furnished to do so, subject to the following conditions:                     }
{                                                                              }
{ The above copyright notice and this permission notice shall be included in   }
{ all copies or substantial portions of the Software.                          }
{                                                                              }
{ THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   }
{ IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     }
{ FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  }
{ AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       }
{ LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,}
{ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN    }
{ THE SOFTWARE.                                                                }
{                                                                              }
{ email: softforyou@inbox.ru                                                   }
{ skype: dimandevil                                                            }
{******************************************************************************}

// compiler directives
{$ifdef FPC}
  {$MESSAGE ERROR 'FreePascal not supported'}
{$else}
  {$ifNdef UNICODE}
    {$MESSAGE ERROR 'Old version compiler not supported'} // CompilerVersion < 20
  {$endif}
  {$WARN UNSAFE_CODE OFF}
  {$WARN UNSAFE_TYPE OFF}
  {$WARN UNSAFE_CAST OFF}
  {$WARN SYMBOL_DEPRECATED OFF}
  {$if CompilerVersion >= 24}
    {$LEGACYIFEND ON}
  {$ifend}
  {$if CompilerVersion < 23}
    {$define CPUX86}
  {$ifend}
  {$if CompilerVersion >= 23}
    {$define UNITSCOPENAMES}
  {$ifend}
  {$if CompilerVersion >= 21}
    {$WEAKLINKRTTI ON}
    {$RTTI EXPLICIT METHODS([]) PROPERTIES([]) FIELDS([])}
  {$ifend}
  {$if CompilerVersion >= 28}
    {$define SMARTGENERICS}
  {$ifend}
{$endif}
{$POINTERMATH ON}
{$U-}{$V+}{$B-}{$X+}{$T+}{$P+}{$H+}{$J-}{$Z1}{$A4}
{$O+}{$R-}{$I-}{$Q-}{$W-}
{$if Defined(CPUX86) or Defined(CPUX64)}
  {$define CPUINTEL}
{$ifend}
{$if Defined(CPUX64) or Defined(CPUARM64)}
  {$define LARGEINT}
{$ifend}
{$if (not Defined(CPUX64)) and (not Defined(CPUARM64))}
  {$define SMALLINT}
{$ifend}
(*$HPPEMIT '#pragma option -w-8022'*)

interface
  uses {$ifdef MSWINDOWS}
         {$ifdef UNITSCOPENAMES}Winapi.Windows{$else}Windows{$endif},
       {$endif}
       {$ifdef USE_LIBICU}
         System.Internal.ICU,
       {$endif}
       {$ifdef MACOS}
         Posix.Langinfo, Posix.Locale, Posix.String_,
       {$endif}
       {$ifdef UNITSCOPENAMES}
         System.Types, System.SysUtils, System.TypInfo, System.Variants,
         System.SysConst, System.RTLConsts, System.Math
       {$else}
         Types, SysUtils, TypInfo, Variants, SysConst, RTLConsts, Math
       {$endif};

type

{ TRAIIHelper record
  Low level RTTI routine: initialization/finalization }

  TRAIIHelper = record
  public type
    TClearNativeProc = procedure(P, TypeInfo: Pointer);
    TClearNativeRec = record
      Offset: NativeInt;
      DynTypeInfo: PTypeInfo;
      ClearNativeProc: TClearNativeProc;
    end;
    TClearNatives = record
      Items: TArray<TClearNativeRec>;
      ItemSingle: TClearNativeRec;
      Count: NativeInt;
      procedure Clear; inline;
      procedure Add(AOffset: NativeInt; ADynTypeInfo: PTypeInfo; AClearNativeProc: TClearNativeProc);
    end;
    {$ifdef WEAKREF}
      TInitNativeRec = record
        Offset: NativeInt;
      end;
      TInitNatives = record
        Items: TArray<TInitNativeRec>;
        ItemSingle: TInitNativeRec;
        Count: NativeInt;
        procedure Clear; inline;
        procedure Add(AOffset: NativeInt);
      end;
    {$else}
      TNativeRec = TClearNativeRec;
      TNatives = TClearNatives;
    {$endif}
    TStaticArrayRec = record
      Offset: NativeInt;
      StaticTypeInfo: PTypeInfo;
      Count: NativeUInt;
    end;
    TStaticArrays = record
      Items: TArray<TStaticArrayRec>;
      Count: NativeInt;
      procedure Clear; inline;
      procedure Add(AOffset: NativeInt; AStaticTypeInfo: PTypeInfo; ACount: NativeUInt);
    end;
  public const
    varDeepData = $BFE8;
  private type
    PFieldInfo = ^TFieldInfo;
    TFieldInfo = packed record
      TypeInfo: PPTypeInfo;
      Offset: Cardinal;
      {$ifdef LARGEINT}
      _Padding: Integer;
      {$endif}
    end;
    PFieldTable = ^TFieldTable;
    TFieldTable = packed record
      X: Word;
      Size: Cardinal;
      Count: Cardinal;
      Fields: array [0..0] of TFieldInfo;
    end;
    TData16 = packed record
    case Integer of
      0: (Native: NativeInt);
      1: (Method: TMethod);
      2: (VarData: TVarData);
      3: (Bytes: array[0..15] of Byte);
      4: (Words: array[0..7] of Word);
      5: (Integers: array[0..3] of Integer);
      6: (Int64s: array[0..1] of Int64);
      7: (Natives: array[0..{$ifdef LARGEINT}1{$else .SMALLINT}3{$endif}] of NativeUInt);
    end;
    TData16<TOffset> = packed record
      Offset: TOffset;
    case Integer of
      0: (Native: NativeInt);
      1: (Method: TMethod);
      2: (VarData: TVarData);
      3: (Bytes: array[0..15] of Byte);
      4: (Words: array[0..7] of Word);
      5: (Integers: array[0..3] of Integer);
      6: (Int64s: array[0..1] of Int64);
      7: (Natives: array[0..{$ifdef LARGEINT}1{$else .SMALLINT}3{$endif}] of NativeUInt);
    end;
    T1 = Byte;
    T2 = Word;
    T3 = array[1..3] of Byte;
    T4 = Cardinal;
    T5 = array[1..5] of Byte;
    T6 = array[1..6] of Byte;
    T7 = array[1..7] of Byte;
    T8 = Int64;
    T9 = array[1..9] of Byte;
    T10 = array[1..10] of Byte;
    T11 = array[1..11] of Byte;
    T12 = array[1..12] of Byte;
    T13 = array[1..13] of Byte;
    T14 = array[1..14] of Byte;
    T15 = array[1..15] of Byte;
    T16 = array[1..16] of Byte;
    T17 = array[1..17] of Byte;
    T18 = array[1..18] of Byte;
    T19 = array[1..19] of Byte;
    T20 = array[1..20] of Byte;
    T21 = array[1..21] of Byte;
    T22 = array[1..22] of Byte;
    T23 = array[1..23] of Byte;
    T24 = array[1..24] of Byte;
    T25 = array[1..25] of Byte;
    T26 = array[1..26] of Byte;
    T27 = array[1..27] of Byte;
    T28 = array[1..28] of Byte;
    T29 = array[1..29] of Byte;
    T30 = array[1..30] of Byte;
    T31 = array[1..31] of Byte;
    T32 = array[1..32] of Byte;
    T33 = array[1..33] of Byte;
    T34 = array[1..34] of Byte;
    T35 = array[1..35] of Byte;
    T36 = array[1..36] of Byte;
    T37 = array[1..37] of Byte;
    T38 = array[1..38] of Byte;
    T39 = array[1..39] of Byte;
    T40 = array[1..40] of Byte;
    TTemp40 = record
    case Integer of
       1: (V1: T1);
       2: (V2: T2);
       3: (V3: T3);
       4: (V4: T4);
       5: (V5: T5);
       6: (V6: T6);
       7: (V7: T7);
       8: (V8: T8);
       9: (V9: T9);
      10: (V10: T10);
      11: (V11: T11);
      12: (V12: T12);
      13: (V13: T13);
      14: (V14: T14);
      15: (V15: T15);
      16: (V16: T16);
      17: (V17: T17);
      18: (V18: T18);
      19: (V19: T19);
      20: (V20: T20);
      21: (V21: T21);
      22: (V22: T22);
      23: (V23: T23);
      24: (V24: T24);
      25: (V25: T25);
      26: (V26: T26);
      27: (V27: T27);
      28: (V28: T28);
      29: (V29: T29);
      30: (V30: T30);
      31: (V31: T31);
      32: (V32: T32);
      33: (V33: T33);
      34: (V34: T34);
      35: (V35: T35);
      36: (V36: T36);
      37: (V37: T37);
      38: (V38: T38);
      39: (V39: T39);
      40: (V40: T40);
    end;
  private
    FTypeInfo: PTypeInfo;
    FSize: NativeInt;

    // initialization/finalization
    procedure Include(AOffset: NativeInt; Value: PTypeInfo);
    procedure Initialize(Value: PTypeInfo);
    class function IsManagedTypeInfo(Value: PTypeInfo): Boolean; static;
    class procedure InitsProcNativeSingle(const Self: TRAIIHelper; P: Pointer); static;
    class procedure InitsProcNatives(const Self: TRAIIHelper; P: Pointer); static;
    class procedure InitsProc(const Self: TRAIIHelper; P: Pointer); static;
    class procedure ClearsProcNativeSingle(const Self: TRAIIHelper; P: Pointer); static;
    class procedure ClearsProcNatives(const Self: TRAIIHelper; P: Pointer); static;
    class procedure ClearsProc(const Self: TRAIIHelper; P: Pointer); static;

    // TClearNativeProc-anonyms
    class procedure ULStrClear(P: Pointer); static;
    {$ifdef MSWINDOWS}
    class procedure WStrClear(P: Pointer); static;
    {$endif}
    class procedure IntfClear(P: Pointer); static;
    class procedure VarClear(P: Pointer); static;
    class procedure DynArrayClear(P, TypeInfo: Pointer); static;
    {$ifdef AUTOREFCOUNT}
    class procedure RefObjClear(P: Pointer); static;
    {$endif}
    {$ifdef WEAKINSTREF}
    class procedure WeakObjClear(P: Pointer); static;
    {$endif}
    {$ifdef WEAKREF}
    class procedure WeakMethodClear(P: Pointer); static;
    {$endif}
    {$ifdef WEAKINTFREF}
    class procedure WeakIntfClear(P: Pointer); static;
    {$endif}
  public
    {$ifdef WEAKREF}
      InitNatives: TInitNatives;
      ClearNatives: TClearNatives;
    {$else}
      Natives: TNatives;
    {$endif}
    StaticArrays: TStaticArrays;
    InitProc: procedure(const Self: TRAIIHelper; P: Pointer);
    ClearProc: procedure(const Self: TRAIIHelper; P: Pointer);

    property TypeInfo: PTypeInfo read FTypeInfo write Initialize;
    property Size: NativeInt read FSize;
  end;

  TRAIIHelper<T> = record
  public type
    P = ^T;
  private class var
    FOptions: TRAIIHelper;
    class function GetIsManaged: Boolean; static; inline;
  public
    class constructor Create;
    class property Options: TRAIIHelper read FOptions;
    class property IsManaged: Boolean read GetIsManaged;
    class procedure Init(Item: P); static; inline;
    class procedure InitArray(Items: P; Count: NativeUInt; Offset: NativeUInt = 0); static;
    class procedure Clear(Item: P); static; inline;
    class procedure ClearArray(Items: P; Count: NativeUInt; Offset: NativeUInt = 0); static;
  end;


{ InterfaceDefaults record
  Default functions/interfaces }

  InterfaceDefaults = record
  public type
    TMethodPtr = procedure of object;
    TTriple = packed record
    case Integer of
      0: (Low: Word; High: Byte);
      1: (Bytes: array[0..2] of Byte);
    end;
    IComparerInst = packed record
      Vtable: Pointer;
      Size: NativeInt;
      QueryInterface,
      AddRef,
      Release,
      Compare: Pointer;
    end;
    IEqualityComparerInst = packed record
      Vtable: Pointer;
      Size: NativeInt;
      QueryInterface,
      AddRef,
      Release,
      Equals,
      GetHashCode: Pointer;
    end;
    TDefaultComparer<T> = record
    public class var
      Instance: IComparerInst;
    public
      class constructor Create;
      class procedure Initialize; static;
    end;
    TDefaultEqualityComparer<T> = record
    public class var
      Instance: IEqualityComparerInst;
    public
      class constructor Create;
      class procedure Initialize; static;
    end;
  private
    class function Compare_Var_Difficult(Equal: Boolean; Left, Right: PVariant): Integer; static;
    class function GetHashCode_Var_Difficult(Value: PVariant): Integer; static;
  public
    // Nop Interface
    class function NopQueryInterface(Inst: Pointer; const IID: TGUID; out Obj): HResult; stdcall; static;
    class function NopAddRef(Inst: Pointer): Integer; stdcall; static;
    class function NopRelease(Inst: Pointer): Integer; stdcall; static;

    // IComparer<T>
    class function Compare_I1(Inst: Pointer; Left, Right: Shortint): Integer; static;
    class function Compare_U1(Inst: Pointer; Left, Right: Byte): Integer; static;
    class function Compare_I2(Inst: Pointer; Left, Right: Smallint): Integer; static;
    class function Compare_U2(Inst: Pointer; Left, Right: Word): Integer; static;
    class function Compare_I4(Inst: Pointer; Left, Right: Integer): Integer; static;
    class function Compare_U4(Inst: Pointer; Left, Right: Cardinal): Integer; static;
    class function Compare_I8(Inst: Pointer; Left, Right: Int64): Integer; static;
    class function Compare_U8(Inst: Pointer; Left, Right: UInt64): Integer; static;
    class function Compare_F4(Inst: Pointer; Left, Right: Single): Integer; static;
    class function Compare_F8(Inst: Pointer; Left, Right: Double): Integer; static;
    class function Compare_FE(Inst: Pointer; Left, Right: Extended): Integer; static;
    class function Compare_Var(Inst: Pointer; Left, Right: PVarData): Integer; static;
    class function Compare_OStr(Inst: Pointer; Left, Right: PByte): Integer; static;
    class function Compare_LStr(Inst: Pointer; Left, Right: PByte): Integer; static;
    class function Compare_UStr(Inst: Pointer; Left, Right: PByte): Integer; static;
    class function Compare_WStr(Inst: Pointer; Left, Right: PByte): Integer; static;
    class function Compare_Method(Inst: Pointer; const Left, Right: TMethodPtr): Integer; static;
    class function Compare_Dyn(const Inst: IComparerInst; Left, Right: PByte): Integer; static;
    class function Compare_Bin2(Inst: Pointer; Left, Right: Word): Integer; static;
    class function Compare_Bin3(Inst: Pointer; const Left, Right: TTriple): Integer; static;
    class function Compare_Bin4(Inst: Pointer; Left, Right: Cardinal): Integer; static;
    class function Compare_Bin8(Inst: Pointer; Left, Right: Int64): Integer; static;
    class function Compare_Bin(const Inst: IComparerInst; Left, Right: PByte): Integer; static;

    // IEqualityComparer<T>
    class function Equals_N1(Inst: Pointer; Left, Right: Byte): Boolean; static;
    class function GetHashCode_N1(Inst: Pointer; Value: Byte): Integer; static;
    class function Equals_N2(Inst: Pointer; Left, Right: Word): Boolean; static;
    class function GetHashCode_N2(Inst: Pointer; Value: Word): Integer; static;
    class function Equals_N4(Inst: Pointer; Left, Right: Integer): Boolean; static;
    class function GetHashCode_N4(Inst: Pointer; Value: Integer): Integer; static;
    class function Equals_N8(Inst: Pointer; Left, Right: Int64): Boolean; static;
    class function GetHashCode_N8(Inst: Pointer; Value: Int64): Integer; static;
    class function Equals_Class(Inst: Pointer; Left, Right: TObject): Boolean; static;
    class function GetHashCode_Class(Inst: Pointer; Value: TObject): Integer; static;
    class function GetHashCode_Ptr(Inst: Pointer; Value: NativeInt): Integer; static;
    class function Equals_F4(Inst: Pointer; Left, Right: Single): Boolean; static;
    class function GetHashCode_F4(Inst: Pointer; Value: Single): Integer; static;
    class function Equals_F8(Inst: Pointer; Left, Right: Double): Boolean; static;
    class function GetHashCode_F8(Inst: Pointer; Value: Double): Integer; static;
    class function Equals_FE(Inst: Pointer; Left, Right: Extended): Boolean; static;
    class function GetHashCode_FE(Inst: Pointer; Value: Extended): Integer; static;
    class function Equals_Var(Inst: Pointer; Left, Right: PVarData): Boolean; static;
    class function GetHashCode_Var(Inst: Pointer; Value: PVarData): Integer; static;
    class function Equals_OStr(Inst: Pointer; Left, Right: PByte): Boolean; static;
    class function GetHashCode_OStr(Inst: Pointer; Value: PByte): Integer; static;
    class function Equals_LStr(Inst: Pointer; Left, Right: PByte): Boolean; static;
    class function GetHashCode_LStr(Inst: Pointer; Value: PByte): Integer; static;
    class function Equals_UStr(Inst: Pointer; Left, Right: PByte): Boolean; static;
    class function GetHashCode_UStr(Inst: Pointer; Value: PByte): Integer; static;
    class function Equals_WStr(Inst: Pointer; Left, Right: PByte): Boolean; static;
    class function GetHashCode_WStr(Inst: Pointer; Value: PByte): Integer; static;
    class function Equals_Method(Inst: Pointer; const Left, Right: TMethodPtr): Boolean; static;
    class function GetHashCode_Method(Inst: Pointer; const Value: TMethodPtr): Integer; static;
    class function Equals_Dyn(const Inst: IEqualityComparerInst; Left, Right: PByte): Boolean; static;
    class function GetHashCode_Dyn(const Inst: IEqualityComparerInst; Value: PByte): Integer; static;
    class function Equals_Bin3(Inst: Pointer; const Left, Right: TTriple): Boolean; static;
    class function GetHashCode_Bin3(Inst: Pointer; const Value: TTriple): Integer; static;
    class function Equals_Bin(const Inst: IEqualityComparerInst; Left, Right: PByte): Boolean; static;
    class function GetHashCode_Bin(const Inst: IEqualityComparerInst; Value: PByte): Integer; static;
  end;


{ System.Generics.Defaults
  Equivalent types }

  IComparer<T> = interface
    function Compare(const Left, Right: T): Integer;
  end;

  IEqualityComparer<T> = interface
    function Equals(const Left, Right: T): Boolean;
    function GetHashCode(const Value: T): Integer;
  end;

  TComparison<T> = reference to function(const Left, Right: T): Integer;

  // Abstract base class for IComparer<T> implementations, and a provider
  // of default IComparer<T> implementations.
  TComparer<T> = class(TInterfacedObject, IComparer<T>)
  public
    class function Default: IComparer<T>;
    class function Construct(const Comparison: TComparison<T>): IComparer<T>;
    function Compare(const Left, Right: T): Integer; virtual; abstract;
  end;

  TEqualityComparison<T> = reference to function(const Left, Right: T): Boolean;
  THasher<T> = reference to function(const Value: T): Integer;

  // Abstract base class for IEqualityComparer<T> implementations, and a provider
  // of default IEqualityComparer<T> implementations.
  TEqualityComparer<T> = class(TInterfacedObject, IEqualityComparer<T>)
  public
    class function Default: IEqualityComparer<T>; static;

    class function Construct(const EqualityComparison: TEqualityComparison<T>;
      const Hasher: THasher<T>): IEqualityComparer<T>;

    function Equals(const Left, Right: T): Boolean;
      reintroduce; overload; virtual; abstract;
    function GetHashCode(const Value: T): Integer;
      reintroduce; overload; virtual; abstract;
  end;

  // A non-reference-counted IInterface implementation.
  TSingletonImplementation = class(TObject, IInterface)
  protected
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
  end;

  TDelegatedComparer<T> = class(TComparer<T>)
  private
    FCompare: TComparison<T>;
  public
    constructor Create(const ACompare: TComparison<T>);
    function Compare(const Left, Right: T): Integer; override;
  end;

  TDelegatedEqualityComparer<T> = class(TEqualityComparer<T>)
  private
    FEquals: TEqualityComparison<T>;
    FGetHashCode: THasher<T>;
  public
    constructor Create(const AEquals: TEqualityComparison<T>;
      const AGetHashCode: THasher<T>);
    function Equals(const Left, Right: T): Boolean; overload; override;
    function GetHashCode(const Value: T): Integer; overload; override;
  end;

  TCustomComparer<T> = class(TSingletonImplementation, IComparer<T>, IEqualityComparer<T>)
  protected
    function Compare(const Left, Right: T): Integer; virtual; abstract;
    function Equals(const Left, Right: T): Boolean;
      reintroduce; overload; virtual; abstract;
    function GetHashCode(const Value: T): Integer;
      reintroduce; overload; virtual; abstract;
  end;

  TStringComparer = class(TCustomComparer<string>)
  private
    class var
      FOrdinal: TCustomComparer<string>;
  public
    class destructor Destroy;
    class function Ordinal: TStringComparer;
  end;

  TOrdinalIStringComparer = class(TStringComparer)
  private
    function CharsLower(Dest, Src: PWideChar; Count: Integer): Boolean;
    function GetHashCodeLower(const Value: string): Integer;
  public
    function Compare(const Left, Right: string): Integer; override;
    function Equals(const Left, Right: string): Boolean;
      reintroduce; overload; override;
    function GetHashCode(const Value: string): Integer;
      reintroduce; overload; override;
  end;

  TIStringComparer = class(TCustomComparer<string>)
  private
    class var
      FOrdinal: TCustomComparer<string>;
  public
    class destructor Destroy;
    class function Ordinal: TStringComparer;
  end;


{ System.Generics.Collections
  Basic types }

  TCollectionNotification = (cnAdded, cnRemoved, cnExtracted);
  TCollectionNotifyEvent<T> = procedure(Sender: TObject; const Item: T;
    Action: TCollectionNotification) of object;

  TEnumerator<T> = class abstract
  protected
    function DoGetCurrent: T; virtual; abstract;
    function DoMoveNext: Boolean; virtual; abstract;
  public
    property Current: T read DoGetCurrent;
    function MoveNext: Boolean;
  end;

  TEnumerable<T> = class abstract
  protected
    function DoGetEnumerator: TEnumerator<T>; virtual; abstract;
  public
    destructor Destroy; override;
    function GetEnumerator: TEnumerator<T>;
    function ToArray: TArray<T>; virtual;
  end;

  TPair<TKey,TValue> = record
    Key: TKey;
    Value: TValue;
    constructor Create(const AKey: TKey; const AValue: TValue);
  end;

  PObject = ^TObject;


{ TCustomDictionary<TKey,TValue> class
  Basic class for TDictionary<TKey,TValue>, TRapidDictionary<TKey,TValue> }

  TCustomDictionary<TKey,TValue> = class(TEnumerable<TPair<TKey,TValue>>)
  public type
    PItem = ^TItem;
    TItem = packed record
    private
      FKey: TKey;
      FValue: TValue;
      FNext: PItem;
      FHashCode: Integer;
    public
      property Key: TKey read FKey;
      property Value: TValue read FValue write FValue;
      property HashCode: Integer read FHashCode;
    end;
    TItemList = array[0..0] of TItem;
    PItemList = ^TItemList;

    TPairEnumerator = class(TEnumerator<TPair<TKey,TValue>>)
    private
      FDictionary: TCustomDictionary<TKey,TValue>;
      FIndex: NativeInt;
      function GetCurrent: TPair<TKey,TValue>; inline;
    protected
      function DoGetCurrent: TPair<TKey,TValue>; override;
      function DoMoveNext: Boolean; override;
    public
      constructor Create(const ADictionary: TCustomDictionary<TKey,TValue>);
      property Current: TPair<TKey,TValue> read GetCurrent;
      function MoveNext: Boolean; inline;
    end;

    TKeyEnumerator = class(TEnumerator<TKey>)
    private
      FDictionary: TCustomDictionary<TKey,TValue>;
      FIndex: NativeInt;
      function GetCurrent: TKey;
    protected
      function DoGetCurrent: TKey; override;
      function DoMoveNext: Boolean; override;
    public
      constructor Create(const ADictionary: TCustomDictionary<TKey,TValue>);
      property Current: TKey read GetCurrent;
      function MoveNext: Boolean;
    end;

    TValueEnumerator = class(TEnumerator<TValue>)
    private
      FDictionary: TCustomDictionary<TKey,TValue>;
      FIndex: NativeInt;
      function GetCurrent: TValue;
    protected
      function DoGetCurrent: TValue; override;
      function DoMoveNext: Boolean; override;
    public
      constructor Create(const ADictionary: TCustomDictionary<TKey,TValue>);
      property Current: TValue read GetCurrent;
      function MoveNext: Boolean;
    end;

    TKeyCollection = class(TEnumerable<TKey>)
    private
      [Weak] FDictionary: TCustomDictionary<TKey,TValue>;
      function GetCount: Integer; inline;
    protected
      function DoGetEnumerator: TEnumerator<TKey>; override;
    public
      constructor Create(const ADictionary: TCustomDictionary<TKey,TValue>);
      function GetEnumerator: TKeyEnumerator; reintroduce;
      function ToArray: TArray<TKey>; override; final;
      property Count: Integer read GetCount;
    end;

    TValueCollection = class(TEnumerable<TValue>)
    private
      [Weak] FDictionary: TCustomDictionary<TKey,TValue>;
      function GetCount: Integer; inline;
    protected
      function DoGetEnumerator: TEnumerator<TValue>; override;
    public
      constructor Create(const ADictionary: TCustomDictionary<TKey,TValue>);
      function GetEnumerator: TValueEnumerator; reintroduce;
      function ToArray: TArray<TValue>; override; final;
      property Count: Integer read GetCount;
    end;
  protected
    FItems: PItemList;
    FCapacity: NativeInt;
    FHashTable: TArray<PItem>;
    FHashTableMask: NativeInt;
    FCount: packed record
    case Boolean of
     False: (Int: Integer);
      True: (Native: NativeInt);
    end;
    FKeyCollection: TKeyCollection;
    FValueCollection: TValueCollection;

    // rehash
    procedure Rehash(NewTableCount: NativeInt);
    procedure SetCapacity(ACapacity: NativeInt);
    function Grow: TCustomDictionary<TKey,TValue>;

    // items
    function NewItem: PItem;
    procedure DisposeItem(Item: PItem);
    procedure DoCleanupItems(Item: PItem; Count: NativeInt); virtual;

    // enumarators
    function DoGetEnumerator: TEnumerator<TPair<TKey,TValue>>; override;
    function GetKeys: TKeyCollection; inline;
    function GetValues: TValueCollection; inline;

    // helpers
    class function IntfMethod(Intf: Pointer; MethodNum: NativeUInt = 3): TMethod; inline; static;
    class procedure ClearMethod(var Method); inline; static;
  protected
    FOnKeyNotify: TCollectionNotifyEvent<TKey>;
    FOnValueNotify: TCollectionNotifyEvent<TValue>;
    FInternalKeyNotify: TCollectionNotifyEvent<TKey>;
    FInternalValueNotify: TCollectionNotifyEvent<TValue>;
    FInternalItemNotify: procedure(const Item: TItem; Action: TCollectionNotification) of object;

    procedure SetKeyNotify(const Value: TCollectionNotifyEvent<TKey>);
    procedure SetValueNotify(const Value: TCollectionNotifyEvent<TValue>);
    procedure KeyNotify(const Key: TKey; Action: TCollectionNotification); virtual;
    procedure ValueNotify(const Value: TValue; Action: TCollectionNotification); virtual;
    procedure KeyNotifyCaller(Sender: TObject; const Item: TKey; Action: TCollectionNotification);
    procedure ValueNotifyCaller(Sender: TObject; const Item: TValue; Action: TCollectionNotification);
    procedure ItemNotifyCaller(const Item: TItem; Action: TCollectionNotification);
    procedure ItemNotifyEvents(const Item: TItem; Action: TCollectionNotification);
    procedure ItemNotifyKey(const Item: TItem; Action: TCollectionNotification);
    procedure ItemNotifyValue(const Item: TItem; Action: TCollectionNotification);
    procedure SetNotifyMethods; virtual;
  protected const
    FOUND_NONE = 0;
    FOUND_EXCEPTION = 1;
    FOUND_DELETE = 2;
    FOUND_REPLACE = 3;
    FOUND_MASK = 3;

    EMPTY_NONE = 0 shl 2;
    EMPTY_EXCEPTION = 1 shl 2;
    EMPTY_NEW = 2 shl 2;
    EMPTY_MASK = 3 shl 2;
  protected
    FInternalFindValue: ^TValue;
    FDefaultValue: TValue;
  public
    constructor Create(ACapacity: Integer = 0);
    destructor Destroy; override;
    procedure Clear;
    procedure TrimExcess;
    function ContainsValue(const Value: TValue): Boolean;

    function GetEnumerator: TPairEnumerator; reintroduce;
    function ToArray: TArray<TPair<TKey,TValue>>; override; final;

    property Keys: TKeyCollection read GetKeys;
    property Values: TValueCollection read GetValues;
  end;


{ TDictionary<TKey,TValue>
  System.Generics.Collections equivalent

  Included 3 members:
    function Find(const Key: TKey): PItem;
    function FindOrAdd(const Key: TKey): PItem;
    property List: PItemList read FItems; }

  TDictionary<TKey,TValue> = class(TCustomDictionary<TKey,TValue>)
  public type
    TItem = TCustomDictionary<TKey,TValue>.TItem;
    PItem = TCustomDictionary<TKey,TValue>.PItem;
    PItemList = TCustomDictionary<TKey,TValue>.PItemList;
  private type
    TInternalFindStored = record
      HashCode: Integer;
      Parent: ^PItem;
    end;
  protected
    FComparer: IEqualityComparer<TKey>;
    FComparerEquals: function(const Left, Right: TKey): Boolean of object;
    FComparerGetHashCode: function(const Value: TKey): Integer of object;

    function InternalFindItem(const Key: TKey; const FindMode: Integer): PItem;
    function GetItem(const Key: TKey): TValue;
    procedure SetItem(const Key: TKey; const Value: TValue);
  public
    constructor Create(ACapacity: Integer = 0); overload;
    constructor Create(const AComparer: IEqualityComparer<TKey>); overload;
    constructor Create(ACapacity: Integer; const AComparer: IEqualityComparer<TKey>); overload;
    constructor Create(const Collection: TEnumerable<TPair<TKey,TValue>>); overload;
    constructor Create(const Collection: TEnumerable<TPair<TKey,TValue>>; const AComparer: IEqualityComparer<TKey>); overload;
    destructor Destroy; override;

    function Find(const Key: TKey): PItem;
    function FindOrAdd(const Key: TKey): PItem;
    procedure Add(const Key: TKey; const Value: TValue);
    procedure Remove(const Key: TKey);
    function ExtractPair(const Key: TKey): TPair<TKey,TValue>;
    function TryGetValue(const Key: TKey; out Value: TValue): Boolean;
    procedure AddOrSetValue(const Key: TKey; const Value: TValue);
    function ContainsKey(const Key: TKey): Boolean;

    property Items[const Key: TKey]: TValue read GetItem write SetItem; default;
    property List: PItemList read FItems;
    property Count: Integer read FCount.Int;
    property OnKeyNotify: TCollectionNotifyEvent<TKey> read FOnKeyNotify write SetKeyNotify;
    property OnValueNotify: TCollectionNotifyEvent<TValue> read FOnValueNotify write SetValueNotify;
  end;


{ TRapidDictionary<TKey,TValue> class
  Rapid "inline" TDictionary equivalent with default hash code and comparer }

  TRapidDictionary<TKey,TValue> = class(TCustomDictionary<TKey,TValue>)
  public type
    TItem = TCustomDictionary<TKey,TValue>.TItem;
    PItem = TCustomDictionary<TKey,TValue>.PItem;
    PItemList = TCustomDictionary<TKey,TValue>.PItemList;
  private type
    TInternalFindStored = record
      HashCode: Integer;
      Parent: ^PItem;
      {$ifdef SMARTGENERICS}
      Self: TRapidDictionary<TKey,TValue>;
      case Integer of
        0: (SingleRec: packed record
              Exponent: Integer;
              case Integer of
                0: (Mantissa: Single);
                1: (HighInt: Integer);
            end);
        1: (DoubleRec: packed record
              Exponent: Integer;
              case Integer of
                0: (Mantissa: Double);
                1: (LowInt: Integer; HighInt: Integer);
            end);
        2: (ExtendedRec: packed record
              Exponent: Integer;
              case Integer of
                0: (Mantissa: Extended);
                1: (LowInt: Integer; Middle: Word; HighInt: Integer);
            end);
      {$endif}
    end;
  protected
    {$ifNdef SMARTGENERICS}
    FDefaultComparer: IEqualityComparer<TKey>;
    FDefaultEquals: function(const Left, Right: TKey): Boolean of object;
    FDefaultGetHashCode: function(const Value: TKey): Integer of object;
    {$endif}
    function InternalFindItem(const Key: TKey; const FindMode: Integer): PItem;
    function GetItem(const Key: TKey): TValue; inline;
    procedure SetItem(const Key: TKey; const Value: TValue); inline;
  public
    constructor Create(ACapacity: Integer = 0); overload;
    constructor Create(const Collection: TEnumerable<TPair<TKey,TValue>>); overload;
    destructor Destroy; override;

    function Find(const Key: TKey): PItem; inline;
    function FindOrAdd(const Key: TKey): PItem; inline;
    procedure Add(const Key: TKey; const Value: TValue); inline;
    procedure Remove(const Key: TKey); inline;
    function ExtractPair(const Key: TKey): TPair<TKey,TValue>;
    function TryGetValue(const Key: TKey; out Value: TValue): Boolean;
    procedure AddOrSetValue(const Key: TKey; const Value: TValue); inline;
    function ContainsKey(const Key: TKey): Boolean; inline;

    property Items[const Key: TKey]: TValue read GetItem write SetItem; default;
    property List: PItemList read FItems;
    property Count: Integer read FCount.Int;
    property OnKeyNotify: TCollectionNotifyEvent<TKey> read FOnKeyNotify write SetKeyNotify;
    property OnValueNotify: TCollectionNotifyEvent<TValue> read FOnValueNotify write SetValueNotify;
  end;


{ TArray class }

  TArray = class
  protected const
    HIGH_NATIVE = {$ifdef LARGEINT}63{$else}31{$endif};
    HIGH_NATIVE_BIT = NativeInt(1) shl HIGH_NATIVE;
    BUFFER_SIZE = 1024;
    RADIX_BUFFER_SIZE = 2048;
    INSERTION_SORT_LIMIT = 45;
  protected type
    TItemList<T> = array[0..15] of T;
    TSortStackItem<T> = record
      First: ^T;
      Last: ^T;
    end;
    TSortStack<T> = array[0..63] of TSortStackItem<T>;

    HugeByteArray = array[0..High(Integer) div SizeOf(Byte) - 1] of Byte;
    HugeWordArray = array[0..High(Integer) div SizeOf(Word) - 1] of Word;
    HugeCardinalArray = array[0..High(Integer) div SizeOf(Cardinal) - 1] of Cardinal;
    HugeNativeArray = array[0..High(Integer) div SizeOf(NativeUInt) - 1] of NativeUInt;

    TLMemory = packed record
    case Integer of
      0: (LBytes: HugeByteArray);
      1: (LWords: HugeWordArray);
      2: (LCardinals: HugeCardinalArray);
      3: (LNatives: HugeNativeArray);
      4: (L1: array[1..1] of Byte;
          case Integer of
            0: (LWords1: HugeWordArray);
            1: (LCardinals1: HugeCardinalArray);
            2: (LNatives1: HugeNativeArray);
          );
      5: (L2: array[1..2] of Byte;
          case Integer of
            0: (LCardinals2: HugeCardinalArray);
            1: (LNatives2: HugeNativeArray);
          );
      6: (L3: array[1..3] of Byte;
          case Integer of
            0: (LCardinals3: HugeCardinalArray);
            1: (LNatives3: HugeNativeArray);
          );
    {$ifdef LARGEINT}
      7: (L4: array[1..4] of Byte; LNatives4: HugeNativeArray);
      8: (L5: array[1..5] of Byte; LNatives5: HugeNativeArray);
      9: (L6: array[1..6] of Byte; LNatives6: HugeNativeArray);
     10: (L7: array[1..7] of Byte; LNatives7: HugeNativeArray);
    {$endif}
    end;
    PLMemory = ^TLMemory;

    TRMemory = packed record
    case Integer of
      0: (RBytes: HugeByteArray);
      1: (RWords: HugeWordArray);
      2: (RCardinals: HugeCardinalArray);
      3: (RNatives: HugeNativeArray);
      4: (R1: array[1..1] of Byte;
          case Integer of
            0: (RWords1: HugeWordArray);
            1: (RCardinals1: HugeCardinalArray);
            2: (RNatives1: HugeNativeArray);
          );
      5: (R2: array[1..2] of Byte;
          case Integer of
            0: (RCardinals2: HugeCardinalArray);
            1: (RNatives2: HugeNativeArray);
          );
      6: (R3: array[1..3] of Byte;
          case Integer of
            0: (RCardinals3: HugeCardinalArray);
            1: (RNatives3: HugeNativeArray);
          );
    {$ifdef LARGEINT}
      7: (R4: array[1..4] of Byte; RNatives4: HugeNativeArray);
      8: (R5: array[1..5] of Byte; RNatives5: HugeNativeArray);
      9: (R6: array[1..6] of Byte; RNatives6: HugeNativeArray);
     10: (R7: array[1..7] of Byte; RNatives7: HugeNativeArray);
    {$endif}
    end;
    PRMemory = ^TRMemory;

    TSortHelper<T> = record
      Pivot: T;
      Inst: Pointer;
      Compare: function(Inst: Pointer; const Left, Right: T): Integer;

      procedure Init(const Comparer: IComparer<T>); overload;
      procedure Init(const Comparison: TComparison<T>); overload;
      procedure Init; overload;
      procedure FillZero;
    end;

    TFloat = packed record
    case Integer of
      0: (VSingle: Single);
      1: (VDouble: Double);
      2: (VExtended: Extended);
      3: (SSingle: Integer);
      4: (B1: array[1..SizeOf(Double) - SizeOf(Integer)] of Byte; SDouble: Integer);
      5: (B2: array[1..SizeOf(Extended) - SizeOf(Integer)] of Byte; SExtended: Integer);
    end;

    TSearchHelper = record
      Count: NativeInt;
      Comparer: Pointer;
    end;
  protected
    class procedure CheckArrays(Source, Destination: Pointer; SourceIndex, SourceLength, DestIndex, DestLength, Count: NativeInt); static;
    class function SortItemPivot<T>(const I, J: Pointer): Pointer; static; inline;
    class function SortItemNext<T>(const StackItem, I, J: Pointer): Pointer; static; inline;

    {$ifdef SMARTGENERICS}
    class function SortItemCount<T>(const I, J: Pointer): NativeInt; static; inline;
    class function SortBinaryMarker<T>(const Binary: Pointer): NativeUInt; static; inline;
    class function SortBinaryComparer<T>(const Pivot, Binary: Pointer): Integer; static; inline;

    class procedure RadixSort<T>(const Values: Pointer; const Count, Flags: NativeInt); static;
    class function RadixSortSigneds<T>(var StackItem: TSortStackItem<T>): Pointer; static;
    class function RadixSortDescendingSigneds<T>(var StackItem: TSortStackItem<T>): Pointer; static;
    class function RadixSortUnsigneds<T>(var StackItem: TSortStackItem<T>): Pointer; static;
    class function RadixSortDescendingUnsigneds<T>(var StackItem: TSortStackItem<T>): Pointer; static;
    class function RadixSortFloats<T>(var StackItem: TSortStackItem<T>): Pointer; static;
    class function RadixSortDescendingFloats<T>(var StackItem: TSortStackItem<T>): Pointer; static;

    class procedure SortSigneds<T>(const Values: Pointer; const Count: NativeInt); static;
    class procedure SortDescendingSigneds<T>(const Values: Pointer; const Count: NativeInt); static;
    class procedure SortUnsigneds<T>(const Values: Pointer; const Count: NativeInt); static;
    class procedure SortDescendingUnsigneds<T>(const Values: Pointer; const Count: NativeInt); static;
    class procedure SortFloats<T>(const Values: Pointer; const Count: NativeInt); static;
    class procedure SortDescendingFloats<T>(const Values: Pointer; const Count: NativeInt); static;
    class procedure SortBinaries<T>(const Values: Pointer; const Count: NativeInt; var PivotBig: T); static;
    class procedure SortDescendingBinaries<T>(const Values: Pointer; const Count: NativeInt; var PivotBig: T); static;
    {$endif}
    class procedure SortUniversals<T>(const Values: Pointer; const Count: NativeInt; var Helper: TSortHelper<T>); static;
    class procedure SortDescendingUniversals<T>(const Values: Pointer; const Count: NativeInt; var Helper: TSortHelper<T>); static;

    {$ifdef SMARTGENERICS}
    class function SearchSigneds<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt; static;
    class function SearchUnsigneds<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt; static;
    class function SearchFloats<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt; static;
    class function SearchBinaries<T>(Values: Pointer; Count: NativeInt; const Item: T): NativeInt; static;

    class function SearchDescendingSigneds<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt; static;
    class function SearchDescendingUnsigneds<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt; static;
    class function SearchDescendingFloats<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt; static;
    class function SearchDescendingBinaries<T>(Values: Pointer; Count: NativeInt; const Item: T): NativeInt; static;
    {$endif}
    class function SearchUniversals<T>(Values: Pointer; const Helper: TSearchHelper; const Item: T): NativeInt; static;
    class function SearchDescendingUniversals<T>(Values: Pointer; const Helper: TSearchHelper; const Item: T): NativeInt; static;

    class function InternalSearch<T>(Values: Pointer; Index, Count: Integer; const Item: T;
      out FoundIndex: Integer): Boolean; overload; static; inline;
    class function InternalSearch<T>(Values: Pointer; Index, Count: Integer; const Item: T;
      out FoundIndex: Integer; Comparer: Pointer): Boolean; overload; static; inline;
    class function InternalSearchDescending<T>(Values: Pointer; Index, Count: Integer; const Item: T;
      out FoundIndex: Integer): Boolean; overload; static; inline;
    class function InternalSearchDescending<T>(Values: Pointer; Index, Count: Integer; const Item: T;
      out FoundIndex: Integer; Comparer: Pointer): Boolean; overload; static; inline;
  public
    class procedure Exchange<T>(const Left, Right: Pointer); static; inline;
    class procedure CopyMemory<T>(const Destination, Source: Pointer); static; inline;
    class procedure ZeroMemory<T>(const Values: Pointer); static; inline;
    class procedure Reverse<T>(const Values: Pointer; const Count: NativeInt); overload; static;
    class procedure Reverse<T>(var Values: array of T); overload; static;
    class procedure Copy<T>(const Source: array of T; var Destination: array of T; SourceIndex, DestIndex, Count: NativeInt); overload; static;
    class procedure Copy<T>(const Source: array of T; var Destination: array of T; Count: NativeInt); overload; static;

    class procedure Sort<T>(var Values: T; const Count: Integer); overload; static;
    class procedure Sort<T>(var Values: T; const Count: Integer; const Comparer: IComparer<T>); overload; static;
    class procedure Sort<T>(var Values: T; const Count: Integer; const Comparison: TComparison<T>); overload; static;
    class procedure Sort<T>(var Values: array of T); overload; static;
    class procedure Sort<T>(var Values: array of T; const Comparer: IComparer<T>); overload; static;
    class procedure Sort<T>(var Values: array of T; const Comparer: IComparer<T>; Index, Count: Integer); overload; static;
    class procedure Sort<T>(var Values: array of T; const Comparison: TComparison<T>); overload; static;
    class procedure Sort<T>(var Values: array of T; Index, Count: Integer; const Comparison: TComparison<T>); overload; static;

    class procedure SortDescending<T>(var Values: T; const Count: Integer); overload; static;
    class procedure SortDescending<T>(var Values: T; const Count: Integer; const Comparer: IComparer<T>); overload; static;
    class procedure SortDescending<T>(var Values: T; const Count: Integer; const Comparison: TComparison<T>); overload; static;
    class procedure SortDescending<T>(var Values: array of T); overload; static;
    class procedure SortDescending<T>(var Values: array of T; const Comparer: IComparer<T>); overload; static;
    class procedure SortDescending<T>(var Values: array of T; const Comparer: IComparer<T>; Index, Count: Integer); overload; static;
    class procedure SortDescending<T>(var Values: array of T; const Comparison: TComparison<T>); overload; static;
    class procedure SortDescending<T>(var Values: array of T; Index, Count: Integer; const Comparison: TComparison<T>); overload; static;

    class function BinarySearch<T>(var Values: T; const Item: T; out FoundIndex: Integer; Count: Integer): Boolean; overload; static;
    class function BinarySearch<T>(const Values: array of T; const Item: T; out FoundIndex: Integer): Boolean; overload; static;
    class function BinarySearch<T>(const Values: array of T; const Item: T; out FoundIndex: Integer;
      Index, Count: Integer): Boolean; overload; static;

    class function BinarySearch<T>(var Values: T; const Item: T; out FoundIndex: Integer; Count: Integer; const Comparer: IComparer<T>): Boolean; overload; static;
    class function BinarySearch<T>(const Values: array of T; const Item: T; out FoundIndex: Integer; const Comparer: IComparer<T>): Boolean; overload; static;
    class function BinarySearch<T>(const Values: array of T; const Item: T; out FoundIndex: Integer; const Comparer: IComparer<T>;
      Index, Count: Integer): Boolean; overload; static;
    class function BinarySearch<T>(var Values: T; const Item: T; out FoundIndex: Integer; Count: Integer; const Comparison: TComparison<T>): Boolean; overload; static;
    class function BinarySearch<T>(const Values: array of T; const Item: T; out FoundIndex: Integer; const Comparison: TComparison<T>): Boolean; overload; static;
    class function BinarySearch<T>(const Values: array of T; const Item: T; out FoundIndex: Integer;
      Index, Count: Integer; const Comparison: TComparison<T>): Boolean; overload; static;

    class function BinarySearchDescending<T>(var Values: T; const Item: T; out FoundIndex: Integer; Count: Integer): Boolean; overload; static;
    class function BinarySearchDescending<T>(const Values: array of T; const Item: T; out FoundIndex: Integer): Boolean; overload; static;
    class function BinarySearchDescending<T>(const Values: array of T; const Item: T; out FoundIndex: Integer;
      Index, Count: Integer): Boolean; overload; static;

    class function BinarySearchDescending<T>(var Values: T; const Item: T; out FoundIndex: Integer; Count: Integer; const Comparer: IComparer<T>): Boolean; overload; static;
    class function BinarySearchDescending<T>(const Values: array of T; const Item: T; out FoundIndex: Integer; const Comparer: IComparer<T>): Boolean; overload; static;
    class function BinarySearchDescending<T>(const Values: array of T; const Item: T; out FoundIndex: Integer; const Comparer: IComparer<T>;
      Index, Count: Integer): Boolean; overload; static;
    class function BinarySearchDescending<T>(var Values: T; const Item: T; out FoundIndex: Integer; Count: Integer; const Comparison: TComparison<T>): Boolean; overload; static;
    class function BinarySearchDescending<T>(const Values: array of T; const Item: T; out FoundIndex: Integer; const Comparison: TComparison<T>): Boolean; overload; static;
    class function BinarySearchDescending<T>(const Values: array of T; const Item: T; out FoundIndex: Integer;
      Index, Count: Integer; const Comparison: TComparison<T>): Boolean; overload; static;
  end;


{ TCustomList<T> class
  Basic class for TList<T>, TQueue<T>, TStack<T> }

  TCustomList<T> = class(TEnumerable<T>)
  public type
    TItem = T;
    PItem = ^TItem;
    TItemList = array[0..0] of TItem;
    PItemList = ^TItemList;

    TEnumerator = class(TEnumerator<T>)
    private
      FList: TCustomList<T>;
      FIndex: NativeInt;
      function GetCurrent: T; inline;
    protected
      function DoGetCurrent: T; override;
      function DoMoveNext: Boolean; override;
    public
      constructor Create(const AList: TCustomList<T>);
      property Current: T read GetCurrent;
      function MoveNext: Boolean; inline;
    end;
  protected
    FItems: PItemList;
    FCapacity, FCount: packed record
    case Boolean of
     False: (Int: Integer);
      True: (Native: NativeInt);
    end;
    FTail: NativeInt;
    FHead: NativeInt;
    FOnNotify: TCollectionNotifyEvent<T>;
    FInternalNotify: TCollectionNotifyEvent<T>;

    class procedure ClearMethod(var Method); static; inline;
    class function EmptyException: Exception; static;
    procedure SetCapacity(Value: Integer);
    procedure Grow;
    procedure GrowTo(Value: Integer);
    function DoGetEnumerator: TEnumerator<T>; override;
    procedure SetOnNotify(const Value: TCollectionNotifyEvent<T>);
    procedure Notify(const Item: T; Action: TCollectionNotification); virtual;
    procedure NotifyCaller(Sender: TObject; const Item: T; Action: TCollectionNotification);
    procedure SetNotifyMethods; virtual;

    property List: PItemList read FItems;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure TrimExcess; inline;
    function ToArray: TArray<T>; override; final;
    function GetEnumerator: TEnumerator; reintroduce; inline;

    property Count: Integer read FCount.Int;
    property Capacity: Integer read FCapacity.Int write SetCapacity;
    property OnNotify: TCollectionNotifyEvent<T> read FOnNotify write SetOnNotify;
  end;


{ TList<T>
  System.Generics.Collections equivalent }

  TList<T> = class(TCustomList<T>)
  private type
    TCompare = function(Inst: Pointer; const Left, Right: T): Integer;
    TEquals = function(Inst: Pointer; const Left, Right: T): Boolean;
    TInternalStored = packed record
      Self: Pointer;
      InternalNotify: TMethod;
      Count: NativeInt;
      Item: PItem;
      ACount: Integer;
    end;
    TComparerInst = packed record
      Vtable: Pointer;
      Method: TMethod;
      QueryInterface,
      AddRef,
      Release,
      Call: Pointer;
    end;
  protected
    FComparer: IComparer<T>;

    function GetItem(Index: Integer): T; inline;
    procedure SetItem(Index: Integer; const Value: T); inline;
    function ItemValue(const Item: T): NativeInt;
    function InternalInsert(Index: NativeInt; const Value: T): Integer;
    procedure InternalDelete(Index: NativeInt; Action: TCollectionNotification);
    procedure SetCount(Value: Integer);
    procedure InternalMove(CurIndex, NewIndex: Integer);
    procedure InternalMove40(CurIndex, NewIndex: Integer);
    function InternalIndexOf(const Value: T): NativeInt; overload;
    function InternalIndexOf(const Value: T; const Comparer: IComparer<T>): NativeInt; overload;
    function InternalIndexOfRev(const Value: T): NativeInt; overload;
    function InternalIndexOfRev(const Value: T; const Comparer: IComparer<T>): NativeInt; overload;
    class function InternalIsEmpty(const Inst: TComparerInst; const Left, Right: T): Boolean; static;
    {$ifdef SMARTGENERICS}
    procedure InternalPackDifficults;
    {$endif}
  public
    type
      TDirection = System.Types.TDirection;
      TEmptyFunc = reference to function (const L, R: T): Boolean;
      TListCompareFunc = reference to function (const L, R: T): Integer;

    constructor Create; overload;
    constructor Create(const AComparer: IComparer<T>); overload;
    constructor Create(const Collection: TEnumerable<T>); overload;

    class procedure Error(const Msg: string; Data: NativeInt); overload; virtual;
    {$ifNdef NEXTGEN}
    class procedure Error(Msg: PResStringRec; Data: NativeInt); overload;
    {$endif}

    function First: T; inline;
    function Last: T; inline;

    function Add(const Value: T): Integer; inline;
    procedure AddRange(const Values: array of T); overload;
    procedure AddRange(const Collection: IEnumerable<T>); overload;
    procedure AddRange(const Collection: TEnumerable<T>); overload;

    procedure Insert(Index: Integer; const Value: T); inline;
    procedure InsertRange(Index: Integer; const Values: array of T); overload;
    procedure InsertRange(Index: Integer; const Collection: IEnumerable<T>); overload;
    procedure InsertRange(Index: Integer; const Collection: TEnumerable<T>); overload;

    procedure Delete(Index: Integer); inline;
    procedure DeleteRange(AIndex, ACount: Integer);

    function Expand: TList<T>; inline;
    procedure Exchange(Index1, Index2: Integer);
    procedure Move(CurIndex, NewIndex: Integer); inline;
    procedure Reverse; 

    procedure Sort; overload;
    procedure Sort(const AComparer: IComparer<T>); overload;
    procedure Sort(const AComparison: TComparison<T>); overload;
    procedure Sort(Index, Count: Integer); overload;
    procedure Sort(Index, Count: Integer; const AComparer: IComparer<T>); overload;
    procedure Sort(Index, Count: Integer; const AComparison: TComparison<T>); overload;
    procedure SortDescending; overload;
    procedure SortDescending(const AComparer: IComparer<T>); overload;
    procedure SortDescending(const AComparison: TComparison<T>); overload;
    procedure SortDescending(Index, Count: Integer); overload;
    procedure SortDescending(Index, Count: Integer; const AComparer: IComparer<T>); overload;
    procedure SortDescending(Index, Count: Integer; const AComparison: TComparison<T>); overload;

    function BinarySearch(const Item: T; out FoundIndex: Integer): Boolean; overload;
    function BinarySearch(const Item: T; out FoundIndex: Integer; const AComparer: IComparer<T>): Boolean; overload;
    function BinarySearch(const Item: T; out FoundIndex: Integer; const AComparison: TComparison<T>): Boolean; overload;
    function BinarySearch(const Item: T; out FoundIndex: Integer; Index, Count: Integer): Boolean; overload;
    function BinarySearch(const Item: T; out FoundIndex: Integer; const AComparer: IComparer<T>; Index, Count: Integer): Boolean; overload;
    function BinarySearch(const Item: T; out FoundIndex: Integer; Index, Count: Integer; const AComparison: TComparison<T>): Boolean; overload;
    function BinarySearchDescending(const Item: T; out FoundIndex: Integer): Boolean; overload;
    function BinarySearchDescending(const Item: T; out FoundIndex: Integer; const AComparer: IComparer<T>): Boolean; overload;
    function BinarySearchDescending(const Item: T; out FoundIndex: Integer; const AComparison: TComparison<T>): Boolean; overload;
    function BinarySearchDescending(const Item: T; out FoundIndex: Integer; Index, Count: Integer): Boolean; overload;
    function BinarySearchDescending(const Item: T; out FoundIndex: Integer; const AComparer: IComparer<T>; Index, Count: Integer): Boolean; overload;
    function BinarySearchDescending(const Item: T; out FoundIndex: Integer; Index, Count: Integer; const AComparison: TComparison<T>): Boolean; overload;

    function Contains(const Value: T): Boolean; inline;
    function IndexOf(const Value: T): Integer; inline;
    function IndexOfItem(const Value: T; Direction: TDirection): Integer; inline;
    function LastIndexOf(const Value: T): Integer; inline;
    function Remove(const Value: T): Integer;
    function RemoveItem(const Value: T; Direction: TDirection): Integer;
    function Extract(const Value: T): T;
    function ExtractItem(const Value: T; Direction: TDirection): T;

    procedure Pack; overload;
    procedure Pack(const IsEmpty: TEmptyFunc); overload;

    property Count: Integer read FCount.Int write SetCount;
    property Items[Index: Integer]: T read GetItem write SetItem; default;
    property List;
  end;


{ TStack<T>
  System.Generics.Collections equivalent }

  TStack<T> = class(TCustomList<T>)
  protected
    procedure InternalPush(const Value: T);
    function InternalPop(const Action: TCollectionNotification): T;
  public
    constructor Create; overload;
    constructor Create(const Collection: TEnumerable<T>); overload;

    procedure Push(const Value: T); inline;
    function Pop: T; inline;
    function Extract: T; inline;
    function Peek: T; inline;
  end;

{ TQueue<T>
  System.Generics.Collections equivalent }

  TQueue<T> = class(TCustomList<T>)
  protected
    procedure InternalEnqueue(const Value: T);
    function InternalDequeue(const Action: TCollectionNotification): T;
  public
    constructor Create; overload;
    constructor Create(const Collection: TEnumerable<T>); overload;

    procedure Enqueue(const Value: T); inline;
    function Dequeue: T; inline;
    function Extract: T; inline;
    function Peek: T; inline;
  end;



  TThreadList<T> = class
  private
    FList: TList<T>;
    FLock: TObject;
    FDuplicates: TDuplicates;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(const Item: T);
    procedure Clear;
    function LockList: TList<T>;
    procedure Remove(const Item: T); inline;
    procedure RemoveItem(const Item: T; Direction: TDirection);
    procedure UnlockList; inline;
    property Duplicates: TDuplicates read FDuplicates write FDuplicates;
  end;

  TThreadedQueue<T> = class
  private
    FQueue: array of T;
    FQueueSize, FQueueOffset: Integer;
    FQueueNotEmpty,
    FQueueNotFull: TObject;
    FQueueLock: TObject;
    FShutDown: Boolean;
    FPushTimeout, FPopTimeout: LongWord;
    FTotalItemsPushed, FTotalItemsPopped: LongWord;
  public
    constructor Create(AQueueDepth: Integer = 10; PushTimeout: LongWord = INFINITE; PopTimeout: LongWord = INFINITE);
    destructor Destroy; override;

    procedure Grow(ADelta: Integer);
    function PushItem(const AItem: T): TWaitResult; overload;
    function PushItem(const AItem: T; var AQueueSize: Integer): TWaitResult; overload;
    function PopItem: T; overload;
    function PopItem(var AQueueSize: Integer): T; overload;
    function PopItem(var AQueueSize: Integer; var AItem: T): TWaitResult; overload;
    function PopItem(var AItem: T): TWaitResult; overload;
    procedure DoShutDown;

    property QueueSize: Integer read FQueueSize;
    property ShutDown: Boolean read FShutDown;
    property TotalItemsPushed: LongWord read FTotalItemsPushed;
    property TotalItemsPopped: LongWord read FTotalItemsPopped;
  end;

  TObjectList<T: class> = class(TList<T>)
  protected
    FOwnsObjects: Boolean;
    procedure SetOwnsObjects(const Value: Boolean);
    procedure DisposeNotifyEvent(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
    procedure DisposeOnly(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
    procedure SetNotifyMethods; override;
  public
    constructor Create(AOwnsObjects: Boolean = True); overload;
    constructor Create(const AComparer: IComparer<T>; AOwnsObjects: Boolean = True); overload;
    constructor Create(const Collection: TEnumerable<T>; AOwnsObjects: Boolean = True); overload;
    property OwnsObjects: Boolean read FOwnsObjects write SetOwnsObjects;
  end;

  TObjectStack<T: class> = class(TStack<T>)
  protected
    FOwnsObjects: Boolean;
    procedure SetOwnsObjects(const Value: Boolean);
    procedure DisposeNotifyEvent(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
    procedure DisposeOnly(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
    procedure SetNotifyMethods; override;
  public
    constructor Create(AOwnsObjects: Boolean = True); overload;
    constructor Create(const Collection: TEnumerable<T>; AOwnsObjects: Boolean = True); overload;
    procedure Pop;
    property OwnsObjects: Boolean read FOwnsObjects write SetOwnsObjects;
  end;

  TObjectQueue<T: class> = class(TQueue<T>)
  protected
    FOwnsObjects: Boolean;
    procedure SetOwnsObjects(const Value: Boolean);
    procedure DisposeNotifyEvent(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
    procedure DisposeOnly(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
    procedure SetNotifyMethods; override;
  public
    constructor Create(AOwnsObjects: Boolean = True); overload;
    constructor Create(const Collection: TEnumerable<T>; AOwnsObjects: Boolean = True); overload;
    procedure Dequeue;
    property OwnsObjects: Boolean read FOwnsObjects write SetOwnsObjects;
  end;

  TDictionaryOwnerships = set of (doOwnsKeys, doOwnsValues);

  TObjectDictionary<TKey,TValue> = class(TDictionary<TKey,TValue>)
  public type
    TItem = TCustomDictionary<TKey,TValue>.TItem;
  protected
    FOwnerships: TDictionaryOwnerships;
    procedure DisposeKeyEvent(Sender: TObject; const Key: TObject; Action: TCollectionNotification);
    procedure DisposeKeyOnly(Sender: TObject; const Key: TObject; Action: TCollectionNotification);
    procedure DisposeValueEvent(Sender: TObject; const Value: TObject; Action: TCollectionNotification);
    procedure DisposeValueOnly(Sender: TObject; const Value: TObject; Action: TCollectionNotification);
    procedure DisposeItemNotifyKeyCaller(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyKeyEvent(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyKeyOnly(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyValueCaller(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyValueEvent(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyValueOnly(const Item: TItem; Action: TCollectionNotification);
    procedure SetNotifyMethods; override;
  public
    constructor Create(Ownerships: TDictionaryOwnerships; ACapacity: Integer = 0); overload;
    constructor Create(Ownerships: TDictionaryOwnerships;
      const AComparer: IEqualityComparer<TKey>); overload;
    constructor Create(Ownerships: TDictionaryOwnerships; ACapacity: Integer;
      const AComparer: IEqualityComparer<TKey>); overload;
  end;

  TRapidObjectDictionary<TKey,TValue> = class(TRapidDictionary<TKey,TValue>)
  public type
    TItem = TCustomDictionary<TKey,TValue>.TItem;
  protected
    FOwnerships: TDictionaryOwnerships;
    procedure DisposeKeyEvent(Sender: TObject; const Key: TObject; Action: TCollectionNotification);
    procedure DisposeKeyOnly(Sender: TObject; const Key: TObject; Action: TCollectionNotification);
    procedure DisposeValueEvent(Sender: TObject; const Value: TObject; Action: TCollectionNotification);
    procedure DisposeValueOnly(Sender: TObject; const Value: TObject; Action: TCollectionNotification);
    procedure DisposeItemNotifyKeyCaller(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyKeyEvent(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyKeyOnly(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyValueCaller(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyValueEvent(const Item: TItem; Action: TCollectionNotification);
    procedure DisposeItemNotifyValueOnly(const Item: TItem; Action: TCollectionNotification);
    procedure SetNotifyMethods; override;
  public
    constructor Create(Ownerships: TDictionaryOwnerships; ACapacity: Integer = 0);
  end;


implementation

// x86 architecture compatibility (Word mode)
{$ifNdef CPUX86}
function Swap(const X: NativeUInt): NativeUInt; inline;
begin
  Result := (Byte(X) shl 8) + Byte(X shr 8);
end;
{$endif}


{ TRAIIHelper.TClearNatives }

procedure TRAIIHelper.TClearNatives.Clear;
begin
  Items := nil;
  Count := 0;
end;

procedure TRAIIHelper.TClearNatives.Add(AOffset: NativeInt; ADynTypeInfo: PTypeInfo;
  AClearNativeProc: TClearNativeProc);
begin
  if (Count = 0) then
  begin
    ItemSingle.ClearNativeProc := AClearNativeProc;
    ItemSingle.Offset := AOffset;
    ItemSingle.DynTypeInfo := ADynTypeInfo;
  end;

  Inc(Count);
  SetLength(Items, Count);
  with Items[Count - 1] do
  begin
    ClearNativeProc := AClearNativeProc;
    Offset := AOffset;
    DynTypeInfo := ADynTypeInfo;
  end;
end;

{$ifdef WEAKREF}
{ TRAIIHelper.TInitNatives }

procedure TRAIIHelper.TInitNatives.Clear;
begin
  Items := nil;
  Count := 0;
end;

procedure TRAIIHelper.TInitNatives.Add(AOffset: NativeInt);
begin
  if (Count = 0) then
  begin
    ItemSingle.Offset := AOffset;
  end;

  Inc(Count);
  SetLength(Items, Count);
  Items[Count - 1].Offset := AOffset;
end;
{$endif}

{ TRAIIHelper.TStaticArrays }

procedure TRAIIHelper.TStaticArrays.Clear;
begin
  Items := nil;
  Count := 0;
end;

procedure TRAIIHelper.TStaticArrays.Add(AOffset: NativeInt; AStaticTypeInfo: PTypeInfo; ACount: NativeUInt);
begin
  Inc(Count);
  SetLength(Items, Count);
  with Items[Count - 1] do
  begin
    Offset := AOffset;
    StaticTypeInfo := AStaticTypeInfo;
    Count := ACount;
  end;
end;

{ TRAIIHelper }

class procedure TRAIIHelper.ULStrClear(P: Pointer);
type
  PStrRec = ^StrRec;
  StrRec = packed record
    {$ifdef LARGEINT}
    _Padding: Integer;
    {$endif}
    codePage: Word;
    elemSize: Word;
    refCnt: Integer;
    length: Integer;
  end;
var
  Rec: PStrRec;
  RefCnt: Integer;
begin
  Rec := PPointer(P)^;
  Dec(Rec);
  RefCnt := Rec.refCnt;
  if (RefCnt > 0) then
  begin
    if (RefCnt = 1) or (AtomicDecrement(Rec.refCnt) = 0) then
      FreeMem(Rec);
  end;
end;

{$ifdef MSWINDOWS}
procedure SysFreeString(P: Pointer); stdcall; external 'oleaut32.dll';

class procedure TRAIIHelper.WStrClear(P: Pointer);
begin
  SysFreeString(PPointer(P)^);
end;
{$endif}

class procedure TRAIIHelper.IntfClear(P: Pointer);
begin
  IInterface(PPointer(P)^)._Release;
end;

class procedure TRAIIHelper.VarClear(P: Pointer);
var
  VType: Integer;
begin
  VType := PVarData(P).VType;
  if (VType and varDeepData <> 0) then
  case VType of
    varBoolean, varUnknown+1..varUInt64: ;
  else
    System.VarClear(PVariant(P)^);
  end;
end;

class procedure TRAIIHelper.DynArrayClear(P, TypeInfo: Pointer);
type
  PDynArrayRec = ^TDynArrayRec;
  TDynArrayRec = packed record
    {$ifdef LARGEINT}
    _Padding: Integer;
    {$endif}
    RefCnt: Integer;
    Length: NativeInt;
  end;
var
  Rec: PDynArrayRec;
  RefCnt: Integer;
begin
  Rec := PPointer(P)^;
  Dec(Rec);
  RefCnt := Rec.RefCnt;
  if (RefCnt > 0) then
  begin
    if (RefCnt = 1) or (AtomicDecrement(Rec.RefCnt) = 0) then
    begin
      Inc(PByte(TypeInfo), PDynArrayTypeInfo(TypeInfo).name);
      TypeInfo := PDynArrayTypeInfo(TypeInfo).elType;
      if (TypeInfo <> nil) and (Rec.Length <> 0) then
      begin
        System.FinalizeArray(Pointer(NativeUInt(Rec) + SizeOf(TDynArrayRec)),
          PPointer(TypeInfo)^, Rec.Length);
      end;

      FreeMem(Rec);
    end;
  end;
end;

{$ifdef AUTOREFCOUNT}
class procedure TRAIIHelper.RefObjClear(P: Pointer);
var
  L: TObject;
begin
  L := PPointer(P)^;
  PPointer(P)^ := nil;
  L.__ObjRelease;
end;
{$endif}

{$ifdef WEAKINSTREF}
class procedure TRAIIHelper.WeakObjClear(P: Pointer);
var
  L: TObject;
begin
  L := PPointer(P)^;
  PPointer(P)^ := nil;
  UnregisterWeakRef(P, L);
end;
{$endif}

{$ifdef WEAKREF}
class procedure TRAIIHelper.WeakMethodClear(P: Pointer);
var
  L: TObject;
begin
  L := PPointer(P)^;
  PPointer(P)^ := nil;
  Dec(NativeUInt(P), SizeOf(Pointer));
  PPointer(P)^ := nil;
  UnregisterWeakMethodRef(P, L);
end;
{$endif}

{$ifdef WEAKINTFREF}
class procedure TRAIIHelper.WeakIntfClear(P: Pointer);
var
  L: TObject;
begin
  L := IInterface(PPointer(P)^) as TObject;
  PPointer(P)^ := nil;
  UnregisterWeakRef(P, L);
end;
{$endif}

procedure TRAIIHelper.Include(AOffset: NativeInt; Value: PTypeInfo);
var
  i: Cardinal;
  {$ifdef WEAKREF}
  Weak: Boolean;
  {$endif}
  FieldTable: PFieldTable;
  ChildSize, ChildOffset: NativeInt;
begin
  case Value.Kind of
    tkVariant:
    begin
      {$ifdef WEAKREF}
        Self.InitNatives.Add(AOffset);
        Self.ClearNatives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.VarClear));
      {$else}
        Self.Natives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.VarClear));
      {$endif}
    end;
    {$ifdef AUTOREFCOUNT}
    tkClass:
    begin
      {$ifdef WEAKREF}
        Self.InitNatives.Add(AOffset);
        Self.ClearNatives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.RefObjClear));
      {$else}
        Self.Natives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.RefObjClear));
      {$endif}
    end;
    {$endif}
    {$ifdef WEAKREF}
    tkMethod:
    begin
      Self.InitNatives.Add(AOffset);
      Self.InitNatives.Add(AOffset + SizeOf(Pointer));

      Self.ClearNatives.Add(AOffset + SizeOf(Pointer), nil, TClearNativeProc(@TRAIIHelper.WeakMethodClear));
    end;
    {$endif}
    {$ifdef MSWINDOWS}
    tkWString:
    begin
      {$ifdef WEAKREF}
        Self.InitNatives.Add(AOffset);
        Self.ClearNatives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.WStrClear));
      {$else}
        Self.Natives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.WStrClear));
      {$endif}
    end;
    {$else}
    tkWString,
    {$endif}
    tkLString, tkUString:
    begin
      {$ifdef WEAKREF}
        Self.InitNatives.Add(AOffset);
        Self.ClearNatives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.ULStrClear));
      {$else}
        Self.Natives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.ULStrClear));
      {$endif}
    end;
    tkInterface:
    begin
      {$ifdef WEAKREF}
        Self.InitNatives.Add(AOffset);
        Self.ClearNatives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.IntfClear));
      {$else}
        Self.Natives.Add(AOffset, nil, TClearNativeProc(@TRAIIHelper.IntfClear));
      {$endif}
    end;
    tkDynArray:
    begin
      {$ifdef WEAKREF}
        Self.InitNatives.Add(AOffset);
        Self.ClearNatives.Add(AOffset, Value, TClearNativeProc(@TRAIIHelper.DynArrayClear));
      {$else}
        Self.Natives.Add(AOffset, Value, TClearNativeProc(@TRAIIHelper.DynArrayClear));
      {$endif}
    end;
    tkArray{static array}:
    begin
      FieldTable := PFieldTable(NativeUInt(Value) + PByte(@Value.Name)^);
      if (FieldTable.Fields[0].TypeInfo <> nil) and IsManagedTypeInfo(FieldTable.Fields[0].TypeInfo^) then
      begin
        if (FieldTable.Count > 4) then
        begin
          Self.StaticArrays.Add(AOffset, FieldTable.Fields[0].TypeInfo^, FieldTable.Count);
        end else
        begin
          ChildSize := FieldTable.Size div FieldTable.Count;
          for i := 1 to FieldTable.Count do
          begin
            Self.Include(AOffset, FieldTable.Fields[0].TypeInfo^);
            Inc(AOffset, ChildSize);
          end;
        end;
      end;
    end;
    tkRecord:
    begin
      FieldTable := PFieldTable(NativeUInt(Value) + PByte(@Value.Name)^);
      if FieldTable.Count > 0 then
      begin
        {$ifdef WEAKREF}
        Weak := False;
        {$endif}
        for i := 0 to FieldTable.Count - 1 do
        begin
          ChildOffset := AOffset + NativeInt(FieldTable.Fields[i].Offset);
        {$ifdef WEAKREF}
          if FieldTable.Fields[i].TypeInfo = nil then
          begin
            Weak := True;
            Continue;
          end;
          if not Weak then
          begin
          {$endif}
            Self.Include(ChildOffset, FieldTable.Fields[i].TypeInfo^);
          {$ifdef WEAKREF}
          end else
          begin
            case FieldTable.Fields[i].TypeInfo^.Kind of
            {$ifdef WEAKINTFREF}
              tkInterface:
              begin
                {$ifdef WEAKREF}
                  Self.InitNatives.Add(ChildOffset);
                  Self.ClearNatives.Add(ChildOffset, nil, TClearProc(@TRAIIHelper.WeakIntfClear));
                {$else}
                  Self.Natives.Add(ChildOffset, nil, TClearProc(@TRAIIHelper.WeakIntfClear));
                {$endif}
              end;
            {$endif}
            {$ifdef WEAKINSTREF}
              tkClass:
              begin
                {$ifdef WEAKREF}
                  Self.InitNatives.Add(ChildOffset);
                  Self.ClearNatives.Add(ChildOffset, nil, TClearProc(@TRAIIHelper.WeakObjClear));
                {$else}
                  Self.Natives.Add(ChildOffset, nil, TClearProc(@TRAIIHelper.WeakObjClear));
                {$endif}
              end;
            {$endif}
            {$ifdef WEAKREF}
              tkMethod:
              begin
                Self.InitNatives.Add(ChildOffset);
                Self.InitNatives.Add(ChildOffset + SizeOf(Pointer));

                Self.ClearNatives.Add(ChildOffset + SizeOf(Pointer), nil, TClearProc(@TRAIIHelper.WeakMethodClear));
              end;
            {$endif}
            end;
          end;
        {$endif}
        end;
      end;
    end;
  end;
end;

procedure TRAIIHelper.Initialize(Value: PTypeInfo);
var
  FieldTable: PFieldTable;
begin
  // clear
  FTypeInfo := Value;
  {$ifdef WEAKREF}
  InitNatives.Clear;
  ClearNatives.Clear;
  {$else}
  Natives.Clear;
  {$endif}
  StaticArrays.Clear;
  InitProc := nil;
  ClearProc := nil;

  // type kind
  case Value.Kind of
    tkInteger, tkChar, tkEnumeration, tkWChar:
    begin
      case (Value.TypeData.OrdType) of
        otSByte, otUByte: FSize := SizeOf(Byte);
        otSWord, otUWord: FSize := SizeOf(Word);
        otSLong, otULong: FSize := SizeOf(Cardinal);
      end;
      Exit;
    end;
    tkSet:
    begin
      with Value.TypeData.CompType^.TypeData^ do
      begin
        FSize := (((MaxValue + 7 + 1) and ($FF shl 3)) - (MinValue and ($FF shl 3))) shr 3;
        if (FSize = 3) then FSize := 4;
      end;
      Exit;
    end;
    tkFloat:
    begin
      case (Value.TypeData.FloatType) of
        ftSingle: FSize := SizeOf(Single);
        ftDouble: FSize := SizeOf(Double);
      ftExtended: FSize := SizeOf(Extended);
          ftComp: FSize := SizeOf(Comp);
          ftCurr: FSize := SizeOf(Currency);
      end;
      Exit;
    end;
    tkInt64:
    begin
      FSize := SizeOf(Int64);
      Exit;
    end;
    {$ifNdef NEXTGEN}
    tkString:
    begin
      FSize := Value.TypeData.MaxLength + 1;
      Exit;
    end;
    {$endif}
    tkClassRef, tkPointer, tkProcedure:
    begin
      FSize := SizeOf(Pointer);
      Exit;
    end;

    tkVariant:
    begin
      FSize := SizeOf(Variant);
      Self.Include(0, Value);
    end;
    tkMethod:
    begin
      FSize := SizeOf(TMethod);
      Self.Include(0, Value);
    end;
    tkClass, tkLString, tkWString, tkInterface, tkDynArray, tkUString:
    begin
      FSize := SizeOf(Pointer);
      Self.Include(0, Value);
    end;
    tkArray:
    begin
      FieldTable := PFieldTable(NativeUInt(Value) + PByte(@Value.Name)^);
      FSize := NativeInt(FieldTable.Size) * NativeInt(FieldTable.Count);
      Self.Include(0, Value);
    end;
    tkRecord:
    begin
      FieldTable := PFieldTable(NativeUInt(Value) + PByte(@Value.Name)^);
      FSize := FieldTable.Size;
      Self.Include(0, Value);
    end;
  else
    System.Error(reInvalidPtr);
  end;

  // initialization
  if (StaticArrays.Count <> 0) then
  begin
    InitProc := Self.InitsProc;
  end else
  if ({$ifdef WEAKREF}InitNatives{$else}Natives{$endif}.Count <> 0) then
  begin
    if ({$ifdef WEAKREF}InitNatives{$else}Natives{$endif}.Count > 1) then
    begin
      InitProc := Self.InitsProcNatives;
    end else
    begin
      InitProc := Self.InitsProcNativeSingle;
    end;
  end;

  // finalization
  if (StaticArrays.Count <> 0) then
  begin
    ClearProc := Self.ClearsProc;
  end else
  if ({$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.Count <> 0) then
  begin
    if ({$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.Count > 1) then
    begin
      ClearProc := Self.ClearsProcNatives;
    end else
    begin
      ClearProc := Self.ClearsProcNativeSingle;
    end;
  end;
end;

class function TRAIIHelper.IsManagedTypeInfo(Value: PTypeInfo): Boolean;
var
  i: Cardinal;
  {$ifdef WEAKREF}
  Weak: Boolean;
  {$endif}
  FieldTable: PFieldTable;
begin
  Result := False;

  if Assigned(Value) then
  case Value.Kind of
    tkVariant:
    begin
      Exit(True);
    end;
    {$ifdef AUTOREFCOUNT}
    tkClass:
    begin
      Exit(True);
    end;
    {$endif}
    {$ifdef WEAKREF}
    tkMethod:
    begin
      Exit(True);
    end;
    {$endif}
    tkWString, tkLString, tkUString, tkInterface, tkDynArray:
    begin
      Exit(True);
    end;
    tkArray{static array}:
    begin
      FieldTable := PFieldTable(NativeUInt(Value) + PByte(@Value.Name)^);
      if (FieldTable.Fields[0].TypeInfo <> nil) then
        Result := IsManagedTypeInfo(FieldTable.Fields[0].TypeInfo^);
    end;
    tkRecord:
    begin
      FieldTable := PFieldTable(NativeUInt(Value) + PByte(@Value.Name)^);
      if FieldTable.Count > 0 then
      begin
        {$ifdef WEAKREF}
        Weak := False;
        {$endif}
        for i := 0 to FieldTable.Count - 1 do
        begin
         {$ifdef WEAKREF}
          if FieldTable.Fields[i].TypeInfo = nil then
          begin
            Weak := True;
            Continue;
          end;
          if not Weak then
          begin
          {$endif}
            if (IsManagedTypeInfo(FieldTable.Fields[i].TypeInfo^)) then
              Exit(True);
          {$ifdef WEAKREF}
          end else
          begin
            case FieldTable.Fields[i].TypeInfo^.Kind of
            {$ifdef WEAKINTFREF}
              tkInterface:
              begin
                Exit(True);
              end;
            {$endif}
            {$ifdef WEAKINSTREF}
              tkClass:
              begin
                Exit(True);
              end;
            {$endif}
            {$ifdef WEAKREF}
              tkMethod:
              begin
                Exit(True);
              end;
            {$endif}
            end;
          end;
        {$endif}
        end;
      end;
    end;
  end;
end;

class procedure TRAIIHelper.InitsProcNativeSingle(const Self: TRAIIHelper; P: Pointer);
begin
  Inc(NativeInt(P), Self.{$ifdef WEAKREF}InitNatives{$else}Natives{$endif}.ItemSingle.Offset);
  PNativeInt(P)^ := 0;
end;

class procedure TRAIIHelper.InitsProcNatives(const Self: TRAIIHelper; P: Pointer);
label
  _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11;
var
  Count, Null: NativeInt;
  Item: ^{$ifdef WEAKREF}TInitNativeRec{$else}TNativeRec{$endif};
begin
  Item := Pointer(Self.{$ifdef WEAKREF}InitNatives{$else}Natives{$endif}.Items);
  Count := Self.{$ifdef WEAKREF}InitNatives{$else}Natives{$endif}.Count;
  Null := 0;

  case Count of
    11:
    begin
    _11:
      repeat
        PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
        Dec(Count);
        Inc(Item);
      until (Count = 0);
    end;
    10:
    begin
    _10:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _9;
    end;
    9:
    begin
    _9:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _8;
    end;
    8:
    begin
    _8:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _7;
    end;
    7:
    begin
    _7:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _6;
    end;
    6:
    begin
    _6:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _5;
    end;
    5:
    begin
    _5:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _4;
    end;
    4:
    begin
    _4:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _3;
    end;
    3:
    begin
    _3:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _2;
    end;
    2:
    begin
    _2:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
      Inc(Item);
      goto _1;
    end;
    1:
    begin
    _1:
      PNativeInt(NativeInt(P) + Item.Offset)^ := Null;
    end;
    0: ;
  else
    goto _11;
  end;
end;

class procedure TRAIIHelper.InitsProc(const Self: TRAIIHelper; P: Pointer);
var
  i: NativeInt;
  Item: ^{$ifdef WEAKREF}TInitNativeRec{$else}TNativeRec{$endif};
  StaticArrayRec: ^TStaticArrayRec;
begin
  Item := Pointer(Self.{$ifdef WEAKREF}InitNatives{$else}Natives{$endif}.Items);
  for i := 1 to Self.{$ifdef WEAKREF}InitNatives{$else}Natives{$endif}.Count do
  begin
    PNativeInt(NativeInt(P) + Item.Offset)^ := 0;
    Inc(Item);
  end;

  StaticArrayRec := Pointer(Self.StaticArrays.Items);
  for i := 1 to Self.StaticArrays.Count do
  begin
    System.InitializeArray(Pointer(NativeInt(P) + StaticArrayRec.Offset),
      StaticArrayRec.StaticTypeInfo, StaticArrayRec.Count);
    Inc(StaticArrayRec);
  end;
end;

class procedure TRAIIHelper.ClearsProcNativeSingle(const Self: TRAIIHelper; P: Pointer);
var
  Value: PNativeInt;
begin
  Value := Pointer(NativeInt(P) + Self.{$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.ItemSingle.Offset);
  if (Value^ <> 0) then
  begin
    Self.{$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.ItemSingle.ClearNativeProc(Value,
      Self.{$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.ItemSingle.DynTypeInfo);
  end;
end;

class procedure TRAIIHelper.ClearsProcNatives(const Self: TRAIIHelper; P: Pointer);
label
  _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11;
var
  Count: NativeInt;
  Value: PNativeInt;
  Item: ^{$ifdef WEAKREF}TClearNativeRec{$else}TNativeRec{$endif};
begin
  Item := Pointer(Self.{$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.Items);
  Count := Self.{$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.Count;

  case Count of
    11:
    begin
    _11:
      repeat
        Value := Pointer(NativeInt(P) + Item.Offset);
        if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
        Dec(Count);
        Inc(Item);
      until (Count = 0);
    end;
    10:
    begin
    _10:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _9;
    end;
    9:
    begin
    _9:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _8;
    end;
    8:
    begin
    _8:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _7;
    end;
    7:
    begin
    _7:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _6;
    end;
    6:
    begin
    _6:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _5;
    end;
    5:
    begin
    _5:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _4;
    end;
    4:
    begin
    _4:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _3;
    end;
    3:
    begin
    _3:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _2;
    end;
    2:
    begin
    _2:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
      Inc(Item);
      goto _1;
    end;
    1:
    begin
    _1:
      Value := Pointer(NativeInt(P) + Item.Offset);
      if (Value^ <> 0) then Item.ClearNativeProc(Value, Item.DynTypeInfo);
    end;
    0: ;
  else
    goto _11;
  end;
end;

class procedure TRAIIHelper.ClearsProc(const Self: TRAIIHelper; P: Pointer);
var
  i: NativeInt;
  Value: PNativeInt;
  Item: ^TClearNativeRec;
  StaticArrayRec: ^TStaticArrayRec;
begin
  Item := Pointer(Self.{$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.Items);
  for i := 1 to Self.{$ifdef WEAKREF}ClearNatives{$else}Natives{$endif}.Count do
  begin
    Value := Pointer(NativeInt(P) + Item.Offset);
    if (Value^ <> 0) then
    begin
      Item.ClearNativeProc(Value, Item.DynTypeInfo);
    end;
    Inc(Item);
  end;

  StaticArrayRec := Pointer(Self.StaticArrays.Items);
  for i := 1 to Self.StaticArrays.Count do
  begin
    System.FinalizeArray(Pointer(NativeInt(P) + StaticArrayRec.Offset),
      StaticArrayRec.StaticTypeInfo, StaticArrayRec.Count);
    Inc(StaticArrayRec);
  end;
end;

{ TRAIIHelper<T> }

class constructor TRAIIHelper<T>.Create;
begin
  FOptions.TypeInfo := TypeInfo(T);
end;

class function TRAIIHelper<T>.GetIsManaged: Boolean;
begin
  {$ifdef SMARTGENERICS}
  Result := System.IsManagedType(T) or System.HasWeakRef(T);
  {$else}
  Result := Assigned(FOptions.ClearProc);
  {$endif}
end;

class procedure TRAIIHelper<T>.Init(Item: P);
begin
  {$ifdef SMARTGENERICS}
  if (System.IsManagedType(T) or System.HasWeakRef(T)) then
  {$else}
  if (Assigned(FOptions.InitProc)) then
  {$endif}
    FOptions.InitProc(FOptions, Item);
end;

class procedure TRAIIHelper<T>.InitArray(Items: P; Count: NativeUInt; Offset: NativeUInt);
var
  i: NativeUInt;
  Item: ^TRAIIHelper.TData16;
  Null: NativeInt;
begin
  {$ifdef SMARTGENERICS}
  if (System.IsManagedType(T) or System.HasWeakRef(T)) then
  {$else}
  if (Assigned(FOptions.InitProc)) then
  {$endif}
  begin
    if (Offset = 0) then Offset := SizeOf(T);
    Item := Pointer(Items);
    Null := 0;

    {$ifdef SMARTGENERICS}
    if (GetTypeKind(T) = tkVariant) then
    begin
      for i := 1 to Count do
      begin
        Item.Integers[0] := Null;
        Inc(NativeUInt(Item), Offset);
      end;
    end else
    {$endif}
    if (SizeOf(T) <= 16) then
    begin
      if (SizeOf(T) = Offset) then
      begin
        FillChar(Item^, Count * SizeOf(T), #0);
      end else
      for i := 1 to Count do
      begin
        {$ifdef SMALLINT}
          if (SizeOf(T) >= SizeOf(Integer) * 1) then Item.Integers[0] := Null;
          if (SizeOf(T) >= SizeOf(Integer) * 2) then Item.Integers[1] := Null;
          if (SizeOf(T) >= SizeOf(Integer) * 3) then Item.Integers[2] := Null;
          if (SizeOf(T)  = SizeOf(Integer) * 4) then Item.Integers[3] := Null;
        {$else .LARGEINT}
          if (SizeOf(T) >= SizeOf(Int64) * 1) then Item.Int64s[0] := Null;
          if (SizeOf(T)  = SizeOf(Int64) * 2) then Item.Int64s[1] := Null;
          case SizeOf(T) of
             4..7: Item.Integers[0] := Null;
           12..15: Item.Integers[2] := Null;
          end;
        {$endif}
        case SizeOf(T) of
           2,3: Item.Words[0] := 0;
           6,7: Item.Words[2] := 0;
         10,11: Item.Words[4] := 0;
         14,15: Item.Words[6] := 0;
        end;
        case SizeOf(T) of
           1: Item.Bytes[ 1-1] := 0;
           3: Item.Bytes[ 3-1] := 0;
           5: Item.Bytes[ 5-1] := 0;
           7: Item.Bytes[ 7-1] := 0;
           9: Item.Bytes[ 9-1] := 0;
          11: Item.Bytes[11-1] := 0;
          13: Item.Bytes[13-1] := 0;
          15: Item.Bytes[15-1] := 0;
        end;

        Inc(NativeUInt(Item), Offset);
      end;
    end else
    for i := 1 to Count do
    begin
      FOptions.InitProc(FOptions, Item);
      Inc(NativeUInt(Item), Offset);
    end;
  end;
end;

class procedure TRAIIHelper<T>.Clear(Item: P);
begin
  {$ifdef SMARTGENERICS}
  if (System.IsManagedType(T) or System.HasWeakRef(T)) then
  {$else}
  if (Assigned(FOptions.ClearProc)) then
  {$endif}
    FOptions.ClearProc(FOptions, Item);
end;

class procedure TRAIIHelper<T>.ClearArray(Items: P; Count: NativeUInt; Offset: NativeUInt);
var
  i: NativeUInt;
  Item: ^TRAIIHelper.TData16;
  {$ifdef SMARTGENERICS}
  VType: Integer;
  {$endif}
begin
  {$ifdef SMARTGENERICS}
  if (System.IsManagedType(T) or System.HasWeakRef(T)) then
  {$else}
  if (Assigned(FOptions.ClearProc)) then
  {$endif}
  begin
    if (Offset = 0) then Offset := SizeOf(T);
    Item := Pointer(Items);

    for i := 1 to Count do
    begin
      {$ifdef SMARTGENERICS}
      case GetTypeKind(T) of
        {$ifdef AUTOREFCOUNT}
        tkClass,
        {$endif}
        tkWString, tkLString, tkUString, tkInterface, tkDynArray:
        begin
          if (Item.Native <> 0) then
          case GetTypeKind(T) of
            {$ifdef AUTOREFCOUNT}
            tkClass:
            begin
              TRAIIHelper.RefObjClear(@Item.Native);
            end;
            {$endif}
            {$ifdef MSWINDOWS}
            tkWString:
            begin
              TRAIIHelper.WStrClear(@Item.Native);
            end;
            {$else}
            tkWString,
            {$endif}
            tkLString, tkUString:
            begin
              TRAIIHelper.ULStrClear(@Item.Native);
            end;
            tkInterface:
            begin
              IInterface(Item.Native)._Release;
            end;
            tkDynArray:
            begin
              TRAIIHelper.DynArrayClear(@Item.Native, TypeInfo(T));
            end;
          end;
        end;
        {$ifdef WEAKREF}
        tkMethod:
        begin
          if (Item.Method.Data <> nil) then
            TRAIIHelper.WeakMethodClear(@Item.Method.Data);
        end;
        {$endif}
        tkVariant:
        begin
          VType := Item.VarData.VType;
          if (VType and TRAIIHelper.varDeepData <> 0) then
          case VType of
            varBoolean, varUnknown+1..varUInt64: ;
          else
            System.VarClear(Variant(Item.VarData));
          end;
        end;
      else
        FOptions.ClearProc(FOptions, Item);
      end;
      {$else}
      FOptions.ClearProc(FOptions, Item);
      {$endif}

      Inc(NativeUInt(Item), Offset);
    end;
  end;
end;


{ InterfaceDefaults }

class constructor InterfaceDefaults.TDefaultComparer<T>.Create;
begin
  Initialize;
end;

class procedure InterfaceDefaults.TDefaultComparer<T>.Initialize;
var
  TypeData: PTypeData;
begin
  Instance.Vtable := @Instance.QueryInterface;
  Instance.Size := SizeOf(T);
  Instance.QueryInterface := @InterfaceDefaults.NopQueryInterface;
  Instance.AddRef := @InterfaceDefaults.NopAddRef;
  Instance.Release := @InterfaceDefaults.NopRelease;

  // Compare
  TypeData := PTypeInfo(TypeInfo(T)).TypeData;
  case {$ifdef SMARTGENERICS}GetTypeKind(T){$else}PTypeInfo(TypeInfo(T)).Kind{$endif} of
    tkInteger, tkEnumeration, tkChar, tkWChar:
    case TypeData.OrdType of
      otSByte: Instance.Compare := @InterfaceDefaults.Compare_I1;
      otUByte: Instance.Compare := @InterfaceDefaults.Compare_U1;
      otSWord: Instance.Compare := @InterfaceDefaults.Compare_I2;
      otUWord: Instance.Compare := @InterfaceDefaults.Compare_U2;
      otSLong: Instance.Compare := @InterfaceDefaults.Compare_I4;
      otULong: Instance.Compare := @InterfaceDefaults.Compare_U4;
    end;
    tkInt64:
    begin
      if (TypeData.MaxInt64Value > TypeData.MinInt64Value) then
      begin
        Instance.Compare := @InterfaceDefaults.Compare_I8
      end else
      begin
        Instance.Compare := @InterfaceDefaults.Compare_U8;
      end;
    end;
    tkClass, tkInterface, tkClassRef, tkPointer, tkProcedure:
    begin
      {$ifdef LARGEINT}
        Instance.Compare := @InterfaceDefaults.Compare_U8;
      {$else .SMALLINT}
        Instance.Compare := @InterfaceDefaults.Compare_U4;
      {$endif}
    end;
    tkFloat:
    case TypeData.FloatType of
        ftSingle: Instance.Compare := @InterfaceDefaults.Compare_F4;
        ftDouble: Instance.Compare := @InterfaceDefaults.Compare_F8;
      ftExtended: Instance.Compare := @InterfaceDefaults.Compare_FE;
    else
      Instance.Compare := @InterfaceDefaults.Compare_U8;
    end;
    tkMethod:
    begin
      Instance.Compare := @InterfaceDefaults.Compare_Method;
    end;
    tkVariant:
    begin
      Instance.Compare := @InterfaceDefaults.Compare_Var;
    end;
    tkString:
    begin
      Instance.Compare := @InterfaceDefaults.Compare_OStr;
    end;
    tkLString:
    begin
      Instance.Compare := @InterfaceDefaults.Compare_LStr;
    end;
    tkWString:
    begin
      Instance.Compare := @InterfaceDefaults.Compare_WStr;
    end;
    tkUString:
    begin
      Instance.Compare := @InterfaceDefaults.Compare_UStr;
    end;
    tkDynArray:
    begin
      Instance.Size := TypeData.elSize;
      Instance.Compare := @InterfaceDefaults.Compare_Dyn;
    end;
  else
    // binary
    case SizeOf(T) of
      1: Instance.Compare := @InterfaceDefaults.Compare_U1;
      2: Instance.Compare := @InterfaceDefaults.Compare_Bin2;
      3: Instance.Compare := @InterfaceDefaults.Compare_Bin3;
      4: Instance.Compare := @InterfaceDefaults.Compare_Bin4;
      {$ifdef LARGEINT}
      8: Instance.Compare := @InterfaceDefaults.Compare_Bin8;
      {$endif}
    else
      Instance.Compare := @InterfaceDefaults.Compare_Bin;
    end;
  end;
end;

class constructor InterfaceDefaults.TDefaultEqualityComparer<T>.Create;
begin
  Initialize;
end;

class procedure InterfaceDefaults.TDefaultEqualityComparer<T>.Initialize;
var
  TypeData: PTypeData;
begin
  Instance.Vtable := @Instance.QueryInterface;
  Instance.Size := SizeOf(T);
  Instance.QueryInterface := @InterfaceDefaults.NopQueryInterface;
  Instance.AddRef := @InterfaceDefaults.NopAddRef;
  Instance.Release := @InterfaceDefaults.NopRelease;

  // Equals/GetHashCode
  TypeData := PTypeInfo(TypeInfo(T)).TypeData;
  case {$ifdef SMARTGENERICS}GetTypeKind(T){$else}PTypeInfo(TypeInfo(T)).Kind{$endif} of
    tkClass:
    begin
      Instance.Equals := @InterfaceDefaults.Equals_Class;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_Class;
    end;
    tkInterface, tkClassRef, tkPointer, tkProcedure:
    begin
      {$ifdef LARGEINT}
        Instance.Equals := @InterfaceDefaults.Equals_N8;
      {$else .SMALLINT}
        Instance.Equals := @InterfaceDefaults.Equals_N4;
      {$endif}
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_Ptr;
    end;
    tkInt64:
    begin
      Instance.Equals := @InterfaceDefaults.Equals_N8;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_N8;
    end;
    tkFloat:
    case TypeData.FloatType of
      ftSingle:
      begin
        Instance.Equals := @InterfaceDefaults.Equals_F4;
        Instance.GetHashCode := @InterfaceDefaults.GetHashCode_F4;
      end;
      ftDouble:
      begin
        Instance.Equals := @InterfaceDefaults.Equals_F8;
        Instance.GetHashCode := @InterfaceDefaults.GetHashCode_F8;
      end;
      ftExtended:
      begin
        Instance.Equals := @InterfaceDefaults.Equals_FE;
        Instance.GetHashCode := @InterfaceDefaults.GetHashCode_FE;
      end;
    else
      Instance.Size := -SizeOf(Int64);
      Instance.Equals := @InterfaceDefaults.Equals_N8;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_N8;
    end;
    tkMethod:
    begin
      Instance.Equals := @InterfaceDefaults.Equals_Method;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_Method;
    end;
    tkVariant:
    begin
      Instance.Equals := @InterfaceDefaults.Equals_Var;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_Var;
    end;
    tkString:
    begin
      Instance.Equals := @InterfaceDefaults.Equals_OStr;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_OStr;
    end;
    tkLString:
    begin
      Instance.Equals := @InterfaceDefaults.Equals_LStr;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_LStr;
    end;
    tkWString:
    begin
      Instance.Equals := @InterfaceDefaults.Equals_WStr;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_WStr;
    end;
    tkUString:
    begin
      Instance.Equals := @InterfaceDefaults.Equals_UStr;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_UStr;
    end;
    tkDynArray:
    begin
      Instance.Size := TypeData.elSize;
      Instance.Equals := @InterfaceDefaults.Equals_Dyn;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_Dyn;
    end;
  else
    // binary
    case SizeOf(T) of
      1:
      begin
        Instance.Equals := @InterfaceDefaults.Equals_N1;
        Instance.GetHashCode := @InterfaceDefaults.GetHashCode_N1;
      end;
      2:
      begin
        Instance.Equals := @InterfaceDefaults.Equals_N2;
        Instance.GetHashCode := @InterfaceDefaults.GetHashCode_N2;
      end;
      3:
      begin
        Instance.Equals := @InterfaceDefaults.Equals_Bin3;
        Instance.GetHashCode := @InterfaceDefaults.GetHashCode_Bin3;
      end;
      4:
      begin
        Instance.Equals := @InterfaceDefaults.Equals_N4;
        Instance.GetHashCode := @InterfaceDefaults.GetHashCode_N4;
      end;
      {$ifdef LARGEINT}
      8:
      begin
        Instance.Equals := @InterfaceDefaults.Equals_N8;
        Instance.GetHashCode := @InterfaceDefaults.GetHashCode_N8;
      end;
      {$endif}
    else
      Instance.Equals := @InterfaceDefaults.Equals_Bin;
      Instance.GetHashCode := @InterfaceDefaults.GetHashCode_Bin;
    end;
  end;
end;

class function InterfaceDefaults.NopQueryInterface(Inst: Pointer; const IID: TGUID; out Obj): HResult; stdcall;
begin
  Result := E_NOINTERFACE;
end;

class function InterfaceDefaults.NopAddRef(Inst: Pointer): Integer; stdcall;
begin
  Result := -1;
end;

class function InterfaceDefaults.NopRelease(Inst: Pointer): Integer; stdcall;
begin
  Result := -1;
end;

class function InterfaceDefaults.Compare_I1(Inst: Pointer; Left, Right: Shortint): Integer;
begin
  Result := Integer(Left) - Integer(Right);
end;

class function InterfaceDefaults.Compare_U1(Inst: Pointer; Left, Right: Byte): Integer;
begin
  Result := Integer(Left) - Integer(Right);
end;

class function InterfaceDefaults.Equals_N1(Inst: Pointer; Left, Right: Byte): Boolean;
begin
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_N1(Inst: Pointer; Value: Byte): Integer;
begin
  Result := Value;
  Inc(Result, + (Result shr 4) * 63689);
end;

class function InterfaceDefaults.Compare_I2(Inst: Pointer; Left, Right: Smallint): Integer;
begin
  Result := Integer(Left) - Integer(Right);
end;

class function InterfaceDefaults.Compare_U2(Inst: Pointer; Left, Right: Word): Integer;
begin
  Result := Integer(Left) - Integer(Right);
end;

class function InterfaceDefaults.Equals_N2(Inst: Pointer; Left, Right: Word): Boolean;
begin
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_N2(Inst: Pointer; Value: Word): Integer;
begin
  Result := Byte(Value);
  Inc(Result, (Result shr 4) * 63689);
  Inc(Result, (Integer(Value) shr 8) * -1660269137);
end;

class function InterfaceDefaults.Compare_I4(Inst: Pointer; Left, Right: Integer): Integer;
{$ifNdef CPUINTEL}
begin
  Result := Shortint(Byte(Left >= Right) - Byte(Left <= Right));
end;
{$else}
asm
  xor eax, eax
  {$ifdef CPUX86}
    cmp edx, ecx
  {$else .CPUX64}
    cmp edx, r8d
  {$endif}
  mov edx, 1
  mov ecx, -1
  cmovg eax, edx
  cmovl eax, ecx
end;
{$endif}

class function InterfaceDefaults.Compare_U4(Inst: Pointer; Left, Right: Cardinal): Integer;
{$ifNdef CPUINTEL}
begin
  Result := Shortint(Byte(Left >= Right) - Byte(Left <= Right));
end;
{$else}
asm
  xor eax, eax
  {$ifdef CPUX86}
    cmp edx, ecx
  {$else .CPUX64}
    cmp edx, r8d
  {$endif}
  mov edx, 1
  mov ecx, -1
  cmova eax, edx
  cmovb eax, ecx
end;
{$endif}

class function InterfaceDefaults.Equals_N4(Inst: Pointer; Left, Right: Integer): Boolean;
begin
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_N4(Inst: Pointer; Value: Integer): Integer;
begin
  Result := Value + ((Value shr 8) * 63689) + ((Value shr 16) * -1660269137) +
    ((Value shr 24) * -1092754919);
end;

class function InterfaceDefaults.Compare_I8(Inst: Pointer; Left, Right: Int64): Integer;
{$if Defined(CPUX64)}
asm
  xor eax, eax
  cmp rdx, r8
  mov edx, 1
  mov ecx, -1
  cmovg eax, edx
  cmovl eax, ecx
end;
{$elseif Defined(LARGEINT)}
begin
  Result := Shortint(Byte(Left >= Right) - Byte(Left <= Right));
end;
{$else .SMALLINT}
var
  IL, IR: Integer;
  UL, UR: Cardinal;
begin
  IL := TPoint(Left).Y;
  IR := TPoint(Right).Y;
  if (IL <> IR) then
  begin
    Result := Shortint(Byte(IL >= IR) - Byte(IL <= IR));
  end else
  begin
    UL := TPoint(Left).X;
    UR := TPoint(Right).X;
    Result := Shortint(Byte(UL >= UR) - Byte(UL <= UR));
  end;
end;
{$ifend}

class function InterfaceDefaults.Compare_U8(Inst: Pointer; Left, Right: UInt64): Integer;
{$if Defined(CPUX64)}
asm
  xor eax, eax
  cmp rdx, r8
  mov edx, 1
  mov ecx, -1
  cmova eax, edx
  cmovb eax, ecx
end;
{$elseif Defined(LARGEINT)}
begin
  Result := Shortint(Byte(Left >= Right) - Byte(Left <= Right));
end;
{$else .SMALLINT}
type
  UInt64Rec = array[0..1] of Cardinal;
var
  Index: NativeUInt;
  L, R: ^UInt64Rec;
  UL, UR: Cardinal;
begin
  Index := Byte(TPoint(Left).Y <> TPoint(Right).Y);
  L := @UInt64Rec(Left);
  R := @UInt64Rec(Right);
  UL := L[Index];
  UR := R[Index];
  Result := Shortint(Byte(UL >= UR) - Byte(UL <= UR));
end;
{$ifend}

class function InterfaceDefaults.Equals_N8(Inst: Pointer; Left, Right: Int64): Boolean;
begin
  {$ifdef LARGEINT}
    Result := (Left = Right);
  {$else .SMALLINT}
    Result := ((TPoint(Left).X - TPoint(Right).X) or (TPoint(Left).Y - TPoint(Right).Y) = 0);
  {$endif}
end;

class function InterfaceDefaults.GetHashCode_N8(Inst: Pointer; Value: Int64): Integer;
begin
  {$ifdef LARGEINT}
    Result := Integer(Value) + Integer(Value shr 32) * 63689;
  {$else .SMALLINT}
    Result := TPoint(Value).X + TPoint(Value).Y * 63689;
  {$endif}

  Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
    ((Result shr 24) * -1092754919));
end;

class function InterfaceDefaults.Equals_Class(Inst: Pointer; Left, Right: TObject): Boolean;
begin
  if (Left <> nil) then
  begin
    if (PPointer(Pointer(Left)^)[vmtEquals div SizeOf(Pointer)] = @TObject.Equals) then
    begin
      Result := (Left = Right);
      Exit;
    end else
    begin
      Result := Left.Equals(Right);
      Exit;
    end;
  end else
  begin
    Result := (Right = nil);
  end;
end;

class function InterfaceDefaults.GetHashCode_Class(Inst: Pointer; Value: TObject): Integer;
begin
  if (Assigned(Value)) then
  begin
    if (PPointer(Pointer(Value)^)[vmtGetHashCode div SizeOf(Pointer)] = @TObject.GetHashCode) then
    begin
      {$ifdef LARGEINT}
        Result := Integer(NativeInt(Value) xor (NativeInt(Value) shr 32));
      {$else .SMALLINT}
        Result := Integer(Value);
      {$endif}
      Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
          ((Result shr 24) * -1092754919));
      Exit;
    end else
    begin
      Result := Value.GetHashCode;
      Exit;
    end;
  end else
  begin
    Result := 0;
  end;
end;

class function InterfaceDefaults.GetHashCode_Ptr(Inst: Pointer; Value: NativeInt): Integer;
begin
  {$ifdef LARGEINT}
    Value := Value xor (Value shr 32);
  {$endif}
  Result := Integer(Value);
  Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
      ((Result shr 24) * -1092754919));
end;

class function InterfaceDefaults.Compare_F4(Inst: Pointer; Left, Right: Single): Integer;
{$if Defined(CPUX86)}
asm
  fld Left
  fcomp Right
  fstsw ax
  xor ecx, ecx
  or ebp, -1
  lea edx, [ecx + 1]
  test ax, 256
  cmovnz edx, ebp
  test ax, 16384
  cmovnz edx, ecx
  xchg eax, edx
end;
{$elseif Defined(CPUX64)}
asm
  or eax, -1
  xor edx, edx
  mov ecx, 1
  comiss xmm1,xmm2
  cmovz eax, edx
  cmova eax, ecx
end;
{$else}
begin
  Result := Shortint(Byte(Left >= Right) - Byte(Left <= Right));
end;
{$ifend}

class function InterfaceDefaults.Equals_F4(Inst: Pointer; Left, Right: Single): Boolean;
begin
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_F4(Inst: Pointer; Value: Single): Integer;
type
  TSingleRec = packed record
    Exponent: Integer;
    case Integer of
      0: (Mantissa: Single);
      1: (HighInt: Integer);
  end;
var
  SingleRec: TSingleRec;
begin
  Result := 0;
  if (Value <> 0) then
  begin
    Frexp(Value, SingleRec.Mantissa, SingleRec.Exponent);
    Result := SingleRec.Exponent + SingleRec.HighInt * 63689;
    Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
        ((Result shr 24) * -1092754919));
  end;
end;

class function InterfaceDefaults.Compare_F8(Inst: Pointer; Left, Right: Double): Integer;
{$if Defined(CPUX86)}
asm
  fld Left
  fcomp Right
  fstsw ax
  xor ecx, ecx
  or ebp, -1
  lea edx, [ecx + 1]
  test ax, 256
  cmovnz edx, ebp
  test ax, 16384
  cmovnz edx, ecx
  xchg eax, edx
end;
{$elseif Defined(CPUX64)}
asm
  or eax, -1
  xor edx, edx
  mov ecx, 1
  comisd xmm1,xmm2
  cmovz eax, edx
  cmova eax, ecx
end;
{$else}
begin
  Result := Shortint(Byte(Left >= Right) - Byte(Left <= Right));
end;
{$ifend}

class function InterfaceDefaults.Equals_F8(Inst: Pointer; Left, Right: Double): Boolean;
begin
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_F8(Inst: Pointer; Value: Double): Integer;
type
  TDoubleRec = packed record
    Exponent: Integer;
    case Integer of
      0: (Mantissa: Double);
      1: (LowInt: Integer; HighInt: Integer);
  end;
var
  DoubleRec: TDoubleRec;
begin
  Result := 0;
  if (Value <> 0) then
  begin
    Frexp(Value, DoubleRec.Mantissa, DoubleRec.Exponent);
    Result := DoubleRec.Exponent + DoubleRec.LowInt * 63689 + DoubleRec.HighInt * -1660269137;
    Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
        ((Result shr 24) * -1092754919));
  end;
end;

class function InterfaceDefaults.Compare_FE(Inst: Pointer; Left, Right: Extended): Integer;
{$if Defined(CPUX86)}
asm
  fld Right
  fld Left
  fcompp
  fstsw ax
  xor ecx, ecx
  or ebp, -1
  lea edx, [ecx + 1]
  test ax, 256
  cmovnz edx, ebp
  test ax, 16384
  cmovnz edx, ecx
  xchg eax, edx
end;
{$elseif Defined(CPUX64)}
asm
  or eax, -1
  xor edx, edx
  mov ecx, 1
  comisd xmm1,xmm2
  cmovz eax, edx
  cmova eax, ecx
end;
{$else}
begin
  Result := Shortint(Byte(Left >= Right) - Byte(Left <= Right));
end;
{$ifend}

class function InterfaceDefaults.Equals_FE(Inst: Pointer; Left, Right: Extended): Boolean;
begin
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_FE(Inst: Pointer; Value: Extended): Integer;
type
  TExtendedRec = packed record
    Exponent: Integer;
    case Integer of
      0: (Mantissa: Extended);
      1: (LowInt: Integer; {$if SizeOf(Extended) = 10}Middle: Word;{$ifend} HighInt: Integer);
  end;
var
  ExtendedRec: TExtendedRec;
begin
  Result := 0;
  if (Value <> 0) then
  begin
    Frexp(Value, ExtendedRec.Mantissa, ExtendedRec.Exponent);
    Result := ExtendedRec.Exponent + ExtendedRec.LowInt * 63689 + ExtendedRec.HighInt * -1660269137
      {$if SizeOf(Extended) = 10}+ Integer(ExtendedRec.Middle) * -1092754919{$ifend};
    Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
        ((Result shr 24) * -1092754919));
  end;
end;

class function InterfaceDefaults.Compare_Var_Difficult(Equal: Boolean; Left, Right: PVariant): Integer;
var
  S1, S2: string;
  Inst: IComparerInst;
begin
  try
    if (Left^ = Right^) then
    begin
      Result := 0;
    end else
    if (not Equal) then
    begin
      Result := 2 * Ord(Left^ > Right^) - 1;
    end else
    begin
      Result := 1;
    end;
  except // if comparison failed with exception, compare as string.
    try
      S1 := Left^;
      S2 := Right^;
      if (Equal) and (Length(S1) <> Length(S2)) then
        Exit(1);
      Result := InterfaceDefaults.Compare_UStr(nil, Pointer(S1), Pointer(S2));
    except  // if comparison fails again, compare bytes.
      Inst.Size := SizeOf(Variant);
      Result := InterfaceDefaults.Compare_Bin(Inst, Pointer(Left), Pointer(Right));
    end;
  end;
end;

class function InterfaceDefaults.Compare_Var(Inst: Pointer; Left, Right: PVarData): Integer;
label
  has_left, has_right, difficult;
var
  VLeft, VRight: Integer;
begin
  VLeft := Left.VType;
  if (VLeft <> varByRef or varVariant) then
  begin
  has_left:
    VRight := Right.VType;
    if (VRight <> varByRef or varVariant) then
    begin
    has_right:
      if (VLeft > varNull) and (VRight > varNull) then
      begin
        if (VLeft <> VRight) then goto difficult;

        case (VLeft) of
          varShortInt:
          begin
            Result := Integer(Left.VShortInt) - Integer(Right.VShortInt);
          end;
          varBoolean, varByte:
          begin
            Result := Integer(Left.VByte) - Integer(Right.VByte);
          end;
          varSmallint:
          begin
            Result := Integer(Left.VSmallInt) - Integer(Right.VSmallInt);
          end;
          varWord:
          begin
            Result := Integer(Left.VWord) - Integer(Right.VWord);
          end;
          varInteger:
          begin
            Result := Shortint(Byte(Left.VInteger >= Right.VInteger) - Byte(Left.VInteger <= Right.VInteger));
          end;
          varLongWord:
          begin
            Result := Shortint(Byte(Left.VLongWord >= Right.VLongWord) - Byte(Left.VLongWord <= Right.VLongWord));
          end;
          varInt64, varCurrency:
          begin
            {$ifdef LARGEINT}
              Result := Shortint(Byte(Left.VInt64 >= Right.VInt64) - Byte(Left.VInt64 <= Right.VInt64));
            {$else .SMALLINT}
              VLeft := Left.VLongs[2];
              VRight := Right.VLongs[2];
              if (VLeft <> VRight) then
              begin
                Result := Shortint(Byte(VLeft >= VRight) - Byte(VLeft <= VRight));
              end else
              begin
                VLeft := Left.VInteger;
                VRight := Right.VInteger;
                Result := Shortint(Byte(Cardinal(VLeft) >= Cardinal(VRight)) - Byte(Cardinal(VLeft) <= Cardinal(VRight)));
              end;
            {$endif}
          end;
          varUInt64:
          begin
            {$ifdef LARGEINT}
              Result := Shortint(Byte(Left.VUInt64 >= Right.VUInt64) - Byte(Left.VUInt64 <= Right.VUInt64));
            {$else .SMALLINT}
              VLeft := Left.VLongs[2];
              VRight := Right.VLongs[2];
              if (VLeft = VRight) then
              begin
                VLeft := Left.VInteger;
                VRight := Right.VInteger;
              end;
              Result := Shortint(Byte(Cardinal(VLeft) >= Cardinal(VRight)) - Byte(Cardinal(VLeft) <= Cardinal(VRight)));
            {$endif}
          end;
          varSingle:
          begin
            Result := Shortint(Byte(Left.VSingle >= Right.VSingle) - Byte(Left.VSingle <= Right.VSingle));
          end;
          varDouble, varDate:
          begin
            Result := Shortint(Byte(Left.VDouble >= Right.VDouble) - Byte(Left.VDouble <= Right.VDouble));
          end;
          varString:
          begin
            Result := InterfaceDefaults.Compare_LStr(nil, Left.VPointer, Right.VPointer);
          end;
          varUString:
          begin
            Result := InterfaceDefaults.Compare_UStr(nil, Left.VPointer, Right.VPointer);
          end;
          varOleStr:
          begin
            Result := InterfaceDefaults.Compare_WStr(nil, Left.VPointer, Right.VPointer);
          end;
        else
        difficult:
          Result := InterfaceDefaults.Compare_Var_Difficult(False, PVariant(Left), PVariant(Right));
        end;
      end else
      begin
        Result := 0;
      end;
    end else
    begin
      repeat
        Right := Right.VPointer;
        VRight := Right.VType;
      until (VRight <> varByRef or varVariant);
      goto has_right;
    end;
  end else
  begin
    repeat
      Left := Left.VPointer;
      VLeft := Left.VType;
    until (VLeft <> varByRef or varVariant);
    goto has_left;
  end;
end;

class function InterfaceDefaults.Equals_Var(Inst: Pointer; Left, Right: PVarData): Boolean;
label
  has_left, has_right, difficult;
var
  VLeft, VRight: Integer;
begin
  VLeft := Left.VType;
  if (VLeft <> varByRef or varVariant) then
  begin
  has_left:
    VRight := Right.VType;
    if (VRight <> varByRef or varVariant) then
    begin
    has_right:
      if (VLeft > varNull) and (VRight > varNull) then
      begin
        if (VLeft <> VRight) then goto difficult;

        case (VLeft) of
          varShortInt, varBoolean, varByte:
          begin
            Result := (Left.VByte = Right.VByte);
          end;
          varSmallint, varWord:
          begin
            Result := (Left.VWord = Right.VWord);
          end;
          varInteger, varLongWord:
          begin
            Result := (Left.VInteger = Right.VInteger);
          end;
          varInt64, varCurrency, varUInt64:
          begin
            {$ifdef LARGEINT}
              Result := (Left.VInt64 = Right.VInt64);
            {$else .SMALLINT}
              Result := ((Left.VLongs[1] - Right.VLongs[1]) or (Left.VLongs[2] - Right.VLongs[2]) = 0);
            {$endif}
          end;
          varSingle:
          begin
            Result := (Left.VSingle >= Right.VSingle) = (Left.VSingle <= Right.VSingle);
          end;
          varDouble, varDate:
          begin
            Result := (Left.VDouble >= Right.VDouble) = (Left.VDouble <= Right.VDouble);
          end;
          varString:
          begin
            Result := InterfaceDefaults.Equals_LStr(nil, Left.VPointer, Right.VPointer);
          end;
          varUString:
          begin
            Result := InterfaceDefaults.Equals_UStr(nil, Left.VPointer, Right.VPointer);
          end;
          varOleStr:
          begin
            Result := InterfaceDefaults.Equals_WStr(nil, Left.VPointer, Right.VPointer);
          end;
        else
        difficult:
          Result := (InterfaceDefaults.Compare_Var_Difficult(True, PVariant(Left), PVariant(Right)) = 0);
        end;
      end else
      begin
        Result := True;
      end;
    end else
    begin
      repeat
        Right := Right.VPointer;
        VRight := Right.VType;
      until (VRight <> varByRef or varVariant);
      goto has_right;
    end;
  end else
  begin
    repeat
      Left := Left.VPointer;
      VLeft := Left.VType;
    until (VLeft <> varByRef or varVariant);
    goto has_left;
  end;
end;

class function InterfaceDefaults.GetHashCode_Var_Difficult(Value: PVariant): Integer;
var
  S: string;
  Instance: IEqualityComparerInst;
begin
  try
    S := Value^;
    Result := GetHashCode_UStr(nil, Pointer(S));
  except
    Instance.Size := SizeOf(Variant);
    Result := GetHashCode_Bin(Instance, Pointer(Value));
  end;
end;

class function InterfaceDefaults.GetHashCode_Var(Inst: Pointer; Value: PVarData): Integer;
label
  hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, hash8, hash9, hash10,
  null, write_uint64, init_write_cardinal, write_cardinal,
  write_ordinal_string, write_terminated_string, write_string;
const
  DIGITS: array[0..99] of array[1..2] of Char = (
    '00','01','02','03','04','05','06','07','08','09',
    '10','11','12','13','14','15','16','17','18','19',
    '20','21','22','23','24','25','26','27','28','29',
    '30','31','32','33','34','35','36','37','38','39',
    '40','41','42','43','44','45','46','47','48','49',
    '50','51','52','53','54','55','56','57','58','59',
    '60','61','62','63','64','65','66','67','68','69',
    '70','71','72','73','74','75','76','77','78','79',
    '80','81','82','83','84','85','86','87','88','89',
    '90','91','92','93','94','95','96','97','98','99');
var
  V32: Cardinal;
  V64, V64Buffer: UInt64;
  VFloat: Extended;
  VSign: Boolean;
  S, Top: PChar;
  Count: NativeUInt;
  M, N: Integer;
  Buffer: array[0..31] of Char;
begin
  VSign := False;

  case Integer(Value.VType) of
    varEmpty, varNull:
    begin
    null:
      Result := 0;
      Exit;
    end;
    varBoolean:
    begin
      if (Value.VBoolean) then
      begin
        Result := -34292652;
        Exit;
      end else
      begin
        Result := 1454490910;
        Exit;
      end;
    end;
    varShortInt:
    begin
      V32 := Cardinal(Integer(Value.VShortInt));
      if (Integer(V32) < 0) then
      begin
        VSign := True;
        V32 := Cardinal(-Integer(V32));
      end;
      goto init_write_cardinal;
    end;
    varByte:
    begin
      V32 := Value.VByte;
      goto init_write_cardinal;
    end;
    varSmallint:
    begin
      V32 := Cardinal(Integer(Value.VSmallInt));
      if (Integer(V32) < 0) then
      begin
        VSign := True;
        V32 := Cardinal(-Integer(V32));
      end;
      goto init_write_cardinal;
    end;
    varWord:
    begin
      V32 := Value.VWord;
      goto init_write_cardinal;
    end;
    varInteger:
    begin
      V32 := Cardinal(Integer(Value.VInteger));
      if (Integer(V32) < 0) then
      begin
        VSign := True;
        V32 := Cardinal(-Integer(V32));
      end;
      goto init_write_cardinal;
    end;
    varLongWord:
    begin
      V32 := Value.VLongWord;
      goto init_write_cardinal;
    end;
    varInt64:
    begin
      V64 := Value.VUInt64;
      if ({$ifdef LARGEINT}Int64(V64){$else}TPoint(V64).Y{$endif} < 0) then
      begin
        VSign := True;
        Int64(V64) := -Int64(V64);
      end;
      goto write_uint64;
    end;
    varUInt64:
    begin
      V64 := Value.VUInt64;
    write_uint64:
      S := @Buffer[High(Buffer)];
      V32 := V64;
      if ({$ifdef LARGEINT}V64 <> V32{$else}TPoint(V64).Y <> 0{$endif}) then
      repeat
        V64Buffer := V64;
        V64 := V64 div 100;
        V64Buffer := V64Buffer - V64 * 100;
        Dec(S, 2);
        PCardinal(S)^ := PCardinal(@DIGITS[NativeUInt(V64Buffer)])^;
        V32 := V64;
      until ({$ifdef LARGEINT}V64 = V32{$else}TPoint(V64).Y = 0{$endif});
      goto write_cardinal;
    end;
    varSingle:
    begin
      VFloat := Value.VSingle;
      Top := @Buffer[FloatToText(Buffer, VFloat, fvExtended, ffGeneral, 15, 0, FormatSettings)];
      S := @Buffer[0];
      goto write_terminated_string;
    end;
    varDouble:
    begin
      VFloat := Value.VDouble;
      Top := @Buffer[FloatToText(Buffer, VFloat, fvExtended, ffGeneral, 15, 0, FormatSettings)];
      S := @Buffer[0];
      goto write_terminated_string;
    end;
    varCurrency:
    begin
      Top := @Buffer[FloatToText(Buffer, Value.VCurrency, fvCurrency, ffGeneral, 0, 0, FormatSettings)];
      S := @Buffer[0];
      goto write_terminated_string;
    end;
    varUString:
    begin
      S := Value.VPointer;
      if (S = nil) then goto null;
      Top := Pointer(@S[PInteger(PByte(S) - SizeOf(Integer))^]);
      goto write_string;
    end;
    varOleStr:
    begin
      S := Value.VPointer;
      if (S = nil) then goto null;
      Top := Pointer(@S[PInteger(PByte(S) - SizeOf(Integer))^ {$ifdef MSWINDOWS}shr 1{$endif}]);
      {$ifdef MSWINDOWS}if (S = Top) then goto null;{$endif}
      goto write_string;
    end;
  else
    Result := GetHashCode_Var_Difficult(Pointer(Value));
    Exit;
  end;

init_write_cardinal:
  S := @Buffer[High(Buffer)];
write_cardinal:
  if (V32 >= 10000) then
  begin
    if (V32 >= 100000000) then
    begin
      M := V32;
      V32 := V32 div 100;
      Dec(M, Integer(V32) * 100);
      Dec(S, 2);
      PCardinal(S)^ := PCardinal(@DIGITS[M])^;
    end;
    M := V32;
    V32 := V32 div 10000;
    Dec(M, Integer(V32) * 10000);
    N := (M * $147B) shr 19; // N := M div 100;
    M := M - (N * 100); // M := M mod 100;
    Dec(S, 2);
    PCardinal(S)^ := PCardinal(@DIGITS[M])^;
    Dec(S, 2);
    PCardinal(S)^ := PCardinal(@DIGITS[N])^;
  end;
  N := (Integer(V32) * $147B) shr 19; // N := M div 100;
  M := Integer(V32) - (N * 100); // M := M mod 100;
  Dec(S, 2);
  PCardinal(S)^ := PCardinal(@DIGITS[M])^;
  Dec(S, 2);
  PCardinal(S)^ := PCardinal(@DIGITS[N])^;
  Inc(S, 3);
  Dec(S, Byte(Byte(V32 > 9) + Byte(V32 > 99) + Byte(V32 > 999)));

write_ordinal_string:
  Top := @Buffer[High(Buffer)];
  if (VSign) then
  begin
    Dec(S);
    S^ := '-';
  end;

  // InterfaceDefaults.GetHashCode_UStr(S..Top)
write_terminated_string:
  Top^ := #0;
write_string:
  Inc(Top);
  Count := NativeInt(Top) - NativeInt(S);
  Count := Count and -4;
  Result := Integer(Count) + PInteger(@PByte(S)[Count - SizeOf(Integer)])^ * 63689;
  case (Count - 1) shr 2 of
   10: goto hash10;
    9: goto hash9;
    8: goto hash8;
    7: goto hash7;
    6: goto hash6;
    5: goto hash5;
    4: goto hash4;
    3: goto hash3;
    2: goto hash2;
    1: goto hash1;
    0: goto hash0;
  else
    Dec(Count);
    M := -1660269137;
    repeat
      Result := Result * M + PInteger(S)^;
      Dec(Count, SizeOf(Integer));
      Inc(PByte(S), SizeOf(Integer));
      M := M * 378551;
    until (Count <= 43);

    hash10:
      Result := Result * 631547855 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash9:
      Result := Result * -1987506439 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash8:
      Result := Result * -1653913089 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash7:
      Result := Result * -186114231 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash6:
      Result := Result * 915264303 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash5:
      Result := Result * -794603367 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash4:
      Result := Result * 135394143 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash3:
      Result := Result * 2012804575 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash2:
      Result := Result * -1092754919 + PInteger(S)^;
      Inc(PByte(S), SizeOf(Integer));
    hash1:
      Result := Result * -1660269137 + PInteger(S)^;
    hash0:
  end;

  Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
    ((Result shr 24) * -1092754919));
end;

class function InterfaceDefaults.Compare_OStr(Inst: Pointer; Left, Right: PByte): Integer;
label
  make_result, make_result_swaped;
var
  X, Y, Count: NativeUInt;
  Modify: Integer;
begin
  X := Left^;
  Y := Right^;
  if (Left <> Right) and (X <> 0) and (Y <> 0) then
  begin
    if (Left[1] = Right[1]) then
    begin
      if (X < Y) then
      begin
        Modify := -1;
        Count := X;
      end else
      begin
        Modify := NativeInt(Y - X) shr {$ifdef SMALLINT}31{$else}63{$endif};
        Count := Y;
      end;
      Inc(Left);
      Inc(Right);

      repeat
        if (Count < SizeOf(NativeUInt)) then Break;
        X := PNativeUInt(Left)^;
        Dec(Count, SizeOf(NativeUInt));
        Y := PNativeUInt(Right)^;
        Inc(Left, SizeOf(NativeUInt));
        Inc(Right, SizeOf(NativeUInt));

        if (X <> Y) then
        begin
          {$ifdef LARGEINT}
            if (Integer(X) = Integer(Y)) then
            begin
              X := X shr 32;
              Y := Y shr 32;
            end else
            begin
              X := Cardinal(X);
              Y := Cardinal(Y);
            end;
          {$endif}

          goto make_result;
        end;
      until (False);

      {$ifdef LARGEINT}
      if (Count and 4 <> 0) then
      begin
        X := PCardinal(Left)^;
        Y := PCardinal(Right)^;
        Inc(Left, SizeOf(Cardinal));
        Inc(Right, SizeOf(Cardinal));

        if (X <> Y) then goto make_result;
      end;
      {$endif}

      case Count of
        1: begin
             X := PByte(Left)^;
             Y := PByte(Right)^;
             if (X <> Y) then goto make_result_swaped;
           end;
        2: begin
             X := Swap(PWord(Left)^);
             Y := Swap(PWord(Right)^);
             if (X <> Y) then goto make_result_swaped;
           end;
        3: begin
             X := Swap(PWord(Left)^);
             Y := Swap(PWord(Right)^);
             Inc(Left, SizeOf(Word));
             Inc(Right, SizeOf(Word));
             X := (X shl 8) or PByte(Left)^;
             Y := (Y shl 8) or PByte(Right)^;
             if (X <> Y) then goto make_result_swaped;
           end;
      end;

      Result := Modify;
      Exit;
    make_result:
      X := (Swap(X) shl 16) + Swap(X shr 16);
      Y := (Swap(Y) shl 16) + Swap(Y shr 16);

    make_result_swaped:
      Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
      Exit;
    end else
    begin
      Inc(Left);
      Inc(Right);
      X := Left^;
      Y := Right^;
    end;
  end;

  Result := Integer(X) - Integer(Y);
end;

class function InterfaceDefaults.Equals_OStr(Inst: Pointer; Left, Right: PByte): Boolean;
label
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5, {$ifdef SMALLINT}cmp6, cmp7, cmp8, cmp9, cmp10,{$endif}
  done;
var
  Count: NativeUInt;
begin
  if (Left = Right) then goto done;
  Count := Left^;
  if (Count <> Right^) then goto done;

  // natives (40 bytes static) compare
  case Count shr {$ifdef LARGEINT}3{$else}2{$endif} of
  {$ifdef SMALLINT}
   10: goto cmp10;
    9: goto cmp9;
    8: goto cmp8;
    7: goto cmp7;
    6: goto cmp6;
  {$endif}
    5: goto cmp5;
    4: goto cmp4;
    3: goto cmp3;
    2: goto cmp2;
    1: goto cmp1;
    0: goto cmp0;
  else
    repeat
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Dec(Count, SizeOf(NativeUInt));
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    until (Count > (40 + SizeOf(NativeUInt) - 1));

    {$ifdef SMALLINT}
    cmp10:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp9:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp8:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp7:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp6:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    {$endif}
    cmp5:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp4:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp3:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp2:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp1:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp0:
  end;

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    if (PCardinal(Left)^ <> PCardinal(Right)^) then goto done;
    Inc(Left, SizeOf(Cardinal));
    Inc(Right, SizeOf(Cardinal));
  end;
  {$endif}

  if (Count and 3 <> 0) then
  begin
    if (Count and 2 <> 0) then
    begin
      if (PWord(Left)^ <> PWord(Right)^) then goto done;
    end;
    Dec(Left);
    Dec(Right);
    if (Left[Count] <> Right[Count]) then goto done;
  end;

// Result := True
  Left := nil;
  Right := nil;
done:
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_OStr(Inst: Pointer; Value: PByte): Integer;
label
  hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, hash8, hash9, hash10;
var
  Count: NativeUInt;
  M: Integer;
begin
  Count := Value^;
  Inc(Count);

  if (Count >= SizeOf(Integer)) then
  begin
    Result := Integer(Count) + PInteger(@Value[Count - SizeOf(Integer)])^ * 63689;

    case (Count - 1) shr 2 of
     10: goto hash10;
      9: goto hash9;
      8: goto hash8;
      7: goto hash7;
      6: goto hash6;
      5: goto hash5;
      4: goto hash4;
      3: goto hash3;
      2: goto hash2;
      1: goto hash1;
      0: goto hash0;
    else
      Dec(Count);
      M := -1660269137;
      repeat
        Result := Result * M + PInteger(Value)^;
        Dec(Count, SizeOf(Integer));
        Inc(Value, SizeOf(Integer));
        M := M * 378551;
      until (Count <= 43);

      hash10:
        Result := Result * 631547855 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash9:
        Result := Result * -1987506439 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash8:
        Result := Result * -1653913089 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash7:
        Result := Result * -186114231 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash6:
        Result := Result * 915264303 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash5:
        Result := Result * -794603367 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash4:
        Result := Result * 135394143 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash3:
        Result := Result * 2012804575 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash2:
        Result := Result * -1092754919 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash1:
        Result := Result * -1660269137 + PInteger(Value)^;
      hash0:
    end;

    Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
      ((Result shr 24) * -1092754919));
    Exit;
  end else
  begin
    Result := Integer(Value[0]);
    Result := Result + (Result shr 4) * 63689;
    if (Count > 1) then
    begin
      Result := Result + Integer(Value[1]) * -1660269137;
      if (Count > 2) then
      begin
        Result := Result + Integer(Value[2]) * -1092754919;
      end;
    end;
  end;
end;

class function InterfaceDefaults.Compare_LStr(Inst: Pointer; Left, Right: PByte): Integer;
label
  make_result, make_result_swaped;
var
  X, Y,Count: NativeUInt;
  Modify: Integer;
begin
  X := NativeUInt(Left);
  Y := NativeUInt(Right);
  if (Left = nil) or (Right = nil) or (Left = Right) then goto make_result_swaped;

  X := Left^;
  Y := Right^;
  if (X <> Y) then goto make_result_swaped;

  Dec(Left, SizeOf(Integer));
  Dec(Right, SizeOf(Integer));
  X := PInteger(Left)^;
  Y := PInteger(Right)^;
  Inc(Left, SizeOf(Integer));
  Inc(Right, SizeOf(Integer));
  if (X < Y) then
  begin
    Modify := -1;
    Count := X + 1;
  end else
  begin
    Modify := NativeInt(Y - X) shr {$ifdef SMALLINT}31{$else}63{$endif};
    Count := Y + 1;
  end;

  repeat
    if (Count < SizeOf(NativeUInt)) then Break;
    X := PNativeUInt(Left)^;
    Dec(Count, SizeOf(NativeUInt));
    Y := PNativeUInt(Right)^;
    Inc(Left, SizeOf(NativeUInt));
    Inc(Right, SizeOf(NativeUInt));

    if (X <> Y) then
    begin
      {$ifdef LARGEINT}
        if (Integer(X) = Integer(Y)) then
        begin
          X := X shr 32;
          Y := Y shr 32;
        end else
        begin
          X := Cardinal(X);
          Y := Cardinal(Y);
        end;
      {$endif}

      goto make_result;
    end;
  until (False);

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    X := PCardinal(Left)^;
    Y := PCardinal(Right)^;
    Inc(Left, SizeOf(Cardinal));
    Inc(Right, SizeOf(Cardinal));

    if (X <> Y) then goto make_result;
  end;
  {$endif}

  if (Count and 2 <> 0) then
  begin
    X := Swap(PWord(Left)^);
    Y := Swap(PWord(Right)^);
    if (X <> Y) then goto make_result_swaped;
  end;

  Result := Modify;
  Exit;
make_result:
  X := (Swap(X) shl 16) + Swap(X shr 16);
  Y := (Swap(Y) shl 16) + Swap(Y shr 16);

make_result_swaped:
  Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
end;

class function InterfaceDefaults.Equals_LStr(Inst: Pointer; Left, Right: PByte): Boolean;
label
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5, {$ifdef SMALLINT}cmp6, cmp7, cmp8, cmp9, cmp10,{$endif}
  done;
var
  Count: NativeUInt;
begin
  if (Left = nil) or (Right = nil) or (Left = Right) then goto done;
  Dec(Left, SizeOf(Integer));
  Dec(Right, SizeOf(Integer));
  Count := PInteger(Left)^;
  if (Integer(Count) <> PInteger(Right)^) then goto done;
  Inc(Count);
  Inc(Left, SizeOf(Integer));
  Inc(Right, SizeOf(Integer));

  // natives (40 bytes static) compare
  case Count shr {$ifdef LARGEINT}3{$else}2{$endif} of
  {$ifdef SMALLINT}
   10: goto cmp10;
    9: goto cmp9;
    8: goto cmp8;
    7: goto cmp7;
    6: goto cmp6;
  {$endif}
    5: goto cmp5;
    4: goto cmp4;
    3: goto cmp3;
    2: goto cmp2;
    1: goto cmp1;
    0: goto cmp0;
  else
    repeat
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Dec(Count, SizeOf(NativeUInt));
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    until (Count > (40 + SizeOf(NativeUInt) - 1));

    {$ifdef SMALLINT}
    cmp10:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp9:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp8:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp7:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp6:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    {$endif}
    cmp5:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp4:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp3:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp2:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp1:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp0:
  end;

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    if (PCardinal(Left)^ <> PCardinal(Right)^) then goto done;
    Inc(Left, SizeOf(Cardinal));
    Inc(Right, SizeOf(Cardinal));
  end;
  {$endif}

  if (Count and 2 <> 0) then
  begin
    if (PWord(Left)^ <> PWord(Right)^) then goto done;
  end;

// Result := True
  Left := nil;
  Right := nil;
done:
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_LStr(Inst: Pointer; Value: PByte): Integer;
label
  hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, hash8, hash9, hash10;
var
  Count: NativeUInt;
  M: Integer;
begin
  if (Value <> nil) then
  begin
    Count := PInteger(@Value[-SizeOf(Integer)])^;

    if (Count >= SizeOf(Integer)) then
    begin
      Result := Integer(Count) + PInteger(@Value[Count - SizeOf(Integer)])^ * 63689;

      case (Count - 1) shr 2 of
       10: goto hash10;
        9: goto hash9;
        8: goto hash8;
        7: goto hash7;
        6: goto hash6;
        5: goto hash5;
        4: goto hash4;
        3: goto hash3;
        2: goto hash2;
        1: goto hash1;
        0: goto hash0;
      else
        Dec(Count);
        M := -1660269137;
        repeat
          Result := Result * M + PInteger(Value)^;
          Dec(Count, SizeOf(Integer));
          Inc(Value, SizeOf(Integer));
          M := M * 378551;
        until (Count <= 43);

        hash10:
          Result := Result * 631547855 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash9:
          Result := Result * -1987506439 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash8:
          Result := Result * -1653913089 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash7:
          Result := Result * -186114231 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash6:
          Result := Result * 915264303 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash5:
          Result := Result * -794603367 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash4:
          Result := Result * 135394143 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash3:
          Result := Result * 2012804575 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash2:
          Result := Result * -1092754919 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash1:
          Result := Result * -1660269137 + PInteger(Value)^;
        hash0:
      end;

      Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
        ((Result shr 24) * -1092754919));
      Exit;
    end else
    begin
      Result := Integer(Value[0]);
      Result := Result + (Result shr 4) * 63689;
      if (Count > 1) then
      begin
        Result := Result + Integer(Value[1]) * -1660269137;
        if (Count > 2) then
        begin
          Result := Result + Integer(Value[2]) * -1092754919;
        end;
      end;
    end;
  end else
  begin
    Result := 0;
  end;
end;

class function InterfaceDefaults.Compare_UStr(Inst: Pointer; Left, Right: PByte): Integer;
label
  make_result, make_result_swaped;
var
  X, Y, Count: NativeUInt;
  Modify: Integer;
begin
  X := NativeUInt(Left);
  Y := NativeUInt(Right);
  if (Left = nil) or (Right = nil) or (Left = Right) then goto make_result_swaped;

  X := PWord(Left)^;
  Y := PWord(Right)^;
  if (X <> Y) then goto make_result_swaped;

  Dec(Left, SizeOf(Integer));
  Dec(Right, SizeOf(Integer));
  X := PInteger(Left)^;
  Y := PInteger(Right)^;
  Inc(Left, SizeOf(Integer));
  Inc(Right, SizeOf(Integer));
  if (X < Y) then
  begin
    Modify := -1;
    Count := X * 2 + 2;
  end else
  begin
    Modify := NativeInt(Y - X) shr {$ifdef SMALLINT}31{$else}63{$endif};
    Count := Y * 2 + 2;
  end;

  repeat
    if (Count < SizeOf(NativeUInt)) then Break;
    X := PNativeUInt(Left)^;
    Dec(Count, SizeOf(NativeUInt));
    Y := PNativeUInt(Right)^;
    Inc(Left, SizeOf(NativeUInt));
    Inc(Right, SizeOf(NativeUInt));

    if (X <> Y) then
    begin
      {$ifdef LARGEINT}
        if (Integer(X) = Integer(Y)) then
        begin
          X := X shr 32;
          Y := Y shr 32;
        end else
        begin
          X := Cardinal(X);
          Y := Cardinal(Y);
        end;
      {$endif}

      goto make_result;
    end;
  until (False);

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    X := PCardinal(Left)^;
    Y := PCardinal(Right)^;
    if (X <> Y) then goto make_result;
  end;
  {$endif}

  Result := Modify;
  Exit;
make_result:
  X := {$ifdef LARGEINT}Cardinal{$endif}(X shl 16) + (X shr 16);
  Y := {$ifdef LARGEINT}Cardinal{$endif}(Y shl 16) + (Y shr 16);
make_result_swaped:
  Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
end;

class function InterfaceDefaults.Equals_UStr(Inst: Pointer; Left, Right: PByte): Boolean;
label
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5, {$ifdef SMALLINT}cmp6, cmp7, cmp8, cmp9, cmp10,{$endif}
  done;
var
  Count: NativeUInt;
begin
  if (Left = nil) or (Right = nil) or (Left = Right) then goto done;
  Dec(Left, SizeOf(Integer));
  Dec(Right, SizeOf(Integer));
  Count := PInteger(Left)^;
  if (Integer(Count) <> PInteger(Right)^) then goto done;
  Count := Count * 2 + 2;
  Inc(Left, SizeOf(Integer));
  Inc(Right, SizeOf(Integer));

  // natives (40 bytes static) compare
  case Count shr {$ifdef LARGEINT}3{$else}2{$endif} of
  {$ifdef SMALLINT}
   10: goto cmp10;
    9: goto cmp9;
    8: goto cmp8;
    7: goto cmp7;
    6: goto cmp6;
  {$endif}
    5: goto cmp5;
    4: goto cmp4;
    3: goto cmp3;
    2: goto cmp2;
    1: goto cmp1;
    0: goto cmp0;
  else
    repeat
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Dec(Count, SizeOf(NativeUInt));
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    until (Count > (40 + SizeOf(NativeUInt) - 1));

    {$ifdef SMALLINT}
    cmp10:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp9:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp8:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp7:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp6:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    {$endif}
    cmp5:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp4:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp3:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp2:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp1:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      {$ifdef LARGEINT}
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
      {$endif}
    cmp0:
  end;

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    if (PCardinal(Left)^ <> PCardinal(Right)^) then goto done;
  end;
  {$endif}

// Result := True
  Left := nil;
  Right := nil;
done:
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_UStr(Inst: Pointer; Value: PByte): Integer;
label
  hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, hash8, hash9, hash10;
var
  Count: NativeUInt;
  M: Integer;
begin
  if (Value <> nil) then
  begin
    Count := PInteger(@Value[-SizeOf(Integer)])^;
    Count := Count * 2 + 2;
    Count := Count and -4;

    Result := Integer(Count) + PInteger(@Value[Count - SizeOf(Integer)])^ * 63689;
    case (Count - 1) shr 2 of
     10: goto hash10;
      9: goto hash9;
      8: goto hash8;
      7: goto hash7;
      6: goto hash6;
      5: goto hash5;
      4: goto hash4;
      3: goto hash3;
      2: goto hash2;
      1: goto hash1;
      0: goto hash0;
    else
      Dec(Count);
      M := -1660269137;
      repeat
        Result := Result * M + PInteger(Value)^;
        Dec(Count, SizeOf(Integer));
        Inc(Value, SizeOf(Integer));
        M := M * 378551;
      until (Count <= 43);

      hash10:
        Result := Result * 631547855 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash9:
        Result := Result * -1987506439 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash8:
        Result := Result * -1653913089 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash7:
        Result := Result * -186114231 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash6:
        Result := Result * 915264303 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash5:
        Result := Result * -794603367 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash4:
        Result := Result * 135394143 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash3:
        Result := Result * 2012804575 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash2:
        Result := Result * -1092754919 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash1:
        Result := Result * -1660269137 + PInteger(Value)^;
      hash0:
    end;

    Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
      ((Result shr 24) * -1092754919));
    Exit;
  end else
  begin
    Result := 0;
  end;
end;

class function InterfaceDefaults.Compare_WStr(Inst: Pointer; Left, Right: PByte): Integer;
label
  {$ifdef MSWINDOWS}left_nil, right_nil,{$endif}
  make_result, make_result_swaped;
var
  X, Y, Count: NativeUInt;
  Modify: Integer;
begin
  X := NativeUInt(Left);
  Y := NativeUInt(Right);
  if (Left = Right) then goto make_result_swaped;
  if (Left = nil) then goto left_nil;
  if (Right = nil) then goto right_nil;

  X := PWord(Left)^;
  Y := PWord(Right)^;
  if (X <> Y) then goto make_result_swaped;

  Dec(Left, SizeOf(Integer));
  Dec(Right, SizeOf(Integer));
  X := PInteger(Left)^;
  Y := PInteger(Right)^;
  Inc(Left, SizeOf(Integer));
  Inc(Right, SizeOf(Integer));
  if (X < Y) then
  begin
    Modify := -1;
    Count := X {$ifNdef MSWINDOWS}* 2{$endif} + 2;
  end else
  begin
    Modify := NativeInt(Y - X) shr {$ifdef SMALLINT}31{$else}63{$endif};
    Count := Y {$ifNdef MSWINDOWS}* 2{$endif} + 2;
  end;

  repeat
    if (Count < SizeOf(NativeUInt)) then Break;
    X := PNativeUInt(Left)^;
    Dec(Count, SizeOf(NativeUInt));
    Y := PNativeUInt(Right)^;
    Inc(Left, SizeOf(NativeUInt));
    Inc(Right, SizeOf(NativeUInt));

    if (X <> Y) then
    begin
      {$ifdef LARGEINT}
        if (Integer(X) = Integer(Y)) then
        begin
          X := X shr 32;
          Y := Y shr 32;
        end else
        begin
          X := Cardinal(X);
          Y := Cardinal(Y);
        end;
      {$endif}

      goto make_result;
    end;
  until (False);

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    X := PCardinal(Left)^;
    Y := PCardinal(Right)^;
    if (X <> Y) then goto make_result;
  end;
  {$endif}

  Result := Modify;
  Exit;
make_result:
  X := {$ifdef LARGEINT}Cardinal{$endif}(X shl 16) + (X shr 16);
  Y := {$ifdef LARGEINT}Cardinal{$endif}(Y shl 16) + (Y shr 16);
make_result_swaped:
  Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
{$ifdef MSWINDOWS}
  Exit;
left_nil:
  Dec(Right, SizeOf(Integer));
  Result := -Ord(PInteger(Right)^ <> 0);
  Exit;
right_nil:
  Dec(Left, SizeOf(Integer));
  Result := Ord(PInteger(Left)^ <> 0);
{$endif}
end;

class function InterfaceDefaults.Equals_WStr(Inst: Pointer; Left, Right: PByte): Boolean;
label
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5, {$ifdef SMALLINT}cmp6, cmp7, cmp8, cmp9, cmp10,{$endif}
  {$ifdef MSWINDOWS}left_nil, right_nil,{$endif}
  done;
var
  Count: NativeUInt;
begin
  if (Left = Right) then goto done;
  if (Left = nil) then goto {$ifdef MSWINDOWS}left_nil{$else}done{$endif};
  if (Right = nil) then goto {$ifdef MSWINDOWS}right_nil{$else}done{$endif};
  Dec(Left, SizeOf(Integer));
  Dec(Right, SizeOf(Integer));
  Count := PInteger(Left)^;
  if (Integer(Count) <> PInteger(Right)^) then goto done;
  Count := Count {$ifNdef MSWINDOWS}* 2{$endif} + 2;
  Inc(Left, SizeOf(Integer));
  Inc(Right, SizeOf(Integer));

  // natives (40 bytes static) compare
  case Count shr {$ifdef LARGEINT}3{$else}2{$endif} of
  {$ifdef SMALLINT}
   10: goto cmp10;
    9: goto cmp9;
    8: goto cmp8;
    7: goto cmp7;
    6: goto cmp6;
  {$endif}
    5: goto cmp5;
    4: goto cmp4;
    3: goto cmp3;
    2: goto cmp2;
    1: goto cmp1;
    0: goto cmp0;
  else
    repeat
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Dec(Count, SizeOf(NativeUInt));
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    until (Count > (40 + SizeOf(NativeUInt) - 1));

    {$ifdef SMALLINT}
    cmp10:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp9:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp8:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp7:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp6:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    {$endif}
    cmp5:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp4:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp3:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp2:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp1:
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      {$ifdef LARGEINT}
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
      {$endif}
    cmp0:
  end;

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    if (PCardinal(Left)^ <> PCardinal(Right)^) then goto done;
  end;
  {$endif}

// Result := True
  Left := nil;
  Right := nil;
done:
  Result := (Left = Right);
{$ifdef MSWINDOWS}
  Exit;
left_nil:
  Dec(Right, SizeOf(Integer));
  Result := (PInteger(Right)^ = 0);
  Exit;
right_nil:
  Dec(Left, SizeOf(Integer));
  Result := (PInteger(Left)^ = 0);
{$endif}
end;

class function InterfaceDefaults.GetHashCode_WStr(Inst: Pointer; Value: PByte): Integer;
label
  {$ifdef MSWINDOWS}null,{$endif}
  hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, hash8, hash9, hash10;
var
  Count: NativeUInt;
  M: Integer;
begin
  if (Value <> nil) then
  begin
    Count := PInteger(@Value[-SizeOf(Integer)])^;
    {$ifdef MSWINDOWS}if (Count = 0) then goto null;{$endif}
    Count := Count {$ifNdef MSWINDOWS}* 2{$endif} + 2;
    Count := Count and -4;

    Result := Integer(Count) + PInteger(@Value[Count - SizeOf(Integer)])^ * 63689;
    case (Count - 1) shr 2 of
     10: goto hash10;
      9: goto hash9;
      8: goto hash8;
      7: goto hash7;
      6: goto hash6;
      5: goto hash5;
      4: goto hash4;
      3: goto hash3;
      2: goto hash2;
      1: goto hash1;
      0: goto hash0;
    else
      Dec(Count);
      M := -1660269137;
      repeat
        Result := Result * M + PInteger(Value)^;
        Dec(Count, SizeOf(Integer));
        Inc(Value, SizeOf(Integer));
        M := M * 378551;
      until (Count <= 43);

      hash10:
        Result := Result * 631547855 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash9:
        Result := Result * -1987506439 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash8:
        Result := Result * -1653913089 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash7:
        Result := Result * -186114231 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash6:
        Result := Result * 915264303 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash5:
        Result := Result * -794603367 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash4:
        Result := Result * 135394143 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash3:
        Result := Result * 2012804575 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash2:
        Result := Result * -1092754919 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash1:
        Result := Result * -1660269137 + PInteger(Value)^;
      hash0:
    end;

    Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
      ((Result shr 24) * -1092754919));
    Exit;
  end else
  begin
  {$ifdef MSWINDOWS}null:{$endif}
    Result := 0;
  end;
end;

class function InterfaceDefaults.Compare_Method(Inst: Pointer; const Left, Right: TMethodPtr): Integer;
var
  X, Y: NativeUInt;
begin
  X := NativeUInt(TMethod(Left).Data);
  Y := NativeUInt(TMethod(Right).Data);
  if (X = Y) then
  begin
    X := NativeUInt(TMethod(Left).Code);
    Y := NativeUInt(TMethod(Right).Code);
  end;

  Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
end;

class function InterfaceDefaults.Equals_Method(Inst: Pointer; const Left, Right: TMethodPtr): Boolean;
begin
  Result := ((NativeInt(TMethod(Left).Data) - NativeInt(TMethod(Right).Data)) or
    (NativeInt(TMethod(Left).Code) - NativeInt(TMethod(Right).Code)) = 0);
end;

class function InterfaceDefaults.GetHashCode_Method(Inst: Pointer; const Value: TMethodPtr): Integer;
{$ifdef LARGEINT}
var
  Data: PByte;
{$endif}
begin
  {$ifdef SMALLINT}
    Result := PPoint(@Value).X + PPoint(@Value).Y * 63689;
  {$else .LARGEINT}
    Data := Pointer(@Value);
    Result := Integer(SizeOf(TMethodPtr)) + PInteger(Data)[3] * 63689;
    Result := Result * 2012804575 + PInteger(Data)[0];
    Result := Result * -1092754919 + PInteger(Data)[1];
    Result := Result * -1660269137 + PInteger(Data)[2];
  {$endif}

  Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
    ((Result shr 24) * -1092754919));
end;

class function InterfaceDefaults.Compare_Dyn(const Inst: IComparerInst; Left, Right: PByte): Integer;
label
  make_result, make_result_swaped;
var
  X, Y,Count: NativeUInt;
  Modify: Integer;
begin
  Count := Inst.Size;
  X := NativeUInt(Left);
  Y := NativeUInt(Right);
  if (Left = nil) or (Right = nil) or (Left = Right) then goto make_result_swaped;

  X := Left^;
  Y := Right^;
  if (X <> Y) then goto make_result_swaped;

  Dec(Left, SizeOf(NativeUInt));
  Dec(Right, SizeOf(NativeUInt));
  X := PNativeUInt(Left)^;
  Y := PNativeUInt(Right)^;
  Inc(Left, SizeOf(NativeUInt));
  Inc(Right, SizeOf(NativeUInt));
  if (X < Y) then
  begin
    Modify := -1;
    NativeInt(Count) := NativeInt(Count) * NativeInt(X);
  end else
  begin
    Modify := NativeInt(Y - X) shr {$ifdef SMALLINT}31{$else}63{$endif};
    NativeInt(Count) := NativeInt(Count) * NativeInt(Y);
  end;

  repeat
    if (Count < SizeOf(NativeUInt)) then Break;
    X := PNativeUInt(Left)^;
    Dec(Count, SizeOf(NativeUInt));
    Y := PNativeUInt(Right)^;
    Inc(Left, SizeOf(NativeUInt));
    Inc(Right, SizeOf(NativeUInt));

    if (X <> Y) then
    begin
      {$ifdef LARGEINT}
        if (Integer(X) = Integer(Y)) then
        begin
          X := X shr 32;
          Y := Y shr 32;
        end else
        begin
          X := Cardinal(X);
          Y := Cardinal(Y);
        end;
      {$endif}

      goto make_result;
    end;
  until (False);

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    X := PCardinal(Left)^;
    Y := PCardinal(Right)^;
    Inc(Left, SizeOf(Cardinal));
    Inc(Right, SizeOf(Cardinal));

    if (X <> Y) then goto make_result;
  end;
  {$endif}

  case Count of
    1: begin
         X := PByte(Left)^;
         Y := PByte(Right)^;
         if (X <> Y) then goto make_result_swaped;
       end;
    2: begin
         X := Swap(PWord(Left)^);
         Y := Swap(PWord(Right)^);
         if (X <> Y) then goto make_result_swaped;
       end;
    3: begin
         X := Swap(PWord(Left)^);
         Y := Swap(PWord(Right)^);
         Inc(Left, SizeOf(Word));
         Inc(Right, SizeOf(Word));
         X := (X shl 8) or PByte(Left)^;
         Y := (Y shl 8) or PByte(Right)^;
         if (X <> Y) then goto make_result_swaped;
       end;
  end;

  Result := Modify;
  Exit;
make_result:
  X := (Swap(X) shl 16) + Swap(X shr 16);
  Y := (Swap(Y) shl 16) + Swap(Y shr 16);

make_result_swaped:
  Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
end;


class function InterfaceDefaults.Equals_Dyn(const Inst: IEqualityComparerInst; Left, Right: PByte): Boolean;
label
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5, {$ifdef SMALLINT}cmp6, cmp7, cmp8, cmp9, cmp10,{$endif}
  done;
var
  Count: NativeUInt;
begin
  if (Left = nil) or (Right = nil) or (Left = Right) then goto done;
  Dec(Left, SizeOf(Integer));
  Dec(Right, SizeOf(Integer));
  Count := PInteger(Left)^;
  if (Integer(Count) <> PInteger(Right)^) then goto done;
  NativeInt(Count) := NativeInt(Count) * Inst.Size;
  Inc(Left, SizeOf(Integer));
  Inc(Right, SizeOf(Integer));

  // natives (40 bytes static) compare
  case Count shr {$ifdef LARGEINT}3{$else}2{$endif} of
  {$ifdef SMALLINT}
   10: goto cmp10;
    9: goto cmp9;
    8: goto cmp8;
    7: goto cmp7;
    6: goto cmp6;
  {$endif}
    5: goto cmp5;
    4: goto cmp4;
    3: goto cmp3;
    2: goto cmp2;
    1: goto cmp1;
    0: goto cmp0;
  else
    repeat
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Dec(Count, SizeOf(NativeUInt));
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    until (Count > (40 + SizeOf(NativeUInt) - 1));

    {$ifdef SMALLINT}
    cmp10:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp9:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp8:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp7:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp6:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    {$endif}
    cmp5:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp4:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp3:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp2:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp1:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp0:
  end;

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    if (PCardinal(Left)^ <> PCardinal(Right)^) then goto done;
    Inc(Left, SizeOf(Cardinal));
    Inc(Right, SizeOf(Cardinal));
  end;
  {$endif}

  if (Count and 3 <> 0) then
  begin
    if (Count and 2 <> 0) then
    begin
      if (PWord(Left)^ <> PWord(Right)^) then goto done;
    end;
    Dec(Left);
    Dec(Right);
    if (Left[Count] <> Right[Count]) then goto done;
  end;

// Result := True
  Left := nil;
  Right := nil;
done:
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_Dyn(const Inst: IEqualityComparerInst; Value: PByte): Integer;
label
  hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, hash8, hash9, hash10;
var
  Count: NativeUInt;
  M: Integer;
begin
  if (Value <> nil) then
  begin
    Count := NativeUInt(PNativeInt(@Value[-SizeOf(NativeInt)])^ * Inst.Size);

    if (Count >= SizeOf(Integer)) then
    begin
      Result := Integer(Count) + PInteger(@Value[Count - SizeOf(Integer)])^ * 63689;

      case (Count - 1) shr 2 of
       10: goto hash10;
        9: goto hash9;
        8: goto hash8;
        7: goto hash7;
        6: goto hash6;
        5: goto hash5;
        4: goto hash4;
        3: goto hash3;
        2: goto hash2;
        1: goto hash1;
        0: goto hash0;
      else
        Dec(Count);
        M := -1660269137;
        repeat
          Result := Result * M + PInteger(Value)^;
          Dec(Count, SizeOf(Integer));
          Inc(Value, SizeOf(Integer));
          M := M * 378551;
        until (Count <= 43);

        hash10:
          Result := Result * 631547855 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash9:
          Result := Result * -1987506439 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash8:
          Result := Result * -1653913089 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash7:
          Result := Result * -186114231 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash6:
          Result := Result * 915264303 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash5:
          Result := Result * -794603367 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash4:
          Result := Result * 135394143 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash3:
          Result := Result * 2012804575 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash2:
          Result := Result * -1092754919 + PInteger(Value)^;
          Inc(Value, SizeOf(Integer));
        hash1:
          Result := Result * -1660269137 + PInteger(Value)^;
        hash0:
      end;

      Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
        ((Result shr 24) * -1092754919));
      Exit;
    end else
    begin
      Result := Integer(Value[0]);
      Result := Result + (Result shr 4) * 63689;
      if (Count > 1) then
      begin
        Result := Result + Integer(Value[1]) * -1660269137;
        if (Count > 2) then
        begin
          Result := Result + Integer(Value[2]) * -1092754919;
        end;
      end;
    end;
  end else
  begin
    Result := 0;
  end;
end;

class function InterfaceDefaults.Compare_Bin2(Inst: Pointer; Left, Right: Word): Integer;
var
  L, R: NativeUInt;
begin
  L := Left;
  R := Right;
  L := Swap(L);
  R := Swap(R);
  Result := Integer(L) - Integer(R);
end;

class function InterfaceDefaults.Compare_Bin3(Inst: Pointer; const Left, Right: TTriple): Integer;
var
  L, R: NativeUInt;
begin
  L := Left.Low;
  R := Right.Low;
  L := (Swap(L) shl 8) + Left.High;
  R := (Swap(R) shl 8) + Right.High;
  Result := Integer(L) - Integer(R);
end;

class function InterfaceDefaults.Equals_Bin3(Inst: Pointer; const Left, Right: TTriple): Boolean;
begin
  Result := ((Integer(Left.High) shl 16) + Left.Low) = ((Integer(Right.High) shl 16) + Right.Low);
end;

class function InterfaceDefaults.GetHashCode_Bin3(Inst: Pointer; const Value: TTriple): Integer;
begin
  Result := Integer(Value.Bytes[0]);
  Result := Result + (Result shr 4) * 63689 + Integer(Value.Bytes[1]) * -1660269137 +
    Integer(Value.Bytes[2]) * -1092754919;
end;

class function InterfaceDefaults.Compare_Bin4(Inst: Pointer; Left, Right: Cardinal): Integer;
var
  X, Y: NativeUInt;
begin
  {$ifdef LARGEINT}
    X := Left;
    Y := Right;
    X := (Swap(X) shl 16) + Swap(X shr 16);
    Y := (Swap(Y) shl 16) + Swap(Y shr 16);
  {$else .SMALLINT}
    X := (Swap(Left) shl 16) + Swap(Left shr 16);
    Y := (Swap(Right) shl 16) + Swap(Right shr 16);
  {$endif}

  Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
end;

class function InterfaceDefaults.Compare_Bin8(Inst: Pointer; Left, Right: Int64): Integer;
var
  X, Y: NativeUInt;
begin
  {$ifdef LARGEINT}
    if (Integer(Left) = Integer(Right)) then
    begin
      Left := Left shr 32;
      Right := Right shr 32;
    end else
    begin
      Left := Cardinal(Left);
      Right := Cardinal(Right);
    end;
    X := (Swap(Left) shl 16) + Swap(Left shr 16);
    Y := (Swap(Right) shl 16) + Swap(Right shr 16);
  {$else .SMALLINT}
    X := TPoint(Left).X;
    Y := TPoint(Right).X;
    if (X = Y) then
    begin
      X := TPoint(Left).Y;
      Y := TPoint(Right).Y;
    end;
    X := (Swap(X) shl 16) + Swap(X shr 16);
    Y := (Swap(Y) shl 16) + Swap(Y shr 16);
  {$endif}

  Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
end;

class function InterfaceDefaults.Compare_Bin(const Inst: IComparerInst; Left, Right: PByte): Integer;
label
  make_result, make_result_swaped;
var
  X, Y, Count: NativeUInt;
begin
  Count := Inst.Size;
  repeat
    if (Count < SizeOf(NativeUInt)) then Break;
    X := PNativeUInt(Left)^;
    Dec(Count, SizeOf(NativeUInt));
    Y := PNativeUInt(Right)^;
    Inc(Left, SizeOf(NativeUInt));
    Inc(Right, SizeOf(NativeUInt));

    if (X <> Y) then
    begin
      {$ifdef LARGEINT}
        if (Integer(X) = Integer(Y)) then
        begin
          X := X shr 32;
          Y := Y shr 32;
        end else
        begin
          X := Cardinal(X);
          Y := Cardinal(Y);
        end;
      {$endif}

      goto make_result;
    end;
  until (False);

  // read last
  {$ifdef LARGEINT}
  if (Count >= SizeOf(Cardinal)) then
  begin
    X := PCardinal(Left)^;
    Dec(Count, SizeOf(Cardinal));
    Y := PCardinal(Right)^;
    Inc(Left, SizeOf(Cardinal));
    Inc(Right, SizeOf(Cardinal));

    if (X <> Y) then goto make_result;
  end;
  {$endif}

  case Count of
    1: begin
         X := PByte(Left)^;
         Y := PByte(Right)^;
         Result := Integer(X) - Integer(Y);
         Exit;
       end;
    2: begin
         X := Swap(PWord(Left)^);
         Y := Swap(PWord(Right)^);
         Result := Integer(X) - Integer(Y);
         Exit;
       end;
    3: begin
         X := Swap(PWord(Left)^);
         Y := Swap(PWord(Right)^);
         Inc(Left, SizeOf(Word));
         Inc(Right, SizeOf(Word));
         X := (X shl 8) or PByte(Left)^;
         Y := (Y shl 8) or PByte(Right)^;
         Result := Integer(X) - Integer(Y);
         Exit;
       end;
  else
    // 0
    Result := 0;
    Exit;
  end;

make_result:
  X := (Swap(X) shl 16) + Swap(X shr 16);
  Y := (Swap(Y) shl 16) + Swap(Y shr 16);

make_result_swaped:
  Result := Shortint(Byte(X >= Y) - Byte(X <= Y));
end;

class function InterfaceDefaults.Equals_Bin(const Inst: IEqualityComparerInst; Left, Right: PByte): Boolean;
label
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5, {$ifdef SMALLINT}cmp6, cmp7, cmp8, cmp9, cmp10,{$endif}
  done;
var
  Count: NativeUInt;
begin
  if (Left = Right) then goto done;
  Count := Inst.Size;

  // natives (40 bytes static) compare
  case Count shr {$ifdef LARGEINT}3{$else}2{$endif} of
  {$ifdef SMALLINT}
   10: goto cmp10;
    9: goto cmp9;
    8: goto cmp8;
    7: goto cmp7;
    6: goto cmp6;
  {$endif}
    5: goto cmp5;
    4: goto cmp4;
    3: goto cmp3;
    2: goto cmp2;
    1: goto cmp1;
    0: goto cmp0;
  else
    repeat
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Dec(Count, SizeOf(NativeUInt));
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    until (Count > (40 + SizeOf(NativeUInt) - 1));

    {$ifdef SMALLINT}
    cmp10:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp9:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp8:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp7:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp6:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    {$endif}
    cmp5:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp4:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp3:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp2:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp1:
      Dec(Count, SizeOf(NativeUInt));
      if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto done;
      Inc(Left, SizeOf(NativeUInt));
      Inc(Right, SizeOf(NativeUInt));
    cmp0:
  end;

  {$ifdef LARGEINT}
  if (Count and 4 <> 0) then
  begin
    if (PCardinal(Left)^ <> PCardinal(Right)^) then goto done;
    Inc(Left, SizeOf(Cardinal));
    Inc(Right, SizeOf(Cardinal));
  end;
  {$endif}

  if (Count and 3 <> 0) then
  begin
    if (Count and 2 <> 0) then
    begin
      if (PWord(Left)^ <> PWord(Right)^) then goto done;
    end;
    Dec(Left);
    Dec(Right);
    if (Left[Count] <> Right[Count]) then goto done;
  end;

// Result := True
  Left := nil;
  Right := nil;
done:
  Result := (Left = Right);
end;

class function InterfaceDefaults.GetHashCode_Bin(const Inst: IEqualityComparerInst; Value: PByte): Integer;
label
  hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, hash8, hash9, hash10;
var
  Count: NativeUInt;
  M: Integer;
begin
  Count := Inst.Size;

  if (Count >= SizeOf(Integer)) then
  begin
    Result := Integer(Count) + PInteger(@Value[Count - SizeOf(Integer)])^ * 63689;

    case (Count - 1) shr 2 of
     10: goto hash10;
      9: goto hash9;
      8: goto hash8;
      7: goto hash7;
      6: goto hash6;
      5: goto hash5;
      4: goto hash4;
      3: goto hash3;
      2: goto hash2;
      1: goto hash1;
      0: goto hash0;
    else
      Dec(Count);
      M := -1660269137;
      repeat
        Result := Result * M + PInteger(Value)^;
        Dec(Count, SizeOf(Integer));
        Inc(Value, SizeOf(Integer));
        M := M * 378551;
      until (Count <= 43);

      hash10:
        Result := Result * 631547855 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash9:
        Result := Result * -1987506439 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash8:
        Result := Result * -1653913089 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash7:
        Result := Result * -186114231 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash6:
        Result := Result * 915264303 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash5:
        Result := Result * -794603367 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash4:
        Result := Result * 135394143 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash3:
        Result := Result * 2012804575 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash2:
        Result := Result * -1092754919 + PInteger(Value)^;
        Inc(Value, SizeOf(Integer));
      hash1:
        Result := Result * -1660269137 + PInteger(Value)^;
      hash0:
    end;

    Inc(Result, ((Result shr 8) * 63689) + ((Result shr 16) * -1660269137) +
      ((Result shr 24) * -1092754919));
    Exit;
  end else
  if (Count <> 0) then
  begin
    Result := Integer(Value[0]);
    Result := Result + (Result shr 4) * 63689;
    if (Count > 1) then
    begin
      Result := Result + Integer(Value[1]) * -1660269137;
      if (Count > 2) then
      begin
        Result := Result + Integer(Value[2]) * -1092754919;
      end;
    end;
  end else
  begin
    Result := 0;
  end;
end;

{ TComparer<T> }

class function TComparer<T>.Default: IComparer<T>;
var
  Instance: ^InterfaceDefaults.IComparerInst;
begin
  { TDefaultComparer<T> constructor bug fix }
  Instance := @InterfaceDefaults.TDefaultComparer<T>.Instance;
  if (not Assigned(Instance.Compare)) then InterfaceDefaults.TDefaultComparer<T>.Initialize;
  Result := IComparer<T>(Instance);
end;

class function TComparer<T>.Construct(const Comparison: TComparison<T>): IComparer<T>;
begin
  { Much faster way to have IComparer<T> interface, than
    TDelegatedComparer<T> instance }

  IInterface(Result) := IInterface(PPointer(@Comparison)^);
end;


{ TEqualityComparer<T> }

class function TEqualityComparer<T>.Default: IEqualityComparer<T>;
var
  Instance: ^InterfaceDefaults.IEqualityComparerInst;
begin
  { TDefaultEqualityComparer<T> constructor bug fix }
  Instance := @InterfaceDefaults.TDefaultEqualityComparer<T>.Instance;
  if (not Assigned(Instance.GetHashCode)) then InterfaceDefaults.TDefaultEqualityComparer<T>.Initialize;
  Result := IEqualityComparer<T>(Instance);
end;

class function TEqualityComparer<T>.Construct(
  const EqualityComparison: TEqualityComparison<T>;
  const Hasher: THasher<T>): IEqualityComparer<T>;
begin
  Result := TDelegatedEqualityComparer<T>.Create(EqualityComparison, Hasher);
end;


{ TSingletonImplementation }

function TSingletonImplementation.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
  if GetInterface(IID, Obj) then
    Result := S_OK
  else
    Result := E_NOINTERFACE;
end;

function TSingletonImplementation._AddRef: Integer;
begin
  Result := -1;
end;

function TSingletonImplementation._Release: Integer;
begin
  Result := -1;
end;


{ Delegated Comparers }

constructor TDelegatedComparer<T>.Create(const ACompare: TComparison<T>);
begin
  FCompare := ACompare;
end;

function TDelegatedComparer<T>.Compare(const Left, Right: T): Integer;
begin
  Result := FCompare(Left, Right);
end;

constructor TDelegatedEqualityComparer<T>.Create(const AEquals: TEqualityComparison<T>; const AGetHashCode: THasher<T>);
begin
  FEquals := AEquals;
  FGetHashCode := AGetHashCode;
end;

function TDelegatedEqualityComparer<T>.Equals(const Left, Right: T): Boolean;
begin
  Result := FEquals(Left, Right);
end;

function TDelegatedEqualityComparer<T>.GetHashCode(const Value: T): Integer;
begin
  Result := FGetHashCode(Value);
end;


{ TOrdinalStringComparer }

type
  TOrdinalStringComparer = class(TStringComparer)
  public
    function Compare(const Left, Right: string): Integer; override;
    function Equals(const Left, Right: string): Boolean;
      reintroduce; overload; override;
    function GetHashCode(const Value: string): Integer;
      reintroduce; overload; override;
  end;

function TOrdinalStringComparer.Compare(const Left, Right: string): Integer;
{$ifNdef CPUINTEL}
begin
  Result := InterfaceDefaults.Compare_UStr(nil, Pointer(Left), Pointer(Right));
end;
{$else}
asm
  jmp InterfaceDefaults.Compare_UStr
end;
{$endif}

function TOrdinalStringComparer.Equals(const Left, Right: string): Boolean;
{$ifNdef CPUINTEL}
begin
  Result := InterfaceDefaults.Equals_UStr(Left, Right);
end;
{$else}
asm
  jmp InterfaceDefaults.Equals_UStr
end;
{$endif}

function TOrdinalStringComparer.GetHashCode(const Value: string): Integer;
{$ifNdef CPUINTEL}
begin
  Result := InterfaceDefaults.GetHashCode_UStr(nil, Pointer(Value));
end;
{$else}
asm
  jmp InterfaceDefaults.GetHashCode_UStr
end;
{$endif}

{ TStringComparer }

class destructor TStringComparer.Destroy;
begin
  FreeAndNil(FOrdinal);
end;

class function TStringComparer.Ordinal: TStringComparer;
begin
  if FOrdinal = nil then
    FOrdinal := TOrdinalStringComparer.Create;
  Result := TStringComparer(FOrdinal);
end;

{ TOrdinalIStringComparer }

function TOrdinalIStringComparer.Compare(const Left, Right: string): Integer;
begin
  Result := AnsiCompareText(Left, Right);
end;

function TOrdinalIStringComparer.Equals(const Left, Right: string): Boolean;
var
  Count: Integer;
begin
  if (NativeInt(Left) <> NativeInt(Right)) and (Pointer(Left) <> nil) and (Pointer(Right) <> nil) then
  begin
    Count := PInteger(NativeUInt(Left) - SizeOf(Integer))^;
    Result := (Count = PInteger(NativeUInt(Right) - SizeOf(Integer))^) and
      (0 = AnsiCompareText(Left, Right));
  end else
  begin
    Result := (NativeUInt(Left) = NativeUInt(Right));
  end;
end;

function TOrdinalIStringComparer.CharsLower(Dest, Src: PWideChar; Count: Integer): Boolean;
{$IF defined(MSWINDOWS)}
begin
  Move(Src^, Dest^, Count * SizeOf(WideChar));
  CharLowerBuff(Dest, Count);
  Result := True;
end;
{$ELSEIF defined(USE_LIBICU)}
var
  ErrorCode: UErrorCode;
begin
  ErrorCode := U_ZERO_ERROR;
  Result := (Count = u_strToLower(Dest, Count, Src, Count, UTF8CompareLocale, ErrorCode)) and
    (ErrorCode <= U_ZERO_ERROR);
end;
{$ELSEIF defined(MACOS)}
var
  MutableStringRef: CFMutableStringRef;
begin
  Move(Src^, Dest^, Count * SizeOf(WideChar));

  MutableStringRef := CFStringCreateMutableWithExternalCharactersNoCopy(kCFAllocatorDefault,
    Dest, Count, Count, kCFAllocatorNull);
  if (MutableStringRef <> nil) then
  begin
    try
      CFStringLowercase(MutableStringRef, UTF8CompareLocale);
      Result := (Count = CFStringGetLength(CFStringRef(MutableStringRef)));
    finally
      CFRelease(MutableStringRef);
    end;
  end else
  begin
    Result := False;
  end;
end;
{$ELSEIF Defined(POSIX)}
begin
  Result := False;
end;
{$IFEND POSIX}

function TOrdinalIStringComparer.GetHashCodeLower(const Value: string): Integer;
var
  S: string;
begin
  S := AnsiLowerCase(Value);
  Result := InterfaceDefaults.GetHashCode_UStr(nil, Pointer(S));
end;

function TOrdinalIStringComparer.GetHashCode(const Value: string): Integer;
var
  Count: Integer;
  Buffer: packed record
    _Padding: array[1..12] of Byte;
    Count: Integer;
    Chars: array[0..1024] of WideChar;
  end;
begin
  if (NativeInt(Value) <> 0) then
  begin
    Count := PInteger(NativeUInt(Value) - SizeOf(Integer))^;
    if (Count <= 1024) and (CharsLower(@Buffer.Chars[0], Pointer(Value), Count)) then
    begin
      Buffer.Count := Count;
      Buffer.Chars[Count] := #0;
      Result := InterfaceDefaults.GetHashCode_UStr(nil, Pointer(@Buffer.Chars[0]));
    end else
    begin
      Result := GetHashCodeLower(Value);
    end;
  end else
  begin
    Result := 0;
  end;
end;


{ TIStringComparer }

class destructor TIStringComparer.Destroy;
begin
  FreeAndNil(FOrdinal);
end;

class function TIStringComparer.Ordinal: TStringComparer;
begin
  if FOrdinal = nil then
    FOrdinal := TOrdinalIStringComparer.Create;
  Result := TStringComparer(FOrdinal);
end;


{ TEnumerator<T> }

function TEnumerator<T>.MoveNext: Boolean;
begin
  Result := DoMoveNext;
end;

// The overridden destructor that simply invokes 'inherited' is
// required to instantiate the destructor for C++ code
destructor TEnumerable<T>.Destroy;
begin
  inherited;
end;

function TEnumerable<T>.GetEnumerator: TEnumerator<T>;
begin
  Result := DoGetEnumerator;
end;

function TEnumerable<T>.ToArray: TArray<T>;
var
  Count, Buffered: NativeUInt;
  Value: T;
begin
  Count := 0;
  Buffered := 16;
  Result := nil;
  SetLength(Result, Buffered);

  for Value in Self do
  begin
    if (Count = Buffered) then
    begin
      Buffered := Buffered * 2;
      SetLength(Result, Buffered);
    end;

    Result[Count] := Value;
    Inc(Count);
  end;

  SetLength(Result, Count);
end;

{ TPair<TKey,TValue> }

constructor TPair<TKey,TValue>.Create(const AKey: TKey; const AValue: TValue);
begin
  Key := AKey;
  Value := AValue;
end;

{ TCustomDictionary<TKey,TValue>.TPairEnumerator }

constructor TCustomDictionary<TKey,TValue>.TPairEnumerator.Create(const ADictionary: TCustomDictionary<TKey,TValue>);
begin
  inherited Create;
  FIndex := -1;
  FDictionary := ADictionary;
end;

function TCustomDictionary<TKey, TValue>.TPairEnumerator.DoGetCurrent: TPair<TKey, TValue>;
begin
  Result := GetCurrent;
end;

function TCustomDictionary<TKey, TValue>.TPairEnumerator.DoMoveNext: Boolean;
begin
  Result := MoveNext;
end;

function TCustomDictionary<TKey,TValue>.TPairEnumerator.GetCurrent: TPair<TKey,TValue>;
var
  Item: PItem;
begin
  Item := @FDictionary.FItems[FIndex];
  Result.Key := Item.Key;
  Result.Value := Item.Value;
end;

function TCustomDictionary<TKey,TValue>.TPairEnumerator.MoveNext: Boolean;
var
  N, R: NativeInt;
begin
  N := FIndex + 1;
  R := Byte(N < FDictionary.FCount.Native);
  Dec(N, R);
  FIndex := N;
  Result := Boolean(R);
end;

{ TCustomDictionary<TKey,TValue>.TKeyEnumerator }

constructor TCustomDictionary<TKey,TValue>.TKeyEnumerator.Create(const ADictionary: TCustomDictionary<TKey,TValue>);
begin
  inherited Create;
  FIndex := -1;
  FDictionary := ADictionary;
end;

function TCustomDictionary<TKey, TValue>.TKeyEnumerator.DoGetCurrent: TKey;
begin
  Result := GetCurrent;
end;

function TCustomDictionary<TKey, TValue>.TKeyEnumerator.DoMoveNext: Boolean;
begin
  Result := MoveNext;
end;

function TCustomDictionary<TKey,TValue>.TKeyEnumerator.GetCurrent: TKey;
begin
  Result := FDictionary.FItems[FIndex].Key;
end;

function TCustomDictionary<TKey,TValue>.TKeyEnumerator.MoveNext: Boolean;
begin
  if (FIndex < FDictionary.FCount.Native) then
  begin
    Inc(FIndex);
    Exit(True);
  end;
  Result := False;
end;

{ TCustomDictionary<TKey,TValue>.TValueEnumerator }

constructor TCustomDictionary<TKey,TValue>.TValueEnumerator.Create(const ADictionary: TCustomDictionary<TKey,TValue>);
begin
  inherited Create;
  FIndex := -1;
  FDictionary := ADictionary;
end;

function TCustomDictionary<TKey, TValue>.TValueEnumerator.DoGetCurrent: TValue;
begin
  Result := GetCurrent;
end;

function TCustomDictionary<TKey, TValue>.TValueEnumerator.DoMoveNext: Boolean;
begin
  Result := MoveNext;
end;

function TCustomDictionary<TKey,TValue>.TValueEnumerator.GetCurrent: TValue;
begin
  Result := FDictionary.FItems[FIndex].Value;
end;

function TCustomDictionary<TKey,TValue>.TValueEnumerator.MoveNext: Boolean;
begin
  if (FIndex < FDictionary.FCount.Native) then
  begin
    Inc(FIndex);
    Exit(True);
  end;
  Result := False;
end;

{ TCustomDictionary<TKey, TValue>.TKeyCollection }

constructor TCustomDictionary<TKey, TValue>.TKeyCollection.Create(const ADictionary: TCustomDictionary<TKey, TValue>);
begin
  inherited Create;
  FDictionary := ADictionary;
end;

function TCustomDictionary<TKey, TValue>.TKeyCollection.DoGetEnumerator: TEnumerator<TKey>;
begin
  Result := GetEnumerator;
end;

function TCustomDictionary<TKey, TValue>.TKeyCollection.GetCount: Integer;
begin
  Result := FDictionary.FCount.Int;
end;

function TCustomDictionary<TKey, TValue>.TKeyCollection.GetEnumerator: TKeyEnumerator;
begin
  Result := TKeyEnumerator.Create(FDictionary);
end;

function TCustomDictionary<TKey, TValue>.TKeyCollection.ToArray: TArray<TKey>;
var
  i, Count: NativeInt;
  Src: TCustomDictionary<TKey, TValue>.PItem;
  Dest: ^TKey;
begin
  Count := Self.FDictionary.FCount.Native;

  SetLength(Result, Count);
  Src := Pointer(FDictionary.FItems);
  Dest := Pointer(Result);
  for i := 0 to Count - 1 do
  begin
    Dest^ := Src.Key;
    Inc(Src);
    Inc(Dest);
  end;
end;

{ TCustomDictionary<TKey, TValue>.TValueCollection }

constructor TCustomDictionary<TKey, TValue>.TValueCollection.Create(const ADictionary: TCustomDictionary<TKey, TValue>);
begin
  inherited Create;
  FDictionary := ADictionary;
end;

function TCustomDictionary<TKey, TValue>.TValueCollection.DoGetEnumerator: TEnumerator<TValue>;
begin
  Result := GetEnumerator;
end;

function TCustomDictionary<TKey, TValue>.TValueCollection.GetCount: Integer;
begin
  Result := FDictionary.FCount.Int;
end;

function TCustomDictionary<TKey, TValue>.TValueCollection.GetEnumerator: TValueEnumerator;
begin
  Result := TValueEnumerator.Create(FDictionary);
end;

function TCustomDictionary<TKey, TValue>.TValueCollection.ToArray: TArray<TValue>;
var
  i, Count: NativeInt;
  Src: TCustomDictionary<TKey, TValue>.PItem;
  Dest: ^TValue;
begin
  Count := Self.FDictionary.FCount.Native;

  SetLength(Result, Count);
  Src := Pointer(FDictionary.FItems);
  Dest := Pointer(Result);
  for i := 0 to Count - 1 do
  begin
    Dest^ := Src.Value;
    Inc(Src);
    Inc(Dest);
  end;
end;

{ TCustomDictionary<TKey, TValue> }

function TCustomDictionary<TKey, TValue>.DoGetEnumerator: TEnumerator<TPair<TKey, TValue>>;
begin
  Result := GetEnumerator;
end;

function TCustomDictionary<TKey, TValue>.GetEnumerator: TPairEnumerator;
begin
  Result := TPairEnumerator.Create(Self);
end;

function TCustomDictionary<TKey, TValue>.GetKeys: TKeyCollection;
begin
  if FKeyCollection = nil then
    FKeyCollection := TKeyCollection.Create(Self);
  Result := FKeyCollection;
end;

function TCustomDictionary<TKey, TValue>.GetValues: TValueCollection;
begin
  if FValueCollection = nil then
    FValueCollection := TValueCollection.Create(Self);
  Result := FValueCollection;
end;

function TCustomDictionary<TKey, TValue>.ToArray: TArray<TPair<TKey, TValue>>;
var
  i, Count: NativeInt;
  Src: PItem;
  Dest: ^TPair<TKey, TValue>;
begin
  Count := Self.FCount.Native;

  SetLength(Result, Count);
  Src := Pointer(FItems);
  Dest := Pointer(Result);
  for i := 0 to Count - 1 do
  begin
    Dest.Key := Src.Key;
    Dest.Value := Src.Value;

    Inc(Src);
    Inc(Dest);
  end;
end;

constructor TCustomDictionary<TKey, TValue>.Create(ACapacity: Integer);
begin
  inherited Create;
  FDefaultValue := Default(TValue);
  FHashTableMask := -1;
  SetNotifyMethods;

  if (ACapacity > 3) then
  begin
    SetCapacity(ACapacity);
  end else
  begin
    Rehash(4);
  end;
end;

destructor TCustomDictionary<TKey, TValue>.Destroy;
begin
  Clear;
  FKeyCollection.Free;
  FValueCollection.Free;
  ClearMethod(FInternalKeyNotify);
  ClearMethod(FInternalValueNotify);
  ClearMethod(FInternalItemNotify);
  inherited;
end;

procedure TCustomDictionary<TKey,TValue>.Rehash(NewTableCount{power of 2}: NativeInt);
type
  THashList = array[0..0] of Pointer{PItem};
var
  NewCapacity: NativeInt;
  NewHashTable: TArray<PItem>;

  i, HashTableMask, Index: NativeInt;
  Item: PItem;
  HashList: ^THashList;
begin
  // grow threshold
  NewCapacity := NewTableCount shr 1 + NewTableCount shr 2; // 75%
  if (NewCapacity < Self.FCount.Native) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  // reallocations
  if (NewTableCount < FHashTableMask + 1{Length(FHashTable)}) then
  begin
    ReallocMem(FItems, NewCapacity * SizeOf(TItem));
    SetLength(FHashTable, NewTableCount);
    NewHashTable := FHashTable;
  end else
  begin
    SetLength(NewHashTable, NewTableCount);
    ReallocMem(FItems, NewCapacity * SizeOf(TItem));
  end;

  // apply new
  FillChar(Pointer(NewHashTable)^, NewTableCount * SizeOf(Pointer), #0);
  FHashTable := NewHashTable;
  FCapacity := NewCapacity;
  FHashTableMask := NewTableCount - 1;

  // regroup items
  Item := Pointer(FItems);
  HashList := Pointer(FHashTable);
  HashTableMask := FHashTableMask;
  for i := 1 to Self.FCount.Native do
  begin
    Index := NativeInt(Cardinal(Item.HashCode)) and HashTableMask;
    Item.FNext := HashList[Index];
    HashList[Index] := Item;
    Inc(Item);
  end;
end;

procedure TCustomDictionary<TKey,TValue>.SetCapacity(ACapacity: NativeInt);
var
  Cap, NewTableCount: NativeInt;
begin
  // 75% threshold
  Cap := 3;
  if (Cap < ACapacity) then
  repeat
    Cap := Cap shl 1;
  until (Cap < 0) or (Cap >= ACapacity);

  // power of 2
  NewTableCount := (Cap and (Cap - 1)) shl 1;
  if (NewTableCount = FHashTableMask + 1{Length(FHashTable)}) then
    Exit;
  if (NativeUInt(NewTableCount) > NativeUInt(High(Integer)){Integer(NewTableCount) < 0}) then
    OutOfMemoryError;

  // rehash
  Rehash(NewTableCount);
end;

function TCustomDictionary<TKey,TValue>.Grow: TCustomDictionary<TKey,TValue>;
begin
  Rehash((FHashTableMask + 1){Length(FHashTable)} * 2);
  Result := Self;
end;

procedure TCustomDictionary<TKey, TValue>.TrimExcess;
begin
  SetCapacity(FCount.Int);
end;

procedure TCustomDictionary<TKey, TValue>.Clear;
begin
  if (FCount.Native <> 0) then
  begin
    Self.DoCleanupItems(Pointer(FItems), FCount.Native);
  end;

  FCount.Native := 0;
  if (FHashTableMask + 1 = 4) then
  begin
    FillChar(Pointer(FHashTable), 4 * SizeOf(Pointer), #0);
  end else
  begin
    Rehash(4);
  end;
end;

procedure TCustomDictionary<TKey, TValue>.DoCleanupItems(Item: PItem; Count: NativeInt);
{$ifdef SMARTGENERICS}
type
  TKeyRec = TRAIIHelper.TData16;
  PKeyRec = ^TKeyRec;
  TValueRec = TRAIIHelper.TData16<TKey>;
  PValueRec = ^TValueRec;
var
  i: NativeInt;
  VType: Integer;
{$else}
var
  i: NativeInt;
{$endif}
  StoredItem: PItem;
begin
  // Key/Value notifies (cnRemoved)
  StoredItem := Item;
  if Assigned(FInternalKeyNotify) then
  begin
    if Assigned(FInternalValueNotify) then
    begin
      if (TMethod(FInternalItemNotify).Code = @TCustomDictionary<TKey,TValue>.ItemNotifyCaller) then
      begin
        for i := 1 to Count do
        begin
          Self.KeyNotify(Item.Key, cnRemoved);
          Self.ValueNotify(Item.Value, cnRemoved);
          Inc(Item);
        end;
      end else
      begin
        for i := 1 to Count do
        begin
          FInternalKeyNotify(Self, Item.Key, cnRemoved);
          FInternalValueNotify(Self, Item.Value, cnRemoved);
          Inc(Item);
        end;
      end;
    end else
    begin
      // Key
      if (TMethod(FInternalKeyNotify).Code = @TCustomDictionary<TKey,TKey>.KeyNotifyCaller) then
      begin
        for i := 1 to Count do
        begin
          Self.KeyNotify(Item.Key, cnRemoved);
          Inc(Item);
        end;
      end else
      begin
        for i := 1 to Count do
        begin
          FInternalKeyNotify(Self, Item.Key, cnRemoved);
          Inc(Item);
        end;
      end;
    end;
  end else
  if Assigned(FInternalValueNotify) then
  begin
    // Value
    if (TMethod(FInternalValueNotify).Code = @TDictionary<TKey,TValue>.ValueNotifyCaller) then
    begin
      for i := 1 to Count do
      begin
        Self.ValueNotify(Item.Value, cnRemoved);
        Inc(Item);
      end;
    end else
    begin
      for i := 1 to Count do
      begin
        FInternalValueNotify(Self, Item.Value, cnRemoved);
        Inc(Item);
      end;
    end;
  end;

  // finalize array
  Item := StoredItem;
  {$ifdef SMARTGENERICS}
  if (System.IsManagedType(TKey) or System.HasWeakRef(TKey)) then
  {$else}
  if Assigned(TRAIIHelper<TKey>.FOptions.ClearProc) then
  {$endif}
  begin
    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(TValue) or System.HasWeakRef(TValue)) then
    {$else}
    if Assigned(TRAIIHelper<TValue>.FOptions.ClearProc) then
    {$endif}
    begin
      // Keys + Values
      for i := 1 to Count do
      begin
        {$ifdef SMARTGENERICS}
        case GetTypeKind(TKey) of
          {$ifdef AUTOREFCOUNT}
          tkClass,
          {$endif}
          tkWString, tkLString, tkUString, tkInterface, tkDynArray:
          begin
            if (PKeyRec(Item).Native <> 0) then
            case GetTypeKind(TKey) of
              {$ifdef AUTOREFCOUNT}
              tkClass:
              begin
                TRAIIHelper.RefObjClear(@PKeyRec(Item).Native);
              end;
              {$endif}
              {$ifdef MSWINDOWS}
              tkWString:
              begin
                TRAIIHelper.WStrClear(@PKeyRec(Item).Native);
              end;
              {$else}
              tkWString,
              {$endif}
              tkLString, tkUString:
              begin
                TRAIIHelper.ULStrClear(@PKeyRec(Item).Native);
              end;
              tkInterface:
              begin
                IInterface(PKeyRec(Item).Native)._Release;
              end;
              tkDynArray:
              begin
                TRAIIHelper.DynArrayClear(@PKeyRec(Item).Native, TypeInfo(TKey));
              end;
            end;
          end;
          {$ifdef WEAKREF}
          tkMethod:
          begin
            if (PKeyRec(Item).Method.Data <> nil) then
              TRAIIHelper.WeakMethodClear(@PKeyRec(Item).Method.Data);
          end;
          {$endif}
          tkVariant:
          begin
            VType := PKeyRec(Item).VarData.VType;
            if (VType and TRAIIHelper.varDeepData <> 0) then
            case VType of
              varBoolean, varUnknown+1..varUInt64: ;
            else
              System.VarClear(Variant(PKeyRec(Item).VarData));
            end;
          end;
        else
          TRAIIHelper<TKey>.FOptions.ClearProc(TRAIIHelper<TKey>.FOptions, @Item.FKey);
        end;
        {$else}
        TRAIIHelper<TKey>.FOptions.ClearProc(TRAIIHelper<TKey>.FOptions, @Item.FKey);
        {$endif}

        {$ifdef SMARTGENERICS}
        case GetTypeKind(TValue) of
          {$ifdef AUTOREFCOUNT}
          tkClass,
          {$endif}
          tkWString, tkLString, tkUString, tkInterface, tkDynArray:
          begin
            if (PValueRec(Item).Native <> 0) then
            case GetTypeKind(TValue) of
              {$ifdef AUTOREFCOUNT}
              tkClass:
              begin
                TRAIIHelper.RefObjClear(@PValueRec(Item).Native);
              end;
              {$endif}
              {$ifdef MSWINDOWS}
              tkWString:
              begin
                TRAIIHelper.WStrClear(@PValueRec(Item).Native);
              end;
              {$else}
              tkWString,
              {$endif}
              tkLString, tkUString:
              begin
                TRAIIHelper.ULStrClear(@PValueRec(Item).Native);
              end;
              tkInterface:
              begin
                IInterface(PValueRec(Item).Native)._Release;
              end;
              tkDynArray:
              begin
                TRAIIHelper.DynArrayClear(@PValueRec(Item).Native, TypeInfo(TValue));
              end;
            end;
          end;
          {$ifdef WEAKREF}
          tkMethod:
          begin
            if (PValueRec(Item).Method.Data <> nil) then
              TRAIIHelper.WeakMethodClear(@PValueRec(Item).Method.Data);
          end;
          {$endif}
          tkVariant:
          begin
            VType := PValueRec(Item).VarData.VType;
            if (VType and TRAIIHelper.varDeepData <> 0) then
            case VType of
              varBoolean, varUnknown+1..varUInt64: ;
            else
              System.VarClear(Variant(PValueRec(Item).VarData));
            end;
          end;
        else
          TRAIIHelper<TValue>.FOptions.ClearProc(TRAIIHelper<TValue>.FOptions, @Item.FValue);
        end;
        {$else}
        TRAIIHelper<TValue>.FOptions.ClearProc(TRAIIHelper<TValue>.FOptions, @Item.FValue);
        {$endif}

        Inc(Item);
      end;
    end else
    begin
      // Keys only
      TRAIIHelper<TKey>.ClearArray(@Item.Key, Count, SizeOf(TItem));
    end;
  end else
  {$ifdef SMARTGENERICS}
  if (System.IsManagedType(TValue) or System.HasWeakRef(TValue)) then
  {$else}
  if Assigned(TRAIIHelper<TValue>.FOptions.ClearProc) then
  {$endif}
  begin
    // Values only
    TRAIIHelper<TValue>.ClearArray(@Item.Value, Count, SizeOf(TItem));
  end;
end;

function TCustomDictionary<TKey, TValue>.NewItem: PItem;
label
  start;
type
  TKeyRec = TRAIIHelper.TData16;
  PKeyRec = ^TKeyRec;
  TValueRec = TRAIIHelper.TData16<TKey>;
  PValueRec = ^TValueRec;
var
  Instance: TCustomDictionary<TKey, TValue>;
  Count: NativeInt;
  Null: NativeUInt;
begin
  Instance := Self;
start:
  Count := Instance.FCount.Native;
  if (Count <> Instance.FCapacity) then
  begin
    Instance.FCount.Native := Count + 1;
    Result := @Instance.FItems[Count];

    if ((SizeOf(TKey) >= SizeOf(NativeInt)) and (SizeOf(TKey) <= 16)) or
      ((SizeOf(TValue) >= SizeOf(NativeInt)) and (SizeOf(TValue) <= 16)) then
    begin
      {$ifdef SMARTGENERICS}
      if (System.IsManagedType(TItem) or System.HasWeakRef(TItem)) then
      {$endif}
      Null := 0;
    end;

    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(TKey) or System.HasWeakRef(TKey)) then
    {$else}
    if (SizeOf(TKey) >= SizeOf(NativeInt)) then
    {$endif}
    begin
      if (SizeOf(TKey) = SizeOf(NativeInt)){$ifdef SMARTGENERICS}or (GetTypeKind(TKey) = tkVariant){$endif} then
      begin
        PKeyRec(Result).Natives[0] := Null;
      end else
      if ((SizeOf(TKey) >= SizeOf(NativeInt)) and (SizeOf(TKey) <= 16)) then
      begin
        PKeyRec(Result).Natives[0] := Null;
        if (SizeOf(TKey) >= 2 * SizeOf(NativeInt)) then PKeyRec(Result).Natives[1] := Null;
        if (SizeOf(TKey) >= 3 * SizeOf(NativeInt)) then PKeyRec(Result).Natives[2] := Null;
        if (SizeOf(TKey)  = 4 * SizeOf(NativeInt)) then PKeyRec(Result).Natives[3] := Null;
      end else
      TRAIIHelper<TKey>.Init(@Result.FKey);
    end;

    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(TValue) or System.HasWeakRef(TValue)) then
    {$else}
    if (SizeOf(TValue) >= SizeOf(NativeInt)) then
    {$endif}
    begin
      if (SizeOf(TValue) = SizeOf(NativeInt)){$ifdef SMARTGENERICS}or (GetTypeKind(TValue) = tkVariant){$endif} then
      begin
        PValueRec(Result).Natives[0] := Null;
      end else
      if ((SizeOf(TValue) >= SizeOf(NativeInt)) and (SizeOf(TValue) <= 16)) then
      begin
        PValueRec(Result).Natives[0] := Null;
        if (SizeOf(TValue) >= 2 * SizeOf(NativeInt)) then PValueRec(Result).Natives[1] := Null;
        if (SizeOf(TValue) >= 3 * SizeOf(NativeInt)) then PValueRec(Result).Natives[2] := Null;
        if (SizeOf(TValue)  = 4 * SizeOf(NativeInt)) then PValueRec(Result).Natives[3] := Null;
      end else
      TRAIIHelper<TValue>.Init(@Result.FValue);
    end;

    Exit;
  end else
  begin
    Instance := Instance.Grow;
    goto start;
  end;
end;

procedure TCustomDictionary<TKey, TValue>.DisposeItem(Item: PItem);
type
  TKeyRec = TRAIIHelper.TData16;
  PKeyRec = ^TKeyRec;
  TValueRec = TRAIIHelper.TData16<TKey>;
  PValueRec = ^TValueRec;
var
  {$ifdef SMARTGENERICS}
  VType: Integer;
  {$endif}
  Count: NativeInt;
  Parent: ^PItem;
  TopItem, Current: PItem;
  Index: NativeInt;
begin
  // Key
  {$ifdef SMARTGENERICS}
  case GetTypeKind(TKey) of
    {$ifdef AUTOREFCOUNT}
    tkClass,
    {$endif}
    tkWString, tkLString, tkUString, tkInterface, tkDynArray:
    begin
      if (PKeyRec(Item).Native <> 0) then
      case GetTypeKind(TKey) of
        {$ifdef AUTOREFCOUNT}
        tkClass:
        begin
          TRAIIHelper.RefObjClear(@PKeyRec(Item).Native);
        end;
        {$endif}
        {$ifdef MSWINDOWS}
        tkWString:
        begin
          TRAIIHelper.WStrClear(@PKeyRec(Item).Native);
        end;
        {$else}
        tkWString,
        {$endif}
        tkLString, tkUString:
        begin
          TRAIIHelper.ULStrClear(@PKeyRec(Item).Native);
        end;
        tkInterface:
        begin
          IInterface(PKeyRec(Item).Native)._Release;
        end;
        tkDynArray:
        begin
          TRAIIHelper.DynArrayClear(@PKeyRec(Item).Native, TypeInfo(TKey));
        end;
      end;
    end;
    {$ifdef WEAKREF}
    tkMethod:
    begin
      if (PKeyRec(Item).Method.Data <> nil) then
        TRAIIHelper.WeakMethodClear(@PKeyRec(Item).Method.Data);
    end;
    {$endif}
    tkVariant:
    begin
      VType := PKeyRec(Item).VarData.VType;
      if (VType and TRAIIHelper.varDeepData <> 0) then
      case VType of
        varBoolean, varUnknown+1..varUInt64: ;
      else
        System.VarClear(Variant(PKeyRec(Item).VarData));
      end;
    end
  else
    TRAIIHelper<TKey>.Clear(@Item.FKey);
  end;
  {$else}
  TRAIIHelper<TKey>.Clear(@Item.FKey);
  {$endif}

  // Value
  {$ifdef SMARTGENERICS}
  case GetTypeKind(TValue) of
    {$ifdef AUTOREFCOUNT}
    tkClass,
    {$endif}
    tkWString, tkLString, tkUString, tkInterface, tkDynArray:
    begin
      if (PValueRec(Item).Native <> 0) then
      case GetTypeKind(TValue) of
        {$ifdef AUTOREFCOUNT}
        tkClass:
        begin
          TRAIIHelper.RefObjClear(@PValueRec(Item).Native);
        end;
        {$endif}
        {$ifdef MSWINDOWS}
        tkWString:
        begin
          TRAIIHelper.WStrClear(@PValueRec(Item).Native);
        end;
        {$else}
        tkWString,
        {$endif}
        tkLString, tkUString:
        begin
          TRAIIHelper.ULStrClear(@PValueRec(Item).Native);
        end;
        tkInterface:
        begin
          IInterface(PValueRec(Item).Native)._Release;
        end;
        tkDynArray:
        begin
          TRAIIHelper.DynArrayClear(@PValueRec(Item).Native, TypeInfo(TValue));
        end;
      end;
    end;
    {$ifdef WEAKREF}
    tkMethod:
    begin
      if (PValueRec(Item).Method.Data <> nil) then
        TRAIIHelper.WeakMethodClear(@PValueRec(Item).Method.Data);
    end;
    {$endif}
    tkVariant:
    begin
      VType := PValueRec(Item).VarData.VType;
      if (VType and TRAIIHelper.varDeepData <> 0) then
      case VType of
        varBoolean, varUnknown+1..varUInt64: ;
      else
        System.VarClear(Variant(PValueRec(Item).VarData));
      end;
    end
  else
    TRAIIHelper<TValue>.Clear(@Item.FValue);
  end;
  {$else}
  TRAIIHelper<TValue>.Clear(@Item.FValue);
  {$endif}

  Count := Self.FCount.Native;
  Dec(Count);
  Self.FCount.Native := Count;
  TopItem := @FItems[Count];
  if (Item <> TopItem) then
  begin
    // change TopItem.Parent.Next --> Item
    Parent := @FHashTable[NativeInt(Cardinal(TopItem.HashCode)) and FHashTableMask];
    repeat
      Current := Parent^;
      if (Current = TopItem) then
      begin
        Parent^ := Item;
        Break;
      end;
      Parent := @Current.FNext;
    until (False);

    // move TopItem --> Item
    case SizeOf(TItem) of
      1: TRAIIHelper.T1(Pointer(Item)^) := TRAIIHelper.T1(Pointer(TopItem)^);
      2: TRAIIHelper.T2(Pointer(Item)^) := TRAIIHelper.T2(Pointer(TopItem)^);
      3: TRAIIHelper.T3(Pointer(Item)^) := TRAIIHelper.T3(Pointer(TopItem)^);
      4: TRAIIHelper.T4(Pointer(Item)^) := TRAIIHelper.T4(Pointer(TopItem)^);
      5: TRAIIHelper.T5(Pointer(Item)^) := TRAIIHelper.T5(Pointer(TopItem)^);
      6: TRAIIHelper.T6(Pointer(Item)^) := TRAIIHelper.T6(Pointer(TopItem)^);
      7: TRAIIHelper.T7(Pointer(Item)^) := TRAIIHelper.T7(Pointer(TopItem)^);
      8: TRAIIHelper.T8(Pointer(Item)^) := TRAIIHelper.T8(Pointer(TopItem)^);
      9: TRAIIHelper.T9(Pointer(Item)^) := TRAIIHelper.T9(Pointer(TopItem)^);
     10: TRAIIHelper.T10(Pointer(Item)^) := TRAIIHelper.T10(Pointer(TopItem)^);
     11: TRAIIHelper.T11(Pointer(Item)^) := TRAIIHelper.T11(Pointer(TopItem)^);
     12: TRAIIHelper.T12(Pointer(Item)^) := TRAIIHelper.T12(Pointer(TopItem)^);
     13: TRAIIHelper.T13(Pointer(Item)^) := TRAIIHelper.T13(Pointer(TopItem)^);
     14: TRAIIHelper.T14(Pointer(Item)^) := TRAIIHelper.T14(Pointer(TopItem)^);
     15: TRAIIHelper.T15(Pointer(Item)^) := TRAIIHelper.T15(Pointer(TopItem)^);
     16: TRAIIHelper.T16(Pointer(Item)^) := TRAIIHelper.T16(Pointer(TopItem)^);
     17: TRAIIHelper.T17(Pointer(Item)^) := TRAIIHelper.T17(Pointer(TopItem)^);
     18: TRAIIHelper.T18(Pointer(Item)^) := TRAIIHelper.T18(Pointer(TopItem)^);
     19: TRAIIHelper.T19(Pointer(Item)^) := TRAIIHelper.T19(Pointer(TopItem)^);
     20: TRAIIHelper.T20(Pointer(Item)^) := TRAIIHelper.T20(Pointer(TopItem)^);
     21: TRAIIHelper.T21(Pointer(Item)^) := TRAIIHelper.T21(Pointer(TopItem)^);
     22: TRAIIHelper.T22(Pointer(Item)^) := TRAIIHelper.T22(Pointer(TopItem)^);
     23: TRAIIHelper.T23(Pointer(Item)^) := TRAIIHelper.T23(Pointer(TopItem)^);
     24: TRAIIHelper.T24(Pointer(Item)^) := TRAIIHelper.T24(Pointer(TopItem)^);
     25: TRAIIHelper.T25(Pointer(Item)^) := TRAIIHelper.T25(Pointer(TopItem)^);
     26: TRAIIHelper.T26(Pointer(Item)^) := TRAIIHelper.T26(Pointer(TopItem)^);
     27: TRAIIHelper.T27(Pointer(Item)^) := TRAIIHelper.T27(Pointer(TopItem)^);
     28: TRAIIHelper.T28(Pointer(Item)^) := TRAIIHelper.T28(Pointer(TopItem)^);
     29: TRAIIHelper.T29(Pointer(Item)^) := TRAIIHelper.T29(Pointer(TopItem)^);
     30: TRAIIHelper.T30(Pointer(Item)^) := TRAIIHelper.T30(Pointer(TopItem)^);
     31: TRAIIHelper.T31(Pointer(Item)^) := TRAIIHelper.T31(Pointer(TopItem)^);
     32: TRAIIHelper.T32(Pointer(Item)^) := TRAIIHelper.T32(Pointer(TopItem)^);
     33: TRAIIHelper.T33(Pointer(Item)^) := TRAIIHelper.T33(Pointer(TopItem)^);
     34: TRAIIHelper.T34(Pointer(Item)^) := TRAIIHelper.T34(Pointer(TopItem)^);
     35: TRAIIHelper.T35(Pointer(Item)^) := TRAIIHelper.T35(Pointer(TopItem)^);
     36: TRAIIHelper.T36(Pointer(Item)^) := TRAIIHelper.T36(Pointer(TopItem)^);
     37: TRAIIHelper.T37(Pointer(Item)^) := TRAIIHelper.T37(Pointer(TopItem)^);
     38: TRAIIHelper.T38(Pointer(Item)^) := TRAIIHelper.T38(Pointer(TopItem)^);
     39: TRAIIHelper.T39(Pointer(Item)^) := TRAIIHelper.T39(Pointer(TopItem)^);
     40: TRAIIHelper.T40(Pointer(Item)^) := TRAIIHelper.T40(Pointer(TopItem)^);
    else
      System.Move(TopItem^, Item^, SizeOf(TItem));
    end;
  end;
end;

function TCustomDictionary<TKey, TValue>.ContainsValue(const Value: TValue): Boolean;
label
  {$ifdef SMARTGENERICS}
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5, {$ifdef SMALLINT}cmp6, cmp7, cmp8, cmp9, cmp10,{$endif}
  next_item,
  {$endif}
  done;
type
  TData16 = TRAIIHelper.TData16;
  PData16 = ^TData16;
  PValue = ^TValue;
var
  i: NativeInt;
  Item: PValue;
  {$ifdef SMARTGENERICS}
  Left, Right: PByte;
  Count, Offset: NativeUInt;
  {$else}
  Comparer: IEqualityComparer<TValue>;
  ComparerEquals: function(const Left, Right: TValue): Boolean of object;
  {$endif}
begin
  Item := @Self.FItems[0].FValue;

  {$ifdef SMARTGENERICS}
  begin
    for i := 1 to Self.FCount.Native do
    begin
      if (GetTypeKind(TValue) = tkVariant) then
      begin
        if (not InterfaceDefaults.Equals_Var(nil, PVarData(@Value), PVarData(Item))) then goto next_item;
      end else
      if (GetTypeKind(TValue) = tkClass) then
      begin
        Left := PPointer(@Value)^;
        Right := PPointer(Item)^;

        if (Assigned(Left)) then
        begin
          if (PPointer(Left^)[vmtEquals div SizeOf(Pointer)] = @TObject.Equals) then
          begin
            if (Left <> Right) then goto next_item;
          end else
          begin
            if (not TObject(Left).Equals(TObject(Right))) then goto next_item;
          end;
        end else
        begin
          if (Right <> nil) then goto next_item;
        end;
      end else
      if (GetTypeKind(TValue) = tkFloat) then
      begin
        case SizeOf(TValue) of
          4:
          begin
            if (PSingle(@Value)^ <> PSingle(Result)^) then goto next_item;
          end;
          10:
          begin
            if (PExtended(@Value)^ <> PExtended(Result)^) then goto next_item;
          end;
        else
        {$ifdef LARGEINT}
          if (PInt64(@Value)^ <> PInt64(Result)^) then
        {$else .SMALLINT}
          if ((PPoint(@Value).X - PPoint(Result).X) or (PPoint(@Value).Y - PPoint(Result).Y) <> 0) then
        {$endif}
          begin
            if (InterfaceDefaults.TDefaultEqualityComparer<TValue>.Instance.Size < 0) then goto next_item;
            if (PDouble(@Value)^ <> PDouble(Result)^) then goto next_item;
          end;
        end;
      end else
      if (not (GetTypeKind(TValue) in [tkDynArray, tkString, tkLString, tkWString, tkUString])) and
        (SizeOf(TValue) <= 16) then
      begin
        // small binary
        if (SizeOf(TValue) <> 0) then
        with PData16(@Value)^ do
        begin
          if (SizeOf(TValue) >= SizeOf(Integer)) then
          begin
            if (SizeOf(TValue) >= SizeOf(Int64)) then
            begin
              {$ifdef LARGEINT}
              if (Int64s[0] <> PData16(Item).Int64s[0]) then goto next_item;
              {$else}
              if (Integers[0] <> PData16(Item).Integers[0]) then goto next_item;
              if (Integers[1] <> PData16(Item).Integers[1]) then goto next_item;
              {$endif}

              if (SizeOf(TValue) = 16) then
              begin
                {$ifdef LARGEINT}
                if (Int64s[1] <> PData16(Item).Int64s[1]) then goto next_item;
                {$else}
                if (Integers[2] <> PData16(Item).Integers[2]) then goto next_item;
                if (Integers[3] <> PData16(Item).Integers[3]) then goto next_item;
                {$endif}
              end else
              if (SizeOf(TValue) >= 12) then
              begin
                if (Integers[2] <> PData16(Item).Integers[2]) then goto next_item;
              end;
            end else
            begin
              if (Integers[0] <> PData16(Item).Integers[0]) then goto next_item;
            end;
          end;

          if (SizeOf(TValue) and 2 <> 0) then
          begin
            if (Words[(SizeOf(TValue) and -4) shr 1] <> PData16(Item).Words[(SizeOf(TValue) and -4) shr 1]) then goto next_item;
          end;
          if (SizeOf(TValue) and 1 <> 0) then
          begin
            if (Bytes[SizeOf(TValue) and -2] <> PData16(Item).Bytes[SizeOf(TValue) and -2]) then goto next_item;
          end;
        end;
      end else
      begin
        if (GetTypeKind(TValue) in [tkDynArray, tkString, tkLString, tkWString, tkUString]) then
        begin
          // dynamic size
          if (GetTypeKind(TValue) = tkString) then
          begin
            Left := Pointer(@Value);
            Right := Pointer(Item);
            if (PValue(Left) = {Right}Item) then goto cmp0;
            Count := Left^;
            if (Count <> Right^) then goto next_item;
            if (Count = 0) then goto cmp0;
            // compare last bytes
            if (Left[Count] <> Right[Count]) then goto next_item;
          end else
          // if (GetTypeKind(TValue) in [tkDynArray, tkLString, tkWString, tkUString]) then
          begin
            Left := PPointer(@Value)^;
            Right := PPointer(Item)^;
            if (Left = Right) then goto cmp0;
            if (Left = nil) then
            begin
              {$ifdef MSWINDOWS}
              if (GetTypeKind(TValue) = tkWString) then
              begin
                Dec(Right, SizeOf(Cardinal));
                if (PCardinal(Right)^ = 0) then goto cmp0;
              end;
              {$endif}
              goto next_item;
            end;
            if (Right = nil) then
            begin
              {$ifdef MSWINDOWS}
              if (GetTypeKind(TValue) = tkWString) then
              begin
                Dec(Left, SizeOf(Cardinal));
                if (PCardinal(Left)^ = 0) then goto cmp0;
              end;
              {$endif}
              goto next_item;
            end;

            if (GetTypeKind(TValue) = tkDynArray) then
            begin
              Dec(Left, SizeOf(NativeUInt));
              Dec(Right, SizeOf(NativeUInt));
              Count := PNativeUInt(Left)^;
              if (Count <> PNativeUInt(Right)^) then goto next_item;
              NativeInt(Count) := NativeInt(Count) * InterfaceDefaults.TDefaultEqualityComparer<TKey>.Instance.Size;
              Inc(Left, SizeOf(NativeUInt));
              Inc(Right, SizeOf(NativeUInt));
            end else
            // if (GetTypeKind(TValue) in [tkLString, tkWString, tkUString]) then
            begin
              Dec(Left, SizeOf(Cardinal));
              Dec(Right, SizeOf(Cardinal));
              Count := PCardinal(Left)^;
              if (Cardinal(Count) <> PCardinal(Right)^) then goto next_item;
              Inc(Left, SizeOf(Cardinal));
              Inc(Right, SizeOf(Cardinal));
            end;
          end;

          // compare last (after cardinal) words
          if (GetTypeKind(TValue) in [tkDynArray, tkString, tkLString]) then
          begin
            if (GetTypeKind(TValue) in [tkString, tkLString]) {ByteStrings + 2} then
            begin
              Inc(Count);
            end;
            if (Count and 2 <> 0) then
            begin
              Offset := Count and -4;
              Inc(Left, Offset);
              Inc(Right, Offset);
              if (PWord(Left)^ <> PWord(Right)^) then goto next_item;
              Offset := Count;
              Offset := Offset and -4;
              Dec(Left, Offset);
              Dec(Right, Offset);
            end;
          end else
          // modify Count to have only cardinals to compare
          // if (GetTypeKind(TValue) in [tkWString, tkUString]) {UnicodeStrings + 2} then
          begin
            {$ifdef MSWINDOWS}
            if (GetTypeKind(TValue) = tkWString) then
            begin
              if (Count = 0) then goto cmp0;
            end else
            {$endif}
            begin
              Inc(Count, Count);
            end;
            Inc(Count, 2);
          end;

          {$ifdef LARGEINT}
          if (Count and 4 <> 0) then
          begin
            Offset := Count and -8;
            Inc(Left, Offset);
            Inc(Right, Offset);
            if (PCardinal(Left)^ <> PCardinal(Right)^) then goto next_item;
            Dec(Left, Offset);
            Dec(Right, Offset);
          end;
          {$endif}
        end else
        begin
          // non-dynamic (constant) size binary > 16
          if (SizeOf(TValue) and {$ifdef LARGEINT}7{$else}3{$endif} <> 0) then
          with PData16(@Value)^ do
          begin
            {$ifdef LARGEINT}
            if (SizeOf(TValue) and 4 <> 0) then
            begin
              if (Integers[(SizeOf(TValue) and -8) shr 2] <> PData16(Item).Integers[(SizeOf(TValue) and -8) shr 2]) then goto next_item;
            end;
            {$endif}
            if (SizeOf(TValue) and 2 <> 0) then
            begin
              if (Words[(SizeOf(TValue) and -4) shr 1] <> PData16(Item).Words[(SizeOf(TValue) and -4) shr 1]) then goto next_item;
            end;
            if (SizeOf(TValue) and 1 <> 0) then
            begin
              if (Bytes[SizeOf(TValue) and -2] <> PData16(Item).Bytes[SizeOf(TValue) and -2]) then goto next_item;
            end;
          end;
          Left := Pointer(@Value);
          Right := Pointer(Item);
          Count := SizeOf(TValue);
        end;

        // natives (40 bytes static) compare
        Count := Count shr {$ifdef LARGEINT}3{$else}2{$endif};
        case Count of
        {$ifdef SMALLINT}
         10: goto cmp10;
          9: goto cmp9;
          8: goto cmp8;
          7: goto cmp7;
          6: goto cmp6;
        {$endif}
          5: goto cmp5;
          4: goto cmp4;
          3: goto cmp3;
          2: goto cmp2;
          1: goto cmp1;
          0: goto cmp0;
        else
          repeat
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Dec(Count);
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          until (Count = {$ifdef LARGEINT}5{$else}10{$endif});

          {$ifdef SMALLINT}
          cmp10:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp9:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp8:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp7:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp6:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          {$endif}
          cmp5:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp4:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp3:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp2:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp1:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          cmp0:
        end;
      end;
      goto done;

    next_item:
      Inc(NativeUInt(Item), SizeOf(TItem));
    end;

    Result := False;
    Exit;
  end;
  {$else}
  begin
    Comparer := TEqualityComparer<TValue>.Default;
    TMethod(ComparerEquals) := IntfMethod(Pointer(Comparer), 3);
    for i := 1 to Self.FCount.Native do
    begin
      if (ComparerEquals(Value, Item^)) then goto done;
      Inc(NativeUInt(Item), SizeOf(TItem));
    end;
    ClearMethod(ComparerEquals);
    Result := False;
    Exit;
  end;
  {$endif}

done:
  {$ifNdef SMARTGENERICS}
  ClearMethod(ComparerEquals);
  {$endif}
  Result := True;
end;

class function TCustomDictionary<TKey, TValue>.IntfMethod(Intf: Pointer; MethodNum: NativeUInt): TMethod;
begin
  Result.Data := Intf;
  Result.Code := PPointer(PNativeUInt(Intf)^ + MethodNum * SizeOf(Pointer))^;
end;

class procedure TCustomDictionary<TKey, TValue>.ClearMethod(var Method);
begin
  {$ifdef WEAKREF}
    TMethod(Method).Data := nil;
  {$endif}
end;

procedure TCustomDictionary<TKey,TValue>.SetKeyNotify(const Value: TCollectionNotifyEvent<TKey>);
begin
  if (TMethod(FOnKeyNotify) <> TMethod(Value)) then
  begin
    FOnKeyNotify := Value;
    SetNotifyMethods;
  end;
end;

procedure TCustomDictionary<TKey,TValue>.SetValueNotify(const Value: TCollectionNotifyEvent<TValue>);
begin
  if (TMethod(FOnValueNotify) <> TMethod(Value)) then
  begin
    FOnValueNotify := Value;
    SetNotifyMethods;
  end;
end;

procedure TCustomDictionary<TKey,TValue>.KeyNotify(const Key: TKey; Action: TCollectionNotification);
begin
  if Assigned(FOnKeyNotify) then
    FOnKeyNotify(Self, Key, Action);
end;

procedure TCustomDictionary<TKey,TValue>.ValueNotify(const Value: TValue; Action: TCollectionNotification);
begin
  if Assigned(FOnValueNotify) then
    FOnValueNotify(Self, Value, Action);
end;

procedure TCustomDictionary<TKey,TValue>.KeyNotifyCaller(Sender: TObject; const Item: TKey; Action: TCollectionNotification);
begin
  Self.KeyNotify(Item, Action);
end;

procedure TCustomDictionary<TKey,TValue>.ValueNotifyCaller(Sender: TObject; const Item: TValue; Action: TCollectionNotification);
begin
  Self.ValueNotify(Item, Action);
end;

procedure TCustomDictionary<TKey,TValue>.ItemNotifyCaller(const Item: TItem; Action: TCollectionNotification);
begin
  Self.KeyNotify(Item.Key, Action);
  Self.ValueNotify(Item.Value, Action);
end;

procedure TCustomDictionary<TKey,TValue>.ItemNotifyEvents(const Item: TItem; Action: TCollectionNotification);
begin
  Self.FInternalKeyNotify(Self, Item.Key, Action);
  Self.FInternalValueNotify(Self, Item.Value, Action);
end;

procedure TCustomDictionary<TKey,TValue>.ItemNotifyKey(const Item: TItem; Action: TCollectionNotification);
begin
  Self.FInternalKeyNotify(Self, Item.Key, Action);
end;

procedure TCustomDictionary<TKey,TValue>.ItemNotifyValue(const Item: TItem; Action: TCollectionNotification);
begin
  Self.FInternalValueNotify(Self, Item.Value, Action);
end;

procedure TCustomDictionary<TKey,TValue>.SetNotifyMethods;
var
  VMTKeyNotify: procedure(const Key: TKey; Action: TCollectionNotification) of object;
  VMTValueNotify: procedure(const Value: TValue; Action: TCollectionNotification) of object;
begin
  // FInternalKeyNotify, FInternalValueNotify
  VMTKeyNotify := Self.KeyNotify;
  VMTValueNotify := Self.ValueNotify;
  if (TMethod(VMTKeyNotify).Code <> @TCustomDictionary<TKey,TValue>.KeyNotify) then
  begin
    // FInternalKeyNotify := Self.KeyNotifyCaller;
    TMethod(FInternalKeyNotify).Data := Pointer(Self);
    TMethod(FInternalKeyNotify).Code := @TCustomDictionary<TKey,TValue>.KeyNotifyCaller;
  end else
  begin
    TMethod(FInternalKeyNotify) := TMethod(Self.FOnKeyNotify);
  end;
  if (TMethod(VMTValueNotify).Code <> @TCustomDictionary<TKey,TValue>.ValueNotify) then
  begin
    // FInternalValueNotify := Self.ValueNotifyCaller;
    TMethod(FInternalValueNotify).Data := Pointer(Self);
    TMethod(FInternalValueNotify).Code := @TCustomDictionary<TKey,TValue>.ValueNotifyCaller;
  end else
  begin
    TMethod(FInternalValueNotify) := TMethod(Self.FOnValueNotify);
  end;

  // FInternalItemNotify
  if Assigned(FInternalKeyNotify) then
  begin
    // FInternalItemNotify := Self.ItemNotifyKey;
    TMethod(FInternalItemNotify).Data := Pointer(Self);
    TMethod(FInternalItemNotify).Code := @TCustomDictionary<TKey,TValue>.ItemNotifyKey;

    if Assigned(FInternalValueNotify) then
    begin
      // FInternalItemNotify := Self.ItemNotifyEvents;
      TMethod(FInternalItemNotify).Code := @TCustomDictionary<TKey,TValue>.ItemNotifyEvents;

      if (TMethod(VMTKeyNotify).Code <> @TCustomDictionary<TKey,TValue>.KeyNotify) or
        (TMethod(VMTValueNotify).Code <> @TCustomDictionary<TKey,TValue>.ValueNotify) then
      begin
        // FInternalItemNotify := Self.ItemNotifyCaller;
        TMethod(FInternalItemNotify).Code := @TCustomDictionary<TKey,TValue>.ItemNotifyCaller;
      end;
    end;
  end else
  if Assigned(FInternalValueNotify) then
  begin
    // FInternalItemNotify := Self.ItemNotifyValue;
    TMethod(FInternalItemNotify).Data := Pointer(Self);
    TMethod(FInternalItemNotify).Code := @TCustomDictionary<TKey,TValue>.ItemNotifyValue;
  end else
  begin
    // FInternalItemNotify := nil;
    TMethod(FInternalItemNotify).Data := nil;
    TMethod(FInternalItemNotify).Code := nil;
  end;
end;


{ TDictionary }

constructor TDictionary<TKey,TValue>.Create(ACapacity: Integer);
begin
  Create(ACapacity, nil);
end;

constructor TDictionary<TKey,TValue>.Create(const AComparer: IEqualityComparer<TKey>);
begin
  Create(0, AComparer);
end;

constructor TDictionary<TKey,TValue>.Create(ACapacity: Integer; const AComparer: IEqualityComparer<TKey>);
begin
  if ACapacity < 0 then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  // comparer
  FComparer := AComparer;
  if (FComparer = nil) then
    IInterface(FComparer) := IInterface(@InterfaceDefaults.TDefaultEqualityComparer<TKey>.Instance);

  // comparer methods
  TMethod(FComparerEquals) := IntfMethod(Pointer(FComparer), 3);
  TMethod(FComparerGetHashCode) := IntfMethod(Pointer(FComparer), 4);

  // initialization
  inherited Create(ACapacity);
end;

constructor TDictionary<TKey, TValue>.Create(const Collection: TEnumerable<TPair<TKey, TValue>>);
begin
  Create(Collection, nil);
end;

constructor TDictionary<TKey, TValue>.Create(const Collection: TEnumerable<TPair<TKey, TValue>>;
  const AComparer: IEqualityComparer<TKey>);
var
  Item: TPair<TKey,TValue>;
begin
  Create(0, AComparer);
  for Item in Collection do
    AddOrSetValue(Item.Key, Item.Value);
end;

destructor TDictionary<TKey,TValue>.Destroy;
begin
  inherited;

  ClearMethod(FComparerEquals);
  ClearMethod(FComparerGetHashCode);
end;

function TDictionary<TKey,TValue>.InternalFindItem(const Key: TKey; const FindMode: Integer): PItem;
label
  next_item, not_found;
var
  Parent: ^PItem;
  HashCode, Mode: Integer;
  Stored: TInternalFindStored;
begin
  // hash code
  HashCode := Self.FComparerGetHashCode(Key);

  // parent
  Pointer(Result{Parent}) := @FHashTable[NativeInt(Cardinal(HashCode)) and FHashTableMask];
  Dec(NativeUInt(Result{Parent}), SizeOf(TKey) + SizeOf(TValue));

  // find
  Stored.HashCode := HashCode;
  repeat
  next_item:
    // hash code item
    repeat
      Parent := @Result.FNext;
      Result := Result.FNext;
      if (not Assigned(Result)) then goto not_found;
    until (Stored.HashCode = Result.HashCode);
    NativeUInt(Stored.Parent) := NativeUInt(Parent);

    // keys comparison
    if (not Self.FComparerEquals(Key, Result.Key)) then goto next_item;

    // found
    Mode := FindMode;
    if (Mode and FOUND_MASK = 0) then Exit;
    Cardinal(Mode) := Cardinal(Mode) and FOUND_MASK;
    if (Mode <> FOUND_EXCEPTION) then
    begin
      if (Mode = FOUND_DELETE) then
      begin
        Stored.Parent^ := Result.FNext;
        if (not Assigned(Self.FInternalItemNotify)) then
        begin
          Self.DisposeItem(Result);
        end else
        begin
          Self.FInternalItemNotify(Result^, cnRemoved);
          Self.DisposeItem(Result);
        end;
      end else
      // if (Mode = FOUND_REPLACE) then
      begin
        if (not Assigned(Self.FInternalValueNotify)) then
        begin
          Result.FValue := FInternalFindValue^;
        end else
        begin
          Self.FInternalValueNotify(Self, Result.Value, cnRemoved);
          Result.FValue := FInternalFindValue^;
          Self.FInternalValueNotify(Self, Result.Value, cnAdded);
        end;
      end;
    end else
    begin
      raise EListError.CreateRes(Pointer(@SGenericDuplicateItem));
    end;
    Exit;

  // not found (Result = nil)
  not_found:
    Mode := FindMode;
    if (Mode and EMPTY_MASK = 0) then Exit;
    if (Mode and EMPTY_EXCEPTION = 0) then
    begin
      // EMPTY_NEW
      Result := Self.NewItem;
      Result.FKey := Key;
      Result.FHashCode := Stored.HashCode;
      Parent := @Self.FHashTable[NativeInt(Cardinal(Stored.HashCode)) and Self.FHashTableMask];
      Result.FNext := Parent^;
      Parent^ := Result;
      Result.FValue := FInternalFindValue^;
      if (Assigned(Self.FInternalItemNotify)) then
      begin
        Self.FInternalItemNotify(Result^, cnAdded);
      end;
    end else
    begin
      raise EListError.CreateRes(Pointer(@SGenericItemNotFound));
    end;
    Exit;
  until (False);
end;

function TDictionary<TKey,TValue>.GetItem(const Key: TKey): TValue;
begin
  Result := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_EXCEPTION).Value;
end;

procedure TDictionary<TKey,TValue>.SetItem(const Key: TKey; const Value: TValue);
begin
  Self.FInternalFindValue := @Value;
  Self.InternalFindItem(Key, FOUND_REPLACE + EMPTY_EXCEPTION);
end;

function TDictionary<TKey,TValue>.Find(const Key: TKey): PItem;
begin
  Result := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NONE);
end;

function TDictionary<TKey,TValue>.FindOrAdd(const Key: TKey): PItem;
begin
  Self.FInternalFindValue := @Self.FDefaultValue;
  Result := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NEW);
end;

procedure TDictionary<TKey,TValue>.Add(const Key: TKey; const Value: TValue);
begin
  Self.FInternalFindValue := @Value;
  Self.InternalFindItem(Key, FOUND_EXCEPTION + EMPTY_NEW);
end;

procedure TDictionary<TKey,TValue>.Remove(const Key: TKey);
begin
  Self.InternalFindItem(Key, FOUND_DELETE + EMPTY_NONE)
end;

function TDictionary<TKey,TValue>.ExtractPair(const Key: TKey): TPair<TKey,TValue>;
var
  Parent: ^PItem;
  Item, Current: PItem;
begin
  Result.Key := Key;
  Item := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NONE);
  if (Item = nil) then
  begin
    Result.Value := Default(TValue);
    Exit;
  end;

  Result.Value := Item.Value;
  Parent := @Self.FHashTable[NativeInt(Cardinal(Item.HashCode)) and Self.FHashTableMask];
  repeat
    Current := Parent^;

    if (Item = Current) then
    begin
      Parent^ := Item.FNext;

      if (not Assigned(Self.FInternalItemNotify)) then
      begin
        Self.DisposeItem(Item);
      end else
      begin
        Self.FInternalItemNotify(Item^, cnExtracted);
        Self.DisposeItem(Item);
      end;

      Exit;
    end;

    Parent := @Current.FNext;
  until (False);
end;

function TDictionary<TKey,TValue>.TryGetValue(const Key: TKey; out Value: TValue): Boolean;
var
  Item: PItem;
begin
  Item := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NONE);
  if Assigned(Item) then
  begin
    Value := Item.Value;
    Result := True;
  end else
  begin
    Value := Default(TValue);
    Result := False;
  end;
end;

procedure TDictionary<TKey,TValue>.AddOrSetValue(const Key: TKey; const Value: TValue);
begin
  Self.FInternalFindValue := @Value;
  Self.InternalFindItem(Key, FOUND_REPLACE + EMPTY_NEW);
end;

function TDictionary<TKey,TValue>.ContainsKey(const Key: TKey): Boolean;
begin
  Result := (Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NONE) <> nil);
end;


{ TRapidDictionary<TKey,TValue> }

constructor TRapidDictionary<TKey,TValue>.Create(ACapacity: Integer);
begin
  {$ifNdef SMARTGENERICS}
  FDefaultComparer := TEqualityComparer<TKey>.Default;
  TMethod(FDefaultEquals) := IntfMethod(Pointer(FDefaultComparer), 3);
  TMethod(FDefaultGetHashCode) := IntfMethod(Pointer(FDefaultComparer), 4);
  {$endif}

  inherited;
end;

constructor TRapidDictionary<TKey,TValue>.Create(const Collection: TEnumerable<TPair<TKey,TValue>>);
var
  Item: TPair<TKey,TValue>;
begin
  inherited Create;
  for Item in Collection do
    AddOrSetValue(Item.Key, Item.Value);
end;

destructor TRapidDictionary<TKey,TValue>.Destroy;
begin
  inherited;

  {$ifNdef SMARTGENERICS}
  ClearMethod(FDefaultEquals);
  ClearMethod(FDefaultGetHashCode);
  {$endif}
end;

function TRapidDictionary<TKey,TValue>.InternalFindItem(const Key: TKey; const FindMode: Integer): PItem;
label
  {$ifdef SMARTGENERICS}
  hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, hash8, hash9, hash10,
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5, {$ifdef SMALLINT}cmp6, cmp7, cmp8, cmp9, cmp10,{$endif}
  {$endif}
  hash_calculated, next_item, not_found;
type
  TData16 = TRAIIHelper.TData16;
  PData16 = ^TData16;
var
  Parent: ^PItem;
  HashCode, Mode, M: Integer;
  Stored: TInternalFindStored;
  {$ifdef SMARTGENERICS}
  Left, Right: PByte;
  Count, Offset: NativeUInt;
  _Self1, _Self2: TRapidDictionary<TKey,TValue>;
  {$endif}
begin
  // stores, hash code
  {$ifdef SMARTGENERICS}
  Stored.Self := Self;
  if (GetTypeKind(TKey) = tkVariant) then
  begin
    HashCode := InterfaceDefaults.GetHashCode_Var(nil, PVarData(@Key));
  end else
  if (GetTypeKind(TKey) = tkClass) then
  begin
    Left := PPointer(@Key)^;
    if (Assigned(Left)) then
    begin
      if (PPointer(Pointer(Left)^)[vmtGetHashCode div SizeOf(Pointer)] = @TObject.GetHashCode) then
      begin
        {$ifdef LARGEINT}
          HashCode := Integer(NativeInt(Left) xor (NativeInt(Left) shr 32));
        {$else .SMALLINT}
          HashCode := Integer(Left);
        {$endif}
        Inc(HashCode, ((HashCode shr 8) * 63689) + ((HashCode shr 16) * -1660269137) +
            ((HashCode shr 24) * -1092754919));
      end else
      begin
        HashCode := TObject(Left).GetHashCode;
      end;
    end else
    begin
      HashCode := 0;
    end;
  end else
  if (GetTypeKind(TKey) in [tkInterface, tkClassRef, tkPointer, tkProcedure]) then
  begin
    {$ifdef LARGEINT}
      HashCode := Integer(PNativeInt(@Key)^ xor (PNativeInt(@Key)^ shr 32));
    {$else .SMALLINT}
      HashCode := PInteger(@Key)^;
    {$endif}
    Inc(HashCode, ((HashCode shr 8) * 63689) + ((HashCode shr 16) * -1660269137) +
        ((HashCode shr 24) * -1092754919));
  end else
  if (GetTypeKind(TKey) = tkFloat) then
  begin
    HashCode := 0;
    case SizeOf(TKey) of
      4:
      begin
        if (PSingle(@Key)^ = 0) then goto hash_calculated;
        Frexp(PSingle(@Key)^, Stored.SingleRec.Mantissa, Stored.SingleRec.Exponent);
        HashCode := Stored.SingleRec.Exponent + Stored.SingleRec.HighInt * 63689;
      end;
      10:
      begin
        if (PExtended(@Key)^ = 0) then goto hash_calculated;
        Frexp(PExtended(@Key)^, Stored.ExtendedRec.Mantissa, Stored.ExtendedRec.Exponent);
        HashCode := Stored.ExtendedRec.Exponent + Stored.ExtendedRec.LowInt * 63689 +
          Stored.ExtendedRec.HighInt * -1660269137 + Integer(Stored.ExtendedRec.Middle) * -1092754919;
      end;
    else
      if (InterfaceDefaults.TDefaultEqualityComparer<TKey>.Instance.Size < 0) then
      begin
        HashCode := PPoint(@Key).X + PPoint(@Key).Y * 63689;
      end else
      begin
        if (PDouble(@Key)^ = 0) then goto hash_calculated;
        Frexp(PDouble(@Key)^, Stored.DoubleRec.Mantissa, Stored.DoubleRec.Exponent);
        HashCode := Stored.DoubleRec.Exponent + Stored.DoubleRec.LowInt * 63689 +
          Stored.DoubleRec.HighInt * -1660269137;
      end;
    end;

    Inc(HashCode, ((HashCode shr 8) * 63689) + ((HashCode shr 16) * -1660269137) +
        ((HashCode shr 24) * -1092754919));
  end else
  if (not (GetTypeKind(TKey) in [tkDynArray, tkString, tkLString, tkWString, tkUString])) and
    (SizeOf(TKey) <= 16) then
  begin
    // small binary
    if (SizeOf(TKey) >= SizeOf(Integer)) then
    begin
      if (SizeOf(TKey) = SizeOf(Integer)) then
      begin
        HashCode := PInteger(@Key)^;
      end else
      if (SizeOf(TKey) = SizeOf(Int64)) then
      begin
        HashCode := PPoint(@Key).X + PPoint(@Key).Y * 63689;
      end else
      begin
        Left := Pointer(@Key);
        HashCode := Integer(SizeOf(TKey)) + PInteger(Left + (SizeOf(TKey) - SizeOf(Integer)))^ * 63689;
        HashCode := HashCode * 2012804575 + PInteger(Left)[0];
        if (SizeOf(TKey) > SizeOf(Integer) * 2) then
        begin
          HashCode := HashCode * -1092754919 + PInteger(Left)[1];
          if (SizeOf(TKey) > SizeOf(Integer) * 3) then
          begin
            HashCode := HashCode * -1660269137 + PInteger(Left)[2];
          end;
        end;
      end;

      Inc(HashCode, ((HashCode shr 8) * 63689) + ((HashCode shr 16) * -1660269137) +
        ((HashCode shr 24) * -1092754919));
    end else
    if (SizeOf(TKey) <> 0) then
    begin
      Left := Pointer(@Key);
      HashCode := Integer(Left[0]);
      HashCode := HashCode + (HashCode shr 4) * 63689;
      if (SizeOf(TKey) > 1) then
      begin
        HashCode := HashCode + Integer(Left[1]) * -1660269137;
        if (SizeOf(TKey) > 2) then
        begin
          HashCode := HashCode + Integer(Left[2]) * -1092754919;
        end;
      end;
    end else
    begin
      HashCode := 0;
    end;
  end else
  begin
    if (GetTypeKind(TKey) in [tkDynArray, tkString, tkLString, tkWString, tkUString]) then
    begin
      // dynamic size
      if (GetTypeKind(TKey) = tkString) then
      begin
        Left := Pointer(@Key);
        Count := Left^;
        Inc(Count);
      end else
      // if (GetTypeKind(TKey) in [tkDynArray, tkLString, tkWString, tkUString]) then
      begin
        Left := PPointer(@Key)^;
        HashCode := 0;
        if (Left = nil) then goto hash_calculated;

        case GetTypeKind(TKey) of
          tkLString:
          begin
            Dec(Left, SizeOf(Integer));
            Count := PInteger(Left)^;
            Inc(Left, SizeOf(Integer));
          end;
          tkWString:
          begin
            Dec(Left, SizeOf(Integer));
            Count := PInteger(Left)^;
            Inc(Left, SizeOf(Integer));
            {$ifdef MSWINDOWS}if (Count = 0) then goto hash_calculated;{$endif}
            Count := Count {$ifNdef MSWINDOWS}* 2{$endif} + 2;
            Count := Count and -4;
          end;
          tkUString:
          begin
            Dec(Left, SizeOf(Integer));
            Count := PInteger(Left)^;
            Inc(Left, SizeOf(Integer));
            Count := Count * 2 + 2;
            Count := Count and -4;
          end;
        else
        // tkDynArray
          Dec(Left, SizeOf(NativeUInt));
          Count := PNativeInt(Left)^ * InterfaceDefaults.TDefaultEqualityComparer<TKey>.Instance.Size;
          Inc(Left, SizeOf(NativeUInt));
        end;
      end;
    end else
    begin
      // non-dynamic (constant) size binary > 16
      Left := Pointer(@Key);
      Count := SizeOf(TKey);
    end;

    if (not (GetTypeKind(TKey) in [tkWString, tkUString])) then
    begin
      if (Count < SizeOf(Integer)) then
      begin
        if (Count <> 0) then
        begin
          HashCode := Integer(Left[0]);
          HashCode := HashCode + (HashCode shr 4) * 63689;
          if (Count > 1) then
          begin
            HashCode := HashCode + Integer(Left[1]) * -1660269137;
            if (Count > 2) then
            begin
              HashCode := HashCode + Integer(Left[2]) * -1092754919;
            end;
          end;
        end else
        begin
          HashCode := 0;
        end;
        goto hash_calculated;
      end;
    end;

    HashCode := Integer(Count);
    Dec(Count, SizeOf(Integer));
    Inc(Left, Count);
    Inc(HashCode, PInteger(Left)^ * 63689);
    Dec(Left, Count);
    case (Count + (SizeOf(Integer) - 1)) shr 2 of
     10: goto hash10;
      9: goto hash9;
      8: goto hash8;
      7: goto hash7;
      6: goto hash6;
      5: goto hash5;
      4: goto hash4;
      3: goto hash3;
      2: goto hash2;
      1: goto hash1;
      0: goto hash0;
    else
      Inc(Count, SizeOf(Integer) - 1);
      M := -1660269137;
      repeat
        HashCode := HashCode * M + PInteger(Left)^;
        Dec(Count, SizeOf(Integer));
        Inc(Left, SizeOf(Integer));
        M := M * 378551;
        if (NativeInt(Count) <= 43) then Break;
      until (False);

      hash10:
        HashCode := HashCode * 631547855 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash9:
        HashCode := HashCode * -1987506439 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash8:
        HashCode := HashCode * -1653913089 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash7:
        HashCode := HashCode * -186114231 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash6:
        HashCode := HashCode * 915264303 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash5:
        HashCode := HashCode * -794603367 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash4:
        HashCode := HashCode * 135394143 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash3:
        HashCode := HashCode * 2012804575 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash2:
        HashCode := HashCode * -1092754919 + PInteger(Left)^;
        Inc(Left, SizeOf(Integer));
      hash1:
        HashCode := HashCode * -1660269137 + PInteger(Left)^;
      hash0:
    end;

    Inc(HashCode, ((HashCode shr 8) * 63689) + ((HashCode shr 16) * -1660269137) +
      ((HashCode shr 24) * -1092754919));
  end;
  {$else}
    HashCode := Self.FDefaultGetHashCode(Key);
  {$endif}
hash_calculated:

  // parent
  Pointer(Result{Parent}) := @FHashTable[NativeInt(Cardinal(HashCode)) and FHashTableMask];
  Dec(NativeUInt(Result{Parent}), SizeOf(TKey) + SizeOf(TValue));

  // find
  Stored.HashCode := HashCode;
  repeat
  next_item:
    // hash code item
    repeat
      Parent := @Result.FNext;
      Result := Result.FNext;
      if (not Assigned(Result)) then goto not_found;
    until (Stored.HashCode = Result.HashCode);
    NativeUInt(Stored.Parent) := NativeUInt(Parent);

    // default keys comparison
    {$ifdef SMARTGENERICS}
    if (GetTypeKind(TKey) = tkVariant) then
    begin
      if (not InterfaceDefaults.Equals_Var(nil, PVarData(@Key), PVarData(Result))) then goto next_item;
    end else
    if (GetTypeKind(TKey) = tkClass) then
    begin
      Left := PPointer(@Key)^;
      Right := PPointer(Result)^;
      if (Assigned(Left)) then
      begin
        if (PPointer(Left^)[vmtEquals div SizeOf(Pointer)] = @TObject.Equals) then
        begin
          if (Left <> Right) then goto next_item;
        end else
        begin
          if (not TObject(PNativeUInt(@Key)^).Equals(TObject(PNativeUInt(Result)^))) then goto next_item;
        end;
      end else
      begin
        if (Right <> nil) then goto next_item;
      end;
    end else
    if (GetTypeKind(TKey) = tkFloat) then
    begin
      case SizeOf(TKey) of
        4:
        begin
          if (PSingle(@Key)^ <> PSingle(Result)^) then goto next_item;
        end;
        10:
        begin
          if (PExtended(@Key)^ <> PExtended(Result)^) then goto next_item;
        end;
      else
      {$ifdef LARGEINT}
        if (PInt64(@Key)^ <> PInt64(Result)^) then
      {$else .SMALLINT}
        if ((PPoint(@Key).X - PPoint(Result).X) or (PPoint(@Key).Y - PPoint(Result).Y) <> 0) then
      {$endif}
        begin
          if (InterfaceDefaults.TDefaultEqualityComparer<TKey>.Instance.Size < 0) then goto next_item;
          if (PDouble(@Key)^ <> PDouble(Result)^) then goto next_item;
        end;
      end;
    end else
    if (not (GetTypeKind(TKey) in [tkDynArray, tkString, tkLString, tkWString, tkUString])) and
      (SizeOf(TKey) <= 16) then
    begin
      // small binary
      if (SizeOf(TKey) <> 0) then
      with PData16(@Key)^ do
      begin
        if (SizeOf(TKey) >= SizeOf(Integer)) then
        begin
          if (SizeOf(TKey) >= SizeOf(Int64)) then
          begin
            {$ifdef LARGEINT}
            if (Int64s[0] <> PData16(Result).Int64s[0]) then goto next_item;
            {$else}
            if (Integers[0] <> PData16(Result).Integers[0]) then goto next_item;
            if (Integers[1] <> PData16(Result).Integers[1]) then goto next_item;
            {$endif}

            if (SizeOf(TKey) = 16) then
            begin
              {$ifdef LARGEINT}
              if (Int64s[1] <> P16(Result).Int64s[1]) then goto next_item;
              {$else}
              if (Integers[2] <> PData16(Result).Integers[2]) then goto next_item;
              if (Integers[3] <> PData16(Result).Integers[3]) then goto next_item;
              {$endif}
            end else
            if (SizeOf(TKey) >= 12) then
            begin
              if (Integers[2] <> PData16(Result).Integers[2]) then goto next_item;
            end;
          end else
          begin
            if (Integers[0] <> PData16(Result).Integers[0]) then goto next_item;
          end;
        end;

        if (SizeOf(TKey) and 2 <> 0) then
        begin
          if (Words[(SizeOf(TKey) and -4) shr 1] <> PData16(Result).Words[(SizeOf(TKey) and -4) shr 1]) then goto next_item;
        end;
        if (SizeOf(TKey) and 1 <> 0) then
        begin
          if (Bytes[SizeOf(TKey) and -2] <> PData16(Result).Bytes[SizeOf(TKey) and -2]) then goto next_item;
        end;
      end;
    end else
    begin
      if (GetTypeKind(TKey) in [tkDynArray, tkString, tkLString, tkWString, tkUString]) then
      begin
        // dynamic size
        if (GetTypeKind(TKey) = tkString) then
        begin
          Left := Pointer(@Key);
          Right := Pointer(Result);
          if (PItem(Left) = {Right}Result) then goto cmp0;
          Count := Left^;
          if (Count <> Right^) then goto next_item;
          if (Count = 0) then goto cmp0;
          // compare last bytes
          if (Left[Count] <> Right[Count]) then goto next_item;
        end else
        // if (GetTypeKind(TKey) in [tkDynArray, tkLString, tkWString, tkUString]) then
        begin
          Left := PPointer(@Key)^;
          Right := PPointer(Result)^;
          if (Left = Right) then goto cmp0;
          if (Left = nil) then
          begin
            {$ifdef MSWINDOWS}
            if (GetTypeKind(TKey) = tkWString) then
            begin
              Dec(Right, SizeOf(Cardinal));
              if (PCardinal(Right)^ = 0) then goto cmp0;
            end;
            {$endif}
            goto next_item;
          end;
          if (Right = nil) then
          begin
            {$ifdef MSWINDOWS}
            if (GetTypeKind(TKey) = tkWString) then
            begin
              Dec(Left, SizeOf(Cardinal));
              if (PCardinal(Left)^ = 0) then goto cmp0;
            end;
            {$endif}
            goto next_item;
          end;

          if (GetTypeKind(TKey) = tkDynArray) then
          begin
            Dec(Left, SizeOf(NativeUInt));
            Dec(Right, SizeOf(NativeUInt));
            Count := PNativeUInt(Left)^;
            if (Count <> PNativeUInt(Right)^) then goto next_item;
            NativeInt(Count) := NativeInt(Count) * InterfaceDefaults.TDefaultEqualityComparer<TKey>.Instance.Size;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          end else
          // if (GetTypeKind(TKey) in [tkLString, tkWString, tkUString]) then
          begin
            Dec(Left, SizeOf(Cardinal));
            Dec(Right, SizeOf(Cardinal));
            Count := PCardinal(Left)^;
            if (Cardinal(Count) <> PCardinal(Right)^) then goto next_item;
            Inc(Left, SizeOf(Cardinal));
            Inc(Right, SizeOf(Cardinal));
          end;
        end;

        // compare last (after cardinal) words
        if (GetTypeKind(TKey) in [tkDynArray, tkString, tkLString]) then
        begin
          if (GetTypeKind(TKey) in [tkString, tkLString]) {ByteStrings + 2} then
          begin
            Inc(Count);
          end;
          if (Count and 2 <> 0) then
          begin
            Offset := Count and -4;
            Inc(Left, Offset);
            Inc(Right, Offset);
            if (PWord(Left)^ <> PWord(Right)^) then goto next_item;
            Offset := Count;
            Offset := Offset and -4;
            Dec(Left, Offset);
            Dec(Right, Offset);
          end;
        end else
        // modify Count to have only cardinals to compare
        // if (GetTypeKind(TKey) in [tkWString, tkUString]) {UnicodeStrings + 2} then
        begin
          {$ifdef MSWINDOWS}
          if (GetTypeKind(TKey) = tkWString) then
          begin
            if (Count = 0) then goto cmp0;
          end else
          {$endif}
          begin
            Inc(Count, Count);
          end;
          Inc(Count, 2);
        end;

        {$ifdef LARGEINT}
        if (Count and 4 <> 0) then
        begin
          Offset := Count and -8;
          Inc(Left, Offset);
          Inc(Right, Offset);
          if (PCardinal(Left)^ <> PCardinal(Right)^) then goto next_item;
          Dec(Left, Offset);
          Dec(Right, Offset);
        end;
        {$endif}
      end else
      begin
        // non-dynamic (constant) size binary > 16
        if (SizeOf(TKey) and {$ifdef LARGEINT}7{$else}3{$endif} <> 0) then
        with PData16(@Key)^ do
        begin
          {$ifdef LARGEINT}
          if (SizeOf(TKey) and 4 <> 0) then
          begin
            if (Integers[(SizeOf(TKey) and -8) shr 2] <> P1Data6(Result).Integers[(SizeOf(TKey) and -8) shr 2]) then goto next_item;
          end;
          {$endif}
          if (SizeOf(TKey) and 2 <> 0) then
          begin
            if (Words[(SizeOf(TKey) and -4) shr 1] <> PData16(Result).Words[(SizeOf(TKey) and -4) shr 1]) then goto next_item;
          end;
          if (SizeOf(TKey) and 1 <> 0) then
          begin
            if (Bytes[SizeOf(TKey) and -2] <> PData16(Result).Bytes[SizeOf(TKey) and -2]) then goto next_item;
          end;
        end;
        Left := Pointer(@Key);
        Right := Pointer(Result);
        Count := SizeOf(TKey);
      end;

      // natives (40 bytes static) compare
      Count := Count shr {$ifdef LARGEINT}3{$else}2{$endif};
      case Count of
      {$ifdef SMALLINT}
       10: goto cmp10;
        9: goto cmp9;
        8: goto cmp8;
        7: goto cmp7;
        6: goto cmp6;
      {$endif}
        5: goto cmp5;
        4: goto cmp4;
        3: goto cmp3;
        2: goto cmp2;
        1: goto cmp1;
        0: goto cmp0;
      else
        repeat
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Dec(Count);
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        until (Count = {$ifdef LARGEINT}5{$else}10{$endif});

        {$ifdef SMALLINT}
        cmp10:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        cmp9:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        cmp8:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        cmp7:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        cmp6:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        {$endif}
        cmp5:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        cmp4:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        cmp3:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        cmp2:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
          Inc(Left, SizeOf(NativeUInt));
          Inc(Right, SizeOf(NativeUInt));
        cmp1:
          if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then goto next_item;
        cmp0:
      end;
    end;
    {$else}
      if (not Self.FDefaultEquals(Key, Result.Key)) then goto next_item;
    {$endif}

    // found
    Mode := FindMode;
    if (Mode and FOUND_MASK = 0) then Exit;
    Cardinal(Mode) := Cardinal(Mode) and FOUND_MASK;
    if (Mode <> FOUND_EXCEPTION) then
    begin
      if (Mode = FOUND_DELETE) then
      begin
        Stored.Parent^ := Result.FNext;

        {$ifdef SMARTGENERICS}
        _Self1 := Stored.Self;
        with _Self1 do
        {$endif}
        if (not Assigned(FInternalItemNotify)) then
        begin
          DisposeItem(Result);
        end else
        begin
          FInternalItemNotify(Result^, cnRemoved);
          DisposeItem(Result);
        end;
      end else
      // if (Mode = FOUND_REPLACE) then
      begin
        {$ifdef SMARTGENERICS}
        _Self1 := Stored.Self;
        with _Self1 do
        {$endif}
        if (not Assigned(FInternalValueNotify)) then
        begin
          Result.FValue := FInternalFindValue^;
        end else
        begin
          FInternalValueNotify({$ifdef SMARTGENERICS}_Self1{$else}Self{$endif}, Result.Value, cnRemoved);
          Result.FValue := FInternalFindValue^;
          FInternalValueNotify({$ifdef SMARTGENERICS}_Self1{$else}Self{$endif}, Result.Value, cnAdded);
        end;
      end;
    end else
    begin
      raise EListError.CreateRes(Pointer(@SGenericDuplicateItem));
    end;
    Exit;

    // not found (Result = nil)
  not_found:
    Mode := FindMode;
    if (Mode and EMPTY_MASK = 0) then Exit;
    if (Mode and EMPTY_EXCEPTION = 0) then
    begin
      // EMPTY_NEW
      {$ifdef SMARTGENERICS}
      _Self2 := Stored.Self;
      with _Self2 do
      {$endif}
      begin
        Result := NewItem;
        Result.FKey := Key;
        Result.FHashCode := Stored.HashCode;
        Parent := @FHashTable[NativeInt(Cardinal(Stored.HashCode)) and FHashTableMask];
        Result.FNext := Parent^;
        Parent^ := Result;
        Result.FValue := FInternalFindValue^;
        if (Assigned(FInternalItemNotify)) then
        begin
          FInternalItemNotify(Result^, cnAdded);
        end;
      end;
    end else
    begin
      raise EListError.CreateRes(Pointer(@SGenericItemNotFound));
    end;
    Exit;
  until (False);
end;

function TRapidDictionary<TKey,TValue>.GetItem(const Key: TKey): TValue;
begin
  Result := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_EXCEPTION).Value;
end;

procedure TRapidDictionary<TKey,TValue>.SetItem(const Key: TKey; const Value: TValue);
begin
  Self.FInternalFindValue := @Value;
  Self.InternalFindItem(Key, FOUND_REPLACE + EMPTY_EXCEPTION);
end;

function TRapidDictionary<TKey,TValue>.Find(const Key: TKey): PItem;
begin
  Result := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NONE);
end;

function TRapidDictionary<TKey,TValue>.FindOrAdd(const Key: TKey): PItem;
begin
  Self.FInternalFindValue := @Self.FDefaultValue;
  Result := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NEW);
end;

procedure TRapidDictionary<TKey,TValue>.Add(const Key: TKey; const Value: TValue);
begin
  Self.FInternalFindValue := @Value;
  Self.InternalFindItem(Key, FOUND_EXCEPTION + EMPTY_NEW);
end;

procedure TRapidDictionary<TKey,TValue>.Remove(const Key: TKey);
begin
  Self.InternalFindItem(Key, FOUND_DELETE + EMPTY_NONE)
end;

function TRapidDictionary<TKey,TValue>.ExtractPair(const Key: TKey): TPair<TKey,TValue>;
var
  Parent: ^PItem;
  Item, Current: PItem;
begin
  Result.Key := Key;
  Item := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NONE);
  if (Item = nil) then
  begin
    Result.Value := Default(TValue);
    Exit;
  end;

  Result.Value := Item.Value;
  Parent := @Self.FHashTable[NativeInt(Cardinal(Item.HashCode)) and Self.FHashTableMask];
  repeat
    Current := Parent^;

    if (Item = Current) then
    begin
      Parent^ := Item.FNext;

      if (not Assigned(Self.FInternalItemNotify)) then
      begin
        Self.DisposeItem(Item);
      end else
      begin
        Self.FInternalItemNotify(Item^, cnExtracted);
        Self.DisposeItem(Item);
      end;

      Exit;
    end;

    Parent := @Current.FNext;
  until (False);
end;

function TRapidDictionary<TKey,TValue>.TryGetValue(const Key: TKey; out Value: TValue): Boolean;
var
  Item: PItem;
begin
  Item := Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NONE);
  if Assigned(Item) then
  begin
    Value := Item.Value;
    Result := True;
  end else
  begin
    Value := Default(TValue);
    Result := False;
  end;
end;

procedure TRapidDictionary<TKey,TValue>.AddOrSetValue(const Key: TKey; const Value: TValue);
begin
  Self.FInternalFindValue := @Value;
  Self.InternalFindItem(Key, FOUND_REPLACE + EMPTY_NEW);
end;

function TRapidDictionary<TKey,TValue>.ContainsKey(const Key: TKey): Boolean;
begin
  Result := (Self.InternalFindItem(Key, FOUND_NONE + EMPTY_NONE) <> nil);
end;


{ TArray }

class procedure TArray.Exchange<T>(const Left, Right: Pointer);
var
  Index: NativeInt;
  Temp1: Byte;
  Temp2: Word;
  Temp4: Cardinal;
  TempNative: NativeUInt;
begin
  with TLMemory(Left^), TRMemory(Right^) do
  case SizeOf(T) of
    0: ;
    1:
    begin
      Temp1 := LBytes[0];
      LBytes[0] := RBytes[0];
      RBytes[0] := Temp1;
    end;
    2:
    begin
      Temp2 := LWords[0];
      LWords[0] := RWords[0];
      RWords[0] := Temp2;
    end;
    3:
    begin
      Temp2 := LWords[0];
      LWords[0] := RWords[0];
      RWords[0] := Temp2;

      Temp1 := LBytes[2];
      LBytes[2] := RBytes[2];
      RBytes[2] := Temp1;
    end;
    4..7:
    begin
      Temp4 := LCardinals[0];
      LCardinals[0] := RCardinals[0];
      RCardinals[0] := Temp4;

      case SizeOf(T) of
        5:
        begin
          Temp1 := LBytes[4];
          LBytes[4] := RBytes[4];
          RBytes[4] := Temp1;
        end;
        6:
        begin
          Temp2 := LWords[2];
          LWords[2] := RWords[2];
          RWords[2] := Temp2;
        end;
        7:
        begin
          Temp2 := LWords[2];
          LWords[2] := RWords[2];
          RWords[2] := Temp2;
          Temp1 := LBytes[6];
          LBytes[6] := RBytes[6];
          RBytes[6] := Temp1;
        end;
      end;
    end;
    8..16:
    begin
      TempNative := LNatives[0];
      LNatives[0] := RNatives[0];
      RNatives[0] := TempNative;

      if (SizeOf(T) >= SizeOf(NativeUInt) * 2) then
      begin
        TempNative := LNatives[1];
        LNatives[1] := RNatives[1];
        RNatives[1] := TempNative;
      end;

      if (SizeOf(T) >= SizeOf(NativeUInt) * 3) then
      begin
        TempNative := LNatives[2];
        LNatives[2] := RNatives[2];
        RNatives[2] := TempNative;
      end;

      if (SizeOf(T)  = SizeOf(NativeUInt) * 4) then
      begin
        TempNative := LNatives[3];
        LNatives[3] := RNatives[3];
        RNatives[3] := TempNative;
      end;

      {$ifdef LARGEINT}
      case SizeOf(T) of
        12, 13, 14, 15:
        begin
          Temp4 := LCardinals[2];
          LCardinals[2] := RCardinals[2];
          RCardinals[2] := Temp4;
        end;
      end;
      {$endif}

      case SizeOf(T) of
        9:
        begin
          Temp1 := LBytes[8];
          LBytes[8] := RBytes[8];
          RBytes[8] := Temp1;
        end;
        10:
        begin
          Temp2 := LWords[4];
          LWords[4] := RWords[4];
          RWords[4] := Temp2;
        end;
        11:
        begin
          Temp2 := LWords[4];
          LWords[4] := RWords[4];
          RWords[4] := Temp2;
          Temp1 := LBytes[10];
          LBytes[10] := RBytes[10];
          RBytes[10] := Temp1;
        end;
        13:
        begin
          Temp2 := LWords[5];
          LWords[5] := RWords[5];
          RWords[5] := Temp2;
          Temp1 := LBytes[12];
          LBytes[12] := RBytes[12];
          RBytes[12] := Temp1;
        end;
        14:
        begin
          Temp2 := LWords[6];
          LWords[6] := RWords[6];
          RWords[6] := Temp2;
        end;
        15:
        begin
          Temp2 := LWords[6];
          LWords[6] := RWords[6];
          RWords[6] := Temp2;
          Temp1 := LBytes[14];
          LBytes[14] := RBytes[14];
          RBytes[14] := Temp1;
        end;
      end;
    end;
  else
    Index := 0;
    repeat
      TempNative := LNatives[Index];
      LNatives[Index] := RNatives[Index];
      RNatives[Index] := TempNative;
      Inc(Index);
    until (Index = SizeOf(T) div SizeOf(NativeUInt));

    if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
    begin
      {$ifdef LARGEINT}
      if (SizeOf(T) and 4 <> 0) then
      begin
        Index := SizeOf(T) div SizeOf(Cardinal) - 1;
        Temp4 := LCardinals[Index];
        LCardinals[Index] := RCardinals[Index];
        RCardinals[Index] := Temp4;
      end;
      {$endif}

      if (SizeOf(T) and 2 <> 0) then
      begin
        Index := SizeOf(T) div SizeOf(Word) - 1;
        Temp2 := LWords[Index];
        LWords[Index] := RWords[Index];
        RWords[Index] := Temp2;
      end;

      if (SizeOf(T) and 1 <> 0) then
      begin
        Index := SizeOf(T) div SizeOf(Byte) - 1;
        Temp1 := LBytes[Index];
        LBytes[Index] := RBytes[Index];
        RBytes[Index] := Temp1;
      end;
    end;
  end;
end;

class procedure TArray.CopyMemory<T>(const Destination, Source: Pointer);
var
  Index: NativeInt;
begin
  with TLMemory(Destination^), TRMemory(Source^) do
  case SizeOf(T) of
    0: ;
    1: LBytes[0] := RBytes[0];
    2: LWords[0] := RWords[0];
    3:
    begin
      LWords[0] := RWords[0];
      LBytes[2] := RBytes[2];
    end;
    4..7:
    begin
      LCardinals[0] := RCardinals[0];

      case SizeOf(T) of
        5: LCardinals1[0] := RCardinals1[0];
        6: LCardinals2[0] := RCardinals2[0];
        7: LCardinals3[0] := RCardinals3[0];
      end;
    end;
    8..16:
    begin
      LNatives[0] := RNatives[0];

      if (SizeOf(T) >= SizeOf(NativeUInt) * 2) then
        LNatives[1] := RNatives[1];

      if (SizeOf(T) >= SizeOf(NativeUInt) * 3) then
        LNatives[2] := RNatives[2];

      if (SizeOf(T)  = SizeOf(NativeUInt) * 4) then
        LNatives[3] := RNatives[3];

      {$ifdef SMALLINT}
      case SizeOf(T) of
         9: LNatives1[1] := RNatives1[1];
        10: LNatives2[1] := RNatives2[1];
        11: LNatives3[1] := RNatives3[1];
        13: LNatives1[2] := RNatives1[2];
        14: LNatives2[2] := RNatives2[2];
        15: LNatives3[2] := RNatives3[2];
      end;
      {$else .LARGEINT}
      case SizeOf(T) of
         9: LNatives1[1] := RNatives1[1];
        10: LNatives2[1] := RNatives2[1];
        11: LNatives3[1] := RNatives3[1];
        12: LNatives4[1] := RNatives4[1];
        13: LNatives5[1] := RNatives5[1];
        14: LNatives6[1] := RNatives6[1];
        15: LNatives7[1] := RNatives7[1];
      end;
      {$endif}
    end;
  else
    Index := 0;
    repeat
      LNatives[Index] := RNatives[Index];
      Inc(Index);
    until (Index = SizeOf(T) div SizeOf(NativeUInt) - 1);

    if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
    begin
      Index := SizeOf(T) div SizeOf(NativeUInt) - 1;
      case SizeOf(T) and (SizeOf(NativeUInt) - 1) of
        1: LNatives1[Index] := RNatives1[Index];
        2: LNatives2[Index] := RNatives2[Index];
        3: LNatives3[Index] := RNatives3[Index];
      {$ifdef LARGEINT}
        4: LNatives4[Index] := RNatives4[Index];
        5: LNatives5[Index] := RNatives5[Index];
        6: LNatives6[Index] := RNatives6[Index];
        7: LNatives7[Index] := RNatives7[Index];
      {$endif}
      end;
    end;
  end;
end;

class procedure TArray.ZeroMemory<T>(const Values: Pointer);
var
  Null4: Cardinal;
  NullNative: NativeUInt;
  Index: NativeInt;
begin
  with TLMemory(Values^) do
  case SizeOf(T) of
    0: ;
    1: LBytes[0] := 0;
    2: LWords[0] := 0;
    3:
    begin
      LWords[0] := 0;
      LBytes[2] := 0;
    end;
    4..7:
    begin
      Null4 := 0;
      LCardinals[0] := Null4;

      case SizeOf(T) of
        5: LCardinals1[0] := Null4;
        6: LCardinals2[0] := Null4;
        7: LCardinals3[0] := Null4;
      end;
    end;
    8..16:
    begin
      NullNative := 0;
      LNatives[0] := NullNative;

      if (SizeOf(T) >= SizeOf(NativeUInt) * 2) then
        LNatives[1] := NullNative;

      if (SizeOf(T) >= SizeOf(NativeUInt) * 3) then
        LNatives[2] := NullNative;

      if (SizeOf(T)  = SizeOf(NativeUInt) * 4) then
        LNatives[3] := NullNative;

      {$ifdef SMALLINT}
      case SizeOf(T) of
         9: LNatives1[1] := NullNative;
        10: LNatives2[1] := NullNative;
        11: LNatives3[1] := NullNative;
        13: LNatives1[2] := NullNative;
        14: LNatives2[2] := NullNative;
        15: LNatives3[2] := NullNative;
      end;
      {$else .LARGEINT}
      case SizeOf(T) of
         9: LNatives1[1] := NullNative;
        10: LNatives2[1] := NullNative;
        11: LNatives3[1] := NullNative;
        12: LNatives4[1] := NullNative;
        13: LNatives5[1] := NullNative;
        14: LNatives6[1] := NullNative;
        15: LNatives7[1] := NullNative;
      end;
      {$endif}
    end;
  else
    Index := 0;
    repeat
      LNatives[Index] := NullNative;
      Inc(Index);
    until (Index = SizeOf(T) div SizeOf(NativeUInt) - 1);

    if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
    begin
      Index := SizeOf(T) div SizeOf(NativeUInt) - 1;
      case SizeOf(T) and (SizeOf(NativeUInt) - 1) of
        1: LNatives1[Index] := NullNative;
        2: LNatives2[Index] := NullNative;
        3: LNatives3[Index] := NullNative;
      {$ifdef LARGEINT}
        4: LNatives4[Index] := NullNative;
        5: LNatives5[Index] := NullNative;
        6: LNatives6[Index] := NullNative;
        7: LNatives7[Index] := NullNative;
      {$endif}
      end;
    end;
  end;
end;

class procedure TArray.Reverse<T>(const Values: Pointer; const Count: NativeInt);
type
  P = ^T;
var
  X, Y: Pointer;
  Index: NativeInt;
  Temp1: Byte;
  Temp2: Word;
  Temp4: Cardinal;
  TempNative: NativeUInt;
begin
  if (Count > 1) then
  begin
    X := Values;
    Y := P(Values) + Count - 1;

    repeat
      if (SizeOf(T) <= 16) then
      begin
        TArray.Exchange<T>(X, Y);
        Inc(NativeUInt(X), SizeOf(T));
        Dec(NativeUInt(Y), SizeOf(T));
      end else
      begin
        Index := 0;
        repeat
          Inc(Index);
          TempNative := PNativeUInt(X)^;
          PNativeUInt(X)^ := PNativeUInt(Y)^;
          PNativeUInt(Y)^ := TempNative;
          Inc(NativeUInt(X), SizeOf(NativeUInt));
          Inc(NativeUInt(Y), SizeOf(NativeUInt));
        until (Index = SizeOf(T) div SizeOf(NativeUInt));

        {$ifdef LARGEINT}
        if (SizeOf(T) and 4 <> 0) then
        begin
          Temp4 := PCardinal(X)^;
          PCardinal(X)^ := PCardinal(Y)^;
          PCardinal(Y)^ := Temp4;
          Inc(NativeUInt(X), SizeOf(Cardinal));
          Inc(NativeUInt(Y), SizeOf(Cardinal));
        end;
        {$endif}

        if (SizeOf(T) and 2 <> 0) then
        begin
          Temp2 := PWord(X)^;
          PWord(X)^ := PWord(Y)^;
          PWord(Y)^ := Temp2;
          Inc(NativeUInt(X), SizeOf(Word));
          Inc(NativeUInt(Y), SizeOf(Word));
        end;

        if (SizeOf(T) and 1 <> 0) then
        begin
          Temp1 := PByte(X)^;
          PByte(X)^ := PByte(Y)^;
          PByte(Y)^ := Temp1;
          Inc(NativeUInt(X), SizeOf(Byte));
          Inc(NativeUInt(Y), SizeOf(Byte));
        end;

        Dec(NativeUInt(Y), 2 * SizeOf(T));
      end;
    until (NativeUInt(X) >= NativeUInt(Y));
  end;
end;

class procedure TArray.Reverse<T>(var Values: array of T);
begin
  if (High(Values) > 0) then
    TArray.Reverse<T>(@Values[0], Length(Values));
end;

procedure TArray.TSortHelper<T>.Init(const Comparer: IComparer<T>);
begin
  Self.Inst := Pointer(Comparer);
  Self.Compare := PPointer(PNativeUInt(Self.Inst)^ + 3 * SizeOf(Pointer))^;
end;

procedure TArray.TSortHelper<T>.Init(const Comparison: TComparison<T>);
begin
  Self.Inst := PPointer(@Comparison)^;
  Self.Compare := PPointer(PNativeUInt(Self.Inst)^ + 3 * SizeOf(Pointer))^;
end;

procedure TArray.TSortHelper<T>.Init;
begin
  Self.Inst := @InterfaceDefaults.TDefaultComparer<T>.Instance;
  Self.Compare := InterfaceDefaults.TDefaultComparer<T>.Instance.Compare;
end;

procedure TArray.TSortHelper<T>.FillZero;
begin
  FillChar(Self, SizeOf(T), #0);
end;

class procedure TArray.CheckArrays(Source, Destination: Pointer; SourceIndex, SourceLength, DestIndex, DestLength, Count: NativeInt);
begin
  if (SourceIndex < 0) or (DestIndex < 0) or (SourceIndex >= SourceLength) or (DestIndex >= DestLength) or
     (SourceIndex + Count > SourceLength) or (DestIndex + Count > DestLength) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Source = Destination then
    raise EArgumentException.CreateRes(Pointer(@sSameArrays));
end;

class procedure TArray.Copy<T>(const Source: array of T; var Destination: array of T; SourceIndex, DestIndex, Count: NativeInt);
begin
  CheckArrays(Pointer(@Source[0]), Pointer(@Destination[0]), SourceIndex, Length(Source), DestIndex, Length(Destination), Count);
  if IsManagedType(T) then
    System.CopyArray(Pointer(@Destination[DestIndex]), Pointer(@Source[SourceIndex]), TypeInfo(T), Count)
  else
    System.Move(Pointer(@Source[SourceIndex])^, Pointer(@Destination[DestIndex])^, Count * SizeOf(T));
end;

class procedure TArray.Copy<T>(const Source: array of T; var Destination: array of T; Count: NativeInt);
begin
  Copy<T>(Source, Destination, 0, 0, Count);
end;

class function TArray.SortItemPivot<T>(const I, J: Pointer): Pointer;
type
  P = ^T;
  TItemList = TItemList<T>;
  PItemList = ^TItemList;
var
  Index: NativeInt;
begin
  if (SizeOf(T) and (SizeOf(T) - 1) = 0) and (SizeOf(T) <= 256) then
  begin
    Index := NativeInt(J) - NativeInt(I);
    case SizeOf(T) of
    0, 1: Index := Index shr 1;
       2: Index := Index shr 2;
       4: Index := Index shr 3;
       8: Index := Index shr 4;
      16: Index := Index shr 5;
      32: Index := Index shr 6;
      64: Index := Index shr 7;
     128: Index := Index shr 8;
    else
    // 256:
      Index := Index shr 9;
    end;
  end else
  begin
    Index := NativeInt(Round((NativeInt(J) - NativeInt(I)) * (1 / SizeOf(T)))) shr 1;
  end;

  Result := P(I) + Index;
end;

class function TArray.SortItemNext<T>(const StackItem, I, J: Pointer): Pointer;
var
  Item: ^TSortStackItem<T>;
  DiffI, DiffJ: NativeInt;
  Buf: Pointer;
begin
  Item := StackItem;

  // next "recursion" iteration
  // if (i < last) qs(s_arr, i, last);
  // if (first < j) qs(s_arr, first, j);
  DiffI := NativeInt(Item.Last) - NativeInt(I);
  DiffJ := NativeInt(J) - NativeInt(Item.First);
  if (DiffI > 0) then
  begin
    if (DiffJ <= 0) then
    begin
      Item.First := I;
      // Item.Last := Item.Last;
    end else
    if (DiffI >= DiffJ) then
    begin
      // i..last, first..j
      Buf := Item.First;
      Item.First := I;
      Inc(Item);
      Item.First := Buf;
      Item.Last := J;
    end else
    begin
      // first..j, i..last
      Buf := Item.Last;
      Item.Last := J;
      Inc(Item);
      Item.First := I;
      Item.Last := Buf;
    end;
  end else
  if (DiffJ > 0) then
  begin
    // Item.First := Item.First;
    Item.Last := J;
  end else
  begin
    Inc(NativeInt(Item), HIGH_NATIVE_BIT);
  end;

  Result := Item;
end;

{$ifdef SMARTGENERICS}
class function TArray.SortItemCount<T>(const I, J: Pointer): NativeInt;
begin
  if (SizeOf(T) and (SizeOf(T) - 1) = 0) and (SizeOf(T) <= 256) then
  begin
    Result := NativeInt(J) + SizeOf(T) - NativeInt(I);
    case SizeOf(T) of
       2: Result := Result shr 1;
       4: Result := Result shr 2;
       8: Result := Result shr 3;
      16: Result := Result shr 4;
      32: Result := Result shr 5;
      64: Result := Result shr 6;
     128: Result := Result shr 7;
     256: Result := Result shr 8;
    end;
  end else
  begin
    Result := Round((NativeInt(J) + SizeOf(T) - NativeInt(I)) * (1 / SizeOf(T)));
  end;
end;

class function TArray.SortBinaryMarker<T>(const Binary: Pointer): NativeUInt;
begin
  case GetTypeKind(T) of
    tkMethod: Result := NativeUInt(TMethod(Binary^).Data);
    tkLString, tkWString, tkUString, tkDynArray:
    begin
      Result := PNativeUInt(Binary)^;
      if (Result <> 0) then
      case GetTypeKind(T) of
        tkLString:
        begin
          Result := PWord(Result)^;
          Result := Swap(Result);
        end;
        {$ifdef MSWINDOWS}
        tkWString:
        begin
          Dec(Result, SizeOf(Integer));
          if (PInteger(Result)^ = 0) then
          begin
            Result := 0;
            Exit;
          end else
          begin
            Inc(Result, SizeOf(Integer));
            Result := PCardinal(Result)^;
            Result := Cardinal((Result shl 16) + (Result shr 16));
          end;
        end;
        {$else}
        tkWString,
        {$endif}
        tkUString:
        begin
          Result := PCardinal(Result)^;
          Result := Cardinal((Result shl 16) + (Result shr 16));
        end;
        tkDynArray:
        begin
          Result := PByte(Result)^;
        end;
      end;
    end;
    tkString:
    begin
      Result := PWord(Binary)^;
      if (Result and $ff = 0) then
      begin
        Result := 0;
      end else
      begin
        Result := Result shr 8;
      end;
    end;
  else
    with TLMemory(Binary^) do
    case SizeOf(T) of
      1: Result := LBytes[0];
      2:
      begin
        Result := LWords[0];
        Result := Swap(Result);
      end;
      3:
      begin
        Result := LWords[0];
        Result := Swap(Result);
        Result := Result shl 8;
        Inc(Result, LBytes[2]);
      end;
    else
      Result := LCardinals[0];
      Result := (Swap(Result) shl 16) + Swap(Result shr 16);
    end;
  end;
end;

class function TArray.SortBinaryComparer<T>(const Pivot, Binary: Pointer): Integer;
begin
  if (GetTypeKind(T) = tkString) then
  begin
    Result := InterfaceDefaults.Compare_OStr(nil, Pivot, Binary);
  end else
  if (SizeOf(T) <= SizeOf(Cardinal)) then
  begin
    Result := 0;
  end else
  {$ifdef LARGEINT}
  if (SizeOf(T) = SizeOf(Int64)) then
  begin
    Result := InterfaceDefaults.Compare_Bin8(nil, PInt64(Pivot)^, PInt64(Binary)^);
  end else
  {$endif}
  begin
    Result := InterfaceDefaults.Compare_Bin(InterfaceDefaults.TDefaultComparer<T>.Instance, Pivot, Binary);
  end;
end;

class procedure TArray.RadixSort<T>(const Values: Pointer; const Count, Flags: NativeInt);
label
  clean_radixes;
type
  PItem = ^T;
  TRadixes = array[Byte] of Word;
  PRadixes = ^TRadixes;
var
  i, Mask, Index, TargetIndex: NativeInt;
  Source, TopSourceFirst, TopSource: PByte;
  Radixes: PRadixes;
  Radix, TopRadix: PWord;
  Target: NativeInt;
  Buffers: record
    Radixes: TRadixes;
    Data: array[0..RADIX_BUFFER_SIZE - 1] of T;
    Index: NativeInt;
    Ptr: array[0..1] of Pointer;
    Mask: NativeInt;
    SingleRadix: Word;
  end;
begin
  Buffers.SingleRadix := Count * SizeOf(T);
  Buffers.Ptr[0] := Values;
  Buffers.Ptr[1] := @Buffers.Data;
  Buffers.Index := 0;

  i := 0;
  repeat
    // clear radixes
    FillChar(Buffers.Radixes, SizeOf(Buffers.Radixes), #0);
  clean_radixes:

    // xor mask
    Mask := Flags;
    if (i = SizeOf(T) - 1) then Mask := Mask shr 8;
    Mask := Byte(Mask);

    // fill radixes basics
    Source := Pointer(Buffers.Ptr[Buffers.Index]);
    Inc(Source, i);
    TopSourceFirst := Pointer(PItem(Source) + Count);
    repeat
      Index := Source^;
      Index := Index xor Mask;
      Radixes := @Buffers.Radixes;
      Inc(Radixes[Index], SizeOf(T));
      Inc(Source, SizeOf(T));
    until (Source = TopSourceFirst);

    // check the same radixes
    Dec(Source, SizeOf(T));
    Index := Source^;
    Index := Index xor Mask;
    if (Buffers.Radixes[Index] = Buffers.SingleRadix) then
    begin
      Inc(i);
      if (i = SizeOf(T)) then Break;

      Buffers.Radixes[Index] := 0;
      goto clean_radixes;
    end;
    Buffers.Mask := Mask;

    // calculate offsets
    Radix := @Buffers.Radixes[0];
    TopRadix := Radix + 256;
    Index := 0;
    repeat
      TargetIndex := Radix^;
      Inc(Radix);
      if (TargetIndex = 0) then
      begin
        if (Radix = TopRadix) then Break;
      end else
      begin
        Dec(Radix);
        Radix^ := Index;
        Inc(Index, TargetIndex);
        Inc(Radix);
        if (Radix = TopRadix) then Break;
      end;
    until (False);

    // fill target
    Source := Pointer(Buffers.Ptr[Buffers.Index]);
    Inc(Source, i);
    Buffers.Index := Buffers.Index xor 1;
    TopSource := Pointer(PItem(Source) + Count);
    Target := NativeInt(Buffers.Ptr[Buffers.Index]) - SizeOf(T);
    case Buffers.Mask of
      $80:
      begin
        repeat
          Index := Source^;
          Index := Index xor $80;

          Radixes := @Buffers.Radixes;
          TargetIndex := Radixes[Index];
          Inc(TargetIndex, SizeOf(T));
          Radixes[Index] := TargetIndex;

          PItem(TargetIndex + Target)^ := PItem(Source - i)^;
          Inc(Source, SizeOf(T));
        until (Source = TopSource);
      end;
      $ff:
      begin
        repeat
          Index := Source^;
          Index := Index xor $ff;

          Radixes := @Buffers.Radixes;
          TargetIndex := Radixes[Index];
          Inc(TargetIndex, SizeOf(T));
          Radixes[Index] := TargetIndex;

          PItem(TargetIndex + Target)^ := PItem(Source - i)^;
          Inc(Source, SizeOf(T));
        until (Source = TopSource);
      end;
      $7f:
      begin
        repeat
          Index := Source^;
          Index := Index xor $7f;

          Radixes := @Buffers.Radixes;
          TargetIndex := Radixes[Index];
          Inc(TargetIndex, SizeOf(T));
          Radixes[Index] := TargetIndex;

          PItem(TargetIndex + Target)^ := PItem(Source - i)^;
          Inc(Source, SizeOf(T));
        until (Source = TopSource);
      end;
    else
      // $00
      repeat
        Index := Source^;

        Radixes := @Buffers.Radixes;
        TargetIndex := Radixes[Index];
        Inc(TargetIndex, SizeOf(T));
        Radixes[Index] := TargetIndex;

        PItem(TargetIndex + Target)^ := PItem(Source - i)^;
        Inc(Source, SizeOf(T));
      until (Source = TopSource);
    end;

    // next iteration
    Inc(i);
  until (i = SizeOf(T));

  // ValuesArray guarantee
  if (Buffers.Index <> 0) then
    Move(Buffers.Data, Buffers.Ptr[0]^, Count * SizeOf(T));
end;

class function TArray.RadixSortSigneds<T>(var StackItem: TSortStackItem<T>): Pointer;
const
  Flags = $8000;
begin
  case SizeOf(T) of
    1: TArray.RadixSort<Byte>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    2: TArray.RadixSort<Word>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    4: TArray.RadixSort<Integer>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    8: TArray.RadixSort<Int64>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
  end;
  Result := @StackItem;
end;

class function TArray.RadixSortDescendingSigneds<T>(var StackItem: TSortStackItem<T>): Pointer;
const
  Flags = not $8000;
begin
  case SizeOf(T) of
    1: TArray.RadixSort<Byte>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    2: TArray.RadixSort<Word>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    4: TArray.RadixSort<Integer>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    8: TArray.RadixSort<Int64>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
  end;
  Result := @StackItem;
end;

class function TArray.RadixSortUnsigneds<T>(var StackItem: TSortStackItem<T>): Pointer;
const
  Flags = $0000;
begin
  case SizeOf(T) of
    1: TArray.RadixSort<Byte>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    2: TArray.RadixSort<Word>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    4: TArray.RadixSort<Integer>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    8: TArray.RadixSort<Int64>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
  end;
  Result := @StackItem;
end;

class function TArray.RadixSortDescendingUnsigneds<T>(var StackItem: TSortStackItem<T>): Pointer;
const
  Flags = not $0000;
begin
  case SizeOf(T) of
    1: TArray.RadixSort<Byte>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    2: TArray.RadixSort<Word>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    4: TArray.RadixSort<Integer>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
    8: TArray.RadixSort<Int64>(StackItem.First, SortItemCount<T>(StackItem.First, StackItem.Last), Flags);
  end;
  Result := @StackItem;
end;

class function TArray.RadixSortFloats<T>(var StackItem: TSortStackItem<T>): Pointer;
label
  sign_sorted;
var
  I, J: ^TFloat;
  Count: NativeInt;
begin
  I := Pointer(StackItem.First);
  J := Pointer(StackItem.Last);

  repeat
    Dec(NativeInt(I), SizeOf(T));
    repeat
      Inc(NativeInt(I), SizeOf(T));
      if (I > J) then goto sign_sorted;
      case SizeOf(T) of
        4: if (I.SSingle >= 0) then Break;
        8: if (I.SDouble >= 0) then Break;
       10: if (I.SExtended >= 0) then Break;
      end;
    until (False);

    Inc(NativeInt(J), SizeOf(T));
    repeat
      Dec(NativeInt(J), SizeOf(T));
      if (J < I) then goto sign_sorted;
      case SizeOf(T) of
        4: if (J.SSingle < 0) then Break;
        8: if (J.SDouble < 0) then Break;
       10: if (J.SExtended < 0) then Break;
      end;
    until (False);

    TArray.Exchange<T>(I, J);
    Inc(NativeInt(I), SizeOf(T));
    Dec(NativeInt(J), SizeOf(T));
  until (False);
sign_sorted:

  Count := SortItemCount<T>(StackItem.First, J);
  if (Count > 1) then
  begin
    if (Count <= ((SizeOf(T) div 4) * 125 - (SizeOf(T) div 10) * 1050)) then
    begin
      TArray.SortFloats<T>(StackItem.First, Count);
    end else
    case SizeOf(T) of
      4: TArray.RadixSort<Integer>(StackItem.First, Count, not $0000);
      8: TArray.RadixSort<Int64>(StackItem.First, Count, not $0000);
    else
      TArray.RadixSort<Extended>(StackItem.First, Count, not $0000);
    end;
  end;

  Count := SortItemCount<T>(I, StackItem.Last);
  if (Count > 1) then
  begin
    if (Count <= ((SizeOf(T) div 4) * 125 - (SizeOf(T) div 10) * 1050)) then
    begin
      TArray.SortFloats<T>(I, Count);
    end else
    case SizeOf(T) of
      4: TArray.RadixSort<Integer>(I, Count, $0000);
      8: TArray.RadixSort<Int64>(I, Count, $0000);
    else
      TArray.RadixSort<Extended>(I, Count, $0000);
    end;
  end;

  Result := @StackItem;
end;

class function TArray.RadixSortDescendingFloats<T>(var StackItem: TSortStackItem<T>): Pointer;
label
  sign_sorted;
var
  I, J: ^TFloat;
  Count: NativeInt;
begin
  I := Pointer(StackItem.First);
  J := Pointer(StackItem.Last);

  repeat
    Dec(NativeInt(I), SizeOf(T));
    repeat
      Inc(NativeInt(I), SizeOf(T));
      if (I > J) then goto sign_sorted;
      case SizeOf(T) of
        4: if (I.SSingle < 0) then Break;
        8: if (I.SDouble < 0) then Break;
       10: if (I.SExtended < 0) then Break;
      end;
    until (False);

    Inc(NativeInt(J), SizeOf(T));
    repeat
      Dec(NativeInt(J), SizeOf(T));
      if (J < I) then goto sign_sorted;
      case SizeOf(T) of
        4: if (J.SSingle >= 0) then Break;
        8: if (J.SDouble >= 0) then Break;
       10: if (J.SExtended >= 0) then Break;
      end;
    until (False);

    TArray.Exchange<T>(I, J);
    Inc(NativeInt(I), SizeOf(T));
    Dec(NativeInt(J), SizeOf(T));
  until (False);
sign_sorted:

  Count := SortItemCount<T>(StackItem.First, J);
  if (Count > 1) then
  begin
    if (Count <= ((SizeOf(T) div 4) * 125 - (SizeOf(T) div 10) * 1050)) then
    begin
      TArray.SortDescendingFloats<T>(StackItem.First, Count);
    end else
    case SizeOf(T) of
      4: TArray.RadixSort<Integer>(StackItem.First, Count, not $0000);
      8: TArray.RadixSort<Int64>(StackItem.First, Count, not $0000);
    else
      TArray.RadixSort<Extended>(StackItem.First, Count, not $0000);
    end;
  end;

  Count := SortItemCount<T>(I, StackItem.Last);
  if (Count > 1) then
  begin
    if (Count <= ((SizeOf(T) div 4) * 125 - (SizeOf(T) div 10) * 1050)) then
    begin
      TArray.SortDescendingFloats<T>(I, Count);
    end else
    case SizeOf(T) of
      4: TArray.RadixSort<Integer>(I, Count, $0000);
      8: TArray.RadixSort<Int64>(I, Count, $0000);
    else
      TArray.RadixSort<Extended>(I, Count, $0000);
    end;
  end;

  Result := @StackItem;
end;

class procedure TArray.SortSigneds<T>(const Values: Pointer; const Count: NativeInt);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
  P1 = ^ShortInt;
  P2 = ^SmallInt;
  P4 = ^Integer;
  P8 = ^Int64;
var
  Pivot4: Integer;
  {$ifdef LARGEINT}
    Pivot8: Int64;
  {$else .SMALLINT}
    Pivot8Low: Cardinal;
    Pivot8High, Buffer8High: Integer;
  {$endif}
  Temp: T;
  Temp4: Cardinal;
  Size: NativeInt;

  I, J, PStop: ^T;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion/radix sort
  Size := NativeInt(J) - NativeInt(I) + SizeOf(T);
  if (Size <= RADIX_BUFFER_SIZE * SizeOf(T)) and
    (
    (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) or
    (Size >= (SizeOf(T) * 45 + (SizeOf(T) div 4) * 30 - (SizeOf(T) div 8) * 220) * SizeOf(T))
    ) then
  begin
    if (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) then
    begin
      PStop := StackItem.First;
      goto insertion_init;
      repeat
        if (J = PStop) then Break;
        Dec(J);

        case SizeOf(T) of
          1: if (Pivot4 <= P1(J)^) then Continue;
          2: if (Pivot4 <= P2(J)^) then Continue;
          4: if (Pivot4 <= P4(J)^) then Continue;
        else
          {$ifdef LARGEINT}
            if (Pivot8 <= P8(J)^) then Continue;
          {$else .SMALLINT}
            Buffer8High := PPoint(J).Y;
            if (Pivot8High < Buffer8High) or
              ((Pivot8High = Buffer8High) and (Pivot8Low <= PCardinal(J)^)) then Continue;
          {$endif}
        end;

      insertion_init:
        I := J;
        case SizeOf(T) of
          1: Pivot4 := P1(J)^;
          2: Pivot4 := P2(J)^;
          4: Pivot4 := P4(J)^;
        else
          {$ifdef LARGEINT}
            Pivot8 := P8(J)^;
          {$else .SMALLINT}
            with PPoint(J)^ do
            begin
              Pivot8Low := X;
              Pivot8High := Y;
            end;
          {$endif}
        end;
      until (False);
      I^ := J^;
      case SizeOf(T) of
        1: P1(J)^ := Pivot4;
        2: P2(J)^ := Pivot4;
        4: P4(J)^ := Pivot4;
      else
        {$ifdef LARGEINT}
          P8(J)^ := Pivot8;
        {$else .SMALLINT}
          with PPoint(J)^ do
          begin
            X := Pivot8Low;
            Y := Pivot8High;
          end;
        {$endif}
      end;

      PStop := StackItem.Last;
      repeat
        Inc(J);
        if (J = PStop) then Break;

        case SizeOf(T) of
          1: Pivot4 := P1(J + 1)^;
          2: Pivot4 := P2(J + 1)^;
          4: Pivot4 := P4(J + 1)^;
        else
          {$ifdef LARGEINT}
            Pivot8 := P8(J + 1)^;
          {$else .SMALLINT}
            with PPoint(J + 1)^ do
            begin
              Pivot8Low := X;
              Pivot8High := Y;
            end;
          {$endif}
        end;
        case SizeOf(T) of
          1: if (P1(J)^ <= Pivot4) then Continue;
          2: if (P2(J)^ <= Pivot4) then Continue;
          4: if (P4(J)^ <= Pivot4) then Continue;
        else
          {$ifdef LARGEINT}
            if (P8(J)^ <= Pivot8)  then Continue;
          {$else .SMALLINT}
            Buffer8High := PPoint(J).Y;
            if (Buffer8High < Pivot8High) or
              ((Buffer8High = Pivot8High) and (PCardinal(J)^ <= Pivot8Low)) then Continue;
          {$endif}
        end;

        I := J;
        repeat
          (I + 1)^ := I^;
          Dec(I);
          case SizeOf(T) of
            1: if (P1(I)^ <= Pivot4) then Break;
            2: if (P2(I)^ <= Pivot4) then Break;
            4: if (P4(I)^ <= Pivot4) then Break;
          else
            {$ifdef LARGEINT}
              if (P8(I)^ <= Pivot8)  then Break;
            {$else .SMALLINT}
              Buffer8High := PPoint(I).Y;
              if (Buffer8High < Pivot8High) or
                ((Buffer8High = Pivot8High) and (PCardinal(I)^ <= Pivot8Low)) then Break;
            {$endif}
          end;
        until (False);

        case SizeOf(T) of
          1: P1(I + 1)^ := Pivot4;
          2: P2(I + 1)^ := Pivot4;
          4: P4(I + 1)^ := Pivot4;
        else
          {$ifdef LARGEINT}
            P8(I + 1)^ := Pivot8;
          {$else .SMALLINT}
            with PPoint(I + 1)^ do
            begin
              X := Pivot8Low;
              Y := Pivot8High;
            end;
          {$endif}
        end;
      until (False);

      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end else
    begin
      StackItem := TArray.RadixSortSigneds<T>(StackItem^);
      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end;
  end;

  // pivot
  case SizeOf(T) of
    1: Pivot4 := P1(SortItemPivot<T>(I, J))^;
    2: Pivot4 := P2(SortItemPivot<T>(I, J))^;
    4: Pivot4 := P4(SortItemPivot<T>(I, J))^;
  else
    {$ifdef LARGEINT}
      Pivot8 := P8(SortItemPivot<T>(I, J))^;
    {$else .SMALLINT}
      with PPoint(I + ((NativeInt(J) - NativeInt(I)) shr 4))^ do
      begin
        Pivot8Low := X;
        Pivot8High := Y;
      end;
    {$endif}
  end;

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);
      case SizeOf(T) of
        1: if (Pivot4 <= P1(I)^) then Break;
        2: if (Pivot4 <= P2(I)^) then Break;
        4: if (Pivot4 <= P4(I)^) then Break;
      else
        {$ifdef LARGEINT}
          if (Pivot8 <= P8(I)^) then Break;
        {$else .SMALLINT}
          Buffer8High := PPoint(I).Y;
          if (Pivot8High < Buffer8High) or
            ((Pivot8High = Buffer8High) and (Pivot8Low <= PCardinal(I)^)) then Break;
        {$endif}
      end;
    until (False);

    repeat
      Dec(J);
      case SizeOf(T) of
        1: if (P1(J)^ <= Pivot4) then Break;
        2: if (P2(J)^ <= Pivot4) then Break;
        4: if (P4(J)^ <= Pivot4) then Break;
      else
        {$ifdef LARGEINT}
          if (P8(J)^ <= Pivot8)  then Break;
        {$else .SMALLINT}
          Buffer8High := PPoint(J).Y;
          if (Buffer8High < Pivot8High) or
            ((Buffer8High = Pivot8High) and (PCardinal(J)^ <= Pivot8Low)) then Break;
        {$endif}
      end;
    until (False);

    if (I <= J) then
    begin
      {$ifdef SMALLINT}
      if (SizeOf(T) = 8) then
      begin
        Temp4 := TLMemory(Pointer(I)^).LCardinals[0];
        TLMemory(Pointer(I)^).LCardinals[0] := TLMemory(Pointer(J)^).LCardinals[0];
        TLMemory(Pointer(J)^).LCardinals[0] := Temp4;
        Temp4 := TLMemory(Pointer(I)^).LCardinals[1];
        TLMemory(Pointer(I)^).LCardinals[1] := TLMemory(Pointer(J)^).LCardinals[1];
        TLMemory(Pointer(J)^).LCardinals[1] := Temp4;
      end else
      {$endif}
      begin
        Temp := I^;
        I^ := J^;
        J^ := Temp;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.SortDescendingSigneds<T>(const Values: Pointer; const Count: NativeInt);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
  P1 = ^ShortInt;
  P2 = ^SmallInt;
  P4 = ^Integer;
  P8 = ^Int64;
var
  Pivot4: Integer;
  {$ifdef LARGEINT}
    Pivot8: Int64;
  {$else .SMALLINT}
    Pivot8Low: Cardinal;
    Pivot8High, Buffer8High: Integer;
  {$endif}
  Temp: T;
  Temp4: Cardinal;
  Size: NativeInt;

  I, J, PStop: ^T;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion/radix sort
  Size := NativeInt(J) - NativeInt(I) + SizeOf(T);
  if (Size <= RADIX_BUFFER_SIZE * SizeOf(T)) and
    (
    (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) or
    (Size >= (SizeOf(T) * 45 + (SizeOf(T) div 4) * 30 - (SizeOf(T) div 8) * 220) * SizeOf(T))
    ) then
  begin
    if (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) then
    begin
      PStop := StackItem.First;
      goto insertion_init;
      repeat
        if (J = PStop) then Break;
        Dec(J);

        case SizeOf(T) of
          1: if (Pivot4 >= P1(J)^) then Continue;
          2: if (Pivot4 >= P2(J)^) then Continue;
          4: if (Pivot4 >= P4(J)^) then Continue;
        else
          {$ifdef LARGEINT}
            if (Pivot8 >= P8(J)^) then Continue;
          {$else .SMALLINT}
            Buffer8High := PPoint(J).Y;
            if (Pivot8High > Buffer8High) or
              ((Pivot8High = Buffer8High) and (Pivot8Low >= PCardinal(J)^)) then Continue;
          {$endif}
        end;

      insertion_init:
        I := J;
        case SizeOf(T) of
          1: Pivot4 := P1(J)^;
          2: Pivot4 := P2(J)^;
          4: Pivot4 := P4(J)^;
        else
          {$ifdef LARGEINT}
            Pivot8 := P8(J)^;
          {$else .SMALLINT}
            with PPoint(J)^ do
            begin
              Pivot8Low := X;
              Pivot8High := Y;
            end;
          {$endif}
        end;
      until (False);
      I^ := J^;
      case SizeOf(T) of
        1: P1(J)^ := Pivot4;
        2: P2(J)^ := Pivot4;
        4: P4(J)^ := Pivot4;
      else
        {$ifdef LARGEINT}
          P8(J)^ := Pivot8;
        {$else .SMALLINT}
          with PPoint(J)^ do
          begin
            X := Pivot8Low;
            Y := Pivot8High;
          end;
        {$endif}
      end;

      PStop := StackItem.Last;
      repeat
        Inc(J);
        if (J = PStop) then Break;

        case SizeOf(T) of
          1: Pivot4 := P1(J + 1)^;
          2: Pivot4 := P2(J + 1)^;
          4: Pivot4 := P4(J + 1)^;
        else
          {$ifdef LARGEINT}
            Pivot8 := P8(J + 1)^;
          {$else .SMALLINT}
            with PPoint(J + 1)^ do
            begin
              Pivot8Low := X;
              Pivot8High := Y;
            end;
          {$endif}
        end;
        case SizeOf(T) of
          1: if (P1(J)^ >= Pivot4) then Continue;
          2: if (P2(J)^ >= Pivot4) then Continue;
          4: if (P4(J)^ >= Pivot4) then Continue;
        else
          {$ifdef LARGEINT}
            if (P8(J)^ >= Pivot8)  then Continue;
          {$else .SMALLINT}
            Buffer8High := PPoint(J).Y;
            if (Buffer8High > Pivot8High) or
              ((Buffer8High = Pivot8High) and (PCardinal(J)^ >= Pivot8Low)) then Continue;
          {$endif}
        end;

        I := J;
        repeat
          (I + 1)^ := I^;
          Dec(I);
          case SizeOf(T) of
            1: if (P1(I)^ >= Pivot4) then Break;
            2: if (P2(I)^ >= Pivot4) then Break;
            4: if (P4(I)^ >= Pivot4) then Break;
          else
            {$ifdef LARGEINT}
              if (P8(I)^ >= Pivot8)  then Break;
            {$else .SMALLINT}
              Buffer8High := PPoint(I).Y;
              if (Buffer8High > Pivot8High) or
                ((Buffer8High = Pivot8High) and (PCardinal(I)^ >= Pivot8Low)) then Break;
            {$endif}
          end;
        until (False);

        case SizeOf(T) of
          1: P1(I + 1)^ := Pivot4;
          2: P2(I + 1)^ := Pivot4;
          4: P4(I + 1)^ := Pivot4;
        else
          {$ifdef LARGEINT}
            P8(I + 1)^ := Pivot8;
          {$else .SMALLINT}
            with PPoint(I + 1)^ do
            begin
              X := Pivot8Low;
              Y := Pivot8High;
            end;
          {$endif}
        end;
      until (False);

      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end else
    begin
      StackItem := TArray.RadixSortDescendingSigneds<T>(StackItem^);
      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end;
  end;

  // pivot
  case SizeOf(T) of
    1: Pivot4 := P1(SortItemPivot<T>(I, J))^;
    2: Pivot4 := P2(SortItemPivot<T>(I, J))^;
    4: Pivot4 := P4(SortItemPivot<T>(I, J))^;
  else
    {$ifdef LARGEINT}
      Pivot8 := P8(SortItemPivot<T>(I, J))^;
    {$else .SMALLINT}
      with PPoint(I + ((NativeInt(J) - NativeInt(I)) shr 4))^ do
      begin
        Pivot8Low := X;
        Pivot8High := Y;
      end;
    {$endif}
  end;

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);
      case SizeOf(T) of
        1: if (Pivot4 >= P1(I)^) then Break;
        2: if (Pivot4 >= P2(I)^) then Break;
        4: if (Pivot4 >= P4(I)^) then Break;
      else
        {$ifdef LARGEINT}
          if (Pivot8 >= P8(I)^) then Break;
        {$else .SMALLINT}
          Buffer8High := PPoint(I).Y;
          if (Pivot8High > Buffer8High) or
            ((Pivot8High = Buffer8High) and (Pivot8Low >= PCardinal(I)^)) then Break;
        {$endif}
      end;
    until (False);

    repeat
      Dec(J);
      case SizeOf(T) of
        1: if (P1(J)^ >= Pivot4) then Break;
        2: if (P2(J)^ >= Pivot4) then Break;
        4: if (P4(J)^ >= Pivot4) then Break;
      else
        {$ifdef LARGEINT}
          if (P8(J)^ >= Pivot8)  then Break;
        {$else .SMALLINT}
          Buffer8High := PPoint(J).Y;
          if (Buffer8High > Pivot8High) or
            ((Buffer8High = Pivot8High) and (PCardinal(J)^ >= Pivot8Low)) then Break;
        {$endif}
      end;
    until (False);

    if (I <= J) then
    begin
      {$ifdef SMALLINT}
      if (SizeOf(T) = 8) then
      begin
        Temp4 := TLMemory(Pointer(I)^).LCardinals[0];
        TLMemory(Pointer(I)^).LCardinals[0] := TLMemory(Pointer(J)^).LCardinals[0];
        TLMemory(Pointer(J)^).LCardinals[0] := Temp4;
        Temp4 := TLMemory(Pointer(I)^).LCardinals[1];
        TLMemory(Pointer(I)^).LCardinals[1] := TLMemory(Pointer(J)^).LCardinals[1];
        TLMemory(Pointer(J)^).LCardinals[1] := Temp4;
      end else
      {$endif}
      begin
        Temp := I^;
        I^ := J^;
        J^ := Temp;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.SortUnsigneds<T>(const Values: Pointer; const Count: NativeInt);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
  P1 = ^Byte;
  P2 = ^Word;
  P4 = ^Cardinal;
  P8 = ^UInt64;
var
  Pivot4: Cardinal;
  {$ifdef LARGEINT}
    Pivot8: UInt64;
  {$else .SMALLINT}
    Pivot8Low: Cardinal;
    Pivot8High, Buffer8High: Cardinal;
  {$endif}
  Temp: T;
  Temp4: Cardinal;
  Size: NativeInt;

  I, J, PStop: ^T;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion/radix sort
  Size := NativeInt(J) - NativeInt(I) + SizeOf(T);
  if (Size <= RADIX_BUFFER_SIZE * SizeOf(T)) and
    (
    (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) or
    (Size >= (SizeOf(T) * 45 + (SizeOf(T) div 4) * 30 - (SizeOf(T) div 8) * 220) * SizeOf(T))
    ) then
  begin
    if (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) then
    begin
      PStop := StackItem.First;
      goto insertion_init;
      repeat
        if (J = PStop) then Break;
        Dec(J);

        case SizeOf(T) of
          1: if (Pivot4 <= P1(J)^) then Continue;
          2: if (Pivot4 <= P2(J)^) then Continue;
          4: if (Pivot4 <= P4(J)^) then Continue;
        else
          {$ifdef LARGEINT}
            if (Pivot8 <= P8(J)^) then Continue;
          {$else .SMALLINT}
            Buffer8High := PPoint(J).Y;
            if (Pivot8High < Buffer8High) or
              ((Pivot8High = Buffer8High) and (Pivot8Low <= PCardinal(J)^)) then Continue;
          {$endif}
        end;

      insertion_init:
        I := J;
        case SizeOf(T) of
          1: Pivot4 := P1(J)^;
          2: Pivot4 := P2(J)^;
          4: Pivot4 := P4(J)^;
        else
          {$ifdef LARGEINT}
            Pivot8 := P8(J)^;
          {$else .SMALLINT}
            with PPoint(J)^ do
            begin
              Pivot8Low := X;
              Pivot8High := Y;
            end;
          {$endif}
        end;
      until (False);
      I^ := J^;
      case SizeOf(T) of
        1: P1(J)^ := Pivot4;
        2: P2(J)^ := Pivot4;
        4: P4(J)^ := Pivot4;
      else
        {$ifdef LARGEINT}
          P8(J)^ := Pivot8;
        {$else .SMALLINT}
          with PPoint(J)^ do
          begin
            X := Pivot8Low;
            Y := Pivot8High;
          end;
        {$endif}
      end;

      PStop := StackItem.Last;
      repeat
        Inc(J);
        if (J = PStop) then Break;

        case SizeOf(T) of
          1: Pivot4 := P1(J + 1)^;
          2: Pivot4 := P2(J + 1)^;
          4: Pivot4 := P4(J + 1)^;
        else
          {$ifdef LARGEINT}
            Pivot8 := P8(J + 1)^;
          {$else .SMALLINT}
            with PPoint(J + 1)^ do
            begin
              Pivot8Low := X;
              Pivot8High := Y;
            end;
          {$endif}
        end;
        case SizeOf(T) of
          1: if (P1(J)^ <= Pivot4) then Continue;
          2: if (P2(J)^ <= Pivot4) then Continue;
          4: if (P4(J)^ <= Pivot4) then Continue;
        else
          {$ifdef LARGEINT}
            if (P8(J)^ <= Pivot8)  then Continue;
          {$else .SMALLINT}
            Buffer8High := PPoint(J).Y;
            if (Buffer8High < Pivot8High) or
              ((Buffer8High = Pivot8High) and (PCardinal(J)^ <= Pivot8Low)) then Continue;
          {$endif}
        end;

        I := J;
        repeat
          (I + 1)^ := I^;
          Dec(I);
          case SizeOf(T) of
            1: if (P1(I)^ <= Pivot4) then Break;
            2: if (P2(I)^ <= Pivot4) then Break;
            4: if (P4(I)^ <= Pivot4) then Break;
          else
            {$ifdef LARGEINT}
              if (P8(I)^ <= Pivot8)  then Break;
            {$else .SMALLINT}
              Buffer8High := PPoint(I).Y;
              if (Buffer8High < Pivot8High) or
                ((Buffer8High = Pivot8High) and (PCardinal(I)^ <= Pivot8Low)) then Break;
            {$endif}
          end;
        until (False);

        case SizeOf(T) of
          1: P1(I + 1)^ := Pivot4;
          2: P2(I + 1)^ := Pivot4;
          4: P4(I + 1)^ := Pivot4;
        else
          {$ifdef LARGEINT}
            P8(I + 1)^ := Pivot8;
          {$else .SMALLINT}
            with PPoint(I + 1)^ do
            begin
              X := Pivot8Low;
              Y := Pivot8High;
            end;
          {$endif}
        end;
      until (False);

      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end else
    begin
      StackItem := TArray.RadixSortUnsigneds<T>(StackItem^);
      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end;
  end;

  // pivot
  case SizeOf(T) of
    1: Pivot4 := P1(SortItemPivot<T>(I, J))^;
    2: Pivot4 := P2(SortItemPivot<T>(I, J))^;
    4: Pivot4 := P4(SortItemPivot<T>(I, J))^;
  else
    {$ifdef LARGEINT}
      Pivot8 := P8(SortItemPivot<T>(I, J))^;
    {$else .SMALLINT}
      with PPoint(I + ((NativeInt(J) - NativeInt(I)) shr 4))^ do
      begin
        Pivot8Low := X;
        Pivot8High := Y;
      end;
    {$endif}
  end;

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);
      case SizeOf(T) of
        1: if (Pivot4 <= P1(I)^) then Break;
        2: if (Pivot4 <= P2(I)^) then Break;
        4: if (Pivot4 <= P4(I)^) then Break;
      else
        {$ifdef LARGEINT}
          if (Pivot8 <= P8(I)^) then Break;
        {$else .SMALLINT}
          Buffer8High := PPoint(I).Y;
          if (Pivot8High < Buffer8High) or
            ((Pivot8High = Buffer8High) and (Pivot8Low <= PCardinal(I)^)) then Break;
        {$endif}
      end;
    until (False);

    repeat
      Dec(J);
      case SizeOf(T) of
        1: if (P1(J)^ <= Pivot4) then Break;
        2: if (P2(J)^ <= Pivot4) then Break;
        4: if (P4(J)^ <= Pivot4) then Break;
      else
        {$ifdef LARGEINT}
          if (P8(J)^ <= Pivot8)  then Break;
        {$else .SMALLINT}
          Buffer8High := PPoint(J).Y;
          if (Buffer8High < Pivot8High) or
            ((Buffer8High = Pivot8High) and (PCardinal(J)^ <= Pivot8Low)) then Break;
        {$endif}
      end;
    until (False);

    if (I <= J) then
    begin
      {$ifdef SMALLINT}
      if (SizeOf(T) = 8) then
      begin
        Temp4 := TLMemory(Pointer(I)^).LCardinals[0];
        TLMemory(Pointer(I)^).LCardinals[0] := TLMemory(Pointer(J)^).LCardinals[0];
        TLMemory(Pointer(J)^).LCardinals[0] := Temp4;
        Temp4 := TLMemory(Pointer(I)^).LCardinals[1];
        TLMemory(Pointer(I)^).LCardinals[1] := TLMemory(Pointer(J)^).LCardinals[1];
        TLMemory(Pointer(J)^).LCardinals[1] := Temp4;
      end else
      {$endif}
      begin
        Temp := I^;
        I^ := J^;
        J^ := Temp;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.SortDescendingUnsigneds<T>(const Values: Pointer; const Count: NativeInt);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
  P1 = ^Byte;
  P2 = ^Word;
  P4 = ^Cardinal;
  P8 = ^UInt64;
var
  Pivot4: Cardinal;
  {$ifdef LARGEINT}
    Pivot8: UInt64;
  {$else .SMALLINT}
    Pivot8Low: Cardinal;
    Pivot8High, Buffer8High: Cardinal;
  {$endif}
  Temp: T;
  Temp4: Cardinal;
  Size: NativeInt;

  I, J, PStop: ^T;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion/radix sort
  Size := NativeInt(J) - NativeInt(I) + SizeOf(T);
  if (Size <= RADIX_BUFFER_SIZE * SizeOf(T)) and
    (
    (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) or
    (Size >= (SizeOf(T) * 45 + (SizeOf(T) div 4) * 30 - (SizeOf(T) div 8) * 220) * SizeOf(T))
    ) then
  begin
    if (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) then
    begin
      PStop := StackItem.First;
      goto insertion_init;
      repeat
        if (J = PStop) then Break;
        Dec(J);

        case SizeOf(T) of
          1: if (Pivot4 >= P1(J)^) then Continue;
          2: if (Pivot4 >= P2(J)^) then Continue;
          4: if (Pivot4 >= P4(J)^) then Continue;
        else
          {$ifdef LARGEINT}
            if (Pivot8 >= P8(J)^) then Continue;
          {$else .SMALLINT}
            Buffer8High := PPoint(J).Y;
            if (Pivot8High > Buffer8High) or
              ((Pivot8High = Buffer8High) and (Pivot8Low >= PCardinal(J)^)) then Continue;
          {$endif}
        end;

      insertion_init:
        I := J;
        case SizeOf(T) of
          1: Pivot4 := P1(J)^;
          2: Pivot4 := P2(J)^;
          4: Pivot4 := P4(J)^;
        else
          {$ifdef LARGEINT}
            Pivot8 := P8(J)^;
          {$else .SMALLINT}
            with PPoint(J)^ do
            begin
              Pivot8Low := X;
              Pivot8High := Y;
            end;
          {$endif}
        end;
      until (False);
      I^ := J^;
      case SizeOf(T) of
        1: P1(J)^ := Pivot4;
        2: P2(J)^ := Pivot4;
        4: P4(J)^ := Pivot4;
      else
        {$ifdef LARGEINT}
          P8(J)^ := Pivot8;
        {$else .SMALLINT}
          with PPoint(J)^ do
          begin
            X := Pivot8Low;
            Y := Pivot8High;
          end;
        {$endif}
      end;

      PStop := StackItem.Last;
      repeat
        Inc(J);
        if (J = PStop) then Break;

        case SizeOf(T) of
          1: Pivot4 := P1(J + 1)^;
          2: Pivot4 := P2(J + 1)^;
          4: Pivot4 := P4(J + 1)^;
        else
          {$ifdef LARGEINT}
            Pivot8 := P8(J + 1)^;
          {$else .SMALLINT}
            with PPoint(J + 1)^ do
            begin
              Pivot8Low := X;
              Pivot8High := Y;
            end;
          {$endif}
        end;
        case SizeOf(T) of
          1: if (P1(J)^ >= Pivot4) then Continue;
          2: if (P2(J)^ >= Pivot4) then Continue;
          4: if (P4(J)^ >= Pivot4) then Continue;
        else
          {$ifdef LARGEINT}
            if (P8(J)^ >= Pivot8)  then Continue;
          {$else .SMALLINT}
            Buffer8High := PPoint(J).Y;
            if (Buffer8High > Pivot8High) or
              ((Buffer8High = Pivot8High) and (PCardinal(J)^ >= Pivot8Low)) then Continue;
          {$endif}
        end;

        I := J;
        repeat
          (I + 1)^ := I^;
          Dec(I);
          case SizeOf(T) of
            1: if (P1(I)^ >= Pivot4) then Break;
            2: if (P2(I)^ >= Pivot4) then Break;
            4: if (P4(I)^ >= Pivot4) then Break;
          else
            {$ifdef LARGEINT}
              if (P8(I)^ >= Pivot8)  then Break;
            {$else .SMALLINT}
              Buffer8High := PPoint(I).Y;
              if (Buffer8High > Pivot8High) or
                ((Buffer8High = Pivot8High) and (PCardinal(I)^ >= Pivot8Low)) then Break;
            {$endif}
          end;
        until (False);

        case SizeOf(T) of
          1: P1(I + 1)^ := Pivot4;
          2: P2(I + 1)^ := Pivot4;
          4: P4(I + 1)^ := Pivot4;
        else
          {$ifdef LARGEINT}
            P8(I + 1)^ := Pivot8;
          {$else .SMALLINT}
            with PPoint(I + 1)^ do
            begin
              X := Pivot8Low;
              Y := Pivot8High;
            end;
          {$endif}
        end;
      until (False);

      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end else
    begin
      StackItem := TArray.RadixSortDescendingUnsigneds<T>(StackItem^);
      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end;
  end;

  // pivot
  case SizeOf(T) of
    1: Pivot4 := P1(SortItemPivot<T>(I, J))^;
    2: Pivot4 := P2(SortItemPivot<T>(I, J))^;
    4: Pivot4 := P4(SortItemPivot<T>(I, J))^;
  else
    {$ifdef LARGEINT}
      Pivot8 := P8(SortItemPivot<T>(I, J))^;
    {$else .SMALLINT}
      with PPoint(I + ((NativeInt(J) - NativeInt(I)) shr 4))^ do
      begin
        Pivot8Low := X;
        Pivot8High := Y;
      end;
    {$endif}
  end;

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);
      case SizeOf(T) of
        1: if (Pivot4 >= P1(I)^) then Break;
        2: if (Pivot4 >= P2(I)^) then Break;
        4: if (Pivot4 >= P4(I)^) then Break;
      else
        {$ifdef LARGEINT}
          if (Pivot8 >= P8(I)^) then Break;
        {$else .SMALLINT}
          Buffer8High := PPoint(I).Y;
          if (Pivot8High > Buffer8High) or
            ((Pivot8High = Buffer8High) and (Pivot8Low >= PCardinal(I)^)) then Break;
        {$endif}
      end;
    until (False);

    repeat
      Dec(J);
      case SizeOf(T) of
        1: if (P1(J)^ >= Pivot4) then Break;
        2: if (P2(J)^ >= Pivot4) then Break;
        4: if (P4(J)^ >= Pivot4) then Break;
      else
        {$ifdef LARGEINT}
          if (P8(J)^ >= Pivot8)  then Break;
        {$else .SMALLINT}
          Buffer8High := PPoint(J).Y;
          if (Buffer8High > Pivot8High) or
            ((Buffer8High = Pivot8High) and (PCardinal(J)^ >= Pivot8Low)) then Break;
        {$endif}
      end;
    until (False);

    if (I <= J) then
    begin
      {$ifdef SMALLINT}
      if (SizeOf(T) = 8) then
      begin
        Temp4 := TLMemory(Pointer(I)^).LCardinals[0];
        TLMemory(Pointer(I)^).LCardinals[0] := TLMemory(Pointer(J)^).LCardinals[0];
        TLMemory(Pointer(J)^).LCardinals[0] := Temp4;
        Temp4 := TLMemory(Pointer(I)^).LCardinals[1];
        TLMemory(Pointer(I)^).LCardinals[1] := TLMemory(Pointer(J)^).LCardinals[1];
        TLMemory(Pointer(J)^).LCardinals[1] := Temp4;
      end else
      {$endif}
      begin
        Temp := I^;
        I^ := J^;
        J^ := Temp;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.SortFloats<T>(const Values: Pointer; const Count: NativeInt);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
  P4 = ^Single;
  P8 = ^Double;
  PE = ^Extended;
var
  Pivot4: Single;
  Pivot8: Double;
  PivotE: Extended;
  TempNative: NativeUInt;
  Size: NativeInt;

  I, J, PStop: ^T;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion/radix sort
  Size := NativeInt(J) - NativeInt(I) + SizeOf(T);
  if (Size <= RADIX_BUFFER_SIZE * SizeOf(T)) and
    (
    (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) or
    (Size >= ((SizeOf(T) div 4) * 125 - (SizeOf(T) div 10) * 50) * SizeOf(T))
    ) then
  begin
    if (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) then
    begin
      PStop := StackItem.First;
      goto insertion_init;
      repeat
        if (J = PStop) then Break;
        Dec(J);

        case SizeOf(T) of
          4: if (Pivot4 <= P4(J)^) then Continue;
          8: if (Pivot8 <= P8(J)^) then Continue;
        else
          if (PivotE <= PE(J)^) then Continue;
        end;

      insertion_init:
        I := J;
        case SizeOf(T) of
          4: Pivot4 := P4(J)^;
          8: Pivot8 := P8(J)^;
        else
          PivotE := PE(J)^;
        end;
      until (False);
      I^ := J^;
      case SizeOf(T) of
        4: P4(J)^ := Pivot4;
        8: P8(J)^ := Pivot8;
      else
        PE(J)^ := PivotE;
      end;

      PStop := StackItem.Last;
      repeat
        Inc(J);
        if (J = PStop) then Break;

        case SizeOf(T) of
          4: Pivot4 := P4(J + 1)^;
          8: Pivot8 := P8(J + 1)^;
        else
          PivotE := PE(J + 1)^;
        end;
        case SizeOf(T) of
          4: if (P4(J)^ <= Pivot4) then Continue;
          8: if (P8(J)^ <= Pivot8) then Continue;
        else
          if (PE(J)^ <= PivotE)  then Continue;
        end;

        I := J;
        repeat
          (I + 1)^ := I^;
          Dec(I);
          case SizeOf(T) of
            4: if (P4(I)^ <= Pivot4) then Break;
            8: if (P8(I)^ <= Pivot8) then Break;
          else
            if (PE(I)^ <= PivotE)  then Break;
          end;
        until (False);

        case SizeOf(T) of
          4: P4(I + 1)^ := Pivot4;
          8: P8(I + 1)^ := Pivot8;
        else
          PE(I + 1)^ := PivotE;
        end;
      until (False);

      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end else
    begin
      StackItem := TArray.RadixSortFloats<T>(StackItem^);
      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end;
  end;

  // pivot
  case SizeOf(T) of
    4: Pivot4 := P4(SortItemPivot<T>(I, J))^;
    8: Pivot8 := P8(SortItemPivot<T>(I, J))^;
  else
    PivotE := PE(SortItemPivot<T>(I, J))^;
  end;

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);
      case SizeOf(T) of
        4: if (Pivot4 <= P4(I)^) then Break;
        8: if (Pivot8 <= P8(I)^) then Break;
      else
        if (PivotE <= PE(I)^) then Break;
      end;
    until (False);

    repeat
      Dec(J);
      case SizeOf(T) of
        4: if (P4(J)^ <= Pivot4) then Break;
        8: if (P8(J)^ <= Pivot8) then Break;
      else
        if (PE(J)^ <= PivotE)  then Break;
      end;
    until (False);

    if (I <= J) then
    begin
      if (SizeOf(T) = 4) then
      begin
        TempNative := TLMemory(Pointer(I)^).LCardinals[0];
        TLMemory(Pointer(I)^).LCardinals[0] := TLMemory(Pointer(J)^).LCardinals[0];
        TLMemory(Pointer(J)^).LCardinals[0] := TempNative;
      end else
      begin
        TempNative := TLMemory(Pointer(I)^).LNatives[0];
        TLMemory(Pointer(I)^).LNatives[0] := TLMemory(Pointer(J)^).LNatives[0];
        TLMemory(Pointer(J)^).LNatives[0] := TempNative;

        if (SizeOf(T) >= 2 * SizeOf(NativeUInt)) then
        begin
          TempNative := TLMemory(Pointer(I)^).LNatives[1];
          TLMemory(Pointer(I)^).LNatives[1] := TLMemory(Pointer(J)^).LNatives[1];
          TLMemory(Pointer(J)^).LNatives[1] := TempNative;
        end;

        if (SizeOf(T) = 10) then
        begin
          TempNative := TLMemory(Pointer(I)^).LWords[4];
          TLMemory(Pointer(I)^).LWords[4] := TLMemory(Pointer(J)^).LWords[4];
          TLMemory(Pointer(J)^).LWords[4] := TempNative;
        end;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.SortDescendingFloats<T>(const Values: Pointer; const Count: NativeInt);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
  P4 = ^Single;
  P8 = ^Double;
  PE = ^Extended;
var
  Pivot4: Single;
  Pivot8: Double;
  PivotE: Extended;
  TempNative: NativeUInt;
  Size: NativeInt;

  I, J, PStop: ^T;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion/radix sort
  Size := NativeInt(J) - NativeInt(I) + SizeOf(T);
  if (Size <= RADIX_BUFFER_SIZE * SizeOf(T)) and
    (
    (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) or
    (Size >= ((SizeOf(T) div 4) * 125 - (SizeOf(T) div 10) * 50) * SizeOf(T))
    ) then
  begin
    if (Size <= INSERTION_SORT_LIMIT * SizeOf(T)) then
    begin
      PStop := StackItem.First;
      goto insertion_init;
      repeat
        if (J = PStop) then Break;
        Dec(J);

        case SizeOf(T) of
          4: if (Pivot4 >= P4(J)^) then Continue;
          8: if (Pivot8 >= P8(J)^) then Continue;
        else
          if (PivotE >= PE(J)^) then Continue;
        end;

      insertion_init:
        I := J;
        case SizeOf(T) of
          4: Pivot4 := P4(J)^;
          8: Pivot8 := P8(J)^;
        else
          PivotE := PE(J)^;
        end;
      until (False);
      I^ := J^;
      case SizeOf(T) of
        4: P4(J)^ := Pivot4;
        8: P8(J)^ := Pivot8;
      else
        PE(J)^ := PivotE;
      end;

      PStop := StackItem.Last;
      repeat
        Inc(J);
        if (J = PStop) then Break;

        case SizeOf(T) of
          4: Pivot4 := P4(J + 1)^;
          8: Pivot8 := P8(J + 1)^;
        else
          PivotE := PE(J + 1)^;
        end;
        case SizeOf(T) of
          4: if (P4(J)^ >= Pivot4) then Continue;
          8: if (P8(J)^ >= Pivot8) then Continue;
        else
          if (PE(J)^ >= PivotE)  then Continue;
        end;

        I := J;
        repeat
          (I + 1)^ := I^;
          Dec(I);
          case SizeOf(T) of
            4: if (P4(I)^ >= Pivot4) then Break;
            8: if (P8(I)^ >= Pivot8) then Break;
          else
            if (PE(I)^ >= PivotE)  then Break;
          end;
        until (False);

        case SizeOf(T) of
          4: P4(I + 1)^ := Pivot4;
          8: P8(I + 1)^ := Pivot8;
        else
          PE(I + 1)^ := PivotE;
        end;
      until (False);

      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end else
    begin
      StackItem := TArray.RadixSortDescendingFloats<T>(StackItem^);
      if (StackItem <> @STACK[0]) then goto proc_loop;
      Exit;
    end;
  end;

  // pivot
  case SizeOf(T) of
    4: Pivot4 := P4(SortItemPivot<T>(I, J))^;
    8: Pivot8 := P8(SortItemPivot<T>(I, J))^;
  else
    PivotE := PE(SortItemPivot<T>(I, J))^;
  end;

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);
      case SizeOf(T) of
        4: if (Pivot4 >= P4(I)^) then Break;
        8: if (Pivot8 >= P8(I)^) then Break;
      else
        if (PivotE >= PE(I)^) then Break;
      end;
    until (False);

    repeat
      Dec(J);
      case SizeOf(T) of
        4: if (P4(J)^ >= Pivot4) then Break;
        8: if (P8(J)^ >= Pivot8) then Break;
      else
        if (PE(J)^ >= PivotE)  then Break;
      end;
    until (False);

    if (I <= J) then
    begin
      if (SizeOf(T) = 4) then
      begin
        TempNative := TLMemory(Pointer(I)^).LCardinals[0];
        TLMemory(Pointer(I)^).LCardinals[0] := TLMemory(Pointer(J)^).LCardinals[0];
        TLMemory(Pointer(J)^).LCardinals[0] := TempNative;
      end else
      begin
        TempNative := TLMemory(Pointer(I)^).LNatives[0];
        TLMemory(Pointer(I)^).LNatives[0] := TLMemory(Pointer(J)^).LNatives[0];
        TLMemory(Pointer(J)^).LNatives[0] := TempNative;

        if (SizeOf(T) >= 2 * SizeOf(NativeUInt)) then
        begin
          TempNative := TLMemory(Pointer(I)^).LNatives[1];
          TLMemory(Pointer(I)^).LNatives[1] := TLMemory(Pointer(J)^).LNatives[1];
          TLMemory(Pointer(J)^).LNatives[1] := TempNative;
        end;

        if (SizeOf(T) = 10) then
        begin
          TempNative := TLMemory(Pointer(I)^).LWords[4];
          TLMemory(Pointer(I)^).LWords[4] := TLMemory(Pointer(J)^).LWords[4];
          TLMemory(Pointer(J)^).LWords[4] := TempNative;
        end;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.SortBinaries<T>(const Values: Pointer; const Count: NativeInt; var PivotBig: T);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
var
  Index: NativeInt;
  Temp1: Byte;
  Temp2: Word;
  Temp4: Cardinal;
  TempNative: NativeUInt;

  I, J, PStop: ^T;
  Pivot: packed record
  case Integer of
    0: (Ptr: Pointer);
    1: (Data: array[0..BUFFER_SIZE - 1] of Byte);
  end;
  X, Y: NativeUInt;
  Buffer: Pointer;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion sort
  if (SizeOf(T) <= SizeOf(Pivot)) then
  if ((NativeUInt(J) - NativeUInt(I)) < INSERTION_SORT_LIMIT * SizeOf(T)) then
  begin
    PStop := StackItem.First;
    goto insertion_init;
    repeat
      if (J = PStop) then Break;
      Dec(J);

      // compare pivot/J
      Y := TArray.SortBinaryMarker<T>(J);
      if (X < Y) then Continue;
      if (X = Y) then
      begin
        if (GetTypeKind(T) = tkMethod) then
        begin
          if (NativeUInt(Pivot.Ptr) <= NativeUInt(Pointer(J)^)) then Continue;
        end else
        if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
        begin
          Buffer := Pointer(Pointer(J)^);
          if (Pivot.Ptr = Buffer) then Continue;
          case GetTypeKind(T) of
            tkLString: if (InterfaceDefaults.Compare_LStr(nil, Pivot.Ptr, Buffer) <= 0) then Continue;
            tkWString: if (InterfaceDefaults.Compare_WStr(nil, Pivot.Ptr, Buffer) <= 0) then Continue;
            tkUString: if (InterfaceDefaults.Compare_UStr(nil, Pivot.Ptr, Buffer) <= 0) then Continue;
           tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Pivot.Ptr, Buffer) <= 0) then Continue;
          end;
        end else
        begin
          if (SortBinaryComparer<T>(@Pivot, J) <= 0) then Continue;
        end;
      end;
    insertion_init:
      I := J;
      // pivot := J^
      if (SizeOf(T) = SizeOf(Pointer)) then
      begin
        Pivot.Ptr := Pointer(Pointer(J)^);
      end else
      begin
        TArray.CopyMemory<T>(@Pivot, J);
      end;
      X := TArray.SortBinaryMarker<T>(@Pivot);
    until (False);
    // I^ := J^
    if (SizeOf(T) = SizeOf(Pointer)) then
    begin
      PNativeUInt(I)^ := PNativeUInt(J)^;
    end else
    begin
      TArray.CopyMemory<T>(I, J);
    end;
    // J := pivot
    if (SizeOf(T) = SizeOf(Pointer)) then
    begin
      Pointer(Pointer(J)^) := Pivot.Ptr;
    end else
    begin
      TArray.CopyMemory<T>(J, @Pivot);
    end;

    PStop := StackItem.Last;
    repeat
      Inc(J);
      if (J = PStop) then Break;

      // pivot := J[1]
      if (SizeOf(T) = SizeOf(Pointer)) then
      begin
        Pivot.Ptr := Pointer(Pointer(J + 1)^);
      end else
      begin
        TArray.CopyMemory<T>(@Pivot, J + 1);
      end;
      X := TArray.SortBinaryMarker<T>(@Pivot);

      // compare J/pivot
      Y := TArray.SortBinaryMarker<T>(J);
      if (Y < X) then Continue;
      if (Y = X) then
      begin
        if (GetTypeKind(T) = tkMethod) then
        begin
          if (NativeUInt(Pointer(J)^) <= NativeUInt(Pivot.Ptr)) then Continue;
        end else
        if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
        begin
          Buffer := Pointer(Pointer(J)^);
          if (Buffer = Pivot.Ptr) then Continue;
          case GetTypeKind(T) of
            tkLString: if (InterfaceDefaults.Compare_LStr(nil, Buffer, Pivot.Ptr) <= 0) then Continue;
            tkWString: if (InterfaceDefaults.Compare_WStr(nil, Buffer, Pivot.Ptr) <= 0) then Continue;
            tkUString: if (InterfaceDefaults.Compare_UStr(nil, Buffer, Pivot.Ptr) <= 0) then Continue;
           tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Buffer, Pivot.Ptr) <= 0) then Continue;
          end;
        end else
        begin
          if (SortBinaryComparer<T>(J, @Pivot) <= 0) then Continue;
        end;
      end;

      I := J;
      repeat
        // I[1] := I^;
        if (SizeOf(T) = SizeOf(Pointer)) then
        begin
          PNativeUInt(I + 1)^ := PNativeUInt(I)^;
        end else
        begin
          TArray.CopyMemory<T>(I + 1, I);
        end;
        Dec(I);

        // compare I/pivot
        Y := TArray.SortBinaryMarker<T>(I);
        if (Y < X) then Break;
        if (Y = X) then
        begin
          if (GetTypeKind(T) = tkMethod) then
          begin
            if (NativeUInt(Pointer(I)^) <= NativeUInt(Pivot.Ptr)) then Break;
          end else
          if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
          begin
            Buffer := Pointer(Pointer(I)^);
            if (Buffer = Pivot.Ptr) then Break;
            case GetTypeKind(T) of
              tkLString: if (InterfaceDefaults.Compare_LStr(nil, Buffer, Pivot.Ptr) <= 0) then Break;
              tkWString: if (InterfaceDefaults.Compare_WStr(nil, Buffer, Pivot.Ptr) <= 0) then Break;
              tkUString: if (InterfaceDefaults.Compare_UStr(nil, Buffer, Pivot.Ptr) <= 0) then Break;
             tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Buffer, Pivot.Ptr) <= 0) then Break;
            end;
          end else
          begin
            if (SortBinaryComparer<T>(I, @Pivot) <= 0) then Break;
          end;
        end;
      until (False);

      // I[1] := pivot
      if (SizeOf(T) = SizeOf(Pointer)) then
      begin
        Pointer(Pointer(I + 1)^) := Pivot.Ptr;
      end else
      begin
        TArray.CopyMemory<T>(I + 1, @Pivot);
      end;
    until (False);

    if (StackItem <> @STACK[0]) then goto proc_loop;
    Exit;
  end;

  // pivot
  if (SizeOf(T) <= SizeOf(Pivot)) then
  begin
    if (SizeOf(T) = SizeOf(Pointer)) then
    begin
      Pivot.Ptr := Pointer(SortItemPivot<T>(I, J)^);
    end else
    begin
      TArray.CopyMemory<T>(@Pivot, SortItemPivot<T>(I, J));
    end;
    X := TArray.SortBinaryMarker<T>(@Pivot);
  end else
  begin
    TArray.CopyMemory<T>(@PivotBig, SortItemPivot<T>(I, J));
    X := TArray.SortBinaryMarker<T>(@PivotBig);
  end;

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);

      Y := TArray.SortBinaryMarker<T>(I);
      if (X < Y) then Break;
      if (X = Y) then
      begin
        if (SizeOf(T) <= SizeOf(Pivot)) then
        begin
          // compare pivot/I
          if (GetTypeKind(T) = tkMethod) then
          begin
            if (NativeUInt(Pivot.Ptr) <= NativeUInt(Pointer(I)^)) then Break;
          end else
          if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
          begin
            Buffer := Pointer(Pointer(I)^);
            if (Pivot.Ptr = Buffer) then Break;
            case GetTypeKind(T) of
              tkLString: if (InterfaceDefaults.Compare_LStr(nil, Pivot.Ptr, Buffer) <= 0) then Break;
              tkWString: if (InterfaceDefaults.Compare_WStr(nil, Pivot.Ptr, Buffer) <= 0) then Break;
              tkUString: if (InterfaceDefaults.Compare_UStr(nil, Pivot.Ptr, Buffer) <= 0) then Break;
             tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Pivot.Ptr, Buffer) <= 0) then Break;
            end;
          end else
          begin
            if (SortBinaryComparer<T>(@Pivot, I) <= 0) then Break;
          end;
        end else
        begin
          if (SortBinaryComparer<T>(@PivotBig, I) <= 0) then Break;
        end;
      end;
    until (False);

    repeat
      Dec(J);

      Y := TArray.SortBinaryMarker<T>(J);
      if (Y < X) then Break;
      if (Y = X) then
      begin
        if (SizeOf(T) <= SizeOf(Pivot)) then
        begin
          // compare J/pivot
          if (GetTypeKind(T) = tkMethod) then
          begin
            if (NativeUInt(Pointer(J)^) <= NativeUInt(Pivot.Ptr)) then Break;
          end else
          if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
          begin
            Buffer := Pointer(Pointer(J)^);
            if (Buffer = Pivot.Ptr) then Break;
            case GetTypeKind(T) of
              tkLString: if (InterfaceDefaults.Compare_LStr(nil, Buffer, Pivot.Ptr) <= 0) then Break;
              tkWString: if (InterfaceDefaults.Compare_WStr(nil, Buffer, Pivot.Ptr) <= 0) then Break;
              tkUString: if (InterfaceDefaults.Compare_UStr(nil, Buffer, Pivot.Ptr) <= 0) then Break;
             tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Buffer, Pivot.Ptr) <= 0) then Break;
            end;
          end else
          begin
            if (SortBinaryComparer<T>(J, @Pivot) <= 0) then Break;
          end;
        end else
        begin
          if (SortBinaryComparer<T>(J, @PivotBig) <= 0) then Break;
        end;
      end;
    until (False);

    if (I <= J) then
    begin
      // TArray.Exchange<T>(I, J);
      case SizeOf(T) of
        0: ;
        1:
        begin
          Temp1 := PLMemory(I).LBytes[0];
          PLMemory(I).LBytes[0] := PRMemory(J).RBytes[0];
          PRMemory(J).RBytes[0] := Temp1;
        end;
        2:
        begin
          Temp2 := PLMemory(I).LWords[0];
          PLMemory(I).LWords[0] := PRMemory(J).RWords[0];
          PRMemory(J).RWords[0] := Temp2;
        end;
        3:
        begin
          Temp2 := PLMemory(I).LWords[0];
          PLMemory(I).LWords[0] := PRMemory(J).RWords[0];
          PRMemory(J).RWords[0] := Temp2;

          Temp1 := PLMemory(I).LBytes[2];
          PLMemory(I).LBytes[2] := PRMemory(J).RBytes[2];
          PRMemory(J).RBytes[2] := Temp1;
        end;
        4..7:
        begin
          Temp4 := PLMemory(I).LCardinals[0];
          PLMemory(I).LCardinals[0] := PRMemory(J).RCardinals[0];
          PRMemory(J).RCardinals[0] := Temp4;

          case SizeOf(T) of
            5:
            begin
              Temp1 := PLMemory(I).LBytes[4];
              PLMemory(I).LBytes[4] := PRMemory(J).RBytes[4];
              PRMemory(J).RBytes[4] := Temp1;
            end;
            6:
            begin
              Temp2 := PLMemory(I).LWords[2];
              PLMemory(I).LWords[2] := PRMemory(J).RWords[2];
              PRMemory(J).RWords[2] := Temp2;
            end;
            7:
            begin
              Temp2 := PLMemory(I).LWords[2];
              PLMemory(I).LWords[2] := PRMemory(J).RWords[2];
              PRMemory(J).RWords[2] := Temp2;
              Temp1 := PLMemory(I).LBytes[6];
              PLMemory(I).LBytes[6] := PRMemory(J).RBytes[6];
              PRMemory(J).RBytes[6] := Temp1;
            end;
          end;
        end;
        8..16:
        begin
          TempNative := PLMemory(I).LNatives[0];
          PLMemory(I).LNatives[0] := PRMemory(J).RNatives[0];
          PRMemory(J).RNatives[0] := TempNative;

          if (SizeOf(T) >= SizeOf(NativeUInt) * 2) then
          begin
            TempNative := PLMemory(I).LNatives[1];
            PLMemory(I).LNatives[1] := PRMemory(J).RNatives[1];
            PRMemory(J).RNatives[1] := TempNative;
          end;

          if (SizeOf(T) >= SizeOf(NativeUInt) * 3) then
          begin
            TempNative := PLMemory(I).LNatives[2];
            PLMemory(I).LNatives[2] := PRMemory(J).RNatives[2];
            PRMemory(J).RNatives[2] := TempNative;
          end;

          if (SizeOf(T)  = SizeOf(NativeUInt) * 4) then
          begin
            TempNative := PLMemory(I).LNatives[3];
            PLMemory(I).LNatives[3] := PRMemory(J).RNatives[3];
            PRMemory(J).RNatives[3] := TempNative;
          end;

          {$ifdef LARGEINT}
          case SizeOf(T) of
            12, 13, 14, 15:
            begin
              Temp4 := PLMemory(I).LCardinals[2];
              PLMemory(I).LCardinals[2] := PRMemory(J).RCardinals[2];
              PRMemory(J).RCardinals[2] := Temp4;
            end;
          end;
          {$endif}

          case SizeOf(T) of
            9:
            begin
              Temp1 := PLMemory(I).LBytes[8];
              PLMemory(I).LBytes[8] := PRMemory(J).RBytes[8];
              PRMemory(J).RBytes[8] := Temp1;
            end;
            10:
            begin
              Temp2 := PLMemory(I).LWords[4];
              PLMemory(I).LWords[4] := PRMemory(J).RWords[4];
              PRMemory(J).RWords[4] := Temp2;
            end;
            11:
            begin
              Temp2 := PLMemory(I).LWords[4];
              PLMemory(I).LWords[4] := PRMemory(J).RWords[4];
              PRMemory(J).RWords[4] := Temp2;
              Temp1 := PLMemory(I).LBytes[10];
              PLMemory(I).LBytes[10] := PRMemory(J).RBytes[10];
              PRMemory(J).RBytes[10] := Temp1;
            end;
            13:
            begin
              Temp2 := PLMemory(I).LWords[5];
              PLMemory(I).LWords[5] := PRMemory(J).RWords[5];
              PRMemory(J).RWords[5] := Temp2;
              Temp1 := PLMemory(I).LBytes[12];
              PLMemory(I).LBytes[12] := PRMemory(J).RBytes[12];
              PRMemory(J).RBytes[12] := Temp1;
            end;
            14:
            begin
              Temp2 := PLMemory(I).LWords[6];
              PLMemory(I).LWords[6] := PRMemory(J).RWords[6];
              PRMemory(J).RWords[6] := Temp2;
            end;
            15:
            begin
              Temp2 := PLMemory(I).LWords[6];
              PLMemory(I).LWords[6] := PRMemory(J).RWords[6];
              PRMemory(J).RWords[6] := Temp2;
              Temp1 := PLMemory(I).LBytes[14];
              PLMemory(I).LBytes[14] := PRMemory(J).RBytes[14];
              PRMemory(J).RBytes[14] := Temp1;
            end;
          end;
        end;
      else
        Index := 0;
        repeat
          TempNative := PLMemory(I).LNatives[Index];
          PLMemory(I).LNatives[Index] := PRMemory(J).RNatives[Index];
          PRMemory(J).RNatives[Index] := TempNative;
          Inc(Index);
        until (Index = SizeOf(T) div SizeOf(NativeUInt));

        if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
        begin
          {$ifdef LARGEINT}
          if (SizeOf(T) and 4 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Cardinal) - 1;
            Temp4 := PLMemory(I).LCardinals[Index];
            PLMemory(I).LCardinals[Index] := PRMemory(J).RCardinals[Index];
            PRMemory(J).RCardinals[Index] := Temp4;
          end;
          {$endif}

          if (SizeOf(T) and 2 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Word) - 1;
            Temp2 := PLMemory(I).LWords[Index];
            PLMemory(I).LWords[Index] := PRMemory(J).RWords[Index];
            PRMemory(J).RWords[Index] := Temp2;
          end;

          if (SizeOf(T) and 1 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Byte) - 1;
            Temp1 := PLMemory(I).LBytes[Index];
            PLMemory(I).LBytes[Index] := PRMemory(J).RBytes[Index];
            PRMemory(J).RBytes[Index] := Temp1;
          end;
        end;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.SortDescendingBinaries<T>(const Values: Pointer; const Count: NativeInt; var PivotBig: T);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
var
  Index: NativeInt;
  Temp1: Byte;
  Temp2: Word;
  Temp4: Cardinal;
  TempNative: NativeUInt;

  I, J, PStop: ^T;
  Pivot: packed record
  case Integer of
    0: (Ptr: Pointer);
    1: (Data: array[0..BUFFER_SIZE - 1] of Byte);
  end;
  X, Y: NativeUInt;
  Buffer: Pointer;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion sort
  if (SizeOf(T) <= SizeOf(Pivot)) then
  if ((NativeUInt(J) - NativeUInt(I)) < INSERTION_SORT_LIMIT * SizeOf(T)) then
  begin
    PStop := StackItem.First;
    goto insertion_init;
    repeat
      if (J = PStop) then Break;
      Dec(J);

      // compare pivot/J
      Y := TArray.SortBinaryMarker<T>(J);
      if (X > Y) then Continue;
      if (X = Y) then
      begin
        if (GetTypeKind(T) = tkMethod) then
        begin
          if (NativeUInt(Pivot.Ptr) >= NativeUInt(Pointer(J)^)) then Continue;
        end else
        if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
        begin
          Buffer := Pointer(Pointer(J)^);
          if (Pivot.Ptr = Buffer) then Continue;
          case GetTypeKind(T) of
            tkLString: if (InterfaceDefaults.Compare_LStr(nil, Pivot.Ptr, Buffer) >= 0) then Continue;
            tkWString: if (InterfaceDefaults.Compare_WStr(nil, Pivot.Ptr, Buffer) >= 0) then Continue;
            tkUString: if (InterfaceDefaults.Compare_UStr(nil, Pivot.Ptr, Buffer) >= 0) then Continue;
           tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Pivot.Ptr, Buffer) >= 0) then Continue;
          end;
        end else
        begin
          if (SortBinaryComparer<T>(@Pivot, J) >= 0) then Continue;
        end;
      end;
    insertion_init:
      I := J;
      // pivot := J^
      if (SizeOf(T) = SizeOf(Pointer)) then
      begin
        Pivot.Ptr := Pointer(Pointer(J)^);
      end else
      begin
        TArray.CopyMemory<T>(@Pivot, J);
      end;
      X := TArray.SortBinaryMarker<T>(@Pivot);
    until (False);
    // I^ := J^
    if (SizeOf(T) = SizeOf(Pointer)) then
    begin
      PNativeUInt(I)^ := PNativeUInt(J)^;
    end else
    begin
      TArray.CopyMemory<T>(I, J);
    end;
    // J := pivot
    if (SizeOf(T) = SizeOf(Pointer)) then
    begin
      Pointer(Pointer(J)^) := Pivot.Ptr;
    end else
    begin
      TArray.CopyMemory<T>(J, @Pivot);
    end;

    PStop := StackItem.Last;
    repeat
      Inc(J);
      if (J = PStop) then Break;

      // pivot := J[1]
      if (SizeOf(T) = SizeOf(Pointer)) then
      begin
        Pivot.Ptr := Pointer(Pointer(J + 1)^);
      end else
      begin
        TArray.CopyMemory<T>(@Pivot, J + 1);
      end;
      X := TArray.SortBinaryMarker<T>(@Pivot);

      // compare J/pivot
      Y := TArray.SortBinaryMarker<T>(J);
      if (Y > X) then Continue;
      if (Y = X) then
      begin
        if (GetTypeKind(T) = tkMethod) then
        begin
          if (NativeUInt(Pointer(J)^) >= NativeUInt(Pivot.Ptr)) then Continue;
        end else
        if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
        begin
          Buffer := Pointer(Pointer(J)^);
          if (Buffer = Pivot.Ptr) then Continue;
          case GetTypeKind(T) of
            tkLString: if (InterfaceDefaults.Compare_LStr(nil, Buffer, Pivot.Ptr) >= 0) then Continue;
            tkWString: if (InterfaceDefaults.Compare_WStr(nil, Buffer, Pivot.Ptr) >= 0) then Continue;
            tkUString: if (InterfaceDefaults.Compare_UStr(nil, Buffer, Pivot.Ptr) >= 0) then Continue;
           tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Buffer, Pivot.Ptr) >= 0) then Continue;
          end;
        end else
        begin
          if (SortBinaryComparer<T>(J, @Pivot) >= 0) then Continue;
        end;
      end;

      I := J;
      repeat
        // I[1] := I^;
        if (SizeOf(T) = SizeOf(Pointer)) then
        begin
          PNativeUInt(I + 1)^ := PNativeUInt(I)^;
        end else
        begin
          TArray.CopyMemory<T>(I + 1, I);
        end;
        Dec(I);

        // compare I/pivot
        Y := TArray.SortBinaryMarker<T>(I);
        if (Y > X) then Break;
        if (Y = X) then
        begin
          if (GetTypeKind(T) = tkMethod) then
          begin
            if (NativeUInt(Pointer(I)^) >= NativeUInt(Pivot.Ptr)) then Break;
          end else
          if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
          begin
            Buffer := Pointer(Pointer(I)^);
            if (Buffer = Pivot.Ptr) then Break;
            case GetTypeKind(T) of
              tkLString: if (InterfaceDefaults.Compare_LStr(nil, Buffer, Pivot.Ptr) >= 0) then Break;
              tkWString: if (InterfaceDefaults.Compare_WStr(nil, Buffer, Pivot.Ptr) >= 0) then Break;
              tkUString: if (InterfaceDefaults.Compare_UStr(nil, Buffer, Pivot.Ptr) >= 0) then Break;
             tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Buffer, Pivot.Ptr) >= 0) then Break;
            end;
          end else
          begin
            if (SortBinaryComparer<T>(I, @Pivot) >= 0) then Break;
          end;
        end;
      until (False);

      // I[1] := pivot
      if (SizeOf(T) = SizeOf(Pointer)) then
      begin
        Pointer(Pointer(I + 1)^) := Pivot.Ptr;
      end else
      begin
        TArray.CopyMemory<T>(I + 1, @Pivot);
      end;
    until (False);

    if (StackItem <> @STACK[0]) then goto proc_loop;
    Exit;
  end;

  // pivot
  if (SizeOf(T) <= SizeOf(Pivot)) then
  begin
    if (SizeOf(T) = SizeOf(Pointer)) then
    begin
      Pivot.Ptr := Pointer(SortItemPivot<T>(I, J)^);
    end else
    begin
      TArray.CopyMemory<T>(@Pivot, SortItemPivot<T>(I, J));
    end;
    X := TArray.SortBinaryMarker<T>(@Pivot);
  end else
  begin
    TArray.CopyMemory<T>(@PivotBig, SortItemPivot<T>(I, J));
    X := TArray.SortBinaryMarker<T>(@PivotBig);
  end;

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);

      Y := TArray.SortBinaryMarker<T>(I);
      if (X > Y) then Break;
      if (X = Y) then
      begin
        if (SizeOf(T) >= SizeOf(Pivot)) then
        begin
          // compare pivot/I
          if (GetTypeKind(T) = tkMethod) then
          begin
            if (NativeUInt(Pivot.Ptr) >= NativeUInt(Pointer(I)^)) then Break;
          end else
          if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
          begin
            Buffer := Pointer(Pointer(I)^);
            if (Pivot.Ptr = Buffer) then Break;
            case GetTypeKind(T) of
              tkLString: if (InterfaceDefaults.Compare_LStr(nil, Pivot.Ptr, Buffer) >= 0) then Break;
              tkWString: if (InterfaceDefaults.Compare_WStr(nil, Pivot.Ptr, Buffer) >= 0) then Break;
              tkUString: if (InterfaceDefaults.Compare_UStr(nil, Pivot.Ptr, Buffer) >= 0) then Break;
             tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Pivot.Ptr, Buffer) >= 0) then Break;
            end;
          end else
          begin
            if (SortBinaryComparer<T>(@Pivot, I) >= 0) then Break;
          end;
        end else
        begin
          if (SortBinaryComparer<T>(@PivotBig, I) >= 0) then Break;
        end;
      end;
    until (False);

    repeat
      Dec(J);

      Y := TArray.SortBinaryMarker<T>(J);
      if (Y > X) then Break;
      if (Y = X) then
      begin
        if (SizeOf(T) <= SizeOf(Pivot)) then
        begin
          // compare J/pivot
          if (GetTypeKind(T) = tkMethod) then
          begin
            if (NativeUInt(Pointer(J)^) >= NativeUInt(Pivot.Ptr)) then Break;
          end else
          if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
          begin
            Buffer := Pointer(Pointer(J)^);
            if (Buffer = Pivot.Ptr) then Break;
            case GetTypeKind(T) of
              tkLString: if (InterfaceDefaults.Compare_LStr(nil, Buffer, Pivot.Ptr) >= 0) then Break;
              tkWString: if (InterfaceDefaults.Compare_WStr(nil, Buffer, Pivot.Ptr) >= 0) then Break;
              tkUString: if (InterfaceDefaults.Compare_UStr(nil, Buffer, Pivot.Ptr) >= 0) then Break;
             tkDynArray: if (InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, Buffer, Pivot.Ptr) >= 0) then Break;
            end;
          end else
          begin
            if (SortBinaryComparer<T>(J, @Pivot) >= 0) then Break;
          end;
        end else
        begin
          if (SortBinaryComparer<T>(J, @PivotBig) >= 0) then Break;
        end;
      end;
    until (False);

    if (I <= J) then
    begin
      // TArray.Exchange<T>(I, J);
      case SizeOf(T) of
        0: ;
        1:
        begin
          Temp1 := PLMemory(I).LBytes[0];
          PLMemory(I).LBytes[0] := PRMemory(J).RBytes[0];
          PRMemory(J).RBytes[0] := Temp1;
        end;
        2:
        begin
          Temp2 := PLMemory(I).LWords[0];
          PLMemory(I).LWords[0] := PRMemory(J).RWords[0];
          PRMemory(J).RWords[0] := Temp2;
        end;
        3:
        begin
          Temp2 := PLMemory(I).LWords[0];
          PLMemory(I).LWords[0] := PRMemory(J).RWords[0];
          PRMemory(J).RWords[0] := Temp2;

          Temp1 := PLMemory(I).LBytes[2];
          PLMemory(I).LBytes[2] := PRMemory(J).RBytes[2];
          PRMemory(J).RBytes[2] := Temp1;
        end;
        4..7:
        begin
          Temp4 := PLMemory(I).LCardinals[0];
          PLMemory(I).LCardinals[0] := PRMemory(J).RCardinals[0];
          PRMemory(J).RCardinals[0] := Temp4;

          case SizeOf(T) of
            5:
            begin
              Temp1 := PLMemory(I).LBytes[4];
              PLMemory(I).LBytes[4] := PRMemory(J).RBytes[4];
              PRMemory(J).RBytes[4] := Temp1;
            end;
            6:
            begin
              Temp2 := PLMemory(I).LWords[2];
              PLMemory(I).LWords[2] := PRMemory(J).RWords[2];
              PRMemory(J).RWords[2] := Temp2;
            end;
            7:
            begin
              Temp2 := PLMemory(I).LWords[2];
              PLMemory(I).LWords[2] := PRMemory(J).RWords[2];
              PRMemory(J).RWords[2] := Temp2;
              Temp1 := PLMemory(I).LBytes[6];
              PLMemory(I).LBytes[6] := PRMemory(J).RBytes[6];
              PRMemory(J).RBytes[6] := Temp1;
            end;
          end;
        end;
        8..16:
        begin
          TempNative := PLMemory(I).LNatives[0];
          PLMemory(I).LNatives[0] := PRMemory(J).RNatives[0];
          PRMemory(J).RNatives[0] := TempNative;

          if (SizeOf(T) >= SizeOf(NativeUInt) * 2) then
          begin
            TempNative := PLMemory(I).LNatives[1];
            PLMemory(I).LNatives[1] := PRMemory(J).RNatives[1];
            PRMemory(J).RNatives[1] := TempNative;
          end;

          if (SizeOf(T) >= SizeOf(NativeUInt) * 3) then
          begin
            TempNative := PLMemory(I).LNatives[2];
            PLMemory(I).LNatives[2] := PRMemory(J).RNatives[2];
            PRMemory(J).RNatives[2] := TempNative;
          end;

          if (SizeOf(T)  = SizeOf(NativeUInt) * 4) then
          begin
            TempNative := PLMemory(I).LNatives[3];
            PLMemory(I).LNatives[3] := PRMemory(J).RNatives[3];
            PRMemory(J).RNatives[3] := TempNative;
          end;

          {$ifdef LARGEINT}
          case SizeOf(T) of
            12, 13, 14, 15:
            begin
              Temp4 := PLMemory(I).LCardinals[2];
              PLMemory(I).LCardinals[2] := PRMemory(J).RCardinals[2];
              PRMemory(J).RCardinals[2] := Temp4;
            end;
          end;
          {$endif}

          case SizeOf(T) of
            9:
            begin
              Temp1 := PLMemory(I).LBytes[8];
              PLMemory(I).LBytes[8] := PRMemory(J).RBytes[8];
              PRMemory(J).RBytes[8] := Temp1;
            end;
            10:
            begin
              Temp2 := PLMemory(I).LWords[4];
              PLMemory(I).LWords[4] := PRMemory(J).RWords[4];
              PRMemory(J).RWords[4] := Temp2;
            end;
            11:
            begin
              Temp2 := PLMemory(I).LWords[4];
              PLMemory(I).LWords[4] := PRMemory(J).RWords[4];
              PRMemory(J).RWords[4] := Temp2;
              Temp1 := PLMemory(I).LBytes[10];
              PLMemory(I).LBytes[10] := PRMemory(J).RBytes[10];
              PRMemory(J).RBytes[10] := Temp1;
            end;
            13:
            begin
              Temp2 := PLMemory(I).LWords[5];
              PLMemory(I).LWords[5] := PRMemory(J).RWords[5];
              PRMemory(J).RWords[5] := Temp2;
              Temp1 := PLMemory(I).LBytes[12];
              PLMemory(I).LBytes[12] := PRMemory(J).RBytes[12];
              PRMemory(J).RBytes[12] := Temp1;
            end;
            14:
            begin
              Temp2 := PLMemory(I).LWords[6];
              PLMemory(I).LWords[6] := PRMemory(J).RWords[6];
              PRMemory(J).RWords[6] := Temp2;
            end;
            15:
            begin
              Temp2 := PLMemory(I).LWords[6];
              PLMemory(I).LWords[6] := PRMemory(J).RWords[6];
              PRMemory(J).RWords[6] := Temp2;
              Temp1 := PLMemory(I).LBytes[14];
              PLMemory(I).LBytes[14] := PRMemory(J).RBytes[14];
              PRMemory(J).RBytes[14] := Temp1;
            end;
          end;
        end;
      else
        Index := 0;
        repeat
          TempNative := PLMemory(I).LNatives[Index];
          PLMemory(I).LNatives[Index] := PRMemory(J).RNatives[Index];
          PRMemory(J).RNatives[Index] := TempNative;
          Inc(Index);
        until (Index = SizeOf(T) div SizeOf(NativeUInt));

        if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
        begin
          {$ifdef LARGEINT}
          if (SizeOf(T) and 4 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Cardinal) - 1;
            Temp4 := PLMemory(I).LCardinals[Index];
            PLMemory(I).LCardinals[Index] := PRMemory(J).RCardinals[Index];
            PRMemory(J).RCardinals[Index] := Temp4;
          end;
          {$endif}

          if (SizeOf(T) and 2 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Word) - 1;
            Temp2 := PLMemory(I).LWords[Index];
            PLMemory(I).LWords[Index] := PRMemory(J).RWords[Index];
            PRMemory(J).RWords[Index] := Temp2;
          end;

          if (SizeOf(T) and 1 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Byte) - 1;
            Temp1 := PLMemory(I).LBytes[Index];
            PLMemory(I).LBytes[Index] := PRMemory(J).RBytes[Index];
            PRMemory(J).RBytes[Index] := Temp1;
          end;
        end;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;
{$endif .SMARTGENERICS}

class procedure TArray.SortUniversals<T>(const Values: Pointer; const Count: NativeInt; var Helper: TSortHelper<T>);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
var
  Index: NativeInt;
  Temp1: Byte;
  Temp2: Word;
  Temp4: Cardinal;
  TempNative: NativeUInt;

  I, J, PStop: ^T;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion sort
  if ((NativeUInt(J) - NativeUInt(I)) < INSERTION_SORT_LIMIT * SizeOf(T)) then
  begin
    PStop := StackItem.First;
    goto insertion_init;
    repeat
      if (J = PStop) then Break;
      Dec(J);

      if (Helper.Compare(Helper.Inst, Helper.Pivot, J^) <= 0) then Continue;
    insertion_init:
      I := J;
      TArray.CopyMemory<T>(@Helper.Pivot, J);
    until (False);
    TArray.CopyMemory<T>(I, J);
    TArray.CopyMemory<T>(J, @Helper.Pivot);

    PStop := StackItem.Last;
    repeat
      Inc(J);
      if (J = PStop) then Break;

      TArray.CopyMemory<T>(@Helper.Pivot, J + 1);
      if (Helper.Compare(Helper.Inst, J^, Helper.Pivot) <= 0) then Continue;

      I := J;
      repeat
        TArray.CopyMemory<T>(I + 1, I);
        Dec(I);
        if (Helper.Compare(Helper.Inst, I^, Helper.Pivot) <= 0) then Break;
      until (False);

      TArray.CopyMemory<T>(I + 1, @Helper.Pivot);
    until (False);

    if (StackItem <> @STACK[0]) then goto proc_loop;
    Exit;
  end;

  // pivot
  TArray.CopyMemory<T>(@Helper.Pivot, SortItemPivot<T>(I, J));

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);
      if (Helper.Compare(Helper.Inst, Helper.Pivot, I^) <= 0) then Break;
    until (False);

    repeat
      Dec(J);
      if (Helper.Compare(Helper.Inst, J^, Helper.Pivot) <= 0) then Break;
    until (False);

    if (I <= J) then
    begin
      // TArray.Exchange<T>(I, J);
      case SizeOf(T) of
        0: ;
        1:
        begin
          Temp1 := PLMemory(I).LBytes[0];
          PLMemory(I).LBytes[0] := PRMemory(J).RBytes[0];
          PRMemory(J).RBytes[0] := Temp1;
        end;
        2:
        begin
          Temp2 := PLMemory(I).LWords[0];
          PLMemory(I).LWords[0] := PRMemory(J).RWords[0];
          PRMemory(J).RWords[0] := Temp2;
        end;
        3:
        begin
          Temp2 := PLMemory(I).LWords[0];
          PLMemory(I).LWords[0] := PRMemory(J).RWords[0];
          PRMemory(J).RWords[0] := Temp2;

          Temp1 := PLMemory(I).LBytes[2];
          PLMemory(I).LBytes[2] := PRMemory(J).RBytes[2];
          PRMemory(J).RBytes[2] := Temp1;
        end;
        4..7:
        begin
          Temp4 := PLMemory(I).LCardinals[0];
          PLMemory(I).LCardinals[0] := PRMemory(J).RCardinals[0];
          PRMemory(J).RCardinals[0] := Temp4;

          case SizeOf(T) of
            5:
            begin
              Temp1 := PLMemory(I).LBytes[4];
              PLMemory(I).LBytes[4] := PRMemory(J).RBytes[4];
              PRMemory(J).RBytes[4] := Temp1;
            end;
            6:
            begin
              Temp2 := PLMemory(I).LWords[2];
              PLMemory(I).LWords[2] := PRMemory(J).RWords[2];
              PRMemory(J).RWords[2] := Temp2;
            end;
            7:
            begin
              Temp2 := PLMemory(I).LWords[2];
              PLMemory(I).LWords[2] := PRMemory(J).RWords[2];
              PRMemory(J).RWords[2] := Temp2;
              Temp1 := PLMemory(I).LBytes[6];
              PLMemory(I).LBytes[6] := PRMemory(J).RBytes[6];
              PRMemory(J).RBytes[6] := Temp1;
            end;
          end;
        end;
        8..16:
        begin
          TempNative := PLMemory(I).LNatives[0];
          PLMemory(I).LNatives[0] := PRMemory(J).RNatives[0];
          PRMemory(J).RNatives[0] := TempNative;

          if (SizeOf(T) >= SizeOf(NativeUInt) * 2) then
          begin
            TempNative := PLMemory(I).LNatives[1];
            PLMemory(I).LNatives[1] := PRMemory(J).RNatives[1];
            PRMemory(J).RNatives[1] := TempNative;
          end;

          if (SizeOf(T) >= SizeOf(NativeUInt) * 3) then
          begin
            TempNative := PLMemory(I).LNatives[2];
            PLMemory(I).LNatives[2] := PRMemory(J).RNatives[2];
            PRMemory(J).RNatives[2] := TempNative;
          end;

          if (SizeOf(T)  = SizeOf(NativeUInt) * 4) then
          begin
            TempNative := PLMemory(I).LNatives[3];
            PLMemory(I).LNatives[3] := PRMemory(J).RNatives[3];
            PRMemory(J).RNatives[3] := TempNative;
          end;

          {$ifdef LARGEINT}
          case SizeOf(T) of
            12, 13, 14, 15:
            begin
              Temp4 := PLMemory(I).LCardinals[2];
              PLMemory(I).LCardinals[2] := PRMemory(J).RCardinals[2];
              PRMemory(J).RCardinals[2] := Temp4;
            end;
          end;
          {$endif}

          case SizeOf(T) of
            9:
            begin
              Temp1 := PLMemory(I).LBytes[8];
              PLMemory(I).LBytes[8] := PRMemory(J).RBytes[8];
              PRMemory(J).RBytes[8] := Temp1;
            end;
            10:
            begin
              Temp2 := PLMemory(I).LWords[4];
              PLMemory(I).LWords[4] := PRMemory(J).RWords[4];
              PRMemory(J).RWords[4] := Temp2;
            end;
            11:
            begin
              Temp2 := PLMemory(I).LWords[4];
              PLMemory(I).LWords[4] := PRMemory(J).RWords[4];
              PRMemory(J).RWords[4] := Temp2;
              Temp1 := PLMemory(I).LBytes[10];
              PLMemory(I).LBytes[10] := PRMemory(J).RBytes[10];
              PRMemory(J).RBytes[10] := Temp1;
            end;
            13:
            begin
              Temp2 := PLMemory(I).LWords[5];
              PLMemory(I).LWords[5] := PRMemory(J).RWords[5];
              PRMemory(J).RWords[5] := Temp2;
              Temp1 := PLMemory(I).LBytes[12];
              PLMemory(I).LBytes[12] := PRMemory(J).RBytes[12];
              PRMemory(J).RBytes[12] := Temp1;
            end;
            14:
            begin
              Temp2 := PLMemory(I).LWords[6];
              PLMemory(I).LWords[6] := PRMemory(J).RWords[6];
              PRMemory(J).RWords[6] := Temp2;
            end;
            15:
            begin
              Temp2 := PLMemory(I).LWords[6];
              PLMemory(I).LWords[6] := PRMemory(J).RWords[6];
              PRMemory(J).RWords[6] := Temp2;
              Temp1 := PLMemory(I).LBytes[14];
              PLMemory(I).LBytes[14] := PRMemory(J).RBytes[14];
              PRMemory(J).RBytes[14] := Temp1;
            end;
          end;
        end;
      else
        Index := 0;
        repeat
          TempNative := PLMemory(I).LNatives[Index];
          PLMemory(I).LNatives[Index] := PRMemory(J).RNatives[Index];
          PRMemory(J).RNatives[Index] := TempNative;
          Inc(Index);
        until (Index = SizeOf(T) div SizeOf(NativeUInt));

        if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
        begin
          {$ifdef LARGEINT}
          if (SizeOf(T) and 4 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Cardinal) - 1;
            Temp4 := PLMemory(I).LCardinals[Index];
            PLMemory(I).LCardinals[Index] := PRMemory(J).RCardinals[Index];
            PRMemory(J).RCardinals[Index] := Temp4;
          end;
          {$endif}

          if (SizeOf(T) and 2 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Word) - 1;
            Temp2 := PLMemory(I).LWords[Index];
            PLMemory(I).LWords[Index] := PRMemory(J).RWords[Index];
            PRMemory(J).RWords[Index] := Temp2;
          end;

          if (SizeOf(T) and 1 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Byte) - 1;
            Temp1 := PLMemory(I).LBytes[Index];
            PLMemory(I).LBytes[Index] := PRMemory(J).RBytes[Index];
            PRMemory(J).RBytes[Index] := Temp1;
          end;
        end;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.SortDescendingUniversals<T>(const Values: Pointer; const Count: NativeInt; var Helper: TSortHelper<T>);
label
  proc_loop, proc_loop_current, insertion_init, swap_loop;
type
  P = ^T;
var
  Index: NativeInt;
  Temp1: Byte;
  Temp2: Word;
  Temp4: Cardinal;
  TempNative: NativeUInt;

  I, J, PStop: ^T;
  StackItem: ^TSortStackItem<T>;
  Stack: TSortStack<T>;
begin
  Stack[0].First := Values;
  Stack[0].Last := P(Values) + Count - 1;
  StackItem := @Stack[1];

proc_loop:
  Dec(StackItem);
proc_loop_current:
  I := StackItem.First;
  J := StackItem.Last;

  // insertion sort
  if ((NativeUInt(J) - NativeUInt(I)) < INSERTION_SORT_LIMIT * SizeOf(T)) then
  begin
    PStop := StackItem.First;
    goto insertion_init;
    repeat
      if (J = PStop) then Break;
      Dec(J);

      if (Helper.Compare(Helper.Inst, Helper.Pivot, J^) >= 0) then Continue;
    insertion_init:
      I := J;
      TArray.CopyMemory<T>(@Helper.Pivot, J);
    until (False);
    TArray.CopyMemory<T>(I, J);
    TArray.CopyMemory<T>(J, @Helper.Pivot);

    PStop := StackItem.Last;
    repeat
      Inc(J);
      if (J = PStop) then Break;

      TArray.CopyMemory<T>(@Helper.Pivot, J + 1);
      if (Helper.Compare(Helper.Inst, J^, Helper.Pivot) >= 0) then Continue;

      I := J;
      repeat
        TArray.CopyMemory<T>(I + 1, I);
        Dec(I);
        if (Helper.Compare(Helper.Inst, I^, Helper.Pivot) >= 0) then Break;
      until (False);

      TArray.CopyMemory<T>(I + 1, @Helper.Pivot);
    until (False);

    if (StackItem <> @STACK[0]) then goto proc_loop;
    Exit;
  end;

  // pivot
  TArray.CopyMemory<T>(@Helper.Pivot, SortItemPivot<T>(I, J));

  // quick sort
  Dec(J);
  Dec(I);
  swap_loop:
  begin
    Inc(J, 2);

    repeat
      Inc(I);
      if (Helper.Compare(Helper.Inst, Helper.Pivot, I^) >= 0) then Break;
    until (False);

    repeat
      Dec(J);
      if (Helper.Compare(Helper.Inst, J^, Helper.Pivot) >= 0) then Break;
    until (False);

    if (I <= J) then
    begin
      // TArray.Exchange<T>(I, J);
      case SizeOf(T) of
        0: ;
        1:
        begin
          Temp1 := PLMemory(I).LBytes[0];
          PLMemory(I).LBytes[0] := PRMemory(J).RBytes[0];
          PRMemory(J).RBytes[0] := Temp1;
        end;
        2:
        begin
          Temp2 := PLMemory(I).LWords[0];
          PLMemory(I).LWords[0] := PRMemory(J).RWords[0];
          PRMemory(J).RWords[0] := Temp2;
        end;
        3:
        begin
          Temp2 := PLMemory(I).LWords[0];
          PLMemory(I).LWords[0] := PRMemory(J).RWords[0];
          PRMemory(J).RWords[0] := Temp2;

          Temp1 := PLMemory(I).LBytes[2];
          PLMemory(I).LBytes[2] := PRMemory(J).RBytes[2];
          PRMemory(J).RBytes[2] := Temp1;
        end;
        4..7:
        begin
          Temp4 := PLMemory(I).LCardinals[0];
          PLMemory(I).LCardinals[0] := PRMemory(J).RCardinals[0];
          PRMemory(J).RCardinals[0] := Temp4;

          case SizeOf(T) of
            5:
            begin
              Temp1 := PLMemory(I).LBytes[4];
              PLMemory(I).LBytes[4] := PRMemory(J).RBytes[4];
              PRMemory(J).RBytes[4] := Temp1;
            end;
            6:
            begin
              Temp2 := PLMemory(I).LWords[2];
              PLMemory(I).LWords[2] := PRMemory(J).RWords[2];
              PRMemory(J).RWords[2] := Temp2;
            end;
            7:
            begin
              Temp2 := PLMemory(I).LWords[2];
              PLMemory(I).LWords[2] := PRMemory(J).RWords[2];
              PRMemory(J).RWords[2] := Temp2;
              Temp1 := PLMemory(I).LBytes[6];
              PLMemory(I).LBytes[6] := PRMemory(J).RBytes[6];
              PRMemory(J).RBytes[6] := Temp1;
            end;
          end;
        end;
        8..16:
        begin
          TempNative := PLMemory(I).LNatives[0];
          PLMemory(I).LNatives[0] := PRMemory(J).RNatives[0];
          PRMemory(J).RNatives[0] := TempNative;

          if (SizeOf(T) >= SizeOf(NativeUInt) * 2) then
          begin
            TempNative := PLMemory(I).LNatives[1];
            PLMemory(I).LNatives[1] := PRMemory(J).RNatives[1];
            PRMemory(J).RNatives[1] := TempNative;
          end;

          if (SizeOf(T) >= SizeOf(NativeUInt) * 3) then
          begin
            TempNative := PLMemory(I).LNatives[2];
            PLMemory(I).LNatives[2] := PRMemory(J).RNatives[2];
            PRMemory(J).RNatives[2] := TempNative;
          end;

          if (SizeOf(T)  = SizeOf(NativeUInt) * 4) then
          begin
            TempNative := PLMemory(I).LNatives[3];
            PLMemory(I).LNatives[3] := PRMemory(J).RNatives[3];
            PRMemory(J).RNatives[3] := TempNative;
          end;

          {$ifdef LARGEINT}
          case SizeOf(T) of
            12, 13, 14, 15:
            begin
              Temp4 := PLMemory(I).LCardinals[2];
              PLMemory(I).LCardinals[2] := PRMemory(J).RCardinals[2];
              PRMemory(J).RCardinals[2] := Temp4;
            end;
          end;
          {$endif}

          case SizeOf(T) of
            9:
            begin
              Temp1 := PLMemory(I).LBytes[8];
              PLMemory(I).LBytes[8] := PRMemory(J).RBytes[8];
              PRMemory(J).RBytes[8] := Temp1;
            end;
            10:
            begin
              Temp2 := PLMemory(I).LWords[4];
              PLMemory(I).LWords[4] := PRMemory(J).RWords[4];
              PRMemory(J).RWords[4] := Temp2;
            end;
            11:
            begin
              Temp2 := PLMemory(I).LWords[4];
              PLMemory(I).LWords[4] := PRMemory(J).RWords[4];
              PRMemory(J).RWords[4] := Temp2;
              Temp1 := PLMemory(I).LBytes[10];
              PLMemory(I).LBytes[10] := PRMemory(J).RBytes[10];
              PRMemory(J).RBytes[10] := Temp1;
            end;
            13:
            begin
              Temp2 := PLMemory(I).LWords[5];
              PLMemory(I).LWords[5] := PRMemory(J).RWords[5];
              PRMemory(J).RWords[5] := Temp2;
              Temp1 := PLMemory(I).LBytes[12];
              PLMemory(I).LBytes[12] := PRMemory(J).RBytes[12];
              PRMemory(J).RBytes[12] := Temp1;
            end;
            14:
            begin
              Temp2 := PLMemory(I).LWords[6];
              PLMemory(I).LWords[6] := PRMemory(J).RWords[6];
              PRMemory(J).RWords[6] := Temp2;
            end;
            15:
            begin
              Temp2 := PLMemory(I).LWords[6];
              PLMemory(I).LWords[6] := PRMemory(J).RWords[6];
              PRMemory(J).RWords[6] := Temp2;
              Temp1 := PLMemory(I).LBytes[14];
              PLMemory(I).LBytes[14] := PRMemory(J).RBytes[14];
              PRMemory(J).RBytes[14] := Temp1;
            end;
          end;
        end;
      else
        Index := 0;
        repeat
          TempNative := PLMemory(I).LNatives[Index];
          PLMemory(I).LNatives[Index] := PRMemory(J).RNatives[Index];
          PRMemory(J).RNatives[Index] := TempNative;
          Inc(Index);
        until (Index = SizeOf(T) div SizeOf(NativeUInt));

        if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
        begin
          {$ifdef LARGEINT}
          if (SizeOf(T) and 4 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Cardinal) - 1;
            Temp4 := PLMemory(I).LCardinals[Index];
            PLMemory(I).LCardinals[Index] := PRMemory(J).RCardinals[Index];
            PRMemory(J).RCardinals[Index] := Temp4;
          end;
          {$endif}

          if (SizeOf(T) and 2 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Word) - 1;
            Temp2 := PLMemory(I).LWords[Index];
            PLMemory(I).LWords[Index] := PRMemory(J).RWords[Index];
            PRMemory(J).RWords[Index] := Temp2;
          end;

          if (SizeOf(T) and 1 <> 0) then
          begin
            Index := SizeOf(T) div SizeOf(Byte) - 1;
            Temp1 := PLMemory(I).LBytes[Index];
            PLMemory(I).LBytes[Index] := PRMemory(J).RBytes[Index];
            PRMemory(J).RBytes[Index] := Temp1;
          end;
        end;
      end;

      Dec(J, 2);
      if (I <= J) then goto swap_loop;
      Inc(I);
      Inc(J);
    end;
  end;

  // next iteration
  StackItem := SortItemNext<T>(StackItem, I, J);
  if (NativeInt(StackItem) >= 0) then goto proc_loop_current;
  Dec(NativeInt(StackItem), HIGH_NATIVE_BIT);
  if (StackItem <> @Stack[0]) then goto proc_loop;
end;

class procedure TArray.Sort<T>(var Values: T; const Count: Integer);
{$ifdef SMARTGENERICS}
var
  TypeData: PTypeData;
  PivotBig: ^T;
begin
  if (Count <= 1) then Exit;

  if (GetTypeKind(T) in [tkInteger, tkEnumeration, tkChar, tkWChar, tkInt64]) or
    ((GetTypeKind(T) = tkFloat) and (SizeOf(T) = 8)) then
  begin
    TypeData := Pointer(TypeInfo(T));
    Inc(NativeUInt(TypeData), NativeUInt(PByte(@PTypeInfo(TypeData).Name)^) + 2);
  end;

  case GetTypeKind(T) of
    tkInteger, tkEnumeration, tkChar, tkWChar:
    case SizeOf(T) of
      1:
      begin
        case TypeData.OrdType of
          otSByte: SortSigneds<ShortInt>(@Values, Count);
          otUByte: SortUnsigneds<Byte>(@Values, Count);
        end;
      end;
      2:
      begin
        case TypeData.OrdType of
          otSWord: SortSigneds<SmallInt>(@Values, Count);
          otUWord: SortUnsigneds<Word>(@Values, Count);
        end;
      end;
      4:
      begin
        case TypeData.OrdType of
          otSLong: SortSigneds<Integer>(@Values, Count);
          otULong: SortUnsigneds<Cardinal>(@Values, Count);
        end;
      end;
    end;
    tkInt64:
    begin
      if (TypeData.MaxInt64Value > TypeData.MinInt64Value) then
      begin
        SortSigneds<Int64>(@Values, Count);
      end else
      begin
        SortUnsigneds<UInt64>(@Values, Count);
      end;
    end;
    tkClass, tkInterface, tkClassRef, tkPointer, tkProcedure:
    begin
      {$ifdef LARGEINT}
        SortUnsigneds<UInt64>(@Values, Count);
      {$else .SMALLINT}
        SortUnsigneds<Cardinal>(@Values, Count);
      {$endif}
    end;
    tkFloat:
    case SizeOf(T) of
       4: SortFloats<Single>(@Values, Count);
      10: SortFloats<Extended>(@Values, Count);
    else
      if (TypeData.FloatType = ftDouble) then
      begin
        SortFloats<Double>(@Values, Count);
      end else
      begin
        SortSigneds<Int64>(@Values, Count);
      end;
    end;
    tkMethod:
    begin
      SortBinaries<InterfaceDefaults.TMethodPtr>(@Values, Count, InterfaceDefaults.TMethodPtr(nil^));
    end;
    tkVariant:
    begin
      TArray.Sort<Variant>(PVariant(@Values)^, Count, IComparer<Variant>(@InterfaceDefaults.TDefaultComparer<Variant>.Instance));
    end;
    tkString:
    begin
      SortBinaries<T>(@Values, Count, T(nil^));
    end;
    tkLString:
    begin
      {$ifdef NEXTGEN}
         SortBinaries<T>(@Values, Count, T(nil^));
      {$else}
         SortBinaries<AnsiString>(@Values, Count, AnsiString(nil^));
      {$endif}
    end;
    {$ifdef MSWINDOWS}
    tkWString:
    begin
      SortBinaries<WideString>(@Values, Count, WideString(nil^));
    end;
    {$else}
    tkWString,
    {$endif}
    tkUString:
    begin
      SortBinaries<UnicodeString>(@Values, Count, UnicodeString(nil^));
    end;
    tkDynArray:
    begin
      SortBinaries<T>(@Values, Count, T(nil^));
    end;
  else
    // binary
    case SizeOf(T) of
      0: ;
      1: SortUnsigneds<Byte>(@Values, Count);
      2..BUFFER_SIZE: SortBinaries<T>(@Values, Count, T(nil^));
    else
      GetMem(PivotBig, SizeOf(T));
      try
        SortBinaries<T>(@Values, Count, PivotBig^);
      finally
        FreeMem(PivotBig);
      end;
    end;
  end;
end;
{$else}
begin
  TArray.Sort<T>(Values, Count, IComparer<T>(@InterfaceDefaults.TDefaultComparer<T>.Instance));
end;
{$endif}

class procedure TArray.Sort<T>(var Values: T; const Count: Integer; const Comparer: IComparer<T>);
var
  HelperBuffer: array[0..BUFFER_SIZE - 1] of Byte;
  Helper: ^TSortHelper<T>;
begin
  if (Count <= 1) then Exit;

  if (SizeOf(TSortHelper<T>) <= SizeOf(HelperBuffer)) then
  begin
    Helper := Pointer(@HelperBuffer);
    Helper.Init(Comparer);
    TArray.SortUniversals<T>(@Values, Count, Helper^);
  end else
  begin
    GetMem(Helper, SizeOf(TSortHelper<T>));
    try
      Helper.Init(Comparer);
      TArray.SortUniversals<T>(@Values, Count, Helper^);
    finally
      FreeMem(Helper);
    end;
  end;
end;

class procedure TArray.Sort<T>(var Values: T; const Count: Integer; const Comparison: TComparison<T>);
var
  HelperBuffer: array[0..BUFFER_SIZE - 1] of Byte;
  Helper: ^TSortHelper<T>;
begin
  if (Count <= 1) then Exit;

  if (SizeOf(TSortHelper<T>) <= SizeOf(HelperBuffer)) then
  begin
    Helper := Pointer(@HelperBuffer);
    Helper.Init(Comparison);
    TArray.SortUniversals<T>(@Values, Count, Helper^);
  end else
  begin
    GetMem(Helper, SizeOf(TSortHelper<T>));
    try
      Helper.Init(Comparison);
      TArray.SortUniversals<T>(@Values, Count, Helper^);
    finally
      FreeMem(Helper);
    end;
  end;
end;

class procedure TArray.Sort<T>(var Values: array of T);
begin
  if (High(Values) > 0) then
    Sort<T>(Values[0], Length(Values));
end;

class procedure TArray.Sort<T>(var Values: array of T; const Comparer: IComparer<T>);
begin
  if (High(Values) > 0) then
   Sort<T>(Values[0], Length(Values), Comparer);
end;

class procedure TArray.Sort<T>(var Values: array of T; const Comparer: IComparer<T>; Index, Count: Integer);
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  Sort<T>(Values[Index], Count, Comparer);
end;

class procedure TArray.Sort<T>(var Values: array of T; const Comparison: TComparison<T>);
begin
  if (High(Values) > 0) then
   Sort<T>(Values[0], Length(Values), Comparison);
end;

class procedure TArray.Sort<T>(var Values: array of T; Index, Count: Integer; const Comparison: TComparison<T>);
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  Sort<T>(Values[Index], Count, Comparison);
end;

class procedure TArray.SortDescending<T>(var Values: T; const Count: Integer);
{$ifdef SMARTGENERICS}
var
  TypeData: PTypeData;
  PivotBig: ^T;
begin
  if (Count <= 1) then Exit;

  if (GetTypeKind(T) in [tkInteger, tkEnumeration, tkChar, tkWChar, tkInt64]) or
    ((GetTypeKind(T) = tkFloat) and (SizeOf(T) = 8)) then
  begin
    TypeData := Pointer(TypeInfo(T));
    Inc(NativeUInt(TypeData), NativeUInt(PByte(@PTypeInfo(TypeData).Name)^) + 2);
  end;

  case GetTypeKind(T) of
    tkInteger, tkEnumeration, tkChar, tkWChar:
    case SizeOf(T) of
      1:
      begin
        case TypeData.OrdType of
          otSByte: SortDescendingSigneds<ShortInt>(@Values, Count);
          otUByte: SortDescendingUnsigneds<Byte>(@Values, Count);
        end;
      end;
      2:
      begin
        case TypeData.OrdType of
          otSWord: SortDescendingSigneds<SmallInt>(@Values, Count);
          otUWord: SortDescendingUnsigneds<Word>(@Values, Count);
        end;
      end;
      4:
      begin
        case TypeData.OrdType of
          otSLong: SortDescendingSigneds<Integer>(@Values, Count);
          otULong: SortDescendingUnsigneds<Cardinal>(@Values, Count);
        end;
      end;
    end;
    tkInt64:
    begin
      if (TypeData.MaxInt64Value > TypeData.MinInt64Value) then
      begin
        SortDescendingSigneds<Int64>(@Values, Count);
      end else
      begin
        SortDescendingUnsigneds<UInt64>(@Values, Count);
      end;
    end;
    tkClass, tkInterface, tkClassRef, tkPointer, tkProcedure:
    begin
      {$ifdef LARGEINT}
        SortDescendingUnsigneds<UInt64>(@Values, Count);
      {$else .SMALLINT}
        SortDescendingUnsigneds<Cardinal>(@Values, Count);
      {$endif}
    end;
    tkFloat:
    case SizeOf(T) of
       4: SortDescendingFloats<Single>(@Values, Count);
      10: SortDescendingFloats<Extended>(@Values, Count);
    else
      if (TypeData.FloatType = ftDouble) then
      begin
        SortDescendingFloats<Double>(@Values, Count);
      end else
      begin
        SortDescendingSigneds<Int64>(@Values, Count);
      end;
    end;
    tkMethod:
    begin
      SortDescendingBinaries<InterfaceDefaults.TMethodPtr>(@Values, Count, InterfaceDefaults.TMethodPtr(nil^));
    end;
    tkVariant:
    begin
      TArray.SortDescending<Variant>(PVariant(@Values)^, Count, IComparer<Variant>(@InterfaceDefaults.TDefaultComparer<Variant>.Instance));
    end;
    tkString:
    begin
      SortDescendingBinaries<T>(@Values, Count, T(nil^));
    end;
    tkLString:
    begin
      {$ifdef NEXTGEN}
         SortDescendingBinaries<T>(@Values, Count, T(nil^));
      {$else}
         SortDescendingBinaries<AnsiString>(@Values, Count, AnsiString(nil^));
      {$endif}
    end;
    {$ifdef MSWINDOWS}
    tkWString:
    begin
      SortDescendingBinaries<WideString>(@Values, Count, WideString(nil^));
    end;
    {$else}
    tkWString,
    {$endif}
    tkUString:
    begin
      SortDescendingBinaries<UnicodeString>(@Values, Count, UnicodeString(nil^));
    end;
    tkDynArray:
    begin
      SortDescendingBinaries<T>(@Values, Count, T(nil^));
    end;
  else
    // binary
    case SizeOf(T) of
      0: ;
      1: SortDescendingUnsigneds<Byte>(@Values, Count);
      2..BUFFER_SIZE: SortDescendingBinaries<T>(@Values, Count, T(nil^));
    else
      GetMem(PivotBig, SizeOf(T));
      try
        SortDescendingBinaries<T>(@Values, Count, PivotBig^);
      finally
        FreeMem(PivotBig);
      end;
    end;
  end;
end;
{$else}
begin
  TArray.SortDescending<T>(Values, Count, IComparer<T>(@InterfaceDefaults.TDefaultComparer<T>.Instance));
end;
{$endif}

class procedure TArray.SortDescending<T>(var Values: T; const Count: Integer; const Comparer: IComparer<T>);
var
  HelperBuffer: array[0..BUFFER_SIZE - 1] of Byte;
  Helper: ^TSortHelper<T>;
begin
  if (Count <= 1) then Exit;

  if (SizeOf(TSortHelper<T>) <= SizeOf(HelperBuffer)) then
  begin
    Helper := Pointer(@HelperBuffer);
    Helper.Init(Comparer);
    TArray.SortDescendingUniversals<T>(@Values, Count, Helper^);
  end else
  begin
    GetMem(Helper, SizeOf(TSortHelper<T>));
    try
      Helper.Init(Comparer);
      TArray.SortDescendingUniversals<T>(@Values, Count, Helper^);
    finally
      FreeMem(Helper);
    end;
  end;
end;

class procedure TArray.SortDescending<T>(var Values: T; const Count: Integer; const Comparison: TComparison<T>);
var
  HelperBuffer: array[0..BUFFER_SIZE - 1] of Byte;
  Helper: ^TSortHelper<T>;
begin
  if (Count <= 1) then Exit;

  if (SizeOf(TSortHelper<T>) <= SizeOf(HelperBuffer)) then
  begin
    Helper := Pointer(@HelperBuffer);
    Helper.Init(Comparison);
    TArray.SortDescendingUniversals<T>(@Values, Count, Helper^);
  end else
  begin
    GetMem(Helper, SizeOf(TSortHelper<T>));
    try
      Helper.Init(Comparison);
      TArray.SortDescendingUniversals<T>(@Values, Count, Helper^);
    finally
      FreeMem(Helper);
    end;
  end;
end;

class procedure TArray.SortDescending<T>(var Values: array of T);
begin
  if (High(Values) > 0) then
    SortDescending<T>(Values[0], Length(Values));
end;

class procedure TArray.SortDescending<T>(var Values: array of T; const Comparer: IComparer<T>);
begin
  if (High(Values) > 0) then
   SortDescending<T>(Values[0], Length(Values), Comparer);
end;

class procedure TArray.SortDescending<T>(var Values: array of T; const Comparer: IComparer<T>; Index, Count: Integer);
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  SortDescending<T>(Values[Index], Count, Comparer);
end;

class procedure TArray.SortDescending<T>(var Values: array of T; const Comparison: TComparison<T>);
begin
  if (High(Values) > 0) then
   SortDescending<T>(Values[0], Length(Values), Comparison);
end;

class procedure TArray.SortDescending<T>(var Values: array of T; Index, Count: Integer; const Comparison: TComparison<T>);
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  SortDescending<T>(Values[Index], Count, Comparison);
end;

{$ifdef SMARTGENERICS}
class function TArray.SearchSigneds<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt;
label
  middle_init, not_found;
type
  TArray1 = array[0..High(Integer) div 1 - 1] of ShortInt;
  TArray2 = array[0..High(Integer) div 2 - 1] of SmallInt;
  TArray4 = array[0..High(Integer) div 4 - 1] of Integer;
  {$ifdef LARGEINT}
    TArray8 = array[0..High(Integer) div 8 - 1] of Int64;
  {$else .SMALLINT}
    TArray8 = array[0..High(Integer) div 8 - 1] of TPoint;
  {$endif}
  PArray1 = ^TArray1;
  PArray2 = ^TArray2;
  PArray4 = ^TArray4;
  PArray8 = ^TArray8;
var
  Item1: ShortInt;
  Item2: SmallInt;
  Item4: Integer;
  {$ifdef LARGEINT}
    Item8: Int64;
  {$else .SMALLINT}
    Item8Low: Cardinal;
    Item8High, Buffer8High: Integer;
  {$endif}
  Left, Right, Middle: NativeInt;
begin
  case SizeOf(T) of
    1: Byte(Item1) := PByte(Item)^;
    2: Word(Item2) := PWord(Item)^;
    4: Integer(Item4) := PInteger(Item)^;
  else
    {$ifdef LARGEINT}
      Int64(Item8) := PInt64(Item)^;
    {$else .SMALLINT}
      Item8Low := PPoint(Item).X;
      Item8High := PPoint(Item).Y;
    {$endif}
  end;

  Middle := -1;
  Right := Count + (-1);
  repeat
    Left := Middle + 1;
    if (Middle >= Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    case SizeOf(T) of
      1: if (PArray1(Values)[Middle] < Item1) then Continue;
      2: if (PArray2(Values)[Middle] < Item2) then Continue;
      4: if (PArray4(Values)[Middle] < Item4) then Continue;
    else
      {$ifdef LARGEINT}
        if (PArray8(Values)[Middle] < Item8) then Continue;
      {$else .SMALLINT}
        Inc(NativeUInt(Values), SizeOf(Integer));
        Buffer8High := PArray8(Values)[Middle].X;
        Dec(NativeUInt(Values), SizeOf(Integer));
        if (Buffer8High < Item8High) then Continue;
        if (Buffer8High = Item8High) then
        begin
          if (Cardinal(PArray8(Values)[Middle].X) < Item8Low) then Continue;
        end;
      {$endif}
    end;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Count) then
  begin
    case SizeOf(T) of
      1: if (PArray1(Values)[Left] <> Item1) then goto not_found;
      2: if (PArray2(Values)[Left] <> Item2) then goto not_found;
      4: if (PArray4(Values)[Left] <> Item4) then goto not_found;
    else
      {$ifdef LARGEINT}
        if (PArray8(Values)[Left] <> Item8) then goto not_found;
      {$else .SMALLINT}
        Inc(NativeUInt(Values), SizeOf(Integer));
        Dec(Item8High, PArray8(Values)[Left].X);
        Dec(NativeUInt(Values), SizeOf(Integer));
        Buffer8High := PArray8(Values)[Left].X;
        Dec(Buffer8High, Item8Low);
        if (Buffer8High or Item8High <> 0) then goto not_found;
      {$endif}
    end;
  end else
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.SearchDescendingSigneds<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt;
label
  middle_init, not_found;
type
  TArray1 = array[0..High(Integer) div 1 - 1] of ShortInt;
  TArray2 = array[0..High(Integer) div 2 - 1] of SmallInt;
  TArray4 = array[0..High(Integer) div 4 - 1] of Integer;
  {$ifdef LARGEINT}
    TArray8 = array[0..High(Integer) div 8 - 1] of Int64;
  {$else .SMALLINT}
    TArray8 = array[0..High(Integer) div 8 - 1] of TPoint;
  {$endif}
  PArray1 = ^TArray1;
  PArray2 = ^TArray2;
  PArray4 = ^TArray4;
  PArray8 = ^TArray8;
var
  Item1: ShortInt;
  Item2: SmallInt;
  Item4: Integer;
  {$ifdef LARGEINT}
    Item8: Int64;
  {$else .SMALLINT}
    Item8Low: Cardinal;
    Item8High, Buffer8High: Integer;
  {$endif}
  Left, Right, Middle: NativeInt;
begin
  case SizeOf(T) of
    1: Byte(Item1) := PByte(Item)^;
    2: Word(Item2) := PWord(Item)^;
    4: Integer(Item4) := PInteger(Item)^;
  else
    {$ifdef LARGEINT}
      Int64(Item8) := PInt64(Item)^;
    {$else .SMALLINT}
      Item8Low := PPoint(Item).X;
      Item8High := PPoint(Item).Y;
    {$endif}
  end;

  Middle := -1;
  Right := Count + (-1);
  repeat
    Left := Middle + 1;
    if (Middle >= Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    case SizeOf(T) of
      1: if (PArray1(Values)[Middle] > Item1) then Continue;
      2: if (PArray2(Values)[Middle] > Item2) then Continue;
      4: if (PArray4(Values)[Middle] > Item4) then Continue;
    else
      {$ifdef LARGEINT}
        if (PArray8(Values)[Middle] > Item8) then Continue;
      {$else .SMALLINT}
        Inc(NativeUInt(Values), SizeOf(Integer));
        Buffer8High := PArray8(Values)[Middle].X;
        Dec(NativeUInt(Values), SizeOf(Integer));
        if (Buffer8High > Item8High) then Continue;
        if (Buffer8High = Item8High) then
        begin
          if (Cardinal(PArray8(Values)[Middle].X) > Item8Low) then Continue;
        end;
      {$endif}
    end;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Count) then
  begin
    case SizeOf(T) of
      1: if (PArray1(Values)[Left] <> Item1) then goto not_found;
      2: if (PArray2(Values)[Left] <> Item2) then goto not_found;
      4: if (PArray4(Values)[Left] <> Item4) then goto not_found;
    else
      {$ifdef LARGEINT}
        if (PArray8(Values)[Left] <> Item8) then goto not_found;
      {$else .SMALLINT}
        Inc(NativeUInt(Values), SizeOf(Integer));
        Dec(Item8High, PArray8(Values)[Left].X);
        Dec(NativeUInt(Values), SizeOf(Integer));
        Buffer8High := PArray8(Values)[Left].X;
        Dec(Buffer8High, Item8Low);
        if (Buffer8High or Item8High <> 0) then goto not_found;
      {$endif}
    end;
  end else
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.SearchUnsigneds<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt;
label
  middle_init, not_found;
type
  TArray1 = array[0..High(Integer) div 1 - 1] of Byte;
  TArray2 = array[0..High(Integer) div 2 - 1] of Word;
  TArray4 = array[0..High(Integer) div 4 - 1] of Cardinal;
  {$ifdef LARGEINT}
    TArray8 = array[0..High(Integer) div 8 - 1] of UInt64;
  {$else .SMALLINT}
    TArray8 = array[0..High(Integer) div 8 - 1] of TPoint;
  {$endif}
  PArray1 = ^TArray1;
  PArray2 = ^TArray2;
  PArray4 = ^TArray4;
  PArray8 = ^TArray8;
var
  Item1: Byte;
  Item2: Word;
  Item4: Cardinal;
  {$ifdef LARGEINT}
    Item8: UInt64;
  {$else .SMALLINT}
    Item8Low: Cardinal;
    Item8High, Buffer8High: Cardinal;
  {$endif}
  Left, Right, Middle: NativeInt;
begin
  case SizeOf(T) of
    1: Byte(Item1) := PByte(Item)^;
    2: Word(Item2) := PWord(Item)^;
    4: Integer(Item4) := PInteger(Item)^;
  else
    {$ifdef LARGEINT}
      Int64(Item8) := PInt64(Item)^;
    {$else .SMALLINT}
      Item8Low := PPoint(Item).X;
      Item8High := PPoint(Item).Y;
    {$endif}
  end;

  Middle := -1;
  Right := Count + (-1);
  repeat
    Left := Middle + 1;
    if (Middle >= Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    case SizeOf(T) of
      1: if (PArray1(Values)[Middle] < Item1) then Continue;
      2: if (PArray2(Values)[Middle] < Item2) then Continue;
      4: if (PArray4(Values)[Middle] < Item4) then Continue;
    else
      {$ifdef LARGEINT}
        if (PArray8(Values)[Middle] < Item8) then Continue;
      {$else .SMALLINT}
        Inc(NativeUInt(Values), SizeOf(Integer));
        Buffer8High := PArray8(Values)[Middle].X;
        Dec(NativeUInt(Values), SizeOf(Integer));
        if (Buffer8High < Item8High) then Continue;
        if (Buffer8High = Item8High) then
        begin
          if (Cardinal(PArray8(Values)[Middle].X) < Item8Low) then Continue;
        end;
      {$endif}
    end;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Count) then
  begin
    case SizeOf(T) of
      1: if (PArray1(Values)[Left] <> Item1) then goto not_found;
      2: if (PArray2(Values)[Left] <> Item2) then goto not_found;
      4: if (PArray4(Values)[Left] <> Item4) then goto not_found;
    else
      {$ifdef LARGEINT}
        if (PArray8(Values)[Left] <> Item8) then goto not_found;
      {$else .SMALLINT}
        Inc(NativeUInt(Values), SizeOf(Integer));
        Dec(Item8High, PArray8(Values)[Left].X);
        Dec(NativeUInt(Values), SizeOf(Integer));
        Buffer8High := PArray8(Values)[Left].X;
        Dec(Buffer8High, Item8Low);
        if (Buffer8High or Item8High <> 0) then goto not_found;
      {$endif}
    end;
  end else
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.SearchDescendingUnsigneds<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt;
label
  middle_init, not_found;
type
  TArray1 = array[0..High(Integer) div 1 - 1] of Byte;
  TArray2 = array[0..High(Integer) div 2 - 1] of Word;
  TArray4 = array[0..High(Integer) div 4 - 1] of Cardinal;
  {$ifdef LARGEINT}
    TArray8 = array[0..High(Integer) div 8 - 1] of UInt64;
  {$else .SMALLINT}
    TArray8 = array[0..High(Integer) div 8 - 1] of TPoint;
  {$endif}
  PArray1 = ^TArray1;
  PArray2 = ^TArray2;
  PArray4 = ^TArray4;
  PArray8 = ^TArray8;
var
  Item1: Byte;
  Item2: Word;
  Item4: Cardinal;
  {$ifdef LARGEINT}
    Item8: UInt64;
  {$else .SMALLINT}
    Item8Low: Cardinal;
    Item8High, Buffer8High: Cardinal;
  {$endif}
  Left, Right, Middle: NativeInt;
begin
  case SizeOf(T) of
    1: Byte(Item1) := PByte(Item)^;
    2: Word(Item2) := PWord(Item)^;
    4: Integer(Item4) := PInteger(Item)^;
  else
    {$ifdef LARGEINT}
      Int64(Item8) := PInt64(Item)^;
    {$else .SMALLINT}
      Item8Low := PPoint(Item).X;
      Item8High := PPoint(Item).Y;
    {$endif}
  end;

  Middle := -1;
  Right := Count + (-1);
  repeat
    Left := Middle + 1;
    if (Middle >= Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    case SizeOf(T) of
      1: if (PArray1(Values)[Middle] > Item1) then Continue;
      2: if (PArray2(Values)[Middle] > Item2) then Continue;
      4: if (PArray4(Values)[Middle] > Item4) then Continue;
    else
      {$ifdef LARGEINT}
        if (PArray8(Values)[Middle] > Item8) then Continue;
      {$else .SMALLINT}
        Inc(NativeUInt(Values), SizeOf(Integer));
        Buffer8High := PArray8(Values)[Middle].X;
        Dec(NativeUInt(Values), SizeOf(Integer));
        if (Buffer8High > Item8High) then Continue;
        if (Buffer8High = Item8High) then
        begin
          if (Cardinal(PArray8(Values)[Middle].X) > Item8Low) then Continue;
        end;
      {$endif}
    end;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Count) then
  begin
    case SizeOf(T) of
      1: if (PArray1(Values)[Left] <> Item1) then goto not_found;
      2: if (PArray2(Values)[Left] <> Item2) then goto not_found;
      4: if (PArray4(Values)[Left] <> Item4) then goto not_found;
    else
      {$ifdef LARGEINT}
        if (PArray8(Values)[Left] <> Item8) then goto not_found;
      {$else .SMALLINT}
        Inc(NativeUInt(Values), SizeOf(Integer));
        Dec(Item8High, PArray8(Values)[Left].X);
        Dec(NativeUInt(Values), SizeOf(Integer));
        Buffer8High := PArray8(Values)[Left].X;
        Dec(Buffer8High, Item8Low);
        if (Buffer8High or Item8High <> 0) then goto not_found;
      {$endif}
    end;
  end else
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.SearchFloats<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt;
label
  middle_init, not_found;
type
  TArray4 = array[0..High(Integer) div 4 - 1] of Single;
  TArray8 = array[0..High(Integer) div 8 - 1] of Double;
  TArrayE = array[0..High(Integer) div 10 - 1] of Extended;
  PArray4 = ^TArray4;
  PArray8 = ^TArray8;
  PArrayE = ^TArrayE;
var
  Item4: {$ifdef CPUX86}Extended{$else}Single{$endif};
  Item8: {$ifdef CPUX86}Extended{$else}Double{$endif};
  ItemE: Extended;
  Left, Right, Middle: NativeInt;
begin
  case SizeOf(T) of
    4: Item4 := PSingle(Item)^;
    8: Item8 := PDouble(Item)^;
  else
    ItemE := PExtended(Item)^;
  end;

  Middle := -1;
  Right := Count + (-1);
  repeat
    Inc(Middle);
    Left := Middle;
    if (Middle > Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    case SizeOf(T) of
      4: if (PArray4(Values)[Middle] < Item4) then Continue;
      8: if (PArray8(Values)[Middle] < Item8) then Continue;
    else
      if (PArrayE(Values)[Middle] < ItemE) then Continue;
    end;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Count) then
  begin
    case SizeOf(T) of
      4: if (PArray4(Values)[Left] <> Item4) then goto not_found;
      8: if (PArray8(Values)[Left] <> Item8) then goto not_found;
    else
      if (PArrayE(Values)[Left] <> ItemE) then goto not_found;
    end;
  end else
  if (Left >= Count) then
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.SearchDescendingFloats<T>(Values: Pointer; Count: NativeInt; Item: Pointer): NativeInt;
label
  middle_init, not_found;
type
  TArray4 = array[0..High(Integer) div 4 - 1] of Single;
  TArray8 = array[0..High(Integer) div 8 - 1] of Double;
  TArrayE = array[0..High(Integer) div 10 - 1] of Extended;
  PArray4 = ^TArray4;
  PArray8 = ^TArray8;
  PArrayE = ^TArrayE;
var
  Item4: {$ifdef CPUX86}Extended{$else}Single{$endif};
  Item8: {$ifdef CPUX86}Extended{$else}Double{$endif};
  ItemE: Extended;
  Left, Right, Middle: NativeInt;
begin
  case SizeOf(T) of
    4: Item4 := PSingle(Item)^;
    8: Item8 := PDouble(Item)^;
  else
    ItemE := PExtended(Item)^;
  end;

  Middle := -1;
  Right := Count + (-1);
  repeat
    Inc(Middle);
    Left := Middle;
    if (Middle > Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    case SizeOf(T) of
      4: if (PArray4(Values)[Middle] > Item4) then Continue;
      8: if (PArray8(Values)[Middle] > Item8) then Continue;
    else
      if (PArrayE(Values)[Middle] > ItemE) then Continue;
    end;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Count) then
  begin
    case SizeOf(T) of
      4: if (PArray4(Values)[Left] <> Item4) then goto not_found;
      8: if (PArray8(Values)[Left] <> Item8) then goto not_found;
    else
      if (PArrayE(Values)[Left] <> ItemE) then goto not_found;
    end;
  end else
  if (Left >= Count) then
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.SearchBinaries<T>(Values: Pointer; Count: NativeInt; const Item: T): NativeInt;
label
  middle_init, not_found;
type
  P = ^T;
var
  Left, Right, Middle: NativeInt;
  X, Y: NativeUInt;
  BufferMiddle, BufferLeft: Pointer;
  Cmp: Integer;
  Stored: record
    X: NativeUInt;
    ItemPtr: Pointer;
  end;
begin
  Middle := -1;
  Right := Count + (-1);
  X := TArray.SortBinaryMarker<T>(@Item);
  Stored.X := X;
  if (GetTypeKind(T) in [tkMethod, tkLString, tkWString, tkUString, tkDynArray]) then
  begin
    Stored.ItemPtr := PPointer(@Item)^;
  end;
  repeat
    Left := Middle + 1;
    if (Middle >= Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    case GetTypeKind(T) of
      tkMethod:
      begin
        Y := Middle;
        Y := Y shl {$ifdef LARGEINT}4{$else .SMALLINT}3{$endif};
        Inc(Y, NativeUInt(Values));
        Y := NativeUInt(PMethod(Y).Data);
      end;
      tkLString, tkWString, tkUString, tkDynArray:
      begin
        Y := PNativeUInt(P(Values) + Middle)^;
        if (Y <> 0) then
        case GetTypeKind(T) of
          tkLString:
          begin
            Y := PWord(Y)^;
            Y := Swap(Y);
          end;
          {$ifdef MSWINDOWS}
          tkWString:
          begin
            Dec(Y, SizeOf(Integer));
            if (PInteger(Y)^ = 0) then
            begin
              Y := 0;
            end else
            begin
              Inc(Y, SizeOf(Integer));
              Y := PCardinal(Y)^;
              Y := Cardinal((Y shl 16) + (Y shr 16));
            end;
          end;
          {$else}
          tkWString,
          {$endif}
          tkUString:
          begin
            Y := PCardinal(Y)^;
            Y := Cardinal((Y shl 16) + (Y shr 16));
          end;
          tkDynArray:
          begin
            Y := PByte(Y)^;
          end;
        end;
      end;
    else
      Y := TArray.SortBinaryMarker<T>(P(Values) + Middle);
    end;

    if (Y < X) then Continue;
    if (Y = X) then
    begin
      if (GetTypeKind(T) = tkMethod) then
      begin
        Y := Middle;
        Y := Y shl {$ifdef LARGEINT}4{$else .SMALLINT}3{$endif};
        Inc(Y, NativeInt(Values));
        Y := NativeUInt(PMethod(Y).Data);
        if (Y < NativeUInt(Stored.ItemPtr)) then Continue;
      end else
      if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
      begin
        BufferMiddle := PPointer(P(Values) + Middle)^;
        if (BufferMiddle <> Stored.ItemPtr) then
        begin
          case GetTypeKind(T) of
            tkLString: Cmp := InterfaceDefaults.Compare_LStr(nil, BufferMiddle, Stored.ItemPtr);
            tkWString: Cmp := InterfaceDefaults.Compare_WStr(nil, BufferMiddle, Stored.ItemPtr);
            tkUString: Cmp := InterfaceDefaults.Compare_UStr(nil, BufferMiddle, Stored.ItemPtr);
           tkDynArray: Cmp := InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, BufferMiddle, Stored.ItemPtr);
          end;
          X := Stored.X;
          if (Cmp < 0) then Continue;
        end;
      end else
      begin
        Cmp := SortBinaryComparer<T>(P(Values) + Middle, @Item);
        X := Stored.X;
        if (Cmp < 0) then Continue;
      end;
    end;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Count) then
  begin
    BufferLeft := P(Values) + Left;
    Y := TArray.SortBinaryMarker<T>(BufferLeft);
    if (Y <> X) then goto not_found;
    if (GetTypeKind(T) = tkMethod) then
    begin
      if (PNativeUInt(BufferLeft)^ <> NativeUInt(Stored.ItemPtr)) then goto not_found;
    end else
    if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
    begin
      BufferLeft := PPointer(BufferLeft)^;
      if (BufferLeft <> Stored.ItemPtr) then
      begin
        case GetTypeKind(T) of
          tkLString: Cmp := InterfaceDefaults.Compare_LStr(nil, BufferLeft, Stored.ItemPtr);
          tkWString: Cmp := InterfaceDefaults.Compare_WStr(nil, BufferLeft, Stored.ItemPtr);
          tkUString: Cmp := InterfaceDefaults.Compare_UStr(nil, BufferLeft, Stored.ItemPtr);
         tkDynArray: Cmp := InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, BufferLeft, Stored.ItemPtr);
        end;
        if (Cmp <> 0) then goto not_found;
      end;
    end else
    begin
      Cmp := SortBinaryComparer<T>(BufferLeft, @Item);
      if (Cmp <> 0) then goto not_found;
    end;
  end else
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.SearchDescendingBinaries<T>(Values: Pointer; Count: NativeInt; const Item: T): NativeInt;
label
  middle_init, not_found;
type
  P = ^T;
var
  Left, Right, Middle: NativeInt;
  X, Y: NativeUInt;
  BufferMiddle, BufferLeft: Pointer;
  Cmp: Integer;
  Stored: record
    X: NativeUInt;
    ItemPtr: Pointer;
  end;
begin
  Middle := -1;
  Right := Count + (-1);
  X := TArray.SortBinaryMarker<T>(@Item);
  Stored.X := X;
  if (GetTypeKind(T) in [tkMethod, tkLString, tkWString, tkUString, tkDynArray]) then
  begin
    Stored.ItemPtr := PPointer(@Item)^;
  end;
  repeat
    Left := Middle + 1;
    if (Middle >= Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    case GetTypeKind(T) of
      tkMethod:
      begin
        Y := Middle;
        Y := Y shl {$ifdef LARGEINT}4{$else .SMALLINT}3{$endif};
        Inc(Y, NativeUInt(Values));
        Y := NativeUInt(PMethod(Y).Data);
      end;
      tkLString, tkWString, tkUString, tkDynArray:
      begin
        Y := PNativeUInt(P(Values) + Middle)^;
        if (Y <> 0) then
        case GetTypeKind(T) of
          tkLString:
          begin
            Y := PWord(Y)^;
            Y := Swap(Y);
          end;
          {$ifdef MSWINDOWS}
          tkWString:
          begin
            Dec(Y, SizeOf(Integer));
            if (PInteger(Y)^ = 0) then
            begin
              Y := 0;
            end else
            begin
              Inc(Y, SizeOf(Integer));
              Y := PCardinal(Y)^;
              Y := Cardinal((Y shl 16) + (Y shr 16));
            end;
          end;
          {$else}
          tkWString,
          {$endif}
          tkUString:
          begin
            Y := PCardinal(Y)^;
            Y := Cardinal((Y shl 16) + (Y shr 16));
          end;
          tkDynArray:
          begin
            Y := PByte(Y)^;
          end;
        end;
      end;
    else
      Y := TArray.SortBinaryMarker<T>(P(Values) + Middle);
    end;

    if (Y > X) then Continue;
    if (Y = X) then
    begin
      if (GetTypeKind(T) = tkMethod) then
      begin
        Y := Middle;
        Y := Y shl {$ifdef LARGEINT}4{$else .SMALLINT}3{$endif};
        Inc(Y, NativeInt(Values));
        Y := NativeUInt(PMethod(Y).Data);
        if (Y > NativeUInt(Stored.ItemPtr)) then Continue;
      end else
      if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
      begin
        BufferMiddle := PPointer(P(Values) + Middle)^;
        if (BufferMiddle <> Stored.ItemPtr) then
        begin
          case GetTypeKind(T) of
            tkLString: Cmp := InterfaceDefaults.Compare_LStr(nil, BufferMiddle, Stored.ItemPtr);
            tkWString: Cmp := InterfaceDefaults.Compare_WStr(nil, BufferMiddle, Stored.ItemPtr);
            tkUString: Cmp := InterfaceDefaults.Compare_UStr(nil, BufferMiddle, Stored.ItemPtr);
           tkDynArray: Cmp := InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, BufferMiddle, Stored.ItemPtr);
          end;
          X := Stored.X;
          if (Cmp > 0) then Continue;
        end;
      end else
      begin
        Cmp := SortBinaryComparer<T>(P(Values) + Middle, @Item);
        X := Stored.X;
        if (Cmp > 0) then Continue;
      end;
    end;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Count) then
  begin
    BufferLeft := P(Values) + Left;
    Y := TArray.SortBinaryMarker<T>(BufferLeft);
    if (Y <> X) then goto not_found;
    if (GetTypeKind(T) = tkMethod) then
    begin
      if (PNativeUInt(BufferLeft)^ <> NativeUInt(Stored.ItemPtr)) then goto not_found;
    end else
    if (GetTypeKind(T) in [tkLString, tkWString, tkUString, tkDynArray]) then
    begin
      BufferLeft := PPointer(BufferLeft)^;
      if (BufferLeft <> Stored.ItemPtr) then
      begin
        case GetTypeKind(T) of
          tkLString: Cmp := InterfaceDefaults.Compare_LStr(nil, BufferLeft, Stored.ItemPtr);
          tkWString: Cmp := InterfaceDefaults.Compare_WStr(nil, BufferLeft, Stored.ItemPtr);
          tkUString: Cmp := InterfaceDefaults.Compare_UStr(nil, BufferLeft, Stored.ItemPtr);
         tkDynArray: Cmp := InterfaceDefaults.Compare_Dyn(InterfaceDefaults.TDefaultComparer<T>.Instance, BufferLeft, Stored.ItemPtr);
        end;
        if (Cmp <> 0) then goto not_found;
      end;
    end else
    begin
      Cmp := SortBinaryComparer<T>(BufferLeft, @Item);
      if (Cmp <> 0) then goto not_found;
    end;
  end else
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;
{$endif .SMARTGENERICS}

class function TArray.SearchUniversals<T>(Values: Pointer; const Helper: TSearchHelper; const Item: T): NativeInt;
label
  middle_init, not_found;
type
  TArrayT = array[0..0] of T;
  PArrayT = ^TArrayT;
var
  Left, Right, Middle: NativeInt;
  Stored: record
    Inst: Pointer;
    Compare: function(const Inst: Pointer; const Left, Right: T): NativeInt;
    Count: NativeInt;
  end;
begin
  Stored.Inst := Helper.Comparer;
  Stored.Compare := PPointer(PNativeUInt(Stored.Inst)^ + 3 * SizeOf(Pointer))^;
  Stored.Count := Helper.Count;

  Middle := -1;
  Right := Stored.Count + (-1);
  repeat
    Left := Middle + 1;
    if (Middle >= Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    if (Stored.Compare(Stored.Inst, PArrayT(Values)[Middle], Item) < 0) then Continue;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Stored.Count) then
  begin
    if (Stored.Compare(Stored.Inst, PArrayT(Values)[Left], Item) <> 0) then goto not_found;
  end else
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.SearchDescendingUniversals<T>(Values: Pointer; const Helper: TSearchHelper; const Item: T): NativeInt;
label
  middle_init, not_found;
type
  TArrayT = array[0..0] of T;
  PArrayT = ^TArrayT;
var
  Left, Right, Middle: NativeInt;
  Stored: record
    Inst: Pointer;
    Compare: function(const Inst: Pointer; const Left, Right: T): NativeInt;
    Count: NativeInt;
  end;
begin
  Stored.Inst := Helper.Comparer;
  Stored.Compare := PPointer(PNativeUInt(Stored.Inst)^ + 3 * SizeOf(Pointer))^;
  Stored.Count := Helper.Count;

  Middle := -1;
  Right := Stored.Count + (-1);
  repeat
    Left := Middle + 1;
    if (Middle >= Right) then Break;

  middle_init:
    Middle := Right;
    Dec(Middle, Left);
    Middle := Left + (Middle shr 1);

    if (Stored.Compare(Stored.Inst, PArrayT(Values)[Middle], Item) > 0) then Continue;

    Right := Middle + (-1);
    if (not (Left > Right)) then goto middle_init;
    Break;
  until (False);

  if (Left < Stored.Count) then
  begin
    if (Stored.Compare(Stored.Inst, PArrayT(Values)[Left], Item) <> 0) then goto not_found;
  end else
  begin
  not_found:
    Left := not Left;
  end;

  Result := Left;
end;

class function TArray.InternalSearch<T>(Values: Pointer; Index, Count: Integer; const Item: T;
  out FoundIndex: Integer): Boolean;
type
  P = ^T;
var
  I: Integer;
  Helper: TSearchHelper;
  {$ifdef SMARTGENERICS}
  TypeData: PTypeData;
  {$endif}
begin
  if (Count <= 0) then
  begin
    if (Count = 0) then
    begin
      FoundIndex := Index;
      Result := True;
      Exit;
    end else
    begin
      raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
    end;
  end;

  {$ifdef SMARTGENERICS}
    if (GetTypeKind(T) in [tkInteger, tkEnumeration, tkChar, tkWChar, tkInt64]) or
      ((GetTypeKind(T) = tkFloat) and (SizeOf(T) = 8)) then
    begin
      TypeData := Pointer(TypeInfo(T));
      Inc(NativeUInt(TypeData), NativeUInt(PByte(@PTypeInfo(TypeData).Name)^) + 2);
    end;

    case GetTypeKind(T) of
      tkInteger, tkEnumeration, tkChar, tkWChar:
      case SizeOf(T) of
        1:
        begin
          case TypeData.OrdType of
            otSByte: I := SearchSigneds<ShortInt>(P(Values) + Index, Count, @Item);
            otUByte: I := SearchUnsigneds<Byte>(P(Values) + Index, Count, @Item);
          end;
        end;
        2:
        begin
          case TypeData.OrdType of
            otSWord: I := SearchSigneds<SmallInt>(P(Values) + Index, Count, @Item);
            otUWord: I := SearchUnsigneds<Word>(P(Values) + Index, Count, @Item);
          end;
        end;
        4:
        begin
          case TypeData.OrdType of
            otSLong: I := SearchSigneds<Integer>(P(Values) + Index, Count, @Item);
            otULong: I := SearchUnsigneds<Cardinal>(P(Values) + Index, Count, @Item);
          end;
        end;
      end;
      tkInt64:
      begin
        if (TypeData.MaxInt64Value > TypeData.MinInt64Value) then
        begin
          I := SearchSigneds<Int64>(P(Values) + Index, Count, @Item);
        end else
        begin
          I := SearchUnsigneds<UInt64>(P(Values) + Index, Count, @Item);
        end;
      end;
      tkClass, tkInterface, tkClassRef, tkPointer, tkProcedure:
      begin
        {$ifdef LARGEINT}
          I := SearchUnsigneds<UInt64>(P(Values) + Index, Count, @Item);
        {$else .SMALLINT}
          I := SearchUnsigneds<Cardinal>(P(Values) + Index, Count, @Item);
        {$endif}
      end;
      tkFloat:
      case SizeOf(T) of
         4: I := SearchFloats<Single>(P(Values) + Index, Count, @Item);
        10: I := SearchFloats<Extended>(P(Values) + Index, Count, @Item);
      else
        if (TypeData.FloatType = ftDouble) then
        begin
          I := SearchFloats<Double>(P(Values) + Index, Count, @Item);
        end else
        begin
          I := SearchSigneds<Int64>(P(Values) + Index, Count, @Item);
        end;
      end;
      tkMethod:
      begin
        I := SearchBinaries<InterfaceDefaults.TMethodPtr>(P(Values) + Index, Count,
          InterfaceDefaults.TMethodPtr(Pointer(@Item)^));
      end;
      tkVariant:
      begin
        Helper.Count := Count;
        Helper.Comparer := @InterfaceDefaults.TDefaultComparer<Variant>.Instance;
        I := SearchUniversals<T>(P(Values) + Index, Helper, Item);
      end;
      tkString:
      begin
        I := SearchBinaries<T>(P(Values) + Index, Count, Item);
      end;
      tkLString:
      begin
        {$ifdef NEXTGEN}
           I := SearchBinaries<T>(P(Values) + Index, Count, Item);
        {$else}
           I := SearchBinaries<AnsiString>(P(Values) + Index, Count, AnsiString(Pointer(@Item)^));
        {$endif}
      end;
      {$ifdef MSWINDOWS}
      tkWString:
      begin
        I := SearchBinaries<WideString>(P(Values) + Index, Count, WideString(Pointer(@Item)^));
      end;
      {$else}
      tkWString,
      {$endif}
      tkUString:
      begin
        I := SearchBinaries<UnicodeString>(P(Values) + Index, Count, UnicodeString(Pointer(@Item)^));
      end;
      tkDynArray:
      begin
        I := SearchBinaries<T>(P(Values) + Index, Count, Item);
      end;
    else
      // binary
      I := SearchBinaries<T>(P(Values) + Index, Count, Item);
    end;
  {$else}
    Helper.Count := Count;
    Helper.Comparer := @InterfaceDefaults.TDefaultComparer<T>.Instance;
    I := SearchUniversals<T>(P(Values) + Index, Helper, Item);
  {$endif}

  if (I < 0) then
  begin
    FoundIndex := Index + (not I);
    Result := False;
  end else
  begin
    FoundIndex := Index + I;
    Result := True;
  end;
end;

class function TArray.InternalSearch<T>(Values: Pointer; Index, Count: Integer; const Item: T;
  out FoundIndex: Integer; Comparer: Pointer): Boolean;
type
  P = ^T;
var
  I: Integer;
  Helper: TSearchHelper;
begin
  if (Count <= 0) then
  begin
    if (Count = 0) then
    begin
      FoundIndex := Index;
      Result := True;
      Exit;
    end else
    begin
      raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
    end;
  end;

  Helper.Count := Count;
  Helper.Comparer := Comparer;
  I := TArray.SearchUniversals<T>(P(Values) + Index, Helper, Item);
  if (I < 0) then
  begin
    FoundIndex := Index + (not I);
    Result := False;
  end else
  begin
    FoundIndex := Index + I;
    Result := True;
  end;
end;

class function TArray.BinarySearch<T>(var Values: T; const Item: T;
  out FoundIndex: Integer; Count: Integer): Boolean;
begin
  Result := TArray.InternalSearch<T>(@Values, 0, Count, Item, FoundIndex);
end;

class function TArray.BinarySearch<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer): Boolean;
begin
  Result := TArray.InternalSearch<T>(@Values[0], 0, Length(Values), Item, FoundIndex);
end;

class function TArray.BinarySearch<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; Index, Count: Integer): Boolean;
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearch<T>(@Values[0], Index, Count, Item, FoundIndex);
end;

class function TArray.BinarySearch<T>(var Values: T; const Item: T;
  out FoundIndex: Integer; Count: Integer; const Comparer: IComparer<T>): Boolean;
begin
  Result := TArray.InternalSearch<T>(@Values, 0, Count, Item, FoundIndex, Pointer(Comparer));
end;

class function TArray.BinarySearch<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; const Comparer: IComparer<T>): Boolean;
begin
  Result := TArray.InternalSearch<T>(@Values[0], 0, Length(Values), Item, FoundIndex, Pointer(Comparer));
end;

class function TArray.BinarySearch<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; const Comparer: IComparer<T>; Index, Count: Integer): Boolean;
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearch<T>(@Values[0], Index, Count, Item, FoundIndex, Pointer(Comparer));
end;

class function TArray.BinarySearch<T>(var Values: T; const Item: T;
  out FoundIndex: Integer; Count: Integer; const Comparison: TComparison<T>): Boolean;
begin
  Result := TArray.InternalSearch<T>(@Values, 0, Count, Item, FoundIndex, PPointer(@Comparison)^);
end;

class function TArray.BinarySearch<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; const Comparison: TComparison<T>): Boolean;
begin
  Result := TArray.InternalSearch<T>(@Values[0], 0, Length(Values), Item, FoundIndex, PPointer(@Comparison)^);
end;

class function TArray.BinarySearch<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; Index, Count: Integer; const Comparison: TComparison<T>): Boolean;
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearch<T>(@Values[0], Index, Count, Item, FoundIndex, PPointer(@Comparison)^);
end;

class function TArray.InternalSearchDescending<T>(Values: Pointer; Index, Count: Integer; const Item: T;
  out FoundIndex: Integer): Boolean;
type
  P = ^T;
var
  I: Integer;
  Helper: TSearchHelper;
  {$ifdef SMARTGENERICS}
  TypeData: PTypeData;
  {$endif}
begin
  if (Count <= 0) then
  begin
    if (Count = 0) then
    begin
      FoundIndex := Index;
      Result := True;
      Exit;
    end else
    begin
      raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
    end;
  end;

  {$ifdef SMARTGENERICS}
    if (GetTypeKind(T) in [tkInteger, tkEnumeration, tkChar, tkWChar, tkInt64]) or
      ((GetTypeKind(T) = tkFloat) and (SizeOf(T) = 8)) then
    begin
      TypeData := Pointer(TypeInfo(T));
      Inc(NativeUInt(TypeData), NativeUInt(PByte(@PTypeInfo(TypeData).Name)^) + 2);
    end;

    case GetTypeKind(T) of
      tkInteger, tkEnumeration, tkChar, tkWChar:
      case SizeOf(T) of
        1:
        begin
          case TypeData.OrdType of
            otSByte: I := SearchDescendingSigneds<ShortInt>(P(Values) + Index, Count, @Item);
            otUByte: I := SearchDescendingUnsigneds<Byte>(P(Values) + Index, Count, @Item);
          end;
        end;
        2:
        begin
          case TypeData.OrdType of
            otSWord: I := SearchDescendingSigneds<SmallInt>(P(Values) + Index, Count, @Item);
            otUWord: I := SearchDescendingUnsigneds<Word>(P(Values) + Index, Count, @Item);
          end;
        end;
        4:
        begin
          case TypeData.OrdType of
            otSLong: I := SearchDescendingSigneds<Integer>(P(Values) + Index, Count, @Item);
            otULong: I := SearchDescendingUnsigneds<Cardinal>(P(Values) + Index, Count, @Item);
          end;
        end;
      end;
      tkInt64:
      begin
        if (TypeData.MaxInt64Value > TypeData.MinInt64Value) then
        begin
          I := SearchDescendingSigneds<Int64>(P(Values) + Index, Count, @Item);
        end else
        begin
          I := SearchDescendingUnsigneds<UInt64>(P(Values) + Index, Count, @Item);
        end;
      end;
      tkClass, tkInterface, tkClassRef, tkPointer, tkProcedure:
      begin
        {$ifdef LARGEINT}
          I := SearchDescendingUnsigneds<UInt64>(P(Values) + Index, Count, @Item);
        {$else .SMALLINT}
          I := SearchDescendingUnsigneds<Cardinal>(P(Values) + Index, Count, @Item);
        {$endif}
      end;
      tkFloat:
      case SizeOf(T) of
         4: I := SearchDescendingFloats<Single>(P(Values) + Index, Count, @Item);
        10: I := SearchDescendingFloats<Extended>(P(Values) + Index, Count, @Item);
      else
        if (TypeData.FloatType = ftDouble) then
        begin
          I := SearchDescendingFloats<Double>(P(Values) + Index, Count, @Item);
        end else
        begin
          I := SearchDescendingSigneds<Int64>(P(Values) + Index, Count, @Item);
        end;
      end;
      tkMethod:
      begin
        I := SearchDescendingBinaries<InterfaceDefaults.TMethodPtr>(P(Values) + Index, Count,
          InterfaceDefaults.TMethodPtr(Pointer(@Item)^));
      end;
      tkVariant:
      begin
        Helper.Count := Count;
        Helper.Comparer := @InterfaceDefaults.TDefaultComparer<Variant>.Instance;
        I := SearchDescendingUniversals<T>(P(Values) + Index, Helper, Item);
      end;
      tkString:
      begin
        I := SearchDescendingBinaries<T>(P(Values) + Index, Count, Item);
      end;
      tkLString:
      begin
        {$ifdef NEXTGEN}
           I := SearchDescendingBinaries<T>(P(Values) + Index, Count, Item);
        {$else}
           I := SearchDescendingBinaries<AnsiString>(P(Values) + Index, Count, AnsiString(Pointer(@Item)^));
        {$endif}
      end;
      {$ifdef MSWINDOWS}
      tkWString:
      begin
        I := SearchDescendingBinaries<WideString>(P(Values) + Index, Count, WideString(Pointer(@Item)^));
      end;
      {$else}
      tkWString,
      {$endif}
      tkUString:
      begin
        I := SearchDescendingBinaries<UnicodeString>(P(Values) + Index, Count, UnicodeString(Pointer(@Item)^));
      end;
      tkDynArray:
      begin
        I := SearchDescendingBinaries<T>(P(Values) + Index, Count, Item);
      end;
    else
      // binary
      I := SearchDescendingBinaries<T>(P(Values) + Index, Count, Item);
    end;
  {$else}
    Helper.Count := Count;
    Helper.Comparer := @InterfaceDefaults.TDefaultComparer<T>.Instance;
    I := SearchDescendingUniversals<T>(P(Values) + Index, Helper, Item);
  {$endif}

  if (I < 0) then
  begin
    FoundIndex := Index + (not I);
    Result := False;
  end else
  begin
    FoundIndex := Index + I;
    Result := True;
  end;
end;

class function TArray.InternalSearchDescending<T>(Values: Pointer; Index, Count: Integer; const Item: T;
  out FoundIndex: Integer; Comparer: Pointer): Boolean;
type
  P = ^T;
var
  I: Integer;
  Helper: TSearchHelper;
begin
  if (Count <= 0) then
  begin
    if (Count = 0) then
    begin
      FoundIndex := Index;
      Result := True;
      Exit;
    end else
    begin
      raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
    end;
  end;

  Helper.Count := Count;
  Helper.Comparer := Comparer;
  I := TArray.SearchDescendingUniversals<T>(P(Values) + Index, Helper, Item);
  if (I < 0) then
  begin
    FoundIndex := Index + (not I);
    Result := False;
  end else
  begin
    FoundIndex := Index + I;
    Result := True;
  end;
end;

class function TArray.BinarySearchDescending<T>(var Values: T; const Item: T;
  out FoundIndex: Integer; Count: Integer): Boolean;
begin
  Result := TArray.InternalSearchDescending<T>(@Values, 0, Count, Item, FoundIndex);
end;

class function TArray.BinarySearchDescending<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer): Boolean;
begin
  Result := TArray.InternalSearchDescending<T>(@Values[0], 0, Length(Values), Item, FoundIndex);
end;

class function TArray.BinarySearchDescending<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; Index, Count: Integer): Boolean;
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearchDescending<T>(@Values[0], Index, Count, Item, FoundIndex);
end;

class function TArray.BinarySearchDescending<T>(var Values: T; const Item: T;
  out FoundIndex: Integer; Count: Integer; const Comparer: IComparer<T>): Boolean;
begin
  Result := TArray.InternalSearchDescending<T>(@Values, 0, Count, Item, FoundIndex, Pointer(Comparer));
end;

class function TArray.BinarySearchDescending<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; const Comparer: IComparer<T>): Boolean;
begin
  Result := TArray.InternalSearchDescending<T>(@Values[0], 0, Length(Values), Item, FoundIndex, Pointer(Comparer));
end;

class function TArray.BinarySearchDescending<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; const Comparer: IComparer<T>; Index, Count: Integer): Boolean;
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearchDescending<T>(@Values[0], Index, Count, Item, FoundIndex, Pointer(Comparer));
end;

class function TArray.BinarySearchDescending<T>(var Values: T; const Item: T;
  out FoundIndex: Integer; Count: Integer; const Comparison: TComparison<T>): Boolean;
begin
  Result := TArray.InternalSearchDescending<T>(@Values, 0, Count, Item, FoundIndex, PPointer(@Comparison)^);
end;

class function TArray.BinarySearchDescending<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; const Comparison: TComparison<T>): Boolean;
begin
  Result := TArray.InternalSearchDescending<T>(@Values[0], 0, Length(Values), Item, FoundIndex, PPointer(@Comparison)^);
end;

class function TArray.BinarySearchDescending<T>(const Values: array of T; const Item: T;
  out FoundIndex: Integer; Index, Count: Integer; const Comparison: TComparison<T>): Boolean;
begin
  if (Index < Low(Values)) or ((Index > High(Values)) and (Count > 0))
    or (Index + Count - 1 > High(Values)) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearchDescending<T>(@Values[0], Index, Count, Item, FoundIndex, PPointer(@Comparison)^);
end;


{ TCustomList<T>.TEnumerator }

constructor TCustomList<T>.TEnumerator.Create(const AList: TCustomList<T>);
begin
  inherited Create;
  FIndex := -1;
  FList := AList;
end;

function TCustomList<T>.TEnumerator.DoGetCurrent: T;
begin
  Result := GetCurrent;
end;

function TCustomList<T>.TEnumerator.DoMoveNext: Boolean;
begin
  Result := MoveNext;
end;

function TCustomList<T>.TEnumerator.GetCurrent: T;
var
  Index, Cap: NativeInt;
begin
  with FList do
  begin
    Index := FTail + FIndex;
    Cap := FCapacity.Native;
    if (Index > Cap) then Dec(Index, Cap);
    Result := FItems[Index];
  end;
end;

function TCustomList<T>.TEnumerator.MoveNext: Boolean;
var
  N, R: NativeInt;
begin
  N := FIndex + 1;
  R := Byte(N < FList.FCount.Native);
  Dec(N, R);
  FIndex := N;
  Result := Boolean(R);
end;

{ TCustomList<T> }

constructor TCustomList<T>.Create;
begin
  inherited Create;
  SetNotifyMethods;
end;

destructor TCustomList<T>.Destroy;
begin
  Clear;
  ClearMethod(FInternalNotify);
  inherited;
end;

class procedure TCustomList<T>.ClearMethod(var Method);
begin
  {$ifdef WEAKREF}
    TMethod(Method).Data := nil;
  {$endif}
end;

class function TCustomList<T>.EmptyException: Exception;
begin
  Result := EListError.CreateRes(Pointer(@SUnbalancedOperation));
end;

procedure TCustomList<T>.SetCapacity(Value: Integer);
var
  Dif, NewTail: NativeInt;
begin
  if (Value = FCapacity.Int) then Exit;
  if Value < Count then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  if (FTail <= FHead) then
  begin
    if (FTail <> 0) then
    begin
      Move(FItems[FTail], FItems[0], FCount.Native * SizeOf(T));
      Dec(FHead, FTail);
      FTail := 0;
    end;

    FCapacity.Native := Value;
    ReallocMem(FItems, Value * SizeOf(T));
  end else
  begin
    Dif := NativeInt(Value) - FCapacity.Native;
    NewTail := FTail + Dif;

    if (Dif > 0) then
    begin
      ReallocMem(FItems, Value * SizeOf(T));
      Move(FItems[FTail], FItems[NewTail], FCapacity.Native - FTail - 1);
    end else
    //if (Dif < 0) then
    begin
      Move(FItems[FTail], FItems[NewTail], FCapacity.Native - FTail - 1);
      ReallocMem(FItems, Value * SizeOf(T));
    end;

    FCapacity.Native := Value;
    FTail := NewTail;
  end;
end;

procedure TCustomList<T>.Grow;
var
  OldCapacity, NewCapacity: Integer;
begin
  OldCapacity := FCapacity.Int;
  NewCapacity := OldCapacity * 2;
  if (NewCapacity < 0) then
    OutOfMemoryError;
  if (NewCapacity < 4) then
    NewCapacity := 4;

  SetCapacity(NewCapacity);
end;

procedure TCustomList<T>.GrowTo(Value: Integer);
var
  OldCapacity, NewCapacity: Integer;
begin
  OldCapacity := FCapacity.Int;
  NewCapacity := OldCapacity * 2;
  if (NewCapacity < 0) then
    OutOfMemoryError;
  if (NewCapacity < 4) then
    NewCapacity := 4;

  while (NewCapacity < Value) do
  begin
    NewCapacity := NewCapacity * 2;
    if (NewCapacity < 0) then
      OutOfMemoryError;
  end;

  SetCapacity(NewCapacity)
end;

procedure TCustomList<T>.Clear;
var
  i: NativeInt;
  Item: PItem;
begin
  if (FItems = nil) then Exit;

  if (Assigned(FInternalNotify)) then
  begin
    Item := @FItems[FTail];

    if (FTail <= FHead) then
    begin
      if (TMethod(FInternalNotify).Code = @TCustomList<T>.NotifyCaller) then
      begin
        for i := 1 to FCount.Native do
        begin
          Self.Notify(Item^, cnRemoved);
          Inc(Item);
        end;
      end else
      begin
        for i := 1 to FCount.Native do
        begin
          FInternalNotify(Self, Item^, cnRemoved);
          Inc(Item);
        end;
      end;
    end else
    if (TMethod(FInternalNotify).Code = @TCustomList<T>.NotifyCaller) then
    begin
      for i := FTail to FCapacity.Native - 1 do
      begin
        Self.Notify(Item^, cnRemoved);
        Inc(Item);
      end;
      Item := @FItems[0];
      for i := 0 to FHead - 1 do
      begin
        Self.Notify(Item^, cnRemoved);
        Inc(Item);
      end;
    end else
    begin
      for i := FTail to FCapacity.Native - 1 do
      begin
        FInternalNotify(Self, Item^, cnRemoved);
        Inc(Item);
      end;
      Item := @FItems[0];
      for i := 0 to FHead - 1 do
      begin
        FInternalNotify(Self, Item^, cnRemoved);
        Inc(Item);
      end;
    end;
  end;

  {$ifdef SMARTGENERICS}
  if (System.IsManagedType(T) or System.HasWeakRef(T)) then
  {$else}
  if Assigned(TRAIIHelper<T>.FOptions.ClearProc) then
  {$endif}
  begin
    if (FTail <= FHead) then
    begin
      TRAIIHelper<T>.ClearArray(@FItems[FTail], FCount.Native);
    end else
    begin
      TRAIIHelper<T>.ClearArray(@FItems[FTail], FCapacity.Native - FTail - 1);
      TRAIIHelper<T>.ClearArray(@FItems[0], FHead);
    end;
  end;

  FCount.Native := 0;
  FCapacity.Native := 0;
  FTail := 0;
  FHead := 0;
  ReallocMem(FItems, 0);
end;

procedure TCustomList<T>.TrimExcess;
begin
  SetCapacity(Count);
end;

procedure TCustomList<T>.SetOnNotify(const Value: TCollectionNotifyEvent<T>);
begin
  if (@FOnNotify <> @Value) then
  begin
    FOnNotify := Value;
    SetNotifyMethods;
  end;
end;

procedure TCustomList<T>.Notify(const Item: T; Action: TCollectionNotification);
begin
  if Assigned(FOnNotify) then
    FOnNotify(Self, Item, Action);
end;

procedure TCustomList<T>.NotifyCaller(Sender: TObject; const Item: T; Action: TCollectionNotification);
begin
  Self.Notify(Item, Action);
end;

procedure TCustomList<T>.SetNotifyMethods;
var
  VMTNotify: procedure(const Item: T; Action: TCollectionNotification) of object;
begin
  VMTNotify := Self.Notify;
  if (TMethod(VMTNotify).Code <> @TCustomList<T>.Notify) then
  begin
    TMethod(FInternalNotify).Data := Pointer(Self);
    TMethod(FInternalNotify).Code := @TCustomList<T>.NotifyCaller;
  end else
  begin
    TMethod(FInternalNotify) := TMethod(Self.FOnNotify);
  end;
end;

function TCustomList<T>.DoGetEnumerator: TEnumerator<T>;
begin
  Result := GetEnumerator;
end;

function TCustomList<T>.GetEnumerator: TEnumerator;
begin
  Result := TEnumerator.Create(Self);
end;

function TCustomList<T>.ToArray: TArray<T>;
var
  Count, TailCount: NativeInt;
begin
  Count := FCount.Native;
  if (Count <> 0) then
  begin
    if (Pointer(Result) <> nil) then Result := nil;
    SetLength(Result, Count);

    if (FTail <= FHead) then
    begin
      {$ifdef SMARTGENERICS}
      if (System.IsManagedType(T) or System.HasWeakRef(T)) then
      {$else}
      if Assigned(TRAIIHelper<T>.FOptions.ClearProc) then
      {$endif}
      begin
        System.CopyArray(Pointer(Result), @FItems[FTail], TypeInfo(T), Count);
      end else
      begin
        Move(FItems[FTail], Pointer(Result)^, Count * SizeOf(T));
      end;
    end else
    begin
      TailCount := FCapacity.Native - FTail - 1;

      {$ifdef SMARTGENERICS}
      if (System.IsManagedType(T) or System.HasWeakRef(T)) then
      {$else}
      if Assigned(TRAIIHelper<T>.FOptions.ClearProc) then
      {$endif}
      begin
        System.CopyArray(Pointer(Result), @FItems[FTail], TypeInfo(T), TailCount);
        System.CopyArray(@Result[TailCount], @FItems[0], TypeInfo(T), FHead);
      end else
      begin
        Move(FItems[FTail], Pointer(Result)^, TailCount * SizeOf(T));
        Move(FItems[0], Result[TailCount], FHead * SizeOf(T));
      end;
    end;
  end else
  begin
    if (Pointer(Result) <> nil) then Result := nil;
  end;
end;


{ TList<T> }

constructor TList<T>.Create;
begin
  inherited Create;
end;

constructor TList<T>.Create(const AComparer: IComparer<T>);
begin
  Create;
  if (Pointer(AComparer) <> @InterfaceDefaults.TDefaultComparer<T>.Instance) then
    FComparer := AComparer;
end;

constructor TList<T>.Create(const Collection: TEnumerable<T>);
begin
  Create;
  InsertRange(0, Collection);
end;

class procedure TList<T>.Error(const Msg: string; Data: NativeInt);
begin
  raise EListError.CreateFmt(Msg, [Data]) at ReturnAddress;
end;

{$ifNdef NEXTGEN}
class procedure TList<T>.Error(Msg: PResStringRec; Data: NativeInt);
begin
  raise EListError.CreateFmt(LoadResString(Msg), [Data]) at ReturnAddress;
end;
{$endif}

procedure TList<T>.SetCount(Value: Integer);
var
  Count: Integer;
begin
  Count := FCount.Int;
  if (Value < 0) then
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end else
  if (Value < Count) then
  begin
    DeleteRange(Value, Count - Value);
  end else
  if (Value > Count) then
  begin
    if (Value > FCapacity.Int) then
      GrowTo(Value);

    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(T) or System.HasWeakRef(T)) then
    {$else}
    if (Assigned(TRAIIHelper<T>.Options.InitProc)) then
    {$endif}
    begin
      if (SizeOf(T) <= 16) then
      begin
        FillChar(FItems[FCount.Native], (Value - FCount.Int) * SizeOf(T), #0);
      end else
      begin
        TRAIIHelper<T>.InitArray(@FItems[FCount.Native], Value - FCount.Int);
      end;
    end;

    FCount.Int := Value;
  end;
end;

function TList<T>.GetItem(Index: Integer): T;
begin
  if (Cardinal(Index) < Cardinal(FCount.Int)) then
  begin
    Result := FItems[Index];
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

procedure TList<T>.SetItem(Index: Integer; const Value: T);
begin
  if (Cardinal(Index) < Cardinal(FCount.Int)) then
  begin
    FItems[Index] := Value;
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

function TList<T>.First: T;
begin
  if (FCount.Native <> 0) then
  begin
    Result := FItems[0];
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

function TList<T>.Last: T;
var
  Count: NativeInt;
begin
  Count := FCount.Native;
  if (Count <> 0) then
  begin
    Dec(Count);
    Result := FItems[Count];
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

function TList<T>.ItemValue(const Item: T): NativeInt;
begin
  case SizeOf(T) of
    1: Result := PByte(@Item)^;
    2: Result := PWord(@Item)^;
    3: Result := PWord(@Item)^ + PByte(@Item)[2] shl 16;
    {$ifdef LARGEINT}
    4: Result := PCardinal(@Item)^;
    5: Result := NativeInt(PCardinal(@Item)^) +
       NativeInt(PByte(@Item)[4]) shl 32;
    6: Result := NativeInt(PCardinal(@Item)^) +
       NativeInt(PWord(PByte(@Item) + SizeOf(Cardinal))^) shl 32;
    7: Result := NativeInt(PCardinal(@Item)^) +
       NativeInt(PWord(PByte(@Item) + SizeOf(Cardinal))^) shl 32 + NativeInt(PByte(@Item)[6]) shl 48;
    {$endif}
  else
    Result := PNativeInt(@Item)^;
  end;
end;

function TList<T>.InternalInsert(Index: NativeInt; const Value: T): Integer;
label
  available;
var
  Count, Null: NativeInt;
  Item: ^TRAIIHelper.TData16;
begin
  Result := Index;
  Count := FCount.Native;
  if (NativeUInt(Index) <= NativeUInt(Count)) then
  begin
    if (Count <> FCapacity.Native) then
    begin
    available:
      Inc(Count);
      FCount.Native := Count;
      Dec(Count);
      if (Index <> Count) then
      begin
        Count := (Count - Index) * SizeOf(T);
        Item := Pointer(@FItems[Index]);
        System.Move(Item^, PByte(PByte(Item) + SizeOf(T))^, Count);
        Index := Result;
      end;
      Item := Pointer(@FItems[Index]);

      {$ifdef SMARTGENERICS}
      if (System.IsManagedType(T) or System.HasWeakRef(T)) then
      {$else}
      if (SizeOf(T) >= SizeOf(NativeInt)) then
      {$endif}
      begin
        {$ifdef SMARTGENERICS}
        if (GetTypeKind(T) = tkVariant) then
        begin
          Item.Integers[0] := 0;
        end else
        {$endif}
        if (SizeOf(T) <= 16) then
        begin
          Null := 0;
          {$ifdef SMALLINT}
            if (SizeOf(T) >= SizeOf(Integer) * 1) then Item.Integers[0] := Null;
            if (SizeOf(T) >= SizeOf(Integer) * 2) then Item.Integers[1] := Null;
            if (SizeOf(T) >= SizeOf(Integer) * 3) then Item.Integers[2] := Null;
            if (SizeOf(T)  = SizeOf(Integer) * 4) then Item.Integers[3] := Null;
          {$else .LARGEINT}
            if (SizeOf(T) >= SizeOf(Int64) * 1) then Item.Int64s[0] := Null;
            if (SizeOf(T)  = SizeOf(Int64) * 2) then Item.Int64s[1] := Null;
            case SizeOf(T) of
               4..7: Item.Integers[0] := Null;
             12..15: Item.Integers[2] := Null;
            end;
          {$endif}
          case SizeOf(T) of
             2,3: Item.Words[0] := 0;
             6,7: Item.Words[2] := 0;
           10,11: Item.Words[4] := 0;
           14,15: Item.Words[6] := 0;
          end;
          case SizeOf(T) of
             1: Item.Bytes[ 1-1] := 0;
             3: Item.Bytes[ 3-1] := 0;
             5: Item.Bytes[ 5-1] := 0;
             7: Item.Bytes[ 7-1] := 0;
             9: Item.Bytes[ 9-1] := 0;
            11: Item.Bytes[11-1] := 0;
            13: Item.Bytes[13-1] := 0;
            15: Item.Bytes[15-1] := 0;
          end;
        end else
        begin
          TRAIIHelper<T>.Init(Pointer(Item));
        end;
      end;

      PItem(Item)^ := Value;
      if Assigned(FInternalNotify) then
        FInternalNotify(Self, Value, cnAdded);
    end else
    begin
      Self.Grow;
      Count := FCount.Native;
      Index := Result;
      goto available;
    end;
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

function TList<T>.Add(const Value: T): Integer;
var
  Count, Null: NativeInt;
  Item: ^TRAIIHelper.TData16;
begin
  Count := FCount.Native;
  if (Count <> FCapacity.Native) and (not Assigned(FInternalNotify)) then
  begin
    Inc(Count);
    FCount.Native := Count;
    Dec(Count);
    Item := Pointer(@FItems[Count]);

    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(T) or System.HasWeakRef(T)) then
    {$else}
    if (SizeOf(T) >= SizeOf(NativeInt)) then
    {$endif}
    begin
      {$ifdef SMARTGENERICS}
      if (GetTypeKind(T) = tkVariant) then
      begin
        Item.Integers[0] := 0;
      end else
      {$endif}
      if (SizeOf(T) <= 16) then
      begin
        Null := 0;
        {$ifdef SMALLINT}
          if (SizeOf(T) >= SizeOf(Integer) * 1) then Item.Integers[0] := Null;
          if (SizeOf(T) >= SizeOf(Integer) * 2) then Item.Integers[1] := Null;
          if (SizeOf(T) >= SizeOf(Integer) * 3) then Item.Integers[2] := Null;
          if (SizeOf(T)  = SizeOf(Integer) * 4) then Item.Integers[3] := Null;
        {$else .LARGEINT}
          if (SizeOf(T) >= SizeOf(Int64) * 1) then Item.Int64s[0] := Null;
          if (SizeOf(T)  = SizeOf(Int64) * 2) then Item.Int64s[1] := Null;
          case SizeOf(T) of
             4..7: Item.Integers[0] := Null;
           12..15: Item.Integers[2] := Null;
          end;
        {$endif}
        case SizeOf(T) of
           2,3: Item.Words[0] := 0;
           6,7: Item.Words[2] := 0;
         10,11: Item.Words[4] := 0;
         14,15: Item.Words[6] := 0;
        end;
        case SizeOf(T) of
           1: Item.Bytes[ 1-1] := 0;
           3: Item.Bytes[ 3-1] := 0;
           5: Item.Bytes[ 5-1] := 0;
           7: Item.Bytes[ 7-1] := 0;
           9: Item.Bytes[ 9-1] := 0;
          11: Item.Bytes[11-1] := 0;
          13: Item.Bytes[13-1] := 0;
          15: Item.Bytes[15-1] := 0;
        end;
      end else
      begin
        TRAIIHelper<T>.Init(Pointer(Item));
      end;
    end;

    PItem(Item)^ := Value;
    Result := Count;
    Exit;
  end else
  begin
    Result := InternalInsert(Count, Value);
  end;
end;

procedure TList<T>.Insert(Index: Integer; const Value: T);
begin
  InternalInsert(Index, Value);
end;

procedure TList<T>.AddRange(const Values: array of T);
label
  available, out_of_memory;
var
  Count, ValuesCount, i: NativeInt;
  Item, Source, Buffer: PItem;
  Stored: TInternalStored;
begin
  ValuesCount := High(Values);
  if (ValuesCount < 0) then Exit;
  Inc(ValuesCount);

  Stored.Self := Self;
  Stored.InternalNotify := TMethod(FInternalNotify);

  Count := FCount.Native;
  Inc(Count, ValuesCount);
  if (NativeUInt(Count) > NativeUInt(High(Integer))) then goto out_of_memory;

  if (Count <= FCapacity.Native) then
  begin
  available:
    FCount.Native := Count;
    Dec(Count, ValuesCount);
    Buffer{Item} := @FItems[Count];

    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(T) or System.HasWeakRef(T)) then
    {$else}
    if (Assigned(TRAIIHelper<T>.Options.InitProc)) then
    {$endif}
    begin
      if (SizeOf(T) <= 16) then
      begin
        FillChar(Buffer{Item}^, ValuesCount * SizeOf(T), #0);
      end else
      begin
        TRAIIHelper<T>.InitArray(Buffer{Item}, ValuesCount);
      end;
    end else
    if (not Assigned(Stored.InternalNotify.Code)) then
    begin
      System.Move(Values[0], Buffer{Item}^, ValuesCount * SizeOf(T));
      Exit;
    end;

    Source := @Values[0];
    Item := Buffer{Item};
    if (not Assigned(Stored.InternalNotify.Code)) then
    begin
      for ValuesCount := ValuesCount downto 1 do
      begin
        Item^ := Source^;
        Inc(Source);
        Inc(Item);
      end;
    end else
    begin
      for ValuesCount := ValuesCount downto 1 do
      begin
        Item^ := Source^;
        TCollectionNotifyEvent<T>(Stored.InternalNotify)(Stored.Self, Source^, cnAdded);
        Inc(Source);
        Inc(Item);
      end;
    end;
    Exit;
  end else
  begin
    Self.GrowTo(Count);
    Count := FCount.Native;
    Inc(Count, ValuesCount);
    goto available;
  end;

out_of_memory:
  OutOfMemoryError;
end;

procedure TList<T>.InsertRange(Index: Integer; const Values: array of T);
label
  available, out_of_range, out_of_memory;
var
  Count, ValuesCount, AIndex: NativeInt;
  Item, Source, Buffer: PItem;
  Stored: TInternalStored;
begin
  ValuesCount := High(Values);
  if (ValuesCount < 0) then Exit;
  Inc(ValuesCount);

  Stored.Self := Self;
  Stored.InternalNotify := TMethod(FInternalNotify);
  Stored.Count := ValuesCount;
  AIndex := Index;

  Count := FCount.Native;
  if (NativeUInt(AIndex) > NativeUInt(Count)) then goto out_of_range;

  Inc(Count, ValuesCount);
  if (NativeUInt(Count) > NativeUInt(High(Integer))) then goto out_of_memory;

  if (Count <= FCapacity.Native) then
  begin
  available:
    FCount.Native := Count;
    Dec(Count, AIndex);
    Dec(Count, ValuesCount);
    Buffer{Item} := @FItems[AIndex];
    if (Count <> 0) then
    begin
      Count := Count * SizeOf(T);
      Source := Buffer{Item} + Stored.Count;
      System.Move(Buffer{Item}^, Source^, Count);
    end;

    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(T) or System.HasWeakRef(T)) then
    {$else}
    if (Assigned(TRAIIHelper<T>.Options.InitProc)) then
    {$endif}
    begin
      if (SizeOf(T) <= 16) then
      begin
        FillChar(Buffer{Item}^, Stored.Count * SizeOf(T), #0);
      end else
      begin
        TRAIIHelper<T>.InitArray(Buffer{Item}, Stored.Count);
      end;
    end else
    if (not Assigned(Stored.InternalNotify.Code)) then
    begin
      System.Move(Values[0], Buffer{Item}^, Stored.Count * SizeOf(T));
      Exit;
    end;

    Source := @Values[0];
    Stored.Item := Source + Stored.Count;
    Item := Buffer{Item};
    if (not Assigned(Stored.InternalNotify.Code)) then
    begin
      repeat
        Item^ := Source^;
        Inc(Source);
        Inc(Item);
      until (Source = Stored.Item);
    end else
    begin
      repeat
        Item^ := Source^;
        TCollectionNotifyEvent<T>(Stored.InternalNotify)(Stored.Self, Source^, cnAdded);
        Inc(Source);
        Inc(Item);
      until (Source = Stored.Item);
    end;
    Exit;
  end else
  begin
    Self.GrowTo(Count);
    AIndex := Index;
    ValuesCount := Stored.Count;
    Count := FCount.Native;
    Inc(Count, ValuesCount);
    goto available;
  end;

out_of_range:
  raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
out_of_memory:
  OutOfMemoryError;
end;

procedure TList<T>.AddRange(const Collection: IEnumerable<T>);
var
  Item: T;
  Index: NativeInt;
begin
  if (not Assigned(FInternalNotify)) then
  begin
    for Item in Collection do
    begin
      Add(Item);
    end;
  end else
  begin
    Index := FCount.Native;
    for Item in Collection do
    begin
      InternalInsert(Index, Item);
      Inc(Index);
    end;
  end;
end;

procedure TList<T>.AddRange(const Collection: TEnumerable<T>);
var
  Item: T;
  Index: NativeInt;
begin
  if (not Assigned(FInternalNotify)) then
  begin
    for Item in Collection do
    begin
      Add(Item);
    end;
  end else
  begin
    Index := FCount.Native;
    for Item in Collection do
    begin
      InternalInsert(Index, Item);
      Inc(Index);
    end;
  end;
end;

procedure TList<T>.InsertRange(Index: Integer; const Collection: IEnumerable<T>);
var
  Item: T;
begin
  if (Index = FCount.Int) and (not Assigned(FInternalNotify)) then
  begin
    AddRange(Collection);
    Exit;
  end;

  for Item in Collection do
  begin
    Insert(Index, Item);
    Inc(Index);
  end;
end;

procedure TList<T>.InsertRange(Index: Integer; const Collection: TEnumerable<T>);
var
  Item: T;
begin
  if (Index = FCount.Int) and (not Assigned(FInternalNotify)) then
  begin
    AddRange(Collection);
    Exit;
  end;

  for Item in Collection do
  begin
    Insert(Index, Item);
    Inc(Index);
  end;
end;

procedure TList<T>.InternalDelete(Index: NativeInt; Action: TCollectionNotification);
label
  available;
var
  Count: NativeInt;
  Item: PItem;
  VType: Integer;
  InternalNotify: TMethod;
begin
  Count := FCount.Native;
  if (NativeUInt(Index) < NativeUInt(Count)) then
  begin
    Dec(Count);
    FCount.Native := Count;
    Dec(Count, Index);
    Item := @FItems[Index];

    if (not Assigned(FInternalNotify)) then
    begin
    available:
      {$ifdef SMARTGENERICS}
      case GetTypeKind(T) of
        {$ifdef AUTOREFCOUNT}
        tkClass,
        {$endif}
        tkWString, tkLString, tkUString, tkInterface, tkDynArray:
        begin
          if (PNativeInt(Item)^ <> 0) then
          case GetTypeKind(T) of
            {$ifdef AUTOREFCOUNT}
            tkClass:
            begin
              TRAIIHelper.RefObjClear(Item);
            end;
            {$endif}
            {$ifdef MSWINDOWS}
            tkWString:
            begin
              TRAIIHelper.WStrClear(Item);
            end;
            {$else}
            tkWString,
            {$endif}
            tkLString, tkUString:
            begin
              TRAIIHelper.ULStrClear(Item);
            end;
            tkInterface:
            begin
              IInterface(PPointer(Item)^)._Release;
            end;
            tkDynArray:
            begin
              TRAIIHelper.DynArrayClear(Item, TypeInfo(T));
            end;
          end;
        end;
        {$ifdef WEAKREF}
        tkMethod:
        begin
          if (PMethod(Item).Data <> nil) then
            TRAIIHelper.WeakMethodClear(@PMethod(Item).Data);
        end;
        {$endif}
        tkVariant:
        begin
          VType := PVarData(Item).VType;
          if (VType and TRAIIHelper.varDeepData <> 0) then
          case VType of
            varBoolean, varUnknown+1..varUInt64: ;
          else
            System.VarClear(PVariant(Item)^);
          end;
        end;
      else
        TRAIIHelper<T>.Clear(Item);
      end;
      {$else}
      TRAIIHelper<T>.Clear(Item);
      {$endif}

      if (Count <> 0) then
        System.Move(Pointer(Item + 1)^, Item^, Count * SizeOf(T));

      Exit;
    end else
    begin
      InternalNotify := TMethod(FInternalNotify);
      TCollectionNotifyEvent<T>(InternalNotify)(Self, Item^, Action);
      goto available;
    end;
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

procedure TList<T>.Delete(Index: Integer);
begin
  InternalDelete(Index, cnRemoved);
end;

procedure TList<T>.DeleteRange(AIndex, ACount: Integer);
label
  available;
var
  Count: NativeInt;
  IndexTop: Integer;
  Item: PItem;
  VType: Integer;
  Stored: TInternalStored;
begin
  Stored.Self := Self;
  Stored.InternalNotify := TMethod(FInternalNotify);

  IndexTop := AIndex + ACount;
  Count := FCount.Native;
  if (AIndex >= 0) and (ACount >= 0) and (IndexTop >= 0) and
    (IndexTop <= Integer(Count)) then
  begin
    if (ACount = 0) then Exit;
    Dec(Count, ACount);
    FCount.Native := Count;
    Dec(Count, AIndex);
    Item := @FItems[AIndex];

    if (not Assigned(Stored.InternalNotify.Code)) then
    begin
    available:
      {$ifdef SMARTGENERICS}
      if (System.IsManagedType(T) or System.HasWeakRef(T)) then
      {$else}
      if (Assigned(TRAIIHelper<T>.Options.ClearProc)) then
      {$endif}
      begin
        Stored.ACount := ACount;
        for ACount := ACount downto 1 do
        begin
          {$ifdef SMARTGENERICS}
          case GetTypeKind(T) of
            {$ifdef AUTOREFCOUNT}
            tkClass,
            {$endif}
            tkWString, tkLString, tkUString, tkInterface, tkDynArray:
            begin
              if (PNativeInt(Item)^ <> 0) then
              case GetTypeKind(T) of
                {$ifdef AUTOREFCOUNT}
                tkClass:
                begin
                  TRAIIHelper.RefObjClear(Item);
                end;
                {$endif}
                {$ifdef MSWINDOWS}
                tkWString:
                begin
                  TRAIIHelper.WStrClear(Item);
                end;
                {$else}
                tkWString,
                {$endif}
                tkLString, tkUString:
                begin
                  TRAIIHelper.ULStrClear(Item);
                end;
                tkInterface:
                begin
                  IInterface(PPointer(Item)^)._Release;
                end;
                tkDynArray:
                begin
                  TRAIIHelper.DynArrayClear(Item, TypeInfo(T));
                end;
              end;
            end;
            {$ifdef WEAKREF}
            tkMethod:
            begin
              if (PMethod(Item).Data <> nil) then
                TRAIIHelper.WeakMethodClear(@PMethod(Item).Data);
            end;
            {$endif}
            tkVariant:
            begin
              VType := PVarData(Item).VType;
              if (VType and TRAIIHelper.varDeepData <> 0) then
              case VType of
                varBoolean, varUnknown+1..varUInt64: ;
              else
                System.VarClear(PVariant(Item)^);
              end;
            end;
          else
            TRAIIHelper<T>.Options.ClearProc(TRAIIHelper<T>.Options, Item);
          end;
          {$else}
          TRAIIHelper<T>.Options.ClearProc(TRAIIHelper<T>.Options, Item);
          {$endif}

          Inc(Item);
        end;
        ACount := Stored.ACount;
        Dec(Item, ACount);
      end;

      if (Count <> 0) then
        System.Move(Pointer(Item + ACount)^, Item^, Count * SizeOf(T));
      Exit;
    end else
    begin
      Stored.Count := Count;
      Stored.ACount := ACount;
      for ACount := ACount downto 1 do
      begin
        TCollectionNotifyEvent<T>(Stored.InternalNotify)(Stored.Self, Item^, cnRemoved);
        Inc(Item);
      end;
      Count := Stored.Count;
      ACount := Stored.ACount;
      Dec(Item, ACount);
      goto available;
    end;
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

function TList<T>.Expand: TList<T>;
begin
  if (FCount.Native = FCapacity.Native) then
    Grow;

  Result := Self;
end;

procedure TList<T>.Exchange(Index1, Index2: Integer);
var
  Count: Cardinal;
  X, Y: Pointer;
begin
  Count := FCount.Int;
  if (Cardinal(Index1) < Count) and (Cardinal(Index2) < Count) then
  begin
    if (Index1 <> Index2) then
    begin
      X := Pointer(FItems);
      Y := Pointer(TCustomList<T>.PItem(X) + Index2);
      X := Pointer(TCustomList<T>.PItem(X) + Index1);

      TArray.Exchange<T>(X, Y);
    end;
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

procedure TList<T>.InternalMove(CurIndex, NewIndex: Integer);
var
  Count: Cardinal;
  X, Y: PItem;
  Temp: T;
begin
  Count := FCount.Int;
  if (Cardinal(CurIndex) < Count) and (Cardinal(NewIndex) < Count) then
  begin
    if (CurIndex <> NewIndex) then
    begin
      X := Pointer(FItems);
      Y := X + NewIndex;
      X := X + CurIndex;

      Temp := X^;
      if (X < Y) then
      begin
        System.Move(Pointer(X + 1)^, X^, NativeUInt(Y) - NativeUInt(X));
      end else
      begin
        System.Move(Y^, Pointer(Y + 1)^, NativeUInt(X) - NativeUInt(Y));
      end;
      Y^ := Temp;
    end;
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

procedure TList<T>.InternalMove40(CurIndex, NewIndex: Integer);
var
  Count: Cardinal;
  X, Y: PItem;

  Temp1: TRAIIHelper.T1;
  Temp2: TRAIIHelper.T2;
  Temp4: TRAIIHelper.T4;
  Temp8: TRAIIHelper.T8;
  Temp40: TRAIIHelper.TTemp40;
begin
  Count := FCount.Int;
  if (Cardinal(CurIndex) < Count) and (Cardinal(NewIndex) < Count) then
  begin
    if (CurIndex <> NewIndex) then
    begin
      X := Pointer(FItems);
      Y := X + NewIndex;
      X := X + CurIndex;

      case SizeOf(T) of
         1: Temp1 := TRAIIHelper.T1(Pointer(X)^);
         2: Temp2 := TRAIIHelper.T2(Pointer(X)^);
         3: Temp40.V3 := TRAIIHelper.T3(Pointer(X)^);
         4: Temp4 := TRAIIHelper.T4(Pointer(X)^);
         5: Temp40.V5 := TRAIIHelper.T5(Pointer(X)^);
         6: Temp40.V6 := TRAIIHelper.T6(Pointer(X)^);
         7: Temp40.V7 := TRAIIHelper.T7(Pointer(X)^);
         8: Temp8 := TRAIIHelper.T8(Pointer(X)^);
         9: Temp40.V9 := TRAIIHelper.T9(Pointer(X)^);
        10: Temp40.V10 := TRAIIHelper.T10(Pointer(X)^);
        11: Temp40.V11 := TRAIIHelper.T11(Pointer(X)^);
        12: Temp40.V12 := TRAIIHelper.T12(Pointer(X)^);
        13: Temp40.V13 := TRAIIHelper.T13(Pointer(X)^);
        14: Temp40.V14 := TRAIIHelper.T14(Pointer(X)^);
        15: Temp40.V15 := TRAIIHelper.T15(Pointer(X)^);
        16: Temp40.V16 := TRAIIHelper.T16(Pointer(X)^);
        17: Temp40.V17 := TRAIIHelper.T17(Pointer(X)^);
        18: Temp40.V18 := TRAIIHelper.T18(Pointer(X)^);
        19: Temp40.V19 := TRAIIHelper.T19(Pointer(X)^);
        20: Temp40.V20 := TRAIIHelper.T20(Pointer(X)^);
        21: Temp40.V21 := TRAIIHelper.T21(Pointer(X)^);
        22: Temp40.V22 := TRAIIHelper.T22(Pointer(X)^);
        23: Temp40.V23 := TRAIIHelper.T23(Pointer(X)^);
        24: Temp40.V24 := TRAIIHelper.T24(Pointer(X)^);
        25: Temp40.V25 := TRAIIHelper.T25(Pointer(X)^);
        26: Temp40.V26 := TRAIIHelper.T26(Pointer(X)^);
        27: Temp40.V27 := TRAIIHelper.T27(Pointer(X)^);
        28: Temp40.V28 := TRAIIHelper.T28(Pointer(X)^);
        29: Temp40.V29 := TRAIIHelper.T29(Pointer(X)^);
        30: Temp40.V30 := TRAIIHelper.T30(Pointer(X)^);
        31: Temp40.V31 := TRAIIHelper.T31(Pointer(X)^);
        32: Temp40.V32 := TRAIIHelper.T32(Pointer(X)^);
        33: Temp40.V33 := TRAIIHelper.T33(Pointer(X)^);
        34: Temp40.V34 := TRAIIHelper.T34(Pointer(X)^);
        35: Temp40.V35 := TRAIIHelper.T35(Pointer(X)^);
        36: Temp40.V36 := TRAIIHelper.T36(Pointer(X)^);
        37: Temp40.V37 := TRAIIHelper.T37(Pointer(X)^);
        38: Temp40.V38 := TRAIIHelper.T38(Pointer(X)^);
        39: Temp40.V39 := TRAIIHelper.T39(Pointer(X)^);
        40: Temp40.V40 := TRAIIHelper.T40(Pointer(X)^);
      end;

      if (X < Y) then
      begin
        System.Move(Pointer(X + 1)^, X^, NativeUInt(Y) - NativeUInt(X));
      end else
      begin
        System.Move(Y^, Pointer(Y + 1)^, NativeUInt(X) - NativeUInt(Y));
      end;

      case SizeOf(T) of
         1: TRAIIHelper.T1(Pointer(Y)^) := Temp1;
         2: TRAIIHelper.T2(Pointer(Y)^) := Temp2;
         3: TRAIIHelper.T3(Pointer(Y)^) := Temp40.V3;
         4: TRAIIHelper.T4(Pointer(Y)^) := Temp4;
         5: TRAIIHelper.T5(Pointer(Y)^) := Temp40.V5;
         6: TRAIIHelper.T6(Pointer(Y)^) := Temp40.V6;
         7: TRAIIHelper.T7(Pointer(Y)^) := Temp40.V7;
         8: TRAIIHelper.T8(Pointer(Y)^) := Temp8;
         9: TRAIIHelper.T9(Pointer(Y)^) := Temp40.V9;
        10: TRAIIHelper.T10(Pointer(Y)^) := Temp40.V10;
        11: TRAIIHelper.T11(Pointer(Y)^) := Temp40.V11;
        12: TRAIIHelper.T12(Pointer(Y)^) := Temp40.V12;
        13: TRAIIHelper.T13(Pointer(Y)^) := Temp40.V13;
        14: TRAIIHelper.T14(Pointer(Y)^) := Temp40.V14;
        15: TRAIIHelper.T15(Pointer(Y)^) := Temp40.V15;
        16: TRAIIHelper.T16(Pointer(Y)^) := Temp40.V16;
        17: TRAIIHelper.T17(Pointer(Y)^) := Temp40.V17;
        18: TRAIIHelper.T18(Pointer(Y)^) := Temp40.V18;
        19: TRAIIHelper.T19(Pointer(Y)^) := Temp40.V19;
        20: TRAIIHelper.T20(Pointer(Y)^) := Temp40.V20;
        21: TRAIIHelper.T21(Pointer(Y)^) := Temp40.V21;
        22: TRAIIHelper.T22(Pointer(Y)^) := Temp40.V22;
        23: TRAIIHelper.T23(Pointer(Y)^) := Temp40.V23;
        24: TRAIIHelper.T24(Pointer(Y)^) := Temp40.V24;
        25: TRAIIHelper.T25(Pointer(Y)^) := Temp40.V25;
        26: TRAIIHelper.T26(Pointer(Y)^) := Temp40.V26;
        27: TRAIIHelper.T27(Pointer(Y)^) := Temp40.V27;
        28: TRAIIHelper.T28(Pointer(Y)^) := Temp40.V28;
        29: TRAIIHelper.T29(Pointer(Y)^) := Temp40.V29;
        30: TRAIIHelper.T30(Pointer(Y)^) := Temp40.V30;
        31: TRAIIHelper.T31(Pointer(Y)^) := Temp40.V31;
        32: TRAIIHelper.T32(Pointer(Y)^) := Temp40.V32;
        33: TRAIIHelper.T33(Pointer(Y)^) := Temp40.V33;
        34: TRAIIHelper.T34(Pointer(Y)^) := Temp40.V34;
        35: TRAIIHelper.T35(Pointer(Y)^) := Temp40.V35;
        36: TRAIIHelper.T36(Pointer(Y)^) := Temp40.V36;
        37: TRAIIHelper.T37(Pointer(Y)^) := Temp40.V37;
        38: TRAIIHelper.T38(Pointer(Y)^) := Temp40.V38;
        39: TRAIIHelper.T39(Pointer(Y)^) := Temp40.V39;
        40: TRAIIHelper.T40(Pointer(Y)^) := Temp40.V40;
      end;
    end;
  end else
  begin
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  end;
end;

procedure TList<T>.Move(CurIndex, NewIndex: Integer);
begin
  if (SizeOf(T) > 40) then
  begin
    InternalMove(CurIndex, NewIndex);
  end else
  begin
    InternalMove40(CurIndex, NewIndex);
  end;
end;

procedure TList<T>.Reverse;
var
  Count: Cardinal;
  X, Y: Pointer;
  Index: NativeInt;
  Temp1: Byte;
  Temp2: Word;
  Temp4: Cardinal;
  TempNative: NativeUInt;
begin
  Count := FCount.Native;
  if (Count > 1) then
  begin
    Dec(Count);
    X := FItems;
    Y := TCustomList<T>.PItem(X) + Count;

    repeat
      if (SizeOf(T) <= 16) then
      begin
        TArray.Exchange<T>(X, Y);
        Inc(NativeUInt(X), SizeOf(T));
        Dec(NativeUInt(Y), SizeOf(T));
      end else
      begin
        Index := 0;
        repeat
          Inc(Index);
          TempNative := PNativeUInt(X)^;
          PNativeUInt(X)^ := PNativeUInt(Y)^;
          PNativeUInt(Y)^ := TempNative;
          Inc(NativeUInt(X), SizeOf(NativeUInt));
          Inc(NativeUInt(Y), SizeOf(NativeUInt));
        until (Index = SizeOf(T) div SizeOf(NativeUInt));

        {$ifdef LARGEINT}
        if (SizeOf(T) and 4 <> 0) then
        begin
          Temp4 := PCardinal(X)^;
          PCardinal(X)^ := PCardinal(Y)^;
          PCardinal(Y)^ := Temp4;
          Inc(NativeUInt(X), SizeOf(Cardinal));
          Inc(NativeUInt(Y), SizeOf(Cardinal));
        end;
        {$endif}

        if (SizeOf(T) and 2 <> 0) then
        begin
          Temp2 := PWord(X)^;
          PWord(X)^ := PWord(Y)^;
          PWord(Y)^ := Temp2;
          Inc(NativeUInt(X), SizeOf(Word));
          Inc(NativeUInt(Y), SizeOf(Word));
        end;

        if (SizeOf(T) and 1 <> 0) then
        begin
          Temp1 := PByte(X)^;
          PByte(X)^ := PByte(Y)^;
          PByte(Y)^ := Temp1;
          Inc(NativeUInt(X), SizeOf(Byte));
          Inc(NativeUInt(Y), SizeOf(Byte));
        end;

        Dec(NativeUInt(Y), 2 * SizeOf(T));
      end;
    until (NativeUInt(X) >= NativeUInt(Y));
  end;
end;

procedure TList<T>.Sort;
var
  Count: NativeInt;
begin
  Count := FCount.Native;
  if (Count > 1) then
  begin
    if (Assigned(FComparer)) then
    begin
      TArray.Sort<T>(FItems[0], Count, FComparer);
    end else
    begin
      TArray.Sort<T>(FItems[0], Count);
    end;
  end;
end;

procedure TList<T>.Sort(Index, Count: Integer);
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  if (Assigned(FComparer)) then
  begin
    TArray.Sort<T>(FItems[Index], Count, FComparer);
  end else
  begin
    TArray.Sort<T>(FItems[Index], Count);
  end;
end;

procedure TList<T>.Sort(const AComparer: IComparer<T>);
var
  Count: NativeInt;
begin
  Count := FCount.Native;
  if (Count > 1) then
  begin
    TArray.Sort<T>(FItems[0], Count, AComparer);
  end;
end;

procedure TList<T>.Sort(const AComparison: TComparison<T>);
var
  Count: NativeInt;
begin
  Count := FCount.Native;
  if (Count > 1) then
  begin
    TArray.Sort<T>(FItems[0], Count, AComparison);
  end;
end;

procedure TList<T>.Sort(Index, Count: Integer; const AComparer: IComparer<T>);
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  TArray.Sort<T>(FItems[Index], Count, AComparer);
end;

procedure TList<T>.Sort(Index, Count: Integer; const AComparison: TComparison<T>);
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  TArray.Sort<T>(FItems[Index], Count, AComparison);
end;

procedure TList<T>.SortDescending;
var
  Count: NativeInt;
begin
  Count := FCount.Native;
  if (Count > 1) then
  begin
    if (Assigned(FComparer)) then
    begin
      TArray.SortDescending<T>(FItems[0], Count, FComparer);
    end else
    begin
      TArray.SortDescending<T>(FItems[0], Count);
    end;
  end;
end;

procedure TList<T>.SortDescending(Index, Count: Integer);
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  if (Assigned(FComparer)) then
  begin
    TArray.SortDescending<T>(FItems[Index], Count, FComparer);
  end else
  begin
    TArray.SortDescending<T>(FItems[Index], Count);
  end;
end;

procedure TList<T>.SortDescending(const AComparer: IComparer<T>);
var
  Count: NativeInt;
begin
  Count := FCount.Native;
  if (Count > 1) then
  begin
    TArray.SortDescending<T>(FItems[0], Count, AComparer);
  end;
end;

procedure TList<T>.SortDescending(const AComparison: TComparison<T>);
var
  Count: NativeInt;
begin
  Count := FCount.Native;
  if (Count > 1) then
  begin
    TArray.SortDescending<T>(FItems[0], Count, AComparison);
  end;
end;

procedure TList<T>.SortDescending(Index, Count: Integer; const AComparer: IComparer<T>);
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  TArray.SortDescending<T>(FItems[Index], Count, AComparer);
end;

procedure TList<T>.SortDescending(Index, Count: Integer; const AComparison: TComparison<T>);
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) or (Count < 0)
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));
  if Count <= 1 then
    Exit;

  TArray.SortDescending<T>(FItems[Index], Count, AComparison);
end;

function TList<T>.BinarySearch(const Item: T; out FoundIndex: Integer): Boolean;
begin
  if (Assigned(FComparer)) then
  begin
    Result := TArray.InternalSearch<T>(FItems, 0, FCount.Int, Item, FoundIndex, Pointer(FComparer));
  end else
  begin
    Result := TArray.InternalSearch<T>(FItems, 0, FCount.Int, Item, FoundIndex);
  end;
end;

function TList<T>.BinarySearch(const Item: T; out FoundIndex: Integer; const AComparer: IComparer<T>): Boolean;
begin
  Result := TArray.InternalSearch<T>(FItems, 0, FCount.Int, Item, FoundIndex, Pointer(AComparer));
end;

function TList<T>.BinarySearch(const Item: T; out FoundIndex: Integer; const AComparison: TComparison<T>): Boolean;
begin
  Result := TArray.InternalSearch<T>(FItems, 0, FCount.Int, Item, FoundIndex, PPointer(@AComparison)^);
end;

function TList<T>.BinarySearch(const Item: T; out FoundIndex: Integer; Index, Count: Integer): Boolean;
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  if (Assigned(FComparer)) then
  begin
    Result := TArray.InternalSearch<T>(FItems, Index, Count, Item, FoundIndex, Pointer(FComparer));
  end else
  begin
    Result := TArray.InternalSearch<T>(FItems, Index, Count, Item, FoundIndex);
  end;
end;

function TList<T>.BinarySearch(const Item: T; out FoundIndex: Integer; const AComparer: IComparer<T>;
  Index, Count: Integer): Boolean;
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearch<T>(FItems, Index, Count, Item, FoundIndex, Pointer(AComparer));
end;

function TList<T>.BinarySearch(const Item: T; out FoundIndex: Integer; Index, Count: Integer;
  const AComparison: TComparison<T>): Boolean;
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearch<T>(FItems, Index, Count, Item, FoundIndex, PPointer(@AComparison)^);
end;

function TList<T>.BinarySearchDescending(const Item: T; out FoundIndex: Integer): Boolean;
begin
  if (Assigned(FComparer)) then
  begin
    Result := TArray.InternalSearchDescending<T>(FItems, 0, FCount.Int, Item, FoundIndex, Pointer(FComparer));
  end else
  begin
    Result := TArray.InternalSearchDescending<T>(FItems, 0, FCount.Int, Item, FoundIndex);
  end;
end;

function TList<T>.BinarySearchDescending(const Item: T; out FoundIndex: Integer; const AComparer: IComparer<T>): Boolean;
begin
  Result := TArray.InternalSearchDescending<T>(FItems, 0, FCount.Int, Item, FoundIndex, Pointer(AComparer));
end;

function TList<T>.BinarySearchDescending(const Item: T; out FoundIndex: Integer; const AComparison: TComparison<T>): Boolean;
begin
  Result := TArray.InternalSearchDescending<T>(FItems, 0, FCount.Int, Item, FoundIndex, PPointer(@AComparison)^);
end;

function TList<T>.BinarySearchDescending(const Item: T; out FoundIndex: Integer; Index, Count: Integer): Boolean;
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  if (Assigned(FComparer)) then
  begin
    Result := TArray.InternalSearchDescending<T>(FItems, Index, Count, Item, FoundIndex, Pointer(FComparer));
  end else
  begin
    Result := TArray.InternalSearchDescending<T>(FItems, Index, Count, Item, FoundIndex);
  end;
end;

function TList<T>.BinarySearchDescending(const Item: T; out FoundIndex: Integer; const AComparer: IComparer<T>;
  Index, Count: Integer): Boolean;
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearchDescending<T>(FItems, Index, Count, Item, FoundIndex, Pointer(AComparer));
end;

function TList<T>.BinarySearchDescending(const Item: T; out FoundIndex: Integer; Index, Count: Integer;
  const AComparison: TComparison<T>): Boolean;
begin
  if (Index < 0) or ((Index >= FCount.Int) and (Count > 0))
    or (Index + Count - 1 >= FCount.Int) {or (Count < 0)}
    or (Index + Count < 0) then
    raise EArgumentOutOfRangeException.CreateRes(Pointer(@SArgumentOutOfRange));

  Result := TArray.InternalSearchDescending<T>(FItems, Index, Count, Item, FoundIndex, PPointer(@AComparison)^);
end;

function TList<T>.InternalIndexOf(const Value: T): NativeInt;
{$ifdef SMARTGENERICS}
label
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5{$ifdef SMALLINT}, cmp6, cmp7, cmp8, cmp9, cmp10{$endif};
{$endif}
type
  TData16 = TRAIIHelper.TData16;
  PData16 = ^TData16;
var
  R: NativeInt;
  Item, TopItem: PItem;
  {$ifdef SMARTGENERICS}
  Count: NativeUInt;
  Left, Right: PByte;
  Offset: NativeUInt;
  {$endif}
  Stored: TInternalStored;
begin
  if (not Assigned(FComparer)) then
  begin
    Item := Pointer(FItems);
    Stored.Item := Item;
    Dec(Item);
    TopItem := Item + FCount.Native;

    repeat
      if (Item = TopItem) then Break;
      Inc(Item);
      {$ifdef SMARTGENERICS}
      if (GetTypeKind(T) = tkVariant) then
      begin
        if (not InterfaceDefaults.Equals_Var(nil, PVarData(@Value), PVarData(Item))) then Continue;
      end else
      if (GetTypeKind(T) = tkClass) then
      begin
        Left := PPointer(@Value)^;
        Right := PPointer(Item)^;
        if (Assigned(Left)) then
        begin
          if (PPointer(Left^)[vmtEquals div SizeOf(Pointer)] = @TObject.Equals) then
          begin
            if (Left <> Right) then Continue;
          end else
          begin
            if (not TObject(PNativeUInt(@Value)^).Equals(TObject(PNativeUInt(Item)^))) then Continue;
          end;
        end else
        begin
          if (Right <> nil) then Continue;
        end;
      end else
      if (GetTypeKind(T) = tkFloat) then
      begin
        case SizeOf(T) of
          4:
          begin
            if (PSingle(@Value)^ <> PSingle(Item)^) then Continue;
          end;
          10:
          begin
            if (PExtended(@Value)^ <> PExtended(Item)^) then Continue;
          end;
        else
        {$ifdef LARGEINT}
          if (PInt64(@Value)^ <> PInt64(Item)^) then
        {$else .SMALLINT}
          if ((PPoint(@Value).X - PPoint(Item).X) or (PPoint(@Value).Y - PPoint(Item).Y) <> 0) then
        {$endif}
          begin
            if (InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Size < 0) then Continue;
            if (PDouble(@Value)^ <> PDouble(Item)^) then Continue;
          end;
        end;
      end else
      if (not (GetTypeKind(T) in [tkDynArray, tkString, tkLString, tkWString, tkUString])) and
        (SizeOf(T) <= 16) then
      begin
        // small binary
        if (SizeOf(T) <> 0) then
        with PData16(@Value)^ do
        begin
          if (SizeOf(T) >= SizeOf(Integer)) then
          begin
            if (SizeOf(T) >= SizeOf(Int64)) then
            begin
              {$ifdef LARGEINT}
              if (Int64s[0] <> PData16(Item).Int64s[0]) then Continue;
              {$else}
              if (Integers[0] <> PData16(Item).Integers[0]) then Continue;
              if (Integers[1] <> PData16(Item).Integers[1]) then Continue;
              {$endif}

              if (SizeOf(T) = 16) then
              begin
                {$ifdef LARGEINT}
                if (Int64s[1] <> PData16(Item).Int64s[1]) then Continue;
                {$else}
                if (Integers[2] <> PData16(Item).Integers[2]) then Continue;
                if (Integers[3] <> PData16(Item).Integers[3]) then Continue;
                {$endif}
              end else
              if (SizeOf(T) >= 12) then
              begin
                if (Integers[2] <> PData16(Item).Integers[2]) then Continue;
              end;
            end else
            begin
              if (Integers[0] <> PData16(Item).Integers[0]) then Continue;
            end;
          end;

          if (SizeOf(T) and 2 <> 0) then
          begin
            if (Words[(SizeOf(T) and -4) shr 1] <> PData16(Item).Words[(SizeOf(T) and -4) shr 1]) then Continue;
          end;
          if (SizeOf(T) and 1 <> 0) then
          begin
            if (Bytes[SizeOf(T) and -2] <> PData16(Item).Bytes[SizeOf(T) and -2]) then Continue;
          end;
        end;
      end else
      begin
        if (GetTypeKind(T) in [tkDynArray, tkString, tkLString, tkWString, tkUString]) then
        begin
          // dynamic size
          if (GetTypeKind(T) = tkString) then
          begin
            Left := Pointer(@Value);
            Right := Pointer(Item);
            if (PItem(Left) = {Right}Item) then goto cmp0;
            Count := Left^;
            if (Count <> Right^) then Continue;
            if (Count = 0) then goto cmp0;
            // compare last bytes
            if (Left[Count] <> Right[Count]) then Continue;
          end else
          // if (GetTypeKind(T) in [tkDynArray, tkLString, tkWString, tkUString]) then
          begin
            Left := PPointer(@Value)^;
            Right := PPointer(Item)^;
            if (Left = Right) then goto cmp0;
            if (Left = nil) then
            begin
              {$ifdef MSWINDOWS}
              if (GetTypeKind(T) = tkWString) then
              begin
                Dec(Right, SizeOf(Cardinal));
                if (PCardinal(Right)^ = 0) then goto cmp0;
              end;
              {$endif}
              Continue;
            end;
            if (Right = nil) then
            begin
              {$ifdef MSWINDOWS}
              if (GetTypeKind(T) = tkWString) then
              begin
                Dec(Left, SizeOf(Cardinal));
                if (PCardinal(Left)^ = 0) then goto cmp0;
              end;
              {$endif}
              Continue;
            end;

            if (GetTypeKind(T) = tkDynArray) then
            begin
              Dec(Left, SizeOf(NativeUInt));
              Dec(Right, SizeOf(NativeUInt));
              Count := PNativeUInt(Left)^;
              if (Count <> PNativeUInt(Right)^) then Continue;
              NativeInt(Count) := NativeInt(Count) * InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Size;
              Inc(Left, SizeOf(NativeUInt));
              Inc(Right, SizeOf(NativeUInt));
            end else
            // if (GetTypeKind(T) in [tkLString, tkWString, tkUString]) then
            begin
              Dec(Left, SizeOf(Cardinal));
              Dec(Right, SizeOf(Cardinal));
              Count := PCardinal(Left)^;
              if (Cardinal(Count) <> PCardinal(Right)^) then Continue;
              Inc(Left, SizeOf(Cardinal));
              Inc(Right, SizeOf(Cardinal));
            end;
          end;

          // compare last (after cardinal) words
          if (GetTypeKind(T) in [tkDynArray, tkString, tkLString]) then
          begin
            if (GetTypeKind(T) in [tkString, tkLString]) {ByteStrings + 2} then
            begin
              Inc(Count);
            end;
            if (Count and 2 <> 0) then
            begin
              Offset := Count and -4;
              Inc(Left, Offset);
              Inc(Right, Offset);
              if (PWord(Left)^ <> PWord(Right)^) then Continue;
              Offset := Count;
              Offset := Offset and -4;
              Dec(Left, Offset);
              Dec(Right, Offset);
            end;
          end else
          // modify Count to have only cardinals to compare
          // if (GetTypeKind(T) in [tkWString, tkUString]) {UnicodeStrings + 2} then
          begin
            {$ifdef MSWINDOWS}
            if (GetTypeKind(T) = tkWString) then
            begin
              if (Count = 0) then goto cmp0;
            end else
            {$endif}
            begin
              Inc(Count, Count);
            end;
            Inc(Count, 2);
          end;

          {$ifdef LARGEINT}
          if (Count and 4 <> 0) then
          begin
            Offset := Count and -8;
            Inc(Left, Offset);
            Inc(Right, Offset);
            if (PCardinal(Left)^ <> PCardinal(Right)^) then Continue;
            Dec(Left, Offset);
            Dec(Right, Offset);
          end;
          {$endif}
        end else
        begin
          // non-dynamic (constant) size binary > 16
          if (SizeOf(T) and {$ifdef LARGEINT}7{$else}3{$endif} <> 0) then
          with PData16(@Value)^ do
          begin
            {$ifdef LARGEINT}
            if (SizeOf(T) and 4 <> 0) then
            begin
              if (Integers[(SizeOf(T) and -8) shr 2] <> PData16(Item).Integers[(SizeOf(T) and -8) shr 2]) then Continue;
            end;
            {$endif}
            if (SizeOf(T) and 2 <> 0) then
            begin
              if (Words[(SizeOf(T) and -4) shr 1] <> PData16(Item).Words[(SizeOf(T) and -4) shr 1]) then Continue;
            end;
            if (SizeOf(T) and 1 <> 0) then
            begin
              if (Bytes[SizeOf(T) and -2] <> PData16(Item).Bytes[SizeOf(T) and -2]) then Continue;
            end;
          end;
          Left := Pointer(@Value);
          Right := Pointer(Item);
          Count := SizeOf(T);
        end;

        // natives (40 bytes static) compare
        Count := Count shr {$ifdef LARGEINT}3{$else}2{$endif};
        case Count of
        {$ifdef SMALLINT}
         10: goto cmp10;
          9: goto cmp9;
          8: goto cmp8;
          7: goto cmp7;
          6: goto cmp6;
        {$endif}
          5: goto cmp5;
          4: goto cmp4;
          3: goto cmp3;
          2: goto cmp2;
          1: goto cmp1;
          0: goto cmp0;
        else
          repeat
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Dec(Count);
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          until (Count = {$ifdef LARGEINT}5{$else}10{$endif});

          {$ifdef SMALLINT}
          cmp10:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp9:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp8:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp7:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp6:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          {$endif}
          cmp5:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp4:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp3:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp2:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp1:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
          cmp0:
        end;
      end;
      {$else}
        if (not TEquals(InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Equals)(
          @InterfaceDefaults.TDefaultEqualityComparer<T>.Instance, Item^, Value)) then Continue;
      {$endif}

      R := NativeInt(Item) - NativeInt(Stored.Item);
      case SizeOf(T) of
      0, 1: Exit(R);
         2: Exit(R shr 1);
         4: Exit(R shr 2);
         8: Exit(R shr 3);
        16: Exit(R shr 4);
        32: Exit(R shr 5);
        64: Exit(R shr 6);
       128: Exit(R shr 7);
       256: Exit(R shr 8);
      else
        Exit(Round(R * (1 / SizeOf(T))));
      end;
    until (False);

    Exit(-1);
  end else
  begin
    Exit(InternalIndexOf(Value, FComparer));
  end;
end;

function TList<T>.InternalIndexOf(const Value: T; const Comparer: IComparer<T>): NativeInt;
var
  Count: NativeInt;
  Item: PItem;
  Compare: TMethod;
begin
  Count := FCount.Native;
  Item := Pointer(FItems);
  Compare.Data := Pointer(Comparer);
  Compare.Code := PPointer(PNativeUInt(Comparer)^ + 3 * SizeOf(Pointer))^;
  for Result := 0 to Count - 1 do
  begin
    if (TCompare(Compare.Code)(Compare.Data, Item^, Value) = 0) then Exit;
    Inc(Item);
  end;

  Result := -1;
end;

function TList<T>.InternalIndexOfRev(const Value: T): NativeInt;
{$ifdef SMARTGENERICS}
label
  cmp0, cmp1, cmp2, cmp3, cmp4, cmp5{$ifdef SMALLINT}, cmp6, cmp7, cmp8, cmp9, cmp10{$endif};
{$endif}
type
  TData16 = TRAIIHelper.TData16;
  PData16 = ^TData16;
var
  R: NativeInt;
  Item, LowItem: PItem;
  {$ifdef SMARTGENERICS}
  Count: NativeUInt;
  Left, Right: PByte;
  Offset: NativeUInt;
  {$endif}
  Stored: TInternalStored;
begin
  if (not Assigned(FComparer)) then
  begin
    LowItem := Pointer(FItems);
    Stored.Item := LowItem;
    Item := LowItem + FCount.Native;

    repeat
      if (Item = LowItem) then Break;
      Dec(Item);
      {$ifdef SMARTGENERICS}
      if (GetTypeKind(T) = tkVariant) then
      begin
        if (not InterfaceDefaults.Equals_Var(nil, PVarData(@Value), PVarData(Item))) then Continue;
      end else
      if (GetTypeKind(T) = tkClass) then
      begin
        Left := PPointer(@Value)^;
        Right := PPointer(Item)^;
        if (Assigned(Left)) then
        begin
          if (PPointer(Left^)[vmtEquals div SizeOf(Pointer)] = @TObject.Equals) then
          begin
            if (Left <> Right) then Continue;
          end else
          begin
            if (not TObject(PNativeUInt(@Value)^).Equals(TObject(PNativeUInt(Item)^))) then Continue;
          end;
        end else
        begin
          if (Right <> nil) then Continue;
        end;
      end else
      if (GetTypeKind(T) = tkFloat) then
      begin
        case SizeOf(T) of
          4:
          begin
            if (PSingle(@Value)^ <> PSingle(Item)^) then Continue;
          end;
          10:
          begin
            if (PExtended(@Value)^ <> PExtended(Item)^) then Continue;
          end;
        else
        {$ifdef LARGEINT}
          if (PInt64(@Value)^ <> PInt64(Item)^) then
        {$else .SMALLINT}
          if ((PPoint(@Value).X - PPoint(Item).X) or (PPoint(@Value).Y - PPoint(Item).Y) <> 0) then
        {$endif}
          begin
            if (InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Size < 0) then Continue;
            if (PDouble(@Value)^ <> PDouble(Item)^) then Continue;
          end;
        end;
      end else
      if (not (GetTypeKind(T) in [tkDynArray, tkString, tkLString, tkWString, tkUString])) and
        (SizeOf(T) <= 16) then
      begin
        // small binary
        if (SizeOf(T) <> 0) then
        with PData16(@Value)^ do
        begin
          if (SizeOf(T) >= SizeOf(Integer)) then
          begin
            if (SizeOf(T) >= SizeOf(Int64)) then
            begin
              {$ifdef LARGEINT}
              if (Int64s[0] <> PData16(Item).Int64s[0]) then Continue;
              {$else}
              if (Integers[0] <> PData16(Item).Integers[0]) then Continue;
              if (Integers[1] <> PData16(Item).Integers[1]) then Continue;
              {$endif}

              if (SizeOf(T) = 16) then
              begin
                {$ifdef LARGEINT}
                if (Int64s[1] <> PData16(Item).Int64s[1]) then Continue;
                {$else}
                if (Integers[2] <> PData16(Item).Integers[2]) then Continue;
                if (Integers[3] <> PData16(Item).Integers[3]) then Continue;
                {$endif}
              end else
              if (SizeOf(T) >= 12) then
              begin
                if (Integers[2] <> PData16(Item).Integers[2]) then Continue;
              end;
            end else
            begin
              if (Integers[0] <> PData16(Item).Integers[0]) then Continue;
            end;
          end;

          if (SizeOf(T) and 2 <> 0) then
          begin
            if (Words[(SizeOf(T) and -4) shr 1] <> PData16(Item).Words[(SizeOf(T) and -4) shr 1]) then Continue;
          end;
          if (SizeOf(T) and 1 <> 0) then
          begin
            if (Bytes[SizeOf(T) and -2] <> PData16(Item).Bytes[SizeOf(T) and -2]) then Continue;
          end;
        end;
      end else
      begin
        if (GetTypeKind(T) in [tkDynArray, tkString, tkLString, tkWString, tkUString]) then
        begin
          // dynamic size
          if (GetTypeKind(T) = tkString) then
          begin
            Left := Pointer(@Value);
            Right := Pointer(Item);
            if (PItem(Left) = {Right}Item) then goto cmp0;
            Count := Left^;
            if (Count <> Right^) then Continue;
            if (Count = 0) then goto cmp0;
            // compare last bytes
            if (Left[Count] <> Right[Count]) then Continue;
          end else
          // if (GetTypeKind(T) in [tkDynArray, tkLString, tkWString, tkUString]) then
          begin
            Left := PPointer(@Value)^;
            Right := PPointer(Item)^;
            if (Left = Right) then goto cmp0;
            if (Left = nil) then
            begin
              {$ifdef MSWINDOWS}
              if (GetTypeKind(T) = tkWString) then
              begin
                Dec(Right, SizeOf(Cardinal));
                if (PCardinal(Right)^ = 0) then goto cmp0;
              end;
              {$endif}
              Continue;
            end;
            if (Right = nil) then
            begin
              {$ifdef MSWINDOWS}
              if (GetTypeKind(T) = tkWString) then
              begin
                Dec(Left, SizeOf(Cardinal));
                if (PCardinal(Left)^ = 0) then goto cmp0;
              end;
              {$endif}
              Continue;
            end;

            if (GetTypeKind(T) = tkDynArray) then
            begin
              Dec(Left, SizeOf(NativeUInt));
              Dec(Right, SizeOf(NativeUInt));
              Count := PNativeUInt(Left)^;
              if (Count <> PNativeUInt(Right)^) then Continue;
              NativeInt(Count) := NativeInt(Count) * InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Size;
              Inc(Left, SizeOf(NativeUInt));
              Inc(Right, SizeOf(NativeUInt));
            end else
            // if (GetTypeKind(T) in [tkLString, tkWString, tkUString]) then
            begin
              Dec(Left, SizeOf(Cardinal));
              Dec(Right, SizeOf(Cardinal));
              Count := PCardinal(Left)^;
              if (Cardinal(Count) <> PCardinal(Right)^) then Continue;
              Inc(Left, SizeOf(Cardinal));
              Inc(Right, SizeOf(Cardinal));
            end;
          end;

          // compare last (after cardinal) words
          if (GetTypeKind(T) in [tkDynArray, tkString, tkLString]) then
          begin
            if (GetTypeKind(T) in [tkString, tkLString]) {ByteStrings + 2} then
            begin
              Inc(Count);
            end;
            if (Count and 2 <> 0) then
            begin
              Offset := Count and -4;
              Inc(Left, Offset);
              Inc(Right, Offset);
              if (PWord(Left)^ <> PWord(Right)^) then Continue;
              Offset := Count;
              Offset := Offset and -4;
              Dec(Left, Offset);
              Dec(Right, Offset);
            end;
          end else
          // modify Count to have only cardinals to compare
          // if (GetTypeKind(T) in [tkWString, tkUString]) {UnicodeStrings + 2} then
          begin
            {$ifdef MSWINDOWS}
            if (GetTypeKind(T) = tkWString) then
            begin
              if (Count = 0) then goto cmp0;
            end else
            {$endif}
            begin
              Inc(Count, Count);
            end;
            Inc(Count, 2);
          end;

          {$ifdef LARGEINT}
          if (Count and 4 <> 0) then
          begin
            Offset := Count and -8;
            Inc(Left, Offset);
            Inc(Right, Offset);
            if (PCardinal(Left)^ <> PCardinal(Right)^) then Continue;
            Dec(Left, Offset);
            Dec(Right, Offset);
          end;
          {$endif}
        end else
        begin
          // non-dynamic (constant) size binary > 16
          if (SizeOf(T) and {$ifdef LARGEINT}7{$else}3{$endif} <> 0) then
          with PData16(@Value)^ do
          begin
            {$ifdef LARGEINT}
            if (SizeOf(T) and 4 <> 0) then
            begin
              if (Integers[(SizeOf(T) and -8) shr 2] <> PData16(Item).Integers[(SizeOf(T) and -8) shr 2]) then Continue;
            end;
            {$endif}
            if (SizeOf(T) and 2 <> 0) then
            begin
              if (Words[(SizeOf(T) and -4) shr 1] <> PData16(Item).Words[(SizeOf(T) and -4) shr 1]) then Continue;
            end;
            if (SizeOf(T) and 1 <> 0) then
            begin
              if (Bytes[SizeOf(T) and -2] <> PData16(Item).Bytes[SizeOf(T) and -2]) then Continue;
            end;
          end;
          Left := Pointer(@Value);
          Right := Pointer(Item);
          Count := SizeOf(T);
        end;

        // natives (40 bytes static) compare
        Count := Count shr {$ifdef LARGEINT}3{$else}2{$endif};
        case Count of
        {$ifdef SMALLINT}
         10: goto cmp10;
          9: goto cmp9;
          8: goto cmp8;
          7: goto cmp7;
          6: goto cmp6;
        {$endif}
          5: goto cmp5;
          4: goto cmp4;
          3: goto cmp3;
          2: goto cmp2;
          1: goto cmp1;
          0: goto cmp0;
        else
          repeat
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Dec(Count);
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          until (Count = {$ifdef LARGEINT}5{$else}10{$endif});

          {$ifdef SMALLINT}
          cmp10:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp9:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp8:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp7:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp6:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          {$endif}
          cmp5:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp4:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp3:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp2:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
            Inc(Left, SizeOf(NativeUInt));
            Inc(Right, SizeOf(NativeUInt));
          cmp1:
            if (PNativeUInt(Left)^ <> PNativeUInt(Right)^) then Continue;
          cmp0:
        end;
      end;
      {$else}
        if (not TEquals(InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Equals)(
          @InterfaceDefaults.TDefaultEqualityComparer<T>.Instance, Item^, Value)) then Continue;
      {$endif}

      R := NativeInt(Item) - NativeInt(Stored.Item);
      case SizeOf(T) of
      0, 1: Exit(R);
         2: Exit(R shr 1);
         4: Exit(R shr 2);
         8: Exit(R shr 3);
        16: Exit(R shr 4);
        32: Exit(R shr 5);
        64: Exit(R shr 6);
       128: Exit(R shr 7);
       256: Exit(R shr 8);
      else
        Exit(Round(R * (1 / SizeOf(T))));
      end;
    until (False);

    Exit(-1);
  end else
  begin
    Exit(InternalIndexOfRev(Value, FComparer));
  end;
end;

function TList<T>.InternalIndexOfRev(const Value: T; const Comparer: IComparer<T>): NativeInt;
var
  Count: NativeInt;
  Item: PItem;
  Compare: TMethod;
begin
  Count := FCount.Native;
  Dec(Count);
  Item := @FItems[Count];
  Compare.Data := Pointer(Comparer);
  Compare.Code := PPointer(PNativeUInt(Comparer)^ + 3 * SizeOf(Pointer))^;
  for Result := Count downto 0 do
  begin
    if (TCompare(Compare.Code)(Compare.Data, Item^, Value) = 0) then Exit;
    Dec(Item);
  end;

  Result := -1;
end;

function TList<T>.Contains(const Value: T): Boolean;
begin
  Result := (InternalIndexOf(Value) >= 0);
end;

function TList<T>.IndexOf(const Value: T): Integer;
begin
  Result := InternalIndexOf(Value);
end;

function TList<T>.IndexOfItem(const Value: T; Direction: TDirection): Integer;
begin
  if (Direction = FromBeginning) then
  begin
    Result := InternalIndexOf(Value);
  end else
  begin
    Result := InternalIndexOfRev(Value);
  end;
end;

function TList<T>.LastIndexOf(const Value: T): Integer;
begin
  Result := InternalIndexOfRev(Value);
end;

function TList<T>.Remove(const Value: T): Integer;
var
  Index: NativeInt;
begin
  Index := IndexOf(Value);
  if (Index >= 0) then
    InternalDelete(Index, cnRemoved);

  Result := Index;
end;

function TList<T>.RemoveItem(const Value: T; Direction: TDirection): Integer;
var
  Index: NativeInt;
begin
  Index := IndexOfItem(Value, Direction);
  if (Index >= 0) then
    InternalDelete(Index, cnRemoved);

  Result := Index;
end;

function TList<T>.Extract(const Value: T): T;
var
  Index: NativeInt;
begin
  Index := IndexOf(Value);
  if (Index < 0) then
  begin
    Result := Default(T);
  end else
  begin
    Result := FItems[Index];
    InternalDelete(Index, cnExtracted);
  end;
end;

function TList<T>.ExtractItem(const Value: T; Direction: TDirection): T;
var
  Index: NativeInt;
begin
  Index := IndexOfItem(Value, Direction);
  if (Index < 0) then
  begin
    Result := Default(T);
  end else
  begin
    Result := FItems[Index];
    InternalDelete(Index, cnExtracted);
  end;
end;

class function TList<T>.InternalIsEmpty(const Inst: TComparerInst; const Left, Right: T): Boolean;
begin
  Result := (TCompare(Inst.Method.Code)(Inst.Method.Data, Left, Right) = 0);
end;

{$ifdef SMARTGENERICS}
procedure TList<T>.InternalPackDifficults;
label
  next_item;
var
  R: NativeInt;
  Item, TopItem, DestItem: PItem;
  VarData: PVarData;
  {$ifNdef CPUX86}
  VSingle, VSingleNull: Single;
  VDouble, VDoubleNull: Double;
  VExtended, VExtendedNull: Extended;
  {$endif}
begin
  Item := Pointer(FItems);
  Dec(Item);
  TopItem := Item + FCount.Native;

  {$ifNdef CPUX86}
  case SizeOf(T) of
    4:
    begin
      VSingleNull := 0;
    end;
    10:
    begin
      VExtendedNull := 0;
    end;
  else
    VDoubleNull := 0;
  end;
  {$endif}

  repeat
    if (Item = TopItem) then Exit;
    Inc(Item);

    if (GetTypeKind(T) = tkVariant) then
    begin
      VarData := Pointer(Item);
      while (VarData.VType = varByRef or varVariant) do
        VarData := PVarData(VarData.VPointer);

      if (VarData.VType > varNull) then Continue;
    end else
    if (GetTypeKind(T) = tkFloat) then
    begin
      case SizeOf(T) of
        4:
        begin
          if (PSingle(Item)^ <> {$ifdef CPUX86}0{$else}VSingleNull{$endif}) then Continue;
        end;
        10:
        begin
          if (PExtended(Item)^ <> {$ifdef CPUX86}0{$else}VExtendedNull{$endif}) then Continue;
        end;
      else
        if (PDouble(Item)^ <> {$ifdef CPUX86}0{$else}VDoubleNull{$endif}) then Continue;
      end;
    end else
    //if (GetTypeKind(T) = tkString) then
    begin
      if (PByte(Item)^ <> 0) then Continue;
    end;
    Break;
  until (False);

  DestItem := Item;
  goto next_item;
  repeat
    if (GetTypeKind(T) = tkVariant) then
    begin
      PRect(DestItem)^ := PRect(Item)^;
    end else
    if (GetTypeKind(T) = tkFloat) then
    begin
      {$ifdef CPUX86}
        DestItem^ := Item^;
      {$else !CPUX86}
      case SizeOf(T) of
         4: PSingle(DestItem)^ := VSingle;
        10: PExtended(DestItem)^ := VExtended;
      else
        PDouble(DestItem)^ := VDouble;
      end;
      {$endif}
    end else
    begin
      DestItem^ := Item^;
    end;

    Inc(DestItem);
  next_item:
    if (Item = TopItem) then Break;
    Inc(Item);

    if (GetTypeKind(T) = tkVariant) then
    begin
      VarData := Pointer(Item);
      while (VarData.VType = varByRef or varVariant) do
        VarData := PVarData(VarData.VPointer);

      if (VarData.VType > varNull) then Continue;
    end else
    if (GetTypeKind(T) = tkFloat) then
    begin
      case SizeOf(T) of
        4:
        begin
          {$ifdef CPUX86}
            if (PSingle(Item)^ <> 0) then Continue;
          {$else}
            VSingle := PSingle(Item)^;
            if (VSingle <> VSingleNull) then Continue;
          {$endif}
        end;
        10:
        begin
          {$ifdef CPUX86}
            if (PExtended(Item)^ <> 0) then Continue;
          {$else}
            VExtended := PExtended(Item)^;
            if (VExtended <> VExtendedNull) then Continue;
          {$endif}
        end;
      else
        {$ifdef CPUX86}
          if (PDouble(Item)^ <> 0) then Continue;
        {$else}
          VDouble := PDouble(Item)^;
          if (VDouble <> VDoubleNull) then Continue;
        {$endif}
      end;
    end else
    //if (GetTypeKind(T) = tkString) then
    begin
      if (PByte(Item)^ <> 0) then Continue;
    end;

    goto next_item;
  until (False);

  R := NativeInt(DestItem) - NativeInt(FItems);
  case SizeOf(T) of
  0, 1: FCount.Native := R;
     2: FCount.Native := R shr 1;
     4: FCount.Native := R shr 2;
     8: FCount.Native := R shr 3;
    16: FCount.Native := R shr 4;
    32: FCount.Native := R shr 5;
    64: FCount.Native := R shr 6;
   128: FCount.Native := R shr 7;
   256: FCount.Native := R shr 8;
  else
    FCount.Native := Round(R * (1 / SizeOf(T)));
  end;
end;

procedure TList<T>.Pack;
label
  next_find, next_item, comparer_recall;
var
  R, Index: NativeInt;
  Item, TopItem, DestItem: ^TRAIIHelper.TData16;
  VByte: Byte;
  VWord: Word;
  VInteger: Integer;
  VNative, Flags: NativeUInt;
  ComparerInst: TComparerInst;
begin
  if (Assigned(FComparer)) then goto comparer_recall;

  if (GetTypeKind(T) = tkVariant) or (GetTypeKind(T) = tkString) or
    (GetTypeKind(T) = tkFloat) and
    (
      (SizeOf(T) <> SizeOf(Double)) or (InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Size >= 0)
    ) then
  begin
    Self.InternalPackDifficults;
    Exit;
  end;

  Item := Pointer(FItems);
  Dec(NativeUInt(Item), SizeOf(T));
  TopItem := Pointer(TCustomList<T>.PItem(Item) + FCount.Native);

  // find first empty
  repeat
  next_find:
    if (Item = TopItem) then Exit;
    Inc(NativeUInt(Item), SizeOf(T));

    if (SizeOf(T) <= 16) then
    begin
      {$ifdef SMALLINT}
        if (SizeOf(T) >= SizeOf(Int64) * 1) then
        begin
          if (Item.Integers[0] or Item.Integers[1] <> 0) then Continue;
        end;
        if (SizeOf(T)  = SizeOf(Int64) * 2) then
        begin
          if (Item.Integers[2] or Item.Integers[3] <> 0) then Continue;
        end;
      {$else .LARGEINT}
        if (SizeOf(T) >= SizeOf(Int64) * 1) then
        begin
          if (Item.Int64s[0] <> 0) then Continue;
        end;
        if (SizeOf(T)  = SizeOf(Int64) * 2) then
        begin
          if (Item.Int64s[1] <> 0) then Continue;
        end;
      {$endif}
      case SizeOf(T) of
         4..7: if (Item.Integers[0] <> 0) then Continue;
       12..15: if (Item.Integers[2] <> 0) then Continue;
      end;
      case SizeOf(T) of
         2,3: if (Item.Words[0] <> 0) then Continue;
         6,7: if (Item.Words[2] <> 0) then Continue;
       10,11: if (Item.Words[4] <> 0) then Continue;
       14,15: if (Item.Words[6] <> 0) then Continue;
      end;
      case SizeOf(T) of
         1: if (Item.Bytes[ 1-1] <> 0) then Continue;
         3: if (Item.Bytes[ 3-1] <> 0) then Continue;
         5: if (Item.Bytes[ 5-1] <> 0) then Continue;
         7: if (Item.Bytes[ 7-1] <> 0) then Continue;
         9: if (Item.Bytes[ 9-1] <> 0) then Continue;
        11: if (Item.Bytes[11-1] <> 0) then Continue;
        13: if (Item.Bytes[13-1] <> 0) then Continue;
        15: if (Item.Bytes[15-1] <> 0) then Continue;
      end;
    end else
    begin
      DestItem := Item;
      Index := 0;
      repeat
        Inc(Index);
        if (PNativeUInt(DestItem)^ <> 0) then goto next_find;
        Inc(NativeUInt(DestItem), SizeOf(NativeUInt));
      until (Index = SizeOf(T) div SizeOf(NativeUInt));

      if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
      begin
        Dec(NativeUInt(DestItem), SizeOf(T) - (SizeOf(NativeUInt) - 1));
        if (PNativeUInt(DestItem)^ <> 0) then goto next_find;
      end;
    end;
    Break;
  until (False);

  // compare empty and move
  DestItem := Item;
  goto next_item;
  repeat
    // DestItem^ := Item^;
    case SizeOf(T) of
      1: DestItem.Bytes[0] := VByte;
      2: DestItem.Words[0] := VWord;
      4: DestItem.Integers[0] := VInteger;
      8:
      begin
        {$ifdef LARGEINT}
          DestItem.Int64s[0] := VNative;
        {$else .SMALLINT}
          DestItem.Integers[0] := VInteger;
          DestItem.Integers[1] := VNative;
        {$endif}
      end;
    end;
    Inc(NativeUInt(DestItem), SizeOf(T));
  next_item:
    if (Item = TopItem) then Break;
    Inc(NativeUInt(Item), SizeOf(T));

    case SizeOf(T) of
      1:
      begin
        VByte := Item.Bytes[0];
        if (VByte <> 0) then Continue;
      end;
      2:
      begin
        VWord := Item.Words[0];
        if (VWord <> 0) then Continue;
      end;
      4:
      begin
        VInteger := Item.Integers[0];
        if (VInteger <> 0) then Continue;
      end;
      8:
      begin
        {$ifdef LARGEINT}
          VNative := Item.Int64s[0];
          if (VNative <> 0) then Continue;
        {$else .SMALLINT}
          VInteger := Item.Integers[0];
          VNative := Item.Integers[1];
          if (VInteger or Integer(VNative) <> 0) then Continue;
        {$endif}
      end;
    else
      if (SizeOf(T) < SizeOf(NativeUInt)) then
      begin
        case SizeOf(T) of
          3:
          begin
            Flags := Item.Words[0];
            DestItem.Words[0] := Flags;
            VByte := Item.Bytes[2];
            DestItem.Bytes[2] := VByte;
            Flags := Flags or VByte;
          end;
          5:
          begin
            Flags := Item.Integers[0];
            DestItem.Integers[0] := Flags;
            VByte := Item.Bytes[4];
            DestItem.Bytes[4] := VByte;
            Flags := Flags or VByte;
          end;
          6:
          begin
            Flags := Item.Integers[0];
            DestItem.Integers[0] := Flags;
            VWord := Item.Words[2];
            DestItem.Words[2] := VWord;
            Flags := Flags or VWord;
          end;
          7:
          begin
            Flags := Item.Integers[0];
            DestItem.Integers[0] := Flags;
            VWord := Item.Words[2];
            DestItem.Words[2] := VWord;
            Flags := Flags or VWord;
            VByte := Item.Bytes[6];
            DestItem.Bytes[6] := VByte;
            Flags := Flags or VByte;
          end;
        end;
      end else
      if (SizeOf(T) <= 16) then
      begin
        {$ifdef LARGEINT}
          Flags := Item.Int64s[0];
          DestItem.Int64s[0] := Flags;
        {$else .SMALLINT}
          Flags := Item.Integers[0];
          DestItem.Integers[0] := Flags;
        {$endif}

        {$ifdef SMALLINT}
          if (SizeOf(T) >= SizeOf(Integer) * 2) then
          begin
            VInteger := Item.Integers[1];
            Item.Integers[1] := VInteger;
            Flags := Flags or Cardinal(VInteger);
          end;
          if (SizeOf(T) >= SizeOf(Integer) * 3) then
          begin
            VInteger := Item.Integers[2];
            Item.Integers[2] := VInteger;
            Flags := Flags or Cardinal(VInteger);
          end;
          if (SizeOf(T)  = SizeOf(Integer) * 4) then
          begin
            VInteger := Item.Integers[3];
            Item.Integers[3] := VInteger;
            Flags := Flags or Cardinal(VInteger);
          end;
        {$else .LARGEINT}
          if (SizeOf(T)  = SizeOf(Int64) * 2) then
          begin
            VNative := Item.Int64s[1];
            Item.Int64s[1] := VNative;
            Flags := Flags or VNative;
          end;
        {$endif}

        if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
        begin
          VNative := PNativeUInt(@Item.Bytes[SizeOf(T) - SizeOf(NativeUInt)])^;
          PNativeUInt(@DestItem.Bytes[SizeOf(T) - SizeOf(NativeUInt)])^ := VNative;
          Flags := Flags or VNative;
        end;
      end else
      begin
        Flags := 0;

        Index := 0;
        repeat
          Inc(Index);
          VNative := PNativeUInt(Item)^;
          PNativeUInt(DestItem)^ := VNative;
          Inc(NativeUInt(Item), SizeOf(NativeUInt));
          Inc(NativeUInt(DestItem), SizeOf(NativeUInt));
          Flags := Flags or VNative;
        until (Index = SizeOf(T) div SizeOf(NativeUInt));

        if (SizeOf(T) and (SizeOf(NativeUInt) - 1) <> 0) then
        begin
          Dec(NativeUInt(Item), SizeOf(T) - (SizeOf(NativeUInt) - 1));
          Dec(NativeUInt(DestItem), SizeOf(T) - (SizeOf(NativeUInt) - 1));
          VNative := PNativeUInt(Item)^;
          PNativeUInt(DestItem)^ := VNative;
          Flags := Flags or VNative;
        end;

        Dec(NativeUInt(Item), (SizeOf(T) div SizeOf(NativeUInt)) * SizeOf(NativeUInt));
        Dec(NativeUInt(DestItem), (SizeOf(T) div SizeOf(NativeUInt)) * SizeOf(NativeUInt));
      end;

      if (Flags <> 0) then Continue;
    end;

    goto next_item;
  until (False);

  R := NativeInt(DestItem) - NativeInt(FItems);
  case SizeOf(T) of
  0, 1: FCount.Native := R;
     2: FCount.Native := R shr 1;
     4: FCount.Native := R shr 2;
     8: FCount.Native := R shr 3;
    16: FCount.Native := R shr 4;
    32: FCount.Native := R shr 5;
    64: FCount.Native := R shr 6;
   128: FCount.Native := R shr 7;
   256: FCount.Native := R shr 8;
  else
    FCount.Native := Round(R * (1 / SizeOf(T)));
  end;
  Exit;

comparer_recall:
  ComparerInst.Method.Data := Pointer(FComparer);
  ComparerInst.Method.Code := PPointer(PNativeUInt(FComparer)^ + 3 * SizeOf(Pointer))^;
  ComparerInst.Vtable := @ComparerInst.QueryInterface;
  ComparerInst.Call := @TList<T>.InternalIsEmpty;
  Self.Pack(TEmptyFunc(@ComparerInst));
end;

{$else !SMARTGENERICS}
procedure TList<T>.Pack;
label
  next_item, comparer_recall;
var
  R, i: NativeInt;
  Item, TopItem, DestItem: PItem;
  _Self: TList<T>;
  ComparerInst: TComparerInst;
begin
  if (Assigned(FComparer)) then goto comparer_recall;

  _Self := Self;
  with _Self do
  begin
    Item := Pointer(FItems);
    Dec(Item);
    TopItem := Item + FCount.Native;
  end;

  repeat
    if (Item = TopItem) then Exit;
    Inc(Item);
  until (TEquals(InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Equals)(
        @InterfaceDefaults.TDefaultEqualityComparer<T>.Instance, Item^, Default(T)));

  DestItem := Item;
  goto next_item;
  repeat
    // DestItem^ := Item^;
    case SizeOf(T) of
      1: TRAIIHelper.T1(Pointer(DestItem)^) := TRAIIHelper.T1(Pointer(Item)^);
      2: TRAIIHelper.T2(Pointer(DestItem)^) := TRAIIHelper.T2(Pointer(Item)^);
      3: TRAIIHelper.T3(Pointer(DestItem)^) := TRAIIHelper.T3(Pointer(Item)^);
      4: TRAIIHelper.T4(Pointer(DestItem)^) := TRAIIHelper.T4(Pointer(Item)^);
      5: TRAIIHelper.T5(Pointer(DestItem)^) := TRAIIHelper.T5(Pointer(Item)^);
      6: TRAIIHelper.T6(Pointer(DestItem)^) := TRAIIHelper.T6(Pointer(Item)^);
      7: TRAIIHelper.T7(Pointer(DestItem)^) := TRAIIHelper.T7(Pointer(Item)^);
      8: TRAIIHelper.T8(Pointer(DestItem)^) := TRAIIHelper.T8(Pointer(Item)^);
      9: TRAIIHelper.T9(Pointer(DestItem)^) := TRAIIHelper.T9(Pointer(Item)^);
      10: TRAIIHelper.T10(Pointer(DestItem)^) := TRAIIHelper.T10(Pointer(Item)^);
      11: TRAIIHelper.T11(Pointer(DestItem)^) := TRAIIHelper.T11(Pointer(Item)^);
      12: TRAIIHelper.T12(Pointer(DestItem)^) := TRAIIHelper.T12(Pointer(Item)^);
      13: TRAIIHelper.T13(Pointer(DestItem)^) := TRAIIHelper.T13(Pointer(Item)^);
      14: TRAIIHelper.T14(Pointer(DestItem)^) := TRAIIHelper.T14(Pointer(Item)^);
      15: TRAIIHelper.T15(Pointer(DestItem)^) := TRAIIHelper.T15(Pointer(Item)^);
      16: TRAIIHelper.T16(Pointer(DestItem)^) := TRAIIHelper.T16(Pointer(Item)^);
      17: TRAIIHelper.T17(Pointer(DestItem)^) := TRAIIHelper.T17(Pointer(Item)^);
      18: TRAIIHelper.T18(Pointer(DestItem)^) := TRAIIHelper.T18(Pointer(Item)^);
      19: TRAIIHelper.T19(Pointer(DestItem)^) := TRAIIHelper.T19(Pointer(Item)^);
      20: TRAIIHelper.T20(Pointer(DestItem)^) := TRAIIHelper.T20(Pointer(Item)^);
      21: TRAIIHelper.T21(Pointer(DestItem)^) := TRAIIHelper.T21(Pointer(Item)^);
      22: TRAIIHelper.T22(Pointer(DestItem)^) := TRAIIHelper.T22(Pointer(Item)^);
      23: TRAIIHelper.T23(Pointer(DestItem)^) := TRAIIHelper.T23(Pointer(Item)^);
      24: TRAIIHelper.T24(Pointer(DestItem)^) := TRAIIHelper.T24(Pointer(Item)^);
      25: TRAIIHelper.T25(Pointer(DestItem)^) := TRAIIHelper.T25(Pointer(Item)^);
      26: TRAIIHelper.T26(Pointer(DestItem)^) := TRAIIHelper.T26(Pointer(Item)^);
      27: TRAIIHelper.T27(Pointer(DestItem)^) := TRAIIHelper.T27(Pointer(Item)^);
      28: TRAIIHelper.T28(Pointer(DestItem)^) := TRAIIHelper.T28(Pointer(Item)^);
      29: TRAIIHelper.T29(Pointer(DestItem)^) := TRAIIHelper.T29(Pointer(Item)^);
      30: TRAIIHelper.T30(Pointer(DestItem)^) := TRAIIHelper.T30(Pointer(Item)^);
      31: TRAIIHelper.T31(Pointer(DestItem)^) := TRAIIHelper.T31(Pointer(Item)^);
      32: TRAIIHelper.T32(Pointer(DestItem)^) := TRAIIHelper.T32(Pointer(Item)^);
      33: TRAIIHelper.T33(Pointer(DestItem)^) := TRAIIHelper.T33(Pointer(Item)^);
      34: TRAIIHelper.T34(Pointer(DestItem)^) := TRAIIHelper.T34(Pointer(Item)^);
      35: TRAIIHelper.T35(Pointer(DestItem)^) := TRAIIHelper.T35(Pointer(Item)^);
      36: TRAIIHelper.T36(Pointer(DestItem)^) := TRAIIHelper.T36(Pointer(Item)^);
      37: TRAIIHelper.T37(Pointer(DestItem)^) := TRAIIHelper.T37(Pointer(Item)^);
      38: TRAIIHelper.T38(Pointer(DestItem)^) := TRAIIHelper.T38(Pointer(Item)^);
      39: TRAIIHelper.T39(Pointer(DestItem)^) := TRAIIHelper.T39(Pointer(Item)^);
      40: TRAIIHelper.T40(Pointer(DestItem)^) := TRAIIHelper.T40(Pointer(Item)^);
    else
      for i := 1 to SizeOf(T) div SizeOf(NativeUInt) do
      begin
        NativeUInt(Pointer(DestItem)^) := NativeUInt(Pointer(Item)^);
        Inc(NativeInt(Item), SizeOf(NativeUInt));
        Inc(NativeInt(DestItem), SizeOf(NativeUInt));
      end;

      case SizeOf(T) and (SizeOf(NativeUInt) - 1) of
        1: TRAIIHelper.T1(Pointer(DestItem)^) := TRAIIHelper.T1(Pointer(Item)^);
        2: TRAIIHelper.T2(Pointer(DestItem)^) := TRAIIHelper.T2(Pointer(Item)^);
        3: TRAIIHelper.T3(Pointer(DestItem)^) := TRAIIHelper.T3(Pointer(Item)^);
      {$ifdef LARGEINT}
        4: TRAIIHelper.T4(Pointer(DestItem)^) := TRAIIHelper.T4(Pointer(Item)^);
        5: TRAIIHelper.T5(Pointer(DestItem)^) := TRAIIHelper.T5(Pointer(Item)^);
        6: TRAIIHelper.T6(Pointer(DestItem)^) := TRAIIHelper.T6(Pointer(Item)^);
        7: TRAIIHelper.T7(Pointer(DestItem)^) := TRAIIHelper.T7(Pointer(Item)^);
      {$endif}
      end;

      Dec(NativeInt(DestItem), (SizeOf(T) div SizeOf(NativeUInt)) * SizeOf(NativeUInt));
      Dec(NativeInt(Item), (SizeOf(T) div SizeOf(NativeUInt)) * SizeOf(NativeUInt));
    end;

    Inc(DestItem);
  next_item:
    if (Item = TopItem) then Break;
    Inc(Item);

    if (not TEquals(InterfaceDefaults.TDefaultEqualityComparer<T>.Instance.Equals)(
      @InterfaceDefaults.TDefaultEqualityComparer<T>.Instance, Item^, Default(T))) then Continue;
    goto next_item;
  until (False);

  _Self := Self;
  with _Self do
  begin
    R := NativeInt(DestItem) - NativeInt(FItems);
    case SizeOf(T) of
    0, 1: FCount.Native := R;
       2: FCount.Native := R shr 1;
       4: FCount.Native := R shr 2;
       8: FCount.Native := R shr 3;
      16: FCount.Native := R shr 4;
      32: FCount.Native := R shr 5;
      64: FCount.Native := R shr 6;
     128: FCount.Native := R shr 7;
     256: FCount.Native := R shr 8;
    else
      FCount.Native := Round(R * (1 / SizeOf(T)));
    end;
  end;
  Exit;

comparer_recall:
  ComparerInst.Method.Data := Pointer(FComparer);
  ComparerInst.Method.Code := PPointer(PNativeUInt(FComparer)^ + 3 * SizeOf(Pointer))^;
  ComparerInst.Vtable := @ComparerInst.QueryInterface;
  ComparerInst.Call := @TList<T>.InternalIsEmpty;
  Self.Pack(TEmptyFunc(@ComparerInst));
end;
{$endif}

procedure TList<T>.Pack(const IsEmpty: TEmptyFunc);
label
  next_item;
var
  R, i: NativeInt;
  Item, TopItem, DestItem: PItem;
  Equals: TMethod;
  _Self: TList<T>;
begin
  Equals.Data := PPointer(@IsEmpty)^;
  Equals.Code := PPointer(PNativeUInt(Equals.Data)^ + 3 * SizeOf(Pointer))^;

  _Self := Self;
  with _Self do
  begin
    Item := Pointer(FItems);
    Dec(Item);
    TopItem := Item + FCount.Native;
  end;

  repeat
    if (Item = TopItem) then Exit;
    Inc(Item);
  until (TEquals(Equals.Code)(Equals.Data, Item^, Default(T)));

  DestItem := Item;
  goto next_item;
  repeat
    // DestItem^ := Item^;
    case SizeOf(T) of
      1: TRAIIHelper.T1(Pointer(DestItem)^) := TRAIIHelper.T1(Pointer(Item)^);
      2: TRAIIHelper.T2(Pointer(DestItem)^) := TRAIIHelper.T2(Pointer(Item)^);
      3: TRAIIHelper.T3(Pointer(DestItem)^) := TRAIIHelper.T3(Pointer(Item)^);
      4: TRAIIHelper.T4(Pointer(DestItem)^) := TRAIIHelper.T4(Pointer(Item)^);
      5: TRAIIHelper.T5(Pointer(DestItem)^) := TRAIIHelper.T5(Pointer(Item)^);
      6: TRAIIHelper.T6(Pointer(DestItem)^) := TRAIIHelper.T6(Pointer(Item)^);
      7: TRAIIHelper.T7(Pointer(DestItem)^) := TRAIIHelper.T7(Pointer(Item)^);
      8: TRAIIHelper.T8(Pointer(DestItem)^) := TRAIIHelper.T8(Pointer(Item)^);
      9: TRAIIHelper.T9(Pointer(DestItem)^) := TRAIIHelper.T9(Pointer(Item)^);
      10: TRAIIHelper.T10(Pointer(DestItem)^) := TRAIIHelper.T10(Pointer(Item)^);
      11: TRAIIHelper.T11(Pointer(DestItem)^) := TRAIIHelper.T11(Pointer(Item)^);
      12: TRAIIHelper.T12(Pointer(DestItem)^) := TRAIIHelper.T12(Pointer(Item)^);
      13: TRAIIHelper.T13(Pointer(DestItem)^) := TRAIIHelper.T13(Pointer(Item)^);
      14: TRAIIHelper.T14(Pointer(DestItem)^) := TRAIIHelper.T14(Pointer(Item)^);
      15: TRAIIHelper.T15(Pointer(DestItem)^) := TRAIIHelper.T15(Pointer(Item)^);
      16: TRAIIHelper.T16(Pointer(DestItem)^) := TRAIIHelper.T16(Pointer(Item)^);
      17: TRAIIHelper.T17(Pointer(DestItem)^) := TRAIIHelper.T17(Pointer(Item)^);
      18: TRAIIHelper.T18(Pointer(DestItem)^) := TRAIIHelper.T18(Pointer(Item)^);
      19: TRAIIHelper.T19(Pointer(DestItem)^) := TRAIIHelper.T19(Pointer(Item)^);
      20: TRAIIHelper.T20(Pointer(DestItem)^) := TRAIIHelper.T20(Pointer(Item)^);
      21: TRAIIHelper.T21(Pointer(DestItem)^) := TRAIIHelper.T21(Pointer(Item)^);
      22: TRAIIHelper.T22(Pointer(DestItem)^) := TRAIIHelper.T22(Pointer(Item)^);
      23: TRAIIHelper.T23(Pointer(DestItem)^) := TRAIIHelper.T23(Pointer(Item)^);
      24: TRAIIHelper.T24(Pointer(DestItem)^) := TRAIIHelper.T24(Pointer(Item)^);
      25: TRAIIHelper.T25(Pointer(DestItem)^) := TRAIIHelper.T25(Pointer(Item)^);
      26: TRAIIHelper.T26(Pointer(DestItem)^) := TRAIIHelper.T26(Pointer(Item)^);
      27: TRAIIHelper.T27(Pointer(DestItem)^) := TRAIIHelper.T27(Pointer(Item)^);
      28: TRAIIHelper.T28(Pointer(DestItem)^) := TRAIIHelper.T28(Pointer(Item)^);
      29: TRAIIHelper.T29(Pointer(DestItem)^) := TRAIIHelper.T29(Pointer(Item)^);
      30: TRAIIHelper.T30(Pointer(DestItem)^) := TRAIIHelper.T30(Pointer(Item)^);
      31: TRAIIHelper.T31(Pointer(DestItem)^) := TRAIIHelper.T31(Pointer(Item)^);
      32: TRAIIHelper.T32(Pointer(DestItem)^) := TRAIIHelper.T32(Pointer(Item)^);
      33: TRAIIHelper.T33(Pointer(DestItem)^) := TRAIIHelper.T33(Pointer(Item)^);
      34: TRAIIHelper.T34(Pointer(DestItem)^) := TRAIIHelper.T34(Pointer(Item)^);
      35: TRAIIHelper.T35(Pointer(DestItem)^) := TRAIIHelper.T35(Pointer(Item)^);
      36: TRAIIHelper.T36(Pointer(DestItem)^) := TRAIIHelper.T36(Pointer(Item)^);
      37: TRAIIHelper.T37(Pointer(DestItem)^) := TRAIIHelper.T37(Pointer(Item)^);
      38: TRAIIHelper.T38(Pointer(DestItem)^) := TRAIIHelper.T38(Pointer(Item)^);
      39: TRAIIHelper.T39(Pointer(DestItem)^) := TRAIIHelper.T39(Pointer(Item)^);
      40: TRAIIHelper.T40(Pointer(DestItem)^) := TRAIIHelper.T40(Pointer(Item)^);
    else
      for i := 1 to SizeOf(T) div SizeOf(NativeUInt) do
      begin
        NativeUInt(Pointer(DestItem)^) := NativeUInt(Pointer(Item)^);
        Inc(NativeInt(Item), SizeOf(NativeUInt));
        Inc(NativeInt(DestItem), SizeOf(NativeUInt));
      end;

      case SizeOf(T) and (SizeOf(NativeUInt) - 1) of
        1: TRAIIHelper.T1(Pointer(DestItem)^) := TRAIIHelper.T1(Pointer(Item)^);
        2: TRAIIHelper.T2(Pointer(DestItem)^) := TRAIIHelper.T2(Pointer(Item)^);
        3: TRAIIHelper.T3(Pointer(DestItem)^) := TRAIIHelper.T3(Pointer(Item)^);
      {$ifdef LARGEINT}
        4: TRAIIHelper.T4(Pointer(DestItem)^) := TRAIIHelper.T4(Pointer(Item)^);
        5: TRAIIHelper.T5(Pointer(DestItem)^) := TRAIIHelper.T5(Pointer(Item)^);
        6: TRAIIHelper.T6(Pointer(DestItem)^) := TRAIIHelper.T6(Pointer(Item)^);
        7: TRAIIHelper.T7(Pointer(DestItem)^) := TRAIIHelper.T7(Pointer(Item)^);
      {$endif}
      end;

      Dec(NativeInt(DestItem), (SizeOf(T) div SizeOf(NativeUInt)) * SizeOf(NativeUInt));
      Dec(NativeInt(Item), (SizeOf(T) div SizeOf(NativeUInt)) * SizeOf(NativeUInt));
    end;

    Inc(DestItem);
  next_item:
    if (Item = TopItem) then Break;
    Inc(Item);

    if (not TEquals(Equals.Code)(Equals.Data, Item^, Default(T))) then Continue;
    goto next_item;
  until (False);

  _Self := Self;
  with _Self do
  begin
    R := NativeInt(DestItem) - NativeInt(FItems);
    case SizeOf(T) of
    0, 1: FCount.Native := R;
       2: FCount.Native := R shr 1;
       4: FCount.Native := R shr 2;
       8: FCount.Native := R shr 3;
      16: FCount.Native := R shr 4;
      32: FCount.Native := R shr 5;
      64: FCount.Native := R shr 6;
     128: FCount.Native := R shr 7;
     256: FCount.Native := R shr 8;
    else
      FCount.Native := Round(R * (1 / SizeOf(T)));
    end;
  end;
end;


{ TStack<T> }

constructor TStack<T>.Create;
begin
  inherited Create;
end;

constructor TStack<T>.Create(const Collection: TEnumerable<T>);
var
  Item: T;
begin
  Create;
  for Item in Collection do
    Push(Item);
end;

procedure TStack<T>.InternalPush(const Value: T);
label
  available;
var
  Count, Null: NativeInt;
  Item: ^TRAIIHelper.TData16;
begin
  Count := FCount.Native;
  if (Count <> FCapacity.Native) then
  begin
  available:
    FCount.Native := Count + 1;
    Item := Pointer(@FItems[Count]);

    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(T) or System.HasWeakRef(T)) then
    {$else}
    if (SizeOf(T) >= SizeOf(NativeInt)) then
    {$endif}
    begin
      {$ifdef SMARTGENERICS}
      if (GetTypeKind(T) = tkVariant) then
      begin
        Item.Integers[0] := 0;
      end else
      {$endif}
      if (SizeOf(T) <= 16) then
      begin
        Null := 0;
        {$ifdef SMALLINT}
          if (SizeOf(T) >= SizeOf(Integer) * 1) then Item.Integers[0] := Null;
          if (SizeOf(T) >= SizeOf(Integer) * 2) then Item.Integers[1] := Null;
          if (SizeOf(T) >= SizeOf(Integer) * 3) then Item.Integers[2] := Null;
          if (SizeOf(T)  = SizeOf(Integer) * 4) then Item.Integers[3] := Null;
        {$else .LARGEINT}
          if (SizeOf(T) >= SizeOf(Int64) * 1) then Item.Int64s[0] := Null;
          if (SizeOf(T)  = SizeOf(Int64) * 2) then Item.Int64s[1] := Null;
          case SizeOf(T) of
             4..7: Item.Integers[0] := Null;
           12..15: Item.Integers[2] := Null;
          end;
        {$endif}
        case SizeOf(T) of
           2,3: Item.Words[0] := 0;
           6,7: Item.Words[2] := 0;
         10,11: Item.Words[4] := 0;
         14,15: Item.Words[6] := 0;
        end;
        case SizeOf(T) of
           1: Item.Bytes[ 1-1] := 0;
           3: Item.Bytes[ 3-1] := 0;
           5: Item.Bytes[ 5-1] := 0;
           7: Item.Bytes[ 7-1] := 0;
           9: Item.Bytes[ 9-1] := 0;
          11: Item.Bytes[11-1] := 0;
          13: Item.Bytes[13-1] := 0;
          15: Item.Bytes[15-1] := 0;
        end;
      end else
      begin
        TRAIIHelper<T>.Init(Pointer(Item));
      end;
    end;

    PItem(Item)^ := Value;
    if Assigned(FInternalNotify) then
      FInternalNotify(Self, Value, cnAdded);
    Exit;
  end else
  begin
    Self.Grow;
    goto available;
  end;
end;

procedure TStack<T>.Push(const Value: T);
var
  Count, Null: NativeInt;
  Item: ^TRAIIHelper.TData16;
begin
  Count := FCount.Native;
  if (Count <> FCapacity.Native) and (not Assigned(FInternalNotify)) then
  begin
    Inc(Count);
    FCount.Native := Count;
    Dec(Count);
    Item := Pointer(@FItems[Count]);

    {$ifdef SMARTGENERICS}
    if (System.IsManagedType(T) or System.HasWeakRef(T)) then
    {$else}
    if (SizeOf(T) >= SizeOf(NativeInt)) then
    {$endif}
    begin
      {$ifdef SMARTGENERICS}
      if (GetTypeKind(T) = tkVariant) then
      begin
        Item.Integers[0] := 0;
      end else
      {$endif}
      if (SizeOf(T) <= 16) then
      begin
        Null := 0;
        {$ifdef SMALLINT}
          if (SizeOf(T) >= SizeOf(Integer) * 1) then Item.Integers[0] := Null;
          if (SizeOf(T) >= SizeOf(Integer) * 2) then Item.Integers[1] := Null;
          if (SizeOf(T) >= SizeOf(Integer) * 3) then Item.Integers[2] := Null;
          if (SizeOf(T)  = SizeOf(Integer) * 4) then Item.Integers[3] := Null;
        {$else .LARGEINT}
          if (SizeOf(T) >= SizeOf(Int64) * 1) then Item.Int64s[0] := Null;
          if (SizeOf(T)  = SizeOf(Int64) * 2) then Item.Int64s[1] := Null;
          case SizeOf(T) of
             4..7: Item.Integers[0] := Null;
           12..15: Item.Integers[2] := Null;
          end;
        {$endif}
        case SizeOf(T) of
           2,3: Item.Words[0] := 0;
           6,7: Item.Words[2] := 0;
         10,11: Item.Words[4] := 0;
         14,15: Item.Words[6] := 0;
        end;
        case SizeOf(T) of
           1: Item.Bytes[ 1-1] := 0;
           3: Item.Bytes[ 3-1] := 0;
           5: Item.Bytes[ 5-1] := 0;
           7: Item.Bytes[ 7-1] := 0;
           9: Item.Bytes[ 9-1] := 0;
          11: Item.Bytes[11-1] := 0;
          13: Item.Bytes[13-1] := 0;
          15: Item.Bytes[15-1] := 0;
        end;
      end else
      begin
        TRAIIHelper<T>.Init(Pointer(Item));
      end;
    end;

    PItem(Item)^ := Value;
    Exit;
  end else
  begin
    Self.InternalPush(Value);
  end;
end;

function TStack<T>.InternalPop(const Action: TCollectionNotification): T;
var
  Count: NativeInt;
  Item: PItem;
  VType: Integer;
begin
  Count := FCount.Native;
  if (Count <> 0) then
  begin
    Dec(Count);
    FCount.Native := Count;
    Item := @FItems[Count];
    Result := Item^;

    if Assigned(FInternalNotify) then
      Self.FInternalNotify(Self, Item^, Action);

    {$ifdef SMARTGENERICS}
    case GetTypeKind(T) of
      {$ifdef AUTOREFCOUNT}
      tkClass,
      {$endif}
      tkWString, tkLString, tkUString, tkInterface, tkDynArray:
      begin
        if (PNativeInt(Item)^ <> 0) then
        case GetTypeKind(T) of
          {$ifdef AUTOREFCOUNT}
          tkClass:
          begin
            TRAIIHelper.RefObjClear(Item);
          end;
          {$endif}
          {$ifdef MSWINDOWS}
          tkWString:
          begin
            TRAIIHelper.WStrClear(Item);
          end;
          {$else}
          tkWString,
          {$endif}
          tkLString, tkUString:
          begin
            TRAIIHelper.ULStrClear(Item);
          end;
          tkInterface:
          begin
            IInterface(PPointer(Item)^)._Release;
          end;
          tkDynArray:
          begin
            TRAIIHelper.DynArrayClear(Item, TypeInfo(T));
          end;
        end;
      end;
      {$ifdef WEAKREF}
      tkMethod:
      begin
        if (PMethod(Item).Data <> nil) then
          TRAIIHelper.WeakMethodClear(@PMethod(Item).Data);
      end;
      {$endif}
      tkVariant:
      begin
        VType := PVarData(Item).VType;
        if (VType and TRAIIHelper.varDeepData <> 0) then
        case VType of
          varBoolean, varUnknown+1..varUInt64: ;
        else
          System.VarClear(PVariant(Item)^);
        end;
      end;
    else
      TRAIIHelper<T>.Clear(Item);
    end;
    {$else}
    TRAIIHelper<T>.Clear(Item);
    {$endif}

    Exit;
  end else
  begin
    raise Self.EmptyException;
  end;
end;

function TStack<T>.Pop: T;
var
  Count: NativeInt;
  Item: PItem;
  VType: Integer;
begin
  Count := FCount.Native;
  if (Count <> 0) and (not Assigned(FInternalNotify)) then
  begin
    Dec(Count);
    FCount.Native := Count;
    Item := @FItems[Count];
    Result := Item^;

    {$ifdef SMARTGENERICS}
    case GetTypeKind(T) of
      {$ifdef AUTOREFCOUNT}
      tkClass,
      {$endif}
      tkWString, tkLString, tkUString, tkInterface, tkDynArray:
      begin
        if (PNativeInt(Item)^ <> 0) then
        case GetTypeKind(T) of
          {$ifdef AUTOREFCOUNT}
          tkClass:
          begin
            TRAIIHelper.RefObjClear(Item);
          end;
          {$endif}
          {$ifdef MSWINDOWS}
          tkWString:
          begin
            TRAIIHelper.WStrClear(Item);
          end;
          {$else}
          tkWString,
          {$endif}
          tkLString, tkUString:
          begin
            TRAIIHelper.ULStrClear(Item);
          end;
          tkInterface:
          begin
            IInterface(PPointer(Item)^)._Release;
          end;
          tkDynArray:
          begin
            TRAIIHelper.DynArrayClear(Item, TypeInfo(T));
          end;
        end;
      end;
      {$ifdef WEAKREF}
      tkMethod:
      begin
        if (PMethod(Item).Data <> nil) then
          TRAIIHelper.WeakMethodClear(@PMethod(Item).Data);
      end;
      {$endif}
      tkVariant:
      begin
        VType := PVarData(Item).VType;
        if (VType and TRAIIHelper.varDeepData <> 0) then
        case VType of
          varBoolean, varUnknown+1..varUInt64: ;
        else
          System.VarClear(PVariant(Item)^);
        end;
      end;
    else
      TRAIIHelper<T>.Clear(Item);
    end;
    {$else}
    TRAIIHelper<T>.Clear(Item);
    {$endif}

    Exit;
  end else
  begin
    Result := Self.InternalPop(cnRemoved);
  end;
end;

function TStack<T>.Extract: T;
var
  Count: NativeInt;
  Item: PItem;
  VType: Integer;
begin
  Count := FCount.Native;
  if (Count <> 0) and (not Assigned(FInternalNotify)) then
  begin
    Dec(Count);
    FCount.Native := Count;
    Item := @FItems[Count];
    Result := Item^;

    {$ifdef SMARTGENERICS}
    case GetTypeKind(T) of
      {$ifdef AUTOREFCOUNT}
      tkClass,
      {$endif}
      tkWString, tkLString, tkUString, tkInterface, tkDynArray:
      begin
        if (PNativeInt(Item)^ <> 0) then
        case GetTypeKind(T) of
          {$ifdef AUTOREFCOUNT}
          tkClass:
          begin
            TRAIIHelper.RefObjClear(Item);
          end;
          {$endif}
          {$ifdef MSWINDOWS}
          tkWString:
          begin
            TRAIIHelper.WStrClear(Item);
          end;
          {$else}
          tkWString,
          {$endif}
          tkLString, tkUString:
          begin
            TRAIIHelper.ULStrClear(Item);
          end;
          tkInterface:
          begin
            IInterface(PPointer(Item)^)._Release;
          end;
          tkDynArray:
          begin
            TRAIIHelper.DynArrayClear(Item, TypeInfo(T));
          end;
        end;
      end;
      {$ifdef WEAKREF}
      tkMethod:
      begin
        if (PMethod(Item).Data <> nil) then
          TRAIIHelper.WeakMethodClear(@PMethod(Item).Data);
      end;
      {$endif}
      tkVariant:
      begin
        VType := PVarData(Item).VType;
        if (VType and TRAIIHelper.varDeepData <> 0) then
        case VType of
          varBoolean, varUnknown+1..varUInt64: ;
        else
          System.VarClear(PVariant(Item)^);
        end;
      end;
    else
      TRAIIHelper<T>.Clear(Item);
    end;
    {$else}
    TRAIIHelper<T>.Clear(Item);
    {$endif}

    Exit;
  end else
  begin
    Result := Self.InternalPop(cnExtracted);
  end;
end;

function TStack<T>.Peek: T;
var
  Count: NativeInt;
begin
  Count := FCount.Native;
  if (Count <> 0) then
  begin
    Result := FItems[Count - 1];
    Exit;
  end else
  begin
    raise Self.EmptyException;
  end;
end;


{ TQueue<T> }

constructor TQueue<T>.Create;
begin
  inherited Create;
end;

constructor TQueue<T>.Create(const Collection: TEnumerable<T>);
var
  Item: T;
begin
  Create;

  for Item in Collection do
    Enqueue(Item);
end;

procedure TQueue<T>.InternalEnqueue(const Value: T);
label
  available, actual;
var
  Count, Null: NativeInt;
  Item: ^TRAIIHelper.TData16;
begin
  Count := FCount.Native;
  if (Count <> FCapacity.Native) then
  begin
  available:
    FCount.Native := Count + 1;
    Count := FHead;
    if (Count <> FCapacity.Native) then
    begin
      Inc(Count);
      FHead := Count;
      Dec(Count);
    actual:
      Item := Pointer(@FItems[Count]);

      {$ifdef SMARTGENERICS}
      if (System.IsManagedType(T) or System.HasWeakRef(T)) then
      {$else}
      if (SizeOf(T) >= SizeOf(NativeInt)) then
      {$endif}
      begin
        {$ifdef SMARTGENERICS}
        if (GetTypeKind(T) = tkVariant) then
        begin
          Item.Integers[0] := 0;
        end else
        {$endif}
        if (SizeOf(T) <= 16) then
        begin
          Null := 0;
          {$ifdef SMALLINT}
            if (SizeOf(T) >= SizeOf(Integer) * 1) then Item.Integers[0] := Null;
            if (SizeOf(T) >= SizeOf(Integer) * 2) then Item.Integers[1] := Null;
            if (SizeOf(T) >= SizeOf(Integer) * 3) then Item.Integers[2] := Null;
            if (SizeOf(T)  = SizeOf(Integer) * 4) then Item.Integers[3] := Null;
          {$else .LARGEINT}
            if (SizeOf(T) >= SizeOf(Int64) * 1) then Item.Int64s[0] := Null;
            if (SizeOf(T)  = SizeOf(Int64) * 2) then Item.Int64s[1] := Null;
            case SizeOf(T) of
               4..7: Item.Integers[0] := Null;
             12..15: Item.Integers[2] := Null;
            end;
          {$endif}
          case SizeOf(T) of
             2,3: Item.Words[0] := 0;
             6,7: Item.Words[2] := 0;
           10,11: Item.Words[4] := 0;
           14,15: Item.Words[6] := 0;
          end;
          case SizeOf(T) of
             1: Item.Bytes[ 1-1] := 0;
             3: Item.Bytes[ 3-1] := 0;
             5: Item.Bytes[ 5-1] := 0;
             7: Item.Bytes[ 7-1] := 0;
             9: Item.Bytes[ 9-1] := 0;
            11: Item.Bytes[11-1] := 0;
            13: Item.Bytes[13-1] := 0;
            15: Item.Bytes[15-1] := 0;
          end;
        end else
        begin
          TRAIIHelper<T>.Init(Pointer(Item));
        end;
      end;

      PItem(Item)^ := Value;
      if Assigned(FInternalNotify) then
        FInternalNotify(Self, Value, cnAdded);
      Exit;
    end else
    begin
      FHead := 1;
      Count := 0;
      goto actual;
    end;
  end else
  begin
    Self.Grow;
    goto available;
  end;
end;

procedure TQueue<T>.Enqueue(const Value: T);
label
  actual;
var
  Count, Null: NativeInt;
  Item: ^TRAIIHelper.TData16;
begin
  Count := FCount.Native;
  if (Count <> FCapacity.Native) and (not Assigned(FInternalNotify)) then
  begin
    FCount.Native := Count + 1;
    Count := FHead;
    if (Count <> FCapacity.Native) then
    begin
      Inc(Count);
      FHead := Count;
      Dec(Count);
    actual:
      Item := Pointer(@FItems[Count]);

      {$ifdef SMARTGENERICS}
      if (System.IsManagedType(T) or System.HasWeakRef(T)) then
      {$else}
      if (SizeOf(T) >= SizeOf(NativeInt)) then
      {$endif}
      begin
        {$ifdef SMARTGENERICS}
        if (GetTypeKind(T) = tkVariant) then
        begin
          Item.Integers[0] := 0;
        end else
        {$endif}
        if (SizeOf(T) <= 16) then
        begin
          Null := 0;
          {$ifdef SMALLINT}
            if (SizeOf(T) >= SizeOf(Integer) * 1) then Item.Integers[0] := Null;
            if (SizeOf(T) >= SizeOf(Integer) * 2) then Item.Integers[1] := Null;
            if (SizeOf(T) >= SizeOf(Integer) * 3) then Item.Integers[2] := Null;
            if (SizeOf(T)  = SizeOf(Integer) * 4) then Item.Integers[3] := Null;
          {$else .LARGEINT}
            if (SizeOf(T) >= SizeOf(Int64) * 1) then Item.Int64s[0] := Null;
            if (SizeOf(T)  = SizeOf(Int64) * 2) then Item.Int64s[1] := Null;
            case SizeOf(T) of
               4..7: Item.Integers[0] := Null;
             12..15: Item.Integers[2] := Null;
            end;
          {$endif}
          case SizeOf(T) of
             2,3: Item.Words[0] := 0;
             6,7: Item.Words[2] := 0;
           10,11: Item.Words[4] := 0;
           14,15: Item.Words[6] := 0;
          end;
          case SizeOf(T) of
             1: Item.Bytes[ 1-1] := 0;
             3: Item.Bytes[ 3-1] := 0;
             5: Item.Bytes[ 5-1] := 0;
             7: Item.Bytes[ 7-1] := 0;
             9: Item.Bytes[ 9-1] := 0;
            11: Item.Bytes[11-1] := 0;
            13: Item.Bytes[13-1] := 0;
            15: Item.Bytes[15-1] := 0;
          end;
        end else
        begin
          TRAIIHelper<T>.Init(Pointer(Item));
        end;
      end;

      PItem(Item)^ := Value;
      Exit;
    end else
    begin
      FHead := 1;
      Count := 0;
      goto actual;
    end;
  end else
  begin
    Self.InternalEnqueue(Value);
  end;
end;

function TQueue<T>.InternalDequeue(const Action: TCollectionNotification): T;
label
  actual;
var
  Count: NativeInt;
  Item: PItem;
  VType: Integer;
begin
  Count := FCount.Native;
  if (Count <> 0) then
  begin
    FCount.Native := Count - 1;
    Count := FTail + 1;
    if (Count <> FCapacity.Native) then
    begin
      FTail := Count;
    actual:
      Dec(Count);
      Item := @FItems[Count];
      Result := Item^;

      if Assigned(FInternalNotify) then
        Self.FInternalNotify(Self, Item^, Action);

      {$ifdef SMARTGENERICS}
      case GetTypeKind(T) of
        {$ifdef AUTOREFCOUNT}
        tkClass,
        {$endif}
        tkWString, tkLString, tkUString, tkInterface, tkDynArray:
        begin
          if (PNativeInt(Item)^ <> 0) then
          case GetTypeKind(T) of
            {$ifdef AUTOREFCOUNT}
            tkClass:
            begin
              TRAIIHelper.RefObjClear(Item);
            end;
            {$endif}
            {$ifdef MSWINDOWS}
            tkWString:
            begin
              TRAIIHelper.WStrClear(Item);
            end;
            {$else}
            tkWString,
            {$endif}
            tkLString, tkUString:
            begin
              TRAIIHelper.ULStrClear(Item);
            end;
            tkInterface:
            begin
              IInterface(PPointer(Item)^)._Release;
            end;
            tkDynArray:
            begin
              TRAIIHelper.DynArrayClear(Item, TypeInfo(T));
            end;
          end;
        end;
        {$ifdef WEAKREF}
        tkMethod:
        begin
          if (PMethod(Item).Data <> nil) then
            TRAIIHelper.WeakMethodClear(@PMethod(Item).Data);
        end;
        {$endif}
        tkVariant:
        begin
          VType := PVarData(Item).VType;
          if (VType and TRAIIHelper.varDeepData <> 0) then
          case VType of
            varBoolean, varUnknown+1..varUInt64: ;
          else
            System.VarClear(PVariant(Item)^);
          end;
        end;
      else
        TRAIIHelper<T>.Clear(Item);
      end;
      {$else}
      TRAIIHelper<T>.Clear(Item);
      {$endif}

      Exit;
    end else
    begin
      Count := 0;
      FTail := Count;
      Count := FCapacity.Native;
      goto actual;
    end;
  end else
  begin
    raise Self.EmptyException;
  end;
end;

function TQueue<T>.Dequeue: T;
label
  actual;
var
  Count: NativeInt;
  Item: PItem;
  VType: Integer;
begin
  Count := FCount.Native;
  if (Count <> 0) and (not Assigned(FInternalNotify)) then
  begin
    FCount.Native := Count - 1;
    Count := FTail + 1;
    if (Count <> FCapacity.Native) then
    begin
      FTail := Count;
    actual:
      Dec(Count);
      Item := @FItems[Count];
      Result := Item^;

      {$ifdef SMARTGENERICS}
      case GetTypeKind(T) of
        {$ifdef AUTOREFCOUNT}
        tkClass,
        {$endif}
        tkWString, tkLString, tkUString, tkInterface, tkDynArray:
        begin
          if (PNativeInt(Item)^ <> 0) then
          case GetTypeKind(T) of
            {$ifdef AUTOREFCOUNT}
            tkClass:
            begin
              TRAIIHelper.RefObjClear(Item);
            end;
            {$endif}
            {$ifdef MSWINDOWS}
            tkWString:
            begin
              TRAIIHelper.WStrClear(Item);
            end;
            {$else}
            tkWString,
            {$endif}
            tkLString, tkUString:
            begin
              TRAIIHelper.ULStrClear(Item);
            end;
            tkInterface:
            begin
              IInterface(PPointer(Item)^)._Release;
            end;
            tkDynArray:
            begin
              TRAIIHelper.DynArrayClear(Item, TypeInfo(T));
            end;
          end;
        end;
        {$ifdef WEAKREF}
        tkMethod:
        begin
          if (PMethod(Item).Data <> nil) then
            TRAIIHelper.WeakMethodClear(@PMethod(Item).Data);
        end;
        {$endif}
        tkVariant:
        begin
          VType := PVarData(Item).VType;
          if (VType and TRAIIHelper.varDeepData <> 0) then
          case VType of
            varBoolean, varUnknown+1..varUInt64: ;
          else
            System.VarClear(PVariant(Item)^);
          end;
        end;
      else
        TRAIIHelper<T>.Clear(Item);
      end;
      {$else}
      TRAIIHelper<T>.Clear(Item);
      {$endif}

      Exit;
    end else
    begin
      Count := 0;
      FTail := Count;
      Count := FCapacity.Native;
      goto actual;
    end;
  end else
  begin
    Result := Self.InternalDequeue(cnRemoved);
  end;
end;

function TQueue<T>.Extract: T;
label
  actual;
var
  Count: NativeInt;
  Item: PItem;
  VType: Integer;
begin
  Count := FCount.Native;
  if (Count <> 0) and (not Assigned(FInternalNotify)) then
  begin
    FCount.Native := Count - 1;
    Count := FTail + 1;
    if (Count <> FCapacity.Native) then
    begin
      FTail := Count;
    actual:
      Dec(Count);
      Item := @FItems[Count];
      Result := Item^;

      {$ifdef SMARTGENERICS}
      case GetTypeKind(T) of
        {$ifdef AUTOREFCOUNT}
        tkClass,
        {$endif}
        tkWString, tkLString, tkUString, tkInterface, tkDynArray:
        begin
          if (PNativeInt(Item)^ <> 0) then
          case GetTypeKind(T) of
            {$ifdef AUTOREFCOUNT}
            tkClass:
            begin
              TRAIIHelper.RefObjClear(Item);
            end;
            {$endif}
            {$ifdef MSWINDOWS}
            tkWString:
            begin
              TRAIIHelper.WStrClear(Item);
            end;
            {$else}
            tkWString,
            {$endif}
            tkLString, tkUString:
            begin
              TRAIIHelper.ULStrClear(Item);
            end;
            tkInterface:
            begin
              IInterface(PPointer(Item)^)._Release;
            end;
            tkDynArray:
            begin
              TRAIIHelper.DynArrayClear(Item, TypeInfo(T));
            end;
          end;
        end;
        {$ifdef WEAKREF}
        tkMethod:
        begin
          if (PMethod(Item).Data <> nil) then
            TRAIIHelper.WeakMethodClear(@PMethod(Item).Data);
        end;
        {$endif}
        tkVariant:
        begin
          VType := PVarData(Item).VType;
          if (VType and TRAIIHelper.varDeepData <> 0) then
          case VType of
            varBoolean, varUnknown+1..varUInt64: ;
          else
            System.VarClear(PVariant(Item)^);
          end;
        end;
      else
        TRAIIHelper<T>.Clear(Item);
      end;
      {$else}
      TRAIIHelper<T>.Clear(Item);
      {$endif}

      Exit;
    end else
    begin
      Count := 0;
      FTail := Count;
      Count := FCapacity.Native;
      goto actual;
    end;
  end else
  begin
    Result := Self.InternalDequeue(cnExtracted);
  end;
end;

function TQueue<T>.Peek: T;
begin
  if (FCount.Native <> 0) then
  begin
    Result := FItems[FTail];
    Exit;
  end else
  begin
    raise Self.EmptyException;
  end;
end;


{ TThreadList<T> }

procedure TThreadList<T>.Add(const Item: T);
begin
  LockList;
  try
    if (Duplicates = dupAccept) or
       (FList.IndexOf(Item) = -1) then
      FList.Add(Item)
    else if Duplicates = dupError then
      raise EListError.CreateFmt(SDuplicateItem, [FList.ItemValue(Item)]);
  finally
    UnlockList;
  end;
end;

procedure TThreadList<T>.Clear;
begin
  LockList;
  try
    FList.Clear;
  finally
    UnlockList;
  end;
end;

constructor TThreadList<T>.Create;
begin
  inherited Create;
  FLock := TObject.Create;
  FList := TList<T>.Create;
  FDuplicates := dupIgnore;
end;

destructor TThreadList<T>.Destroy;
begin
  LockList;    // Make sure nobody else is inside the list.
  try
    FList.Free;
    inherited Destroy;
  finally
    UnlockList;
    FLock.Free;
  end;
end;

function TThreadList<T>.LockList: TList<T>;
begin
  TMonitor.Enter(FLock);
  Result := FList;
end;

procedure TThreadList<T>.Remove(const Item: T);
begin
  RemoveItem(Item, TDirection.FromBeginning);
end;

procedure TThreadList<T>.RemoveItem(const Item: T; Direction: TDirection);
begin
  LockList;
  try
    FList.RemoveItem(Item, Direction);
  finally
    UnlockList;
  end;
end;

procedure TThreadList<T>.UnlockList;
begin
  TMonitor.Exit(FLock);
end;

{ TThreadedQueue<T> }

constructor TThreadedQueue<T>.Create(AQueueDepth: Integer = 10; PushTimeout: LongWord = INFINITE; PopTimeout: LongWord = INFINITE);
begin
  inherited Create;
  SetLength(FQueue, AQueueDepth);
  FQueueLock := TObject.Create;
  FQueueNotEmpty := TObject.Create;
  FQueueNotFull := TObject.Create;
  FPushTimeout := PushTimeout;
  FPopTimeout := PopTimeout;
end;

destructor TThreadedQueue<T>.Destroy;
begin
  DoShutDown;
  FQueueNotFull.Free;
  FQueueNotEmpty.Free;
  FQueueLock.Free;
  inherited;
end;

procedure TThreadedQueue<T>.Grow(ADelta: Integer);
begin
  TMonitor.Enter(FQueueLock);
  try
    SetLength(FQueue, Length(FQueue) + ADelta);
  finally
    TMonitor.Exit(FQueueLock);
  end;
  TMonitor.PulseAll(FQueueNotFull);
end;

function TThreadedQueue<T>.PopItem: T;
var
  LQueueSize: Integer;
begin
  PopItem(LQueueSize, Result);
end;

function TThreadedQueue<T>.PopItem(var AQueueSize: Integer; var AItem: T): TWaitResult;
begin
  AItem := Default(T);
  TMonitor.Enter(FQueueLock);
  try
    Result := wrSignaled;
    while (Result = wrSignaled) and (FQueueSize = 0) and not FShutDown do
      if not TMonitor.Wait(FQueueNotEmpty, FQueueLock, FPopTimeout) then
        Result := wrTimeout;

    if (FShutDown and (FQueueSize = 0)) or (Result <> wrSignaled) then
      Exit;

    AItem := FQueue[FQueueOffset];

    FQueue[FQueueOffset] := Default(T);

    Dec(FQueueSize);
    Inc(FQueueOffset);
    Inc(FTotalItemsPopped);

    if FQueueOffset = Length(FQueue) then
      FQueueOffset := 0;

  finally
    AQueueSize := FQueueSize;
    TMonitor.Exit(FQueueLock);
  end;

  TMonitor.Pulse(FQueueNotFull);
end;

function TThreadedQueue<T>.PopItem(var AItem: T): TWaitResult;
var
  LQueueSize: Integer;
begin
  Result := PopItem(LQueueSize, AItem);
end;

function TThreadedQueue<T>.PopItem(var AQueueSize: Integer): T;
begin
  PopItem(AQueueSize, Result);
end;

function TThreadedQueue<T>.PushItem(const AItem: T): TWaitResult;
var
  LQueueSize: Integer;
begin
  Result := PushItem(AItem, LQueueSize);
end;

function TThreadedQueue<T>.PushItem(const AItem: T; var AQueueSize: Integer): TWaitResult;
begin
  TMonitor.Enter(FQueueLock);
  try
    Result := wrSignaled;
    while (Result = wrSignaled) and (FQueueSize = Length(FQueue)) and not FShutDown do
      if not TMonitor.Wait(FQueueNotFull, FQueueLock, FPushTimeout) then
        Result := wrTimeout;

    if FShutDown or (Result <> wrSignaled) then
      Exit;

    FQueue[(FQueueOffset + FQueueSize) mod Length(FQueue)] := AItem;
    Inc(FQueueSize);
    Inc(FTotalItemsPushed);

  finally
    AQueueSize := FQueueSize;
    TMonitor.Exit(FQueueLock);
  end;

  TMonitor.Pulse(FQueueNotEmpty);
end;

procedure TThreadedQueue<T>.DoShutDown;
begin
  TMonitor.Enter(FQueueLock);
  try
    FShutDown := True;
  finally
    TMonitor.Exit(FQueueLock);
  end;
  TMonitor.PulseAll(FQueueNotFull);
  TMonitor.PulseAll(FQueueNotEmpty);
end;

{ TObjectList<T> }

constructor TObjectList<T>.Create(AOwnsObjects: Boolean);
begin
  inherited Create;
  FOwnsObjects := AOwnsObjects;
end;

constructor TObjectList<T>.Create(const AComparer: IComparer<T>; AOwnsObjects: Boolean);
begin
  FOwnsObjects := AOwnsObjects;
  inherited Create(AComparer);
end;

constructor TObjectList<T>.Create(const Collection: TEnumerable<T>; AOwnsObjects: Boolean);
begin
  FOwnsObjects := AOwnsObjects;
  inherited Create(Collection);
end;

procedure TObjectList<T>.SetOwnsObjects(const Value: Boolean);
begin
  if (FOwnsObjects <> Value) then
  begin
    FOwnsObjects := Value;
    SetNotifyMethods;
  end;
end;

procedure TObjectList<T>.DisposeNotifyEvent(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
begin
  Self.FOnNotify(Sender, Item, Action);
  if (Action = cnRemoved) then
    Item.DisposeOf;
end;

procedure TObjectList<T>.DisposeOnly(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    Item.DisposeOf;
end;

procedure TObjectList<T>.SetNotifyMethods;
var
  VMTNotify: procedure(const Item: T; Action: TCollectionNotification) of object;
begin
  if (not FOwnsObjects) then
  begin
    inherited;
    Exit;
  end;

  TMethod(FInternalNotify).Data := Pointer(Self);
  VMTNotify := Self.Notify;
  if (TMethod(VMTNotify).Code <> @TCustomList<T>.Notify) then
  begin
    TMethod(FInternalNotify).Code := @TCustomList<T>.NotifyCaller;
  end else
  if (Assigned(Self.FOnNotify)) then
  begin
    TMethod(FInternalNotify).Code := @TObjectList<T>.DisposeNotifyEvent;
  end else
  begin
    TMethod(FInternalNotify).Code := @TObjectList<T>.DisposeOnly;
  end;
end;


{ TObjectStack<T> }

constructor TObjectStack<T>.Create(AOwnsObjects: Boolean);
begin
  FOwnsObjects := AOwnsObjects;
  inherited Create;
end;

constructor TObjectStack<T>.Create(const Collection: TEnumerable<T>; AOwnsObjects: Boolean);
begin
  FOwnsObjects := AOwnsObjects;
  inherited Create(Collection);
end;

procedure TObjectStack<T>.Pop;
begin
  inherited Pop;
end;

procedure TObjectStack<T>.SetOwnsObjects(const Value: Boolean);
begin
  if (FOwnsObjects <> Value) then
  begin
    FOwnsObjects := Value;
    SetNotifyMethods;
  end;
end;

procedure TObjectStack<T>.DisposeNotifyEvent(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
begin
  Self.FOnNotify(Sender, Item, Action);
  if (Action = cnRemoved) then
    Item.DisposeOf;
end;

procedure TObjectStack<T>.DisposeOnly(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    Item.DisposeOf;
end;

procedure TObjectStack<T>.SetNotifyMethods;
var
  VMTNotify: procedure(const Item: T; Action: TCollectionNotification) of object;
begin
  if (not FOwnsObjects) then
  begin
    inherited;
    Exit;
  end;

  TMethod(FInternalNotify).Data := Pointer(Self);
  VMTNotify := Self.Notify;
  if (TMethod(VMTNotify).Code <> @TCustomList<T>.Notify) then
  begin
    TMethod(FInternalNotify).Code := @TCustomList<T>.NotifyCaller;
  end else
  if (Assigned(Self.FOnNotify)) then
  begin
    TMethod(FInternalNotify).Code := @TObjectStack<T>.DisposeNotifyEvent;
  end else
  begin
    TMethod(FInternalNotify).Code := @TObjectStack<T>.DisposeOnly;
  end;
end;


{ TObjectQueue<T> }

constructor TObjectQueue<T>.Create(AOwnsObjects: Boolean);
begin
  FOwnsObjects := AOwnsObjects;
  inherited Create;
end;

constructor TObjectQueue<T>.Create(const Collection: TEnumerable<T>; AOwnsObjects: Boolean);
begin
  FOwnsObjects := AOwnsObjects;
  inherited Create(Collection);
end;

procedure TObjectQueue<T>.Dequeue;
begin
  inherited Dequeue;
end;

procedure TObjectQueue<T>.SetOwnsObjects(const Value: Boolean);
begin
  if (FOwnsObjects <> Value) then
  begin
    FOwnsObjects := Value;
    SetNotifyMethods;
  end;
end;

procedure TObjectQueue<T>.DisposeNotifyEvent(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
begin
  Self.FOnNotify(Sender, Item, Action);
  if (Action = cnRemoved) then
    Item.DisposeOf;
end;

procedure TObjectQueue<T>.DisposeOnly(Sender: TObject; const Item: TObject; Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    Item.DisposeOf;
end;

procedure TObjectQueue<T>.SetNotifyMethods;
var
  VMTNotify: procedure(const Item: T; Action: TCollectionNotification) of object;
begin
  if (not FOwnsObjects) then
  begin
    inherited;
    Exit;
  end;

  TMethod(FInternalNotify).Data := Pointer(Self);
  VMTNotify := Self.Notify;
  if (TMethod(VMTNotify).Code <> @TCustomList<T>.Notify) then
  begin
    TMethod(FInternalNotify).Code := @TCustomList<T>.NotifyCaller;
  end else
  if (Assigned(Self.FOnNotify)) then
  begin
    TMethod(FInternalNotify).Code := @TObjectQueue<T>.DisposeNotifyEvent;
  end else
  begin
    TMethod(FInternalNotify).Code := @TObjectQueue<T>.DisposeOnly;
  end;
end;


{ TObjectDictionary<TKey,TValue> }

constructor TObjectDictionary<TKey,TValue>.Create(Ownerships: TDictionaryOwnerships;
  ACapacity: Integer = 0);
begin
  Create(Ownerships, ACapacity, nil);
end;

constructor TObjectDictionary<TKey,TValue>.Create(Ownerships: TDictionaryOwnerships;
  const AComparer: IEqualityComparer<TKey>);
begin
  Create(Ownerships, 0, AComparer);
end;

constructor TObjectDictionary<TKey,TValue>.Create(Ownerships: TDictionaryOwnerships;
  ACapacity: Integer; const AComparer: IEqualityComparer<TKey>);
begin
  FOwnerships := Ownerships;
  if (Ownerships = []) then
    raise EInvalidCast.CreateRes(Pointer(@SInvalidCast));

  if (doOwnsKeys in Ownerships) then
  begin
    if (TypeInfo(TKey) = nil) or (PTypeInfo(TypeInfo(TKey))^.Kind <> tkClass) then
      raise EInvalidCast.CreateRes(Pointer(@SInvalidCast));
  end;

  if (doOwnsValues in Ownerships) then
  begin
    if (TypeInfo(TValue) = nil) or (PTypeInfo(TypeInfo(TValue))^.Kind <> tkClass) then
      raise EInvalidCast.CreateRes(Pointer(@SInvalidCast));
  end;

  inherited Create(ACapacity, AComparer);
end;

procedure TObjectDictionary<TKey,TValue>.DisposeKeyEvent(Sender: TObject;
  const Key: TObject; Action: TCollectionNotification);
type
  TOnKeyNotify = procedure(Data, Sender: TObject; const Key: TObject; Action: TCollectionNotification);
begin
  TOnKeyNotify(TMethod(FOnKeyNotify).Code)(TMethod(FOnKeyNotify).Data, Self, Key, Action);
  if (Action = cnRemoved) then
    Key.DisposeOf;
end;

procedure TObjectDictionary<TKey,TValue>.DisposeKeyOnly(Sender: TObject;
  const Key: TObject; Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    Key.DisposeOf;
end;

procedure TObjectDictionary<TKey,TValue>.DisposeValueEvent(Sender: TObject;
  const Value: TObject; Action: TCollectionNotification);
type
  TOnValueNotify = procedure(Data, Sender: TObject; const Value: TObject; Action: TCollectionNotification);
begin
  TOnValueNotify(TMethod(FOnValueNotify).Code)(TMethod(FOnValueNotify).Data, Self, Value, Action);
  if (Action = cnRemoved) then
    Value.DisposeOf;
end;

procedure TObjectDictionary<TKey,TValue>.DisposeValueOnly(Sender: TObject;
  const Value: TObject; Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    Value.DisposeOf;
end;

procedure TObjectDictionary<TKey,TValue>.DisposeItemNotifyKeyCaller(const Item: TItem;
  Action: TCollectionNotification);
begin
  Self.KeyNotify(Item.Key, Action);
end;

procedure TObjectDictionary<TKey,TValue>.DisposeItemNotifyKeyEvent(const Item: TItem;
  Action: TCollectionNotification);
begin
  FOnKeyNotify(Self, Item.Key, Action);
  if (Action = cnRemoved) then
    TObject(Pointer(@Item.Key)^).DisposeOf;
end;

procedure TObjectDictionary<TKey,TValue>.DisposeItemNotifyKeyOnly(const Item: TItem;
  Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    TObject(Pointer(@Item.Key)^).DisposeOf;
end;

procedure TObjectDictionary<TKey,TValue>.DisposeItemNotifyValueCaller(const Item: TItem;
  Action: TCollectionNotification);
begin
  Self.ValueNotify(Item.Value, Action);
end;

procedure TObjectDictionary<TKey,TValue>.DisposeItemNotifyValueEvent(const Item: TItem;
  Action: TCollectionNotification);
begin
  FOnValueNotify(Self, Item.Value, Action);
  if (Action = cnRemoved) then
    TObject(Pointer(@Item.Value)^).DisposeOf;
end;

procedure TObjectDictionary<TKey,TValue>.DisposeItemNotifyValueOnly(const Item: TItem;
  Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    TObject(Pointer(@Item.Value)^).DisposeOf;
end;

procedure TObjectDictionary<TKey,TValue>.SetNotifyMethods;
var
  VMTKeyNotify: procedure(const Key: TKey; Action: TCollectionNotification) of object;
  VMTValueNotify: procedure(const Value: TValue; Action: TCollectionNotification) of object;
begin
  // FInternalKeyNotify
  TMethod(FInternalKeyNotify).Data := Pointer(Self);
  VMTKeyNotify := Self.KeyNotify;
  if (TMethod(VMTKeyNotify).Code <> @TCustomDictionary<TKey,TValue>.KeyNotify) then
  begin
    TMethod(FInternalKeyNotify).Code := @TCustomDictionary<TKey,TValue>.KeyNotifyCaller;
  end else
  if (doOwnsKeys in FOwnerships) then
  begin
    if (Assigned(Self.FOnKeyNotify)) then
    begin
      TMethod(FInternalKeyNotify).Code := @TObjectDictionary<TKey,TValue>.DisposeValueEvent;
    end else
    begin
      TMethod(FInternalKeyNotify).Code := @TObjectDictionary<TKey,TValue>.DisposeValueOnly;
    end;
  end else
  begin
    TMethod(FInternalKeyNotify) := TMethod(Self.FOnKeyNotify);
  end;

  // FInternalValueNotify
  TMethod(FInternalValueNotify).Data := Pointer(Self);
  VMTValueNotify := Self.ValueNotify;
  if (TMethod(VMTValueNotify).Code <> @TCustomDictionary<TValue,TValue>.ValueNotify) then
  begin
    TMethod(FInternalValueNotify).Code := @TCustomDictionary<TValue,TValue>.ValueNotifyCaller;
  end else
  if (doOwnsValues in FOwnerships) then
  begin
    if (Assigned(Self.FOnValueNotify)) then
    begin
      TMethod(FInternalValueNotify).Code := @TObjectDictionary<TValue,TValue>.DisposeValueEvent;
    end else
    begin
      TMethod(FInternalValueNotify).Code := @TObjectDictionary<TValue,TValue>.DisposeValueOnly;
    end;
  end else
  begin
    TMethod(FInternalValueNotify) := TMethod(Self.FOnValueNotify);
  end;

  // FInternalItemNotify
  TMethod(FInternalItemNotify).Data := Self;
  if (TMethod(VMTKeyNotify).Code <> @TCustomDictionary<TKey,TValue>.KeyNotify) and
    (TMethod(VMTValueNotify).Code <> @TCustomDictionary<TValue,TValue>.ValueNotify) then
  begin
    TMethod(FInternalItemNotify).Code := @TCustomDictionary<TValue,TValue>.ItemNotifyCaller;
  end else
  if (Assigned(FInternalKeyNotify)) and (Assigned(FInternalValueNotify)) then
  begin
    TMethod(FInternalItemNotify).Code := @TCustomDictionary<TValue,TValue>.ItemNotifyEvents;
  end else
  if (Assigned(FInternalKeyNotify)) then
  begin
    if (TMethod(VMTKeyNotify).Code <> @TCustomDictionary<TKey,TValue>.KeyNotify) then
    begin
      TMethod(FInternalItemNotify).Code := @TObjectDictionary<TKey,TKey>.DisposeItemNotifyKeyCaller;
    end else
    if (Assigned(Self.FOnKeyNotify)) then
    begin
      TMethod(FInternalItemNotify).Code := @TObjectDictionary<TKey,TKey>.DisposeItemNotifyKeyEvent;
    end else
    begin
      TMethod(FInternalItemNotify).Code := @TObjectDictionary<TKey,TKey>.DisposeItemNotifyKeyOnly;
    end;
  end else
  // if (Assigned(FInternalValueNotify)) then
  begin
    if (TMethod(VMTValueNotify).Code <> @TCustomDictionary<TValue,TValue>.ValueNotify) then
    begin
      TMethod(FInternalItemNotify).Code := @TObjectDictionary<TValue,TValue>.DisposeItemNotifyValueCaller;
    end else
    if (Assigned(Self.FOnValueNotify)) then
    begin
      TMethod(FInternalItemNotify).Code := @TObjectDictionary<TValue,TValue>.DisposeItemNotifyValueEvent;
    end else
    begin
      TMethod(FInternalItemNotify).Code := @TObjectDictionary<TValue,TValue>.DisposeItemNotifyValueOnly;
    end;
  end;
end;


{ TRapidObjectDictionary<TKey,TValue> }

constructor TRapidObjectDictionary<TKey,TValue>.Create(Ownerships: TDictionaryOwnerships;
  ACapacity: Integer);
begin
  FOwnerships := Ownerships;
  if (Ownerships = []) then
    raise EInvalidCast.CreateRes(Pointer(@SInvalidCast));

  if (doOwnsKeys in Ownerships) then
  begin
    if (TypeInfo(TKey) = nil) or (PTypeInfo(TypeInfo(TKey))^.Kind <> tkClass) then
      raise EInvalidCast.CreateRes(Pointer(@SInvalidCast));
  end;

  if (doOwnsValues in Ownerships) then
  begin
    if (TypeInfo(TValue) = nil) or (PTypeInfo(TypeInfo(TValue))^.Kind <> tkClass) then
      raise EInvalidCast.CreateRes(Pointer(@SInvalidCast));
  end;

  inherited Create(ACapacity);
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeKeyEvent(Sender: TObject;
  const Key: TObject; Action: TCollectionNotification);
type
  TOnKeyNotify = procedure(Data, Sender: TObject; const Key: TObject; Action: TCollectionNotification);
begin
  TOnKeyNotify(TMethod(FOnKeyNotify).Code)(TMethod(FOnKeyNotify).Data, Self, Key, Action);
  if (Action = cnRemoved) then
    Key.DisposeOf;
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeKeyOnly(Sender: TObject;
  const Key: TObject; Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    Key.DisposeOf;
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeValueEvent(Sender: TObject;
  const Value: TObject; Action: TCollectionNotification);
type
  TOnValueNotify = procedure(Data, Sender: TObject; const Value: TObject; Action: TCollectionNotification);
begin
  TOnValueNotify(TMethod(FOnValueNotify).Code)(TMethod(FOnValueNotify).Data, Self, Value, Action);
  if (Action = cnRemoved) then
    Value.DisposeOf;
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeValueOnly(Sender: TObject;
  const Value: TObject; Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    Value.DisposeOf;
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeItemNotifyKeyCaller(const Item: TItem;
  Action: TCollectionNotification);
begin
  Self.KeyNotify(Item.Key, Action);
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeItemNotifyKeyEvent(const Item: TItem;
  Action: TCollectionNotification);
begin
  FOnKeyNotify(Self, Item.Key, Action);
  if (Action = cnRemoved) then
    TObject(Pointer(@Item.Key)^).DisposeOf;
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeItemNotifyKeyOnly(const Item: TItem;
  Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    TObject(Pointer(@Item.Key)^).DisposeOf;
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeItemNotifyValueCaller(const Item: TItem;
  Action: TCollectionNotification);
begin
  Self.ValueNotify(Item.Value, Action);
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeItemNotifyValueEvent(const Item: TItem;
  Action: TCollectionNotification);
begin
  FOnValueNotify(Self, Item.Value, Action);
  if (Action = cnRemoved) then
    TObject(Pointer(@Item.Value)^).DisposeOf;
end;

procedure TRapidObjectDictionary<TKey,TValue>.DisposeItemNotifyValueOnly(const Item: TItem;
  Action: TCollectionNotification);
begin
  if (Action = cnRemoved) then
    TObject(Pointer(@Item.Value)^).DisposeOf;
end;

procedure TRapidObjectDictionary<TKey,TValue>.SetNotifyMethods;
var
  VMTKeyNotify: procedure(const Key: TKey; Action: TCollectionNotification) of object;
  VMTValueNotify: procedure(const Value: TValue; Action: TCollectionNotification) of object;
begin
  // FInternalKeyNotify
  TMethod(FInternalKeyNotify).Data := Pointer(Self);
  VMTKeyNotify := Self.KeyNotify;
  if (TMethod(VMTKeyNotify).Code <> @TCustomDictionary<TKey,TValue>.KeyNotify) then
  begin
    TMethod(FInternalKeyNotify).Code := @TCustomDictionary<TKey,TValue>.KeyNotifyCaller;
  end else
  if (doOwnsKeys in FOwnerships) then
  begin
    if (Assigned(Self.FOnKeyNotify)) then
    begin
      TMethod(FInternalKeyNotify).Code := @TRapidObjectDictionary<TKey,TValue>.DisposeValueEvent;
    end else
    begin
      TMethod(FInternalKeyNotify).Code := @TRapidObjectDictionary<TKey,TValue>.DisposeValueOnly;
    end;
  end else
  begin
    TMethod(FInternalKeyNotify) := TMethod(Self.FOnKeyNotify);
  end;

  // FInternalValueNotify
  TMethod(FInternalValueNotify).Data := Pointer(Self);
  VMTValueNotify := Self.ValueNotify;
  if (TMethod(VMTValueNotify).Code <> @TCustomDictionary<TValue,TValue>.ValueNotify) then
  begin
    TMethod(FInternalValueNotify).Code := @TCustomDictionary<TValue,TValue>.ValueNotifyCaller;
  end else
  if (doOwnsValues in FOwnerships) then
  begin
    if (Assigned(Self.FOnValueNotify)) then
    begin
      TMethod(FInternalValueNotify).Code := @TRapidObjectDictionary<TValue,TValue>.DisposeValueEvent;
    end else
    begin
      TMethod(FInternalValueNotify).Code := @TRapidObjectDictionary<TValue,TValue>.DisposeValueOnly;
    end;
  end else
  begin
    TMethod(FInternalValueNotify) := TMethod(Self.FOnValueNotify);
  end;

  // FInternalItemNotify
  TMethod(FInternalItemNotify).Data := Self;
  if (TMethod(VMTKeyNotify).Code <> @TCustomDictionary<TKey,TValue>.KeyNotify) and
    (TMethod(VMTValueNotify).Code <> @TCustomDictionary<TValue,TValue>.ValueNotify) then
  begin
    TMethod(FInternalItemNotify).Code := @TCustomDictionary<TValue,TValue>.ItemNotifyCaller;
  end else
  if (Assigned(FInternalKeyNotify)) and (Assigned(FInternalValueNotify)) then
  begin
    TMethod(FInternalItemNotify).Code := @TCustomDictionary<TValue,TValue>.ItemNotifyEvents;
  end else
  if (Assigned(FInternalKeyNotify)) then
  begin
    if (TMethod(VMTKeyNotify).Code <> @TCustomDictionary<TKey,TValue>.KeyNotify) then
    begin
      TMethod(FInternalItemNotify).Code := @TRapidObjectDictionary<TKey,TKey>.DisposeItemNotifyKeyCaller;
    end else
    if (Assigned(Self.FOnKeyNotify)) then
    begin
      TMethod(FInternalItemNotify).Code := @TRapidObjectDictionary<TKey,TKey>.DisposeItemNotifyKeyEvent;
    end else
    begin
      TMethod(FInternalItemNotify).Code := @TRapidObjectDictionary<TKey,TKey>.DisposeItemNotifyKeyOnly;
    end;
  end else
  // if (Assigned(FInternalValueNotify)) then
  begin
    if (TMethod(VMTValueNotify).Code <> @TCustomDictionary<TValue,TValue>.ValueNotify) then
    begin
      TMethod(FInternalItemNotify).Code := @TRapidObjectDictionary<TValue,TValue>.DisposeItemNotifyValueCaller;
    end else
    if (Assigned(Self.FOnValueNotify)) then
    begin
      TMethod(FInternalItemNotify).Code := @TRapidObjectDictionary<TValue,TValue>.DisposeItemNotifyValueEvent;
    end else
    begin
      TMethod(FInternalItemNotify).Code := @TRapidObjectDictionary<TValue,TValue>.DisposeItemNotifyValueOnly;
    end;
  end;
end;


end.
